#ifndef RAND_MESH_H
#define RAND_MESH_H

#include <stdio.h>
#include <stdlib.h>

typedef struct {
  long next;
  long prev;
  long node;
  long owner;
} link_struct;

typedef struct {
  long link_index;
  long link_count;
  long flags;
  long link_head;
} node_struct;

const long MROOT_LINK_FLAG = -1;
const long MBORDER_FLAG = 1;
const long MINTERIOR_FLAG = 0;

class mesh_storage_class {
public:

  node_struct *nodes; 
  long node_count;
  long free_node_count;

  link_struct *links;
  long link_count;
  long free_link_count;
  long free_link_start;

public:

  void init(long nodes, long links);
  
  //actions on nodes
  long new_link_index();
  long new_node_index();

  //void delete_link(long node);
  void front_link_insert(long owner, long node,  long linknode);
  void back_link_insert(long owner, long node, long linknode);

  long new_border_node(long leftnode, long rightnode);

  long get_first_link_node(long node);

  void set_interior(long node);
  int border_test(long node);
  long count_links(long node);

  long find_link(long owner, long dest);

  void print_summary_data();
  void print_node_data(long node);
  void print_node(long node);
  void print_link(long link);

  int seek_link(long node, long linknode);

};

class tmesh_class {
public:

  mesh_storage_class mdata;
  
  // local geometry data -- probably shouldn't be here
  long node_count;
  long link_count;

public:
  void init(long nodes);

  long get_node_count();
  long get_link_count();

  long border_length();
  long border_node();

  long next_border_node(long node);
  long insert_border_node(long node);
  long join_border_node(long node);

  void output_text();
  void output_node_data(long node);

};


// constants for distribution
const int M_ADD = 1;
const int M_JOIN = 2;
const int M_SKIP = 3;

int distribution();

tmesh_class random_mesh(long size);


#endif

