/* sa_types.h  include file of type declarations for the program CirclePack
   and the CirclePack library routines. */

#ifndef CP_TYPES_H
#define CP_TYPES_H

#include "sa_hooks.h"

#if !defined(USING_INTEL_COMPILER) && !defined(WIN32)
using namespace std;
#endif

#ifdef WIN32
#include <float.h>
#include <process.h>
#endif

#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <pthread.h>
#ifdef WIN32
#define ZLIB_WINAPI
#endif
#include <zlib.h>

#ifdef WIN32
#include <fstream>
#include <iostream>
#include <iomanip>
#else
#include <fstream>
#include <iostream>
#include <iomanip>
#endif

#ifndef WIN32
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif

#ifdef CP_USE_DD  // Use doubledouble
#include "doubledouble.h"
#ifdef double
#undef double  // Make sure the typedef works correctly if both {cp,sa}_types.h are included
#endif
typedef double singledouble;
#define double doubledouble
#else
typedef double singledouble;
#include <math.h>  // We don't want the ordinary math function if using doubledouble
#endif

/*
 * Some compilers (e.g. MSVC++) don't define these
 * math constants.
 */

#ifndef M_E
#define M_E            2.7182818284590452354   /* e */
#endif
#ifndef M_LOG2E
#define M_LOG2E        1.4426950408889634074   /* log_2 e */
#endif
#ifndef M_LOG10E
#define M_LOG10E       0.43429448190325182765  /* log_10 e */
#endif
#ifndef M_LN2
#define M_LN2          0.69314718055994530942  /* log_e 2 */
#endif
#ifndef M_LN10
#define M_LN10         2.30258509299404568402  /* log_e 10 */
#endif
#ifndef M_PI
#define M_PI           3.14159265358979323846  /* pi */
#endif
#ifndef M_PI_2
#define M_PI_2         1.57079632679489661923  /* pi/2 */
#endif
#ifndef M_PI_4
#define M_PI_4         0.78539816339744830962  /* pi/4 */
#endif
#ifndef M_1_PI
#define M_1_PI         0.31830988618379067154  /* 1/pi */
#endif
#ifndef M_2_PI
#define M_2_PI         0.63661977236758134308  /* 2/pi */
#endif
#ifndef M_2_SQRTPI
#define M_2_SQRTPI     1.12837916709551257390  /* 2/sqrt(pi) */
#endif
#ifndef M_SQRT2
#define M_SQRT2        1.41421356237309504880  /* sqrt(2) */
#endif
#ifndef M_SQRT1_2
#define M_SQRT1_2      0.70710678118654752440  /* 1/sqrt(2) */
#endif

#ifndef WIN32

#include <unistd.h>
#include <pwd.h>
#include <sys/times.h>
#include <sys/resource.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>

#endif

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "complex_math_types.h"
#include "complex_math_proto.h"
#include "hyp_math_types.h"
#include "CPXlib.h"
#include "CPXlibint.h"
#include "CPXutil.h"
#include "CPregion.h"
#include "CPpoly.h"

#ifndef WIN32
#include <dirent.h>
#define closesocket(s) close(s)
typedef int SOCKET;
#endif

#define float double
#define NUM_PROC 6		/* max number of remote processes */
#define MAX_PETALS 20000	/* max number of petals in a flower */
#define MAX_COMPONENTS 300	/* number of allowed boundary components */
#define CANV_SIZE 450		/* initial size of canvase 0 */
#define MAX_PIXEL 1000 		/* max canvas size */
#define MSG_LENGTH 10000	/* length of message window backup */
#define NUM_COLORS 255          /* No. colors in color table */
#define COLOR_RAMP 200          /* No. colors in color ramp */
#define BG_COLOR 0		/* background color code */
#define FG_COLOR 255    	/* foreground color code */
#define NAME_MAX 255		/* max length of filenames */
#define SHOW 1
#define NOSHOW 0		/* display flags */
#define PS_UNIT_LINEWIDTH .5    /* default standard line width */
#define BUFSIZE 1024		/* need some big buffers */
#define PASS_INIT   0x1         /* Riffle routines should init superstep values */
#define PASS_ALLOC  0x2         /* Riffle routines should allocate memory */
#define PASS_FIRST  (PASS_INIT | PASS_ALLOC)
#define PASS_MIDDLE 0x4         /* Riffle routines should riffle */
#define PASS_LAST   0x8         /* Riffle routines should clean up memory */
#ifdef CP_USE_DD
#define MAX_ACCUR 30
#ifdef M_PI
#undef M_PI
#endif
#define M_PI       doubledouble::Pi   
#ifdef M_PI_2
#undef M_PI_2
#endif
#define M_PI_2     doubledouble::Pion2
#ifdef M_PI_4
#undef M_PI_4
#endif
#define M_PI_4     doubledouble::Pion4
#ifdef M_1_PI
#undef M_1_PI
#endif
#define M_1_PI     doubledouble::_Pi
#else
#define MAX_ACCUR 15            /* limit on requested accuracy */
#endif
#define TOLER .00000000001      /* default utility threshold */
#define OKERR .0000000000001    /* default okerr, smaller than TOLER */
#define M_TOLER .0000000000001  /* default for machine tolerance */
#define MOD1 .999999            /* threshold to be on unit circle:
				   (fixup: needs more care) */
				   
#define MAX_ALLOWABLE_BAD_CUTS 300

extern char *msgbuf;

/*======================= type definitions =========================*/

typedef struct  {
    double lx,ly;   /* lower left corner */
    double rx,ry;   /* upper right corner */
}  Box;

typedef struct  {
    int lx,ly;              /* box in pixels */
    int rx,ry;              /* remember: canvases think y goes down,
                               but we will take ly<ry. */
}  Pix_box;

struct StrTok
{
  int key;                      /* What type of data?
				   1: separator symbol
				   2. pack number
				   3: command flag
				   4: string data
				   5: integer data
				   6: float data */
  char *str;                    /* string info (must be allocated) */
  int intvalue;                 /* integer data */
  double floatvalue;            /* double value */
};

typedef struct
 {
        int vert[3];		/* ordered triple of vertices */
	int index_flag;		/* which to draw first, 0,1, or 2;
				 (So verts index_flag, (index_flag+1)%3
				 are used to plot (index_flag+2)%3.)*/
        int next_face;		/* next face in drawing order. */
	int next_red;		/* next red face (if this face is red) */
	int plot_flag;		/* >0 indicates whether face location
				   is reliable enough for plotting;
				   <=0 indicates some problem. */
	int color;		/* color index */
	int mark;		/* mark */
  } f_data;

struct RedList
 {
        int face;               /* face number */
        int v_flag;             /* gives index (in 'face') of the vert 
				   this face is responsible for in
				   drawing order of red faces. */
        double rad;             /* radius of v_flag vert */
        complex center;         /* center of v_flag vert */
        int corner_flag[3];     /* mark paired segment corner verts:
				   bits: 1=begin; 2=end */
        struct RedList *cross[3]; /* ptr to side-paired 
				     face in red chain */
        struct RedList *next;
        struct RedList *prev;
 };

struct EdgePair 
 {
        int edge_indx;    /* indx of counterckw end of initial edge */
        int mate_indx;    /* indx to counterckw end of final edge */
        struct RedList *edge,*mate; /* in storing edge pairings, 'mate' 
        	will be NULL if 'edge' is a border (non-paired) edge. */
        Mobius mob;   /* Mobius transform if this is paired edge.
			(Not yet implemented in sph case.) */
        double mob_err;   /* error in the Mobius transform (mainly for hyp
			case; mob is always an automorphism, but roundoff 
			can prevent it from providing exact match). */
        int color;
 };

struct Vertlist
 {
	int v;			/* vertex number */
	struct Vertlist *next;
 };

struct Edgelist
 {
	int v,w;		/* vertex numbers of edge ends */
	struct Edgelist *next;
 };

struct Centlist
 {
        complex z;              /* complex, e.g., circle center */
        struct Centlist *next;
 };

struct Pathlist
 {
	double x,y;		/* coords of point */
	struct Pathlist *next;
 };

struct xyzPoint
 {
        double x,y,z;          /* coords of point in 3-space */
 };

struct Overlaps
{
  int v,w;
  double angle;
  struct Overlaps *next;
};

struct BdryData
{
  int v;                  /* original vert number */
  int new_indx;           /* assigned new vert number (temporary) */
  int num;                /* number of faces in fan */
  int bdry_flag;
  int poison_flag;        /* true if vert is poison (util_flag=-1) */
  int indx1,indx2;        /* first and last indices in face_org of
			     faces of this fan */
  int duty_flag;          /* utility flag */
  struct BdryData *next_v;/* pts to next object sharing v */
  struct BdryData *prev_v;/* pts to prev object sharing v */
  struct BdryData *next;
  struct BdryData *prev;
};

/*==================== structure for circle data ====================*/

struct K_data
 {
	int num;                /* number of faces containg this node.*/
				/* 'flower' has entries from 0 to num */
	int color;		/* gives color value, 0=background */
	int mark;		/* mark */
	int bdry_flag; 		/* true for boundary nodes */
        int qual_flag;          /* keeps info on quality of layout */
	int plot_flag;		/* >0 if node has been successfully 
				   placed in packing layout; <=0 generally
				   reflects some placement problem */
	int *flower;		/* ptr to list of nodes in flower;
				   in positive orientation, last = first if
				   flower is closed */
	double *overlaps;	/* optional ptr to list of prescribed 
				   'inversive distances' to nghb's in flower 
				   (ie cosine of overlap angle/inv dist). */
        int util_flag;          /* utility flag (misc temp uses) */
        int next_vert;          /* next vert in draw order (misc temp uses) */
 };

struct R_data
 {
	complex center;		/* center as complex number */
	double rad;		/* radius of circle for this node. 
				   hyperbolic case: called 'x_radius',
				   x=1-exp(-2h) when hyp radius h is finite.
				   for infinite hyp radius, store negative 
				   of eucl radius (useful for plotting). */
	double curv;		/* angle sum at this vertex. */
	double aim;		/* desired curvature at this vertex */
 };


/* ==================== structure for screen data ====================*/

struct s_data
 {
    Pix_box pix_box;	/* describes pixel box for screen */
	Box box;		/* real world box for screen */
        int pix_per_unit;       /* pixels/real unit (for setting
				   num pts to draw in sph curves.) */
	double factor;		/* display factor from param window */
	double disp_trans[3][3];
				/* 3x3 transforms real x,y,z to those
				   for display (and inverse) */
	int cflag_opt;	        /* display option for circles,
				   standard flags. */
	int fflag_opt;	        /* display option for faces,
				   standard flags. */
	int unitcircle;		/* 0=don't diplay */
	int ms_flag;		/* flags for sensitized mice in canvas.
				   0=unsensitized, 1=rubberband circ,
				   2=add circ, 3=del circ, 4=enfold,
				   5=crosshairs,6=incr rad,7=decr rad */
	int coord_flag;		/* 1=show coord axes */
#ifndef WIN32
	Pixmap xpm;		/* pointer to Pixmap */
	XID xid;		/* canvas xid */
#endif
 };

extern struct s_data *screendata;

/* ==================== structure for pack data ====================*/

struct p_data
 {
   int status;		/* 0 if pack empty, 1 otherwise */
   char fhead[64];		/* pack header <= 40 char*/
   char file_name[NAME_MAX]; /* filename when file was written */
   int hes;		/* curvature of geometry,-1=hyp,0=eucl,1=sph */
   int intrinsic_geom;     /* intrinsic geometry (max pack setting) */
   struct K_data *packK_ptr;	/* pointer to corres complex data */
   struct R_data *packR_ptr;	/* pointer to corres rad data */
   int sizelimit;		/* max number of node space allocated */
   int nodecount;		/* number of nodes */
   int facecount;		/* number of faces = 1+edges-nodes */
   int intnode;		/* number of interior nodes */
   int num_bdry_comp;	/* number of bdry components */
   int num_int_comp;	/* number of components of interior verts */
   int bdry_starts[MAX_COMPONENTS+1];	
   /* indices for nodes marking bdry components */
   int int_starts[MAX_COMPONENTS+1];	
   /* indices of nodes marking int components */
   int first_face;		/* index of first face to plot */
   int first_red_face;	/* index of "red" chain of faces */
   int beta;		/* index of beta node (on boundary) */
   int alpha;		/* index of alpha node (origin) */
   int gamma;		/* index of node to be plotted on y>0 axis */
   int euler;		/* Euler characteristic */
   int genus;		/* genus of complex. Euler+#bdry=2-2g */
   int active_node;	/* currently active_node */
   int locks;		/* locks placed by remote processes;
			   bitwise flags */
   int overlap_status;	/* true if some non-trivial overlaps set */
   f_data *faces;		/* pointer to face data: filled by
				   routine complex_count */
   int *rwb_flags;		/* red/white/blue flags for faces:
				   set in face-order stuff */
   int *f_util;            /* temporary utility flags for processing 
			      faces: user must de/allocate as needed. */
   struct s_data *screen;	/* pointer to screen for display */
   struct Vertlist *vlist;   /* pack utility vert list */
   struct Vertlist *flist;   /* pack utility face list */
   struct Edgelist *elist;   /* pack utility edge list */
   struct Edgelist *vertex_map; /* optional table mapping (some)
				   vertices of this pack to 
				   vertex numbers of another, 
				   e.g. in projection to another
				   canvas.*/
   struct RedList *redfaces;  /* double linked list of data for drawing
				 bdry faces.*/
   struct EdgePair *edge_pair; /* stores pointers to red chain for edges
   	  associated with side pairings for non-simply connected complexes.
   	  Stored from index 1 (not zero); separate pointers for an edge and
   	  for its mate. */
   struct xyzPoint *xyzpoint; /* pointer to associated 3D coords */
 };

extern struct p_data *packdata;

/* IDEA: "light" packing has minimal data for standard repacking
   (assume tangency, no overlaps stored). Data is organized explicitly 
   for efficient communication/repacking. Local indices are sequential, 
   from 1 to nodecount; "variable" verts are those subject to
   repacking adjustment (even if on combinatoric bdry), others are
   "fixed" verts and are given the higher indices; fixed verts don't
   have flowers since they're only there to complete flowers of the
   variable verts. Space is allocated for variable vert aims if they
   are != 2*pi (Eg. a variable vert which is a bdry in parent typically
   has a negative aim and that will require an entry here.)

   Carry 'checkpoint' and 'orig_indices' data from parent
   packing to be able to pass info back. */

struct p_light
 {
   int *counts;           /* create area of size 12 for future
			     expansion. Current contents as follows:
                             0: vertcount
			        = size of radii block
				= size of orig_indices block
			     1: hes=geometry
			     2: vcount= count of variable verts
			     3: size of var_indices block 
                             4: checkcount (= parent nodecount) 
			     5: aimcount= count of non-default aims 
			        = size of aim_index block
				= size of aim block   
			  */

   int *generation;
   int *flower_heads;
   int *var_indices;      /* block of variable vert flower data: a
			     contiguous list of "num p(0)..p(num)". 
			     Note: stored here using local index 
			     (1 to vertcount) */
   int *orig_indices;     /* listing of all vert indices in parent pack */
   int *orig_fixed;       /* utility ptr = orig_indices+counts[2]+1 */
   double *radii;         /* block of  radii: bdry are last */
   double *fixed_radii;   /* utility ptr = radii+counts[2]+1 */
   int *aim_indices;      /* block of indices (local) of verts with 
			     non-default aims for variable verts */
   double *aims;          /* block of corresponding aims */
};

typedef void * IOBUF;

/* =================================================================
Remote process structures */

struct remote_control
{
	int 	op_code;	/* situation op_code */
		/* Options:
		 (bitwise)	0 = communications not yet set.
				1 = remote process up and waiting.
		 */
	int	locks;		/* locks on packs; bitwise, 1=locked */
	int 	pid;		/* process id; 0 if not running */
	int 	pipeto[2];	/* pipe to child; [1] open. */
	int	pipefr[2];	/* pipe from child: [0] open */
	FILE	*fp_to_remote;		/* file pointer */
	FILE	*fp_from_remote;	/* file pointer */
        IOBUF   *client_to;     /* to client buffer */
        IOBUF   *client_from;   /* from client buffer */
	char	program[32];	/* name of remote program */ 
	char 	moniker[32];	/* name for user to use in communications */
};

extern struct remote_control my_remote[NUM_PROC];

/* =================================================================
Global variables: here because often required by library routines. */

extern IOBUF *disp_fp;          /* stdout or socket for exchanging display info */
extern FILE *fp;
extern int dum_int;
extern char buff[BUFSIZE],cmd[2*BUFSIZE],data[2*BUFSIZE],next[BUFSIZE];
extern char buf[BUFSIZE];	/* general purpose utility buffer */
extern int alist[300];
extern double toler;	/* numerical tolerance for computing */
extern double okerr;	/* quantity small enough to set to zero 
		   in calculating */
extern double m_toler; /* machine tolerence = sqrt(machine precision) */
extern unsigned long fgcolor,bgcolor; /* indices of fore/backgound colors */
extern int sc_print_size;	/* size (inches) for postscript output */
extern int color_ramp_size;    /* color ramp depends on GUI being used */
extern unsigned char red[NUM_COLORS+1],green[NUM_COLORS+1],blue[NUM_COLORS+1];
extern int bc_min,bc_max;        /* min/max indices for special colors
			     (those for which bdry color is modified
			     from those of interior) */
extern int bc_offset;            /* When reading color values from rgb_file,
			     store circle/face bdry colors values 
			     at .75 of parent (interior); store 
			     in color map with offset (<0) to 
			     index of parent value. */
extern int line_thick;		/* thickness setting for lines */
extern double ps_linewidth;    /* linewidth for postscript */
extern double graylevel;       /* double value to send to postscript */
extern int fill_mode;		/* mode for filled objects: 0=grey,1=white,
				   2=black */
extern int javaMsgHook;      /* to indicate when messages can be sent to java */				   
extern Box std_real_box;	/* default real world window */
extern int iterates;		/* number Newton iterates in pack computations */
extern unsigned long int rand_next; /* random number generator variable */
extern int num_plot;		/* num plot points for arcs on sphere */
extern double degPI;           /* set to 180.0/M_PI */

extern struct Vertlist *Vlist;   /* global utility vert list */
extern struct Vertlist *Flist;   /* global utility face list */
extern struct Edgelist *Elist;   /* global utility edge list */
extern struct Edgelist *Vertex_map;/* global utility vertex_map */
extern struct Centlist *Clist;   /* globel utility center list */

extern struct Pathlist *pathlist,*newpath; 	
			  /* specify "cookie cutting" paths. */
extern int pathlength;		  /* number of pts in stored pathlist */
extern CPRegion region;            /* for XLib region info */

extern int **face_org;           /* For short-lived data on face indices
			     needed by several routines. */
extern double layout_th;         /* layout threshold: tolerance for placement
			     errors in pack centers. */

extern struct Vertlist *Node_link_parse(),*Face_link_parse();
extern struct Edgelist *Node_pair_link();

/* compression variables */
extern int useCompression;
extern int compressAutoDetect;

#endif
