/* matt_cathey.c 

Code to convert files of H's and V's into Brooks packing (with plug).
Call with

   matt_cathey [BA] {filename}

There are two modes of operation indicated by optional flag: 
  (B) insert new circle in location as in Brooks' original scheme; 
  (A) insert new circle in alternating fashion: verticals alternate
      left/right, horizontals, up/down.
{filename} contains H's and V's

Compile (in linux) with

cc -g -o brooks matt_cathey.c

*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#define MAX_LIST 4000   /* limit on list of H's and V's */

int P,N,*FL[MAX_LIST+6],NUM[MAX_LIST+6];
FILE *fp;

int flower_insert(int v,int flag)
     /* insert new P in flower v before (flag=1) or 
	after (flag=-1) N */
{
  int i,j,done=0,*newflower;

  newflower=(int *)calloc(NUM[v]+2,sizeof(int));
  for (i=0;i<=NUM[v];i++)
    newflower[i]=FL[v][i];
  if (flag==1) /* insert P before N */
    {
      for (i=0;(i<=NUM[v]) && !done;i++)
	{
	  if (FL[v][i]==N)
	    {
	      for (j=i;j<=NUM[v];j++)
		newflower[j+1]=FL[v][j];
	      newflower[i]=P;
	      done=1;
	    }
	}
      if (!done) return 0; /* error: should have found M */
    }
  else /* insert P after N */
    {
      for (i=0;(i<=NUM[v]) && !done;i++)
	{
	  if (FL[v][i]==N)
	    {
	      for (j=i+1;j<=NUM[v];j++)
		newflower[j+1]=FL[v][j];
	      newflower[i+1]=P;
	      done=1;
	    }
	}
      if (!done) return 0; /* error: should have found M */
    }
  NUM[v]++;
  free(FL[v]);
  FL[v]=newflower;
  return 1;
} /* flower_insert */

int main(int argc,char *argv[])
{
  int T,R,B,L,D;
  int v_flips=1,h_flips=0,h_mode=0,v_mode=0;
  int i,j,k,*newfl,mode=1;
  char *c,*s,filename[256];

  /* Get file of input sequence: "H", "V" (horizontal/vertical).
     Conclude with "END" */

  if (argc<2)
    {
      printf("Usage: BrooksHV {packname.p}.\n");
      exit(1);
    }

  if (argc==3)
    {
      if (!strcmp(argv[1],"B")) /* (default) use Brooks' original 
				   insertion spots */
	mode=1; 
      if (!strcmp(argv[1],"A")) /* use "alternating" insertion spots */
	mode=2;
    }
  
  strcpy(filename,argv[argc-1]);
  if (!(fp=fopen(filename,"r")))
    {
      printf("Failed to open %s.",filename);
      exit(1);
    }

  /* initialize the variables we need to track: T,R,B,L are the
     circles forming the current quadrilateral. When in mode 2,
     we interchange T/B, L/R each time we add a new vertex. 
     N is the number of the old plug (now to become the new
     vertex), P is incremented for the new plug. */

  T=1;L=2;B=3;R=4;
  N=P=5;

  /* create flowers */

  for (i=1;i<=4;i++)
    {
      FL[i]=(int *)calloc(3,sizeof(int));
      NUM[i]=2;
    }

  FL[P]=(int *)calloc(5,sizeof(int));
  NUM[P]=4;

   /* flowers */
  FL[T][0]=L;
  FL[T][1]=P;
  FL[T][2]=R;
   
  FL[L][0]=B;
  FL[L][1]=P;
  FL[L][2]=T;

  FL[B][0]=R;
  FL[B][1]=P;
  FL[B][2]=L;

  FL[R][0]=T;
  FL[R][1]=P;
  FL[R][2]=B;
  
  FL[P][0]=T;
  FL[P][1]=L;
  FL[P][2]=B;
  FL[P][3]=R;
  FL[P][4]=T;

  T=3;B=1;L=4;R=2; /* prepare for first flip */
   
  /* Cycle through reading H's and V's from file; stop at end
     of file, encountering 'END', or reaching MAX_LIST cap. */

  k=0;
  s=(char *)calloc(2,sizeof(char));
  while (k<MAX_LIST && (fscanf(fp,"%1s",s)) && *s!='E')
    {
      P++; 
      if (*s=='V') /* next circle is 'vertical' */
	{
	  if (!v_mode) v_flips=0;
	  if (mode==2 && (h_flips%2)==0)
		/* Flip (exchange labels) each V. However, for first 
		   V in chain, may have to compensate for flips done
		   while in previous horizontal chain. */
	    {
	      D=T;T=B;B=D; /* exchange T and B */
	      D=L;L=R;R=D; /* exchange L and R */
	      v_flips++;
	    }
	  h_mode=0;
	  v_mode=1;
	  flower_insert(T,-1);
	  flower_insert(B,1);
	  
	  /* fix flower of R */
	  for (i=0;i<=NUM[R];i++)
	    if (FL[R][i]==N) FL[R][i]=P;
	  
	  /* fix flower of N */
	  FL[N][0]=T;
	  FL[N][1]=L;
	  FL[N][2]=B;
	  FL[N][3]=P;
	  FL[N][4]=T;
	  
	  /* add flower of new temporary 'plug' vert */
	  newfl=(int *)calloc(5,sizeof(int));
	  FL[P]=newfl;
	  NUM[P]=4;
	  FL[P][0]=N;
	  FL[P][1]=B;
	  FL[P][2]=R;
	  FL[P][3]=T;
	  FL[P][4]=N;
	  
	  L=N; /* N is new 'left' of interstice */
	  
	  printf("V ");
	  
	}
      else if (*s=='H') /* next circle is 'horizontal' */
	{
	  if (!h_mode) h_flips=0;
	  if (mode==2 && (v_flips%2)==0)
		/* Flip (exchange labels) each H. However, for first 
		   H in chain, may have to compensate for flips done
		   while in  chain. */
	    {
	      D=T;T=B;B=D; /* exchange T and B */
	      D=L;L=R;R=D; /* exchange L and R */
	      h_flips++;
	    }
	  v_mode=0;
	  h_mode=1;
	  flower_insert(R,-1);
	  flower_insert(L,1);
	  
	  /* fix flower of 3 */
	  for (i=0;i<=NUM[B];i++)
	    if (FL[B][i]==N) FL[B][i]=P;
	  
	  /* fix flower of N */
	  FL[N][0]=T;
	  FL[N][1]=L;
	  FL[N][2]=P;
	  FL[N][3]=R;
	  FL[N][4]=T;
	  
	  /* add flower of new P vert */
	  newfl=(int *)calloc(5,sizeof(int));
	  FL[P]=newfl;
	  NUM[P]=4;
	  FL[P][0]=N;
	  FL[P][1]=L;
	  FL[P][2]=B;
	  FL[P][3]=R;
	  FL[P][4]=N;
	  
	  T=N; /* N is new 'top' of interstice */
	  
	  printf("H ");
	  
	}
      else
	{
	  printf("Exited with unrecognized file entry\n");
	  exit(5);
	}
      
      k++;
      N++;
    } /* end of while, mode==1 */
   printf("\n");
   fclose(fp);

   /* write out the new packing */

   strcat(filename,".p\0");
   if (!(fp=fopen(filename,"w")))
     {
       printf("Error writing output to %s\n",filename);
       exit(10);
     }
   fprintf(fp,"NODECOUNT: %d\n",P);
   fprintf(fp,"GEOMETRY: euclidean\n");
   fprintf(fp,"ALPHA/BETA/GAMMA: %d %d %d\n",5,2,1);
   fprintf(fp,"FLOWERS:\n");
   for (i=1;i<=P;i++)
     {
       fprintf(fp,"%d  %d      ",i,NUM[i]);
       for (j=0;j<=NUM[i];j++)
	 fprintf(fp,"%d ",FL[i][j]);
       fprintf(fp,"\n");
     }
   fprintf(fp,"\nEND\n");
   fclose(fp);
   printf("Wrote packing to %s; nodecount=%d.\n",filename,P);

  exit(0);
} /* main */

/* ================================================================= */

  






