/*
 * CPdtos.c:  provides wrapper functions for the qtoa and atodd functions
 * provided in the doubledouble library.  These functions just call
 * printf/scanf if doubledouble is not being used.
 */

#include "sa_types.h"
#include "sa_proto.h"

#define MAX_FORMAT 256

char *CPdtos(double d, char *buf, int prec, int fmtch)

{
#ifdef CP_USE_DD
    if(isnan(d.h()) || isnan(d.l()))  return strcpy(buf, "NaN");
    else return qtoa(buf, prec, fmtch, d);
#else
    char fstr[MAX_FORMAT];
    int len;

    fstr[0] = '%';
    fstr[1] = '.';
    sprintf(fstr + 2, "%d", MAX_ACCUR);
    len = strlen(fstr);
    fstr[len] = 'l';
    fstr[len + 1] = fmtch;
    fstr[len + 2] = '\0';
    sprintf(buf, fstr, d);
    return buf;
#endif
} /* CPdtos */

double CPstod(char *buf)

{
#ifdef CP_USE_DD
    return atodd(buf);
#else
    return atof(buf);
#endif
} /* CPstod */

int CPgetd(FILE *stream, double *d)

{
    singledouble s;
    int len, used;
    int c;
    char buf[BUFSIZE];
    char tmpbuf[BUFSIZE];

    buf[0] = '\0';
    do  {
        c = getc(stream);
        if(c == EOF)  return -1;
        len = strlen(buf);
        len++;
        buf[len - 1] = c;
        buf[len] = '\0';
        if(len == 1 && (buf[0] == '-' || buf[0] == '+' || buf[0] == '.'))  {
            used = len;
        }
        else if(len == 2 && (buf[0] == '-' || buf[0] == '+') && buf[1] == '.')  {
            used = len;
        }
        else if(len == 1 && isspace(buf[0]))  {
            buf[0] = '\0';
            used = len;
        } 
        else if(len >= 2 && buf[len - 1] == 'e')  {
            strncpy(tmpbuf, buf, len - 1);
            tmpbuf[len - 1] = '\0';
            if(sscanf(tmpbuf, "%lf", &s) < 1)  return -2;
            used = len;
        }
        else if(len >= 3 && (buf[len - 1] == '+' || buf[len - 1] == '-') && buf[len - 2] == 'e')  {
            strncpy(tmpbuf, buf, len - 2);
            tmpbuf[len - 2] = '\0';
            if(sscanf(tmpbuf, "%lf", &s) < 1)  return -3;
            used = len;
        }
        else if(sscanf(buf, "%lf%n", &s, &used) < 1)  {
            return -4;
        }
    }  while(len < BUFSIZE - 1 && used == len);

    ungetc(c, stream);
    len--;
    buf[len] = '\0';

    *d = CPstod(buf);
    return 0;
} /* CPgetd */

int CPputd(FILE *stream, double d, int prec, int fmtch)

{
    char buf[BUFSIZE];

    fputs(CPdtos(d, buf, prec, fmtch), stream);

    return 0;
} /* CPputd */
