#include "cp_types.h"
#include "cp_graphics_proto.h"
#include "cp_proto.h"
#include "cp_head.h"

/* Read a *.vl file of a parent packing and light subpackings.
   Create packings from the light subpackings and put them in
   one big file. Some light packings may be disconnected and 
   will result in more than one packing. */

char *emsgbuf,*msgbuf;

int main(int argc,char *argv[])
{
  int n,v,N,j,w,i,count=0,overlap,next_start=1;
  int *util=NULL,*newold;
  struct p_data *p,*P;
  struct K_data *pK_ptr;
  struct p_light **plight;
  struct Edgelist *trace;
  char cmd[BUFSIZE],filename[BUFSIZE];
  FILE *fpp;

  set_CPack_globals(); /* set up library global variables */

  /* ----------- user code starts here */

  P=init_packing();

  printf("Display_decomp: for reading a *.vl file which gives "
	 "a packing and lists of vertices associated with light "
	 "subpackings and showing postsript images."
	 "\n  Enter a *.vl filename:\n");
  fgets(cmd,sizeof(cmd),stdin);
  if (!sscanf(cmd,"%s",filename)
      || !(N=read_parent_child(filename,P,&plight, &overlap))
      || !(p=init_packing()))
    {
      printf("Failed to open %s.",filename);
      exit(1);
    }

  strcpy(cmd,filename);
  strcat(cmd,".cmd");
  if (!(fpp=fopen(cmd,"w")))
    {
      printf("Failed to open output file %s or to write "
	     "parent packing to it.",cmd);
      exit(2);
    }
  fprintf(fpp,"Created by 'Expand' from file %s.\n\n",filename);

  /* put in some script commands */
  fprintf(fpp,"\t[]:=act 0;infile_read %s;disp -w -c;\n\n\n",filename);

  fprintf(fpp,"%s\n\n",filename);
  writepack(fpp,P,0017,0);

  for (n=1;n<=N;n++)
    {
/* copy parent to p */      
      if (!copy_pack(P,p))
	{
	  printf("Didn't copy correctly, n=%d.",n);
	  exit(3);
	}
      pK_ptr=p->packK_ptr;
      if (util) free(util);
      util=(int *)calloc((p->nodecount+1),sizeof(int));
      for(j=1;j<=p->nodecount;j++) pK_ptr[j].mark=0;

/* mark light packing variable vertices in p in both 'mark' and
   'util' spots. */

      for(j=1;j<=plight[n]->counts[2];j++) /* variable verts */
	pK_ptr[(v=plight[n]->orig_indices[j])].mark=util[v]=1;

      while (next_start<=p->nodecount && !util[next_start]) next_start++;

      while (next_start<=p->nodecount) /* have to cut out another packing */
	{
	  sprintf(cmd,"-v %d mc",next_start);
	  if (!cookie_cutter(p,cmd,&newold))
	    {
	      printf("cookie cutter failed, n = %d.",n);
	    }
	  /* complex_count, facedraw order should be done */ 
	  comp_pack_centers(p,0,0,2,okerr); /* opt 2, averaging */

	  /* store vertex-map and reset flags in util. */
	  pK_ptr=p->packK_ptr;
	  if (newold) /* set up vertex_map */
	    {
	      p->vertex_map=trace=(struct Edgelist *)
		calloc((size_t)1,sizeof(struct Edgelist));
	      for (j=1;j<=p->nodecount;j++)
		if ((w=newold[j]))
		  {
		    trace=trace->next=(struct Edgelist *)
		      calloc((size_t)1,sizeof(struct Edgelist));
		    trace->v=j;
		    trace->w=w;
		    util[w]=0; /* got this one */
		  }
	      trace=p->vertex_map;
	      p->vertex_map=p->vertex_map->next;
	      free(trace);
	    }
	  count++;
	  fprintf(fpp,"\n\nchild.%d.p\n\n",count);
	  writepack(fpp,p,0133,0); /* comb, geom, radii, centers, vert map */
	  next_start=1;
	  while (next_start<=p->nodecount && !util[next_start]) next_start++;
	  if (next_start<=p->nodecount)
	    {
	      copy_pack(P,p);
	      pK_ptr=p->packK_ptr;
	      for (i=1;i<=p->nodecount;i++)
		pK_ptr[i].mark=util[i];
	    }
	} /* end of while */
    } /* end of for loop */

  fclose(fpp);
  printf("\nFinished. File '%s.cmd' contains %d subpackings from the given "
	 "%d light packings.",filename,count,N);

  /* ----------- user code ends here */
  exit(0);
} /* main */


/* ====== communication routines; user may want to modify ====== */

int emsg()
{
  printf(emsgbuf);return 1;
} /* emsg */

int msg()
{
  printf(msgbuf);return 1;
} /* msg */

int repack_activity_msg(char *datastr)
     /* report progress of repacking */
{
  strcpy(msgbuf,datastr);
  msg();
  return 1;
} /* repack_activity_msg */

int refresh_canvas(struct s_data *q)
     /* some drawing routines want to refresh */
{return 1;} /* refresh_canvas */

/* ================================================================= */

