#include "sa_types.h"
#include "sa_proto.h"

/* JamesAshe.c, 4/2003

Idea is to get data on holonomy translation lengths in euclidean
packings with fractional branching. A packing is read in along with
a closed flist (created in CirclePack); An input file and a matlab output 
file are opened. The input file contains, on each line,

N  n1 a1 n2 a2 ... nN aN

where N is the number of pairs on the line and each pair nj aj gives a 
vertex and the extra "aim" -- thus, set aim of vert nj to
(2+aj)*pi. Then repack, fix, normalize, compute info needed, write
to output file.

Compile with:
   g++ -g -Wall -o JA JamesAshe.c -I$HOME/CP/include -L$HOME/CP/lib  -L/usr/X11R6/lib -lCP_libs -lX11 -lm

Run with:
   JA <packing.p> <inputfile> <output.m>
*/

char *emsgbuf,*msgbuf;
int repack_activity_msg(),refresh_canvas();

int main(int argc,char *argv[])
{
  int N,i,v;
  double inc;
  char bff[BUFSIZE],filename[BUFSIZE],outfile[BUFSIZE],infile[BUFSIZE];
  struct p_data *p;
  Mobius mob;
  FILE *fp=NULL,*infp=NULL,*outfp=NULL;

  set_CPack_globals();

  /* ----------- user code starts here */

  if (argc<4)
    {
      printf("Usage: JA <packname.p> <inputfile> <output.m>.\n");
      exit(1);
    }

  /* read the packing */
  if (!strcpy(filename,argv[1])
      || !(p=init_packing())
      || !(fp=fopen(filename,"r"))
      || !readpack(fp,p) 
      || p->hes)
    {
      if (fp) fclose(fp);
      printf("  Failed to get the packing %s or packing is "
	     "not euclidean.\n",filename);
      exit(1);
    } 
  fclose(fp);
  complex_count(p,0);
  facedraworder(p,0);
  fp=NULL;

  /* open the input file */
  if (!strcpy(infile,argv[2])
      || !(infp=fopen(infile,"r"))) {
    if (infp) fclose(infp);
    printf("  Failed to get input file %s.\n",infile);
    exit(2);
  }

  /* open output file */
  if (!strcpy(outfile,argv[3])
      || !(outfp=fopen(outfile,"w"))) {
    if (outfp) {fclose(outfp);outfp=NULL;}
    printf("  Failed to open output file %s.\n",outfile);
    exit(3);
  }
  /* initialize output file for matlab */

  fprintf(outfp,"%% matlab file from JamesAshe: input data %s.\n\n"
	  "M=[\n",outfile);
  
  /* read in line */
  while (fscanf(infp,"%d",&N)==1 && N>0) {
    set_aim_default(p);     /* set aims to default */
    for (i=1;i<=N;i++) { /* set fractional branching data */
      if (fscanf(infp,"%d %lf",&v,&inc)<2) {
	printf(" Failed to get expected data.");
	goto BOMB;
      }
      p->packR_ptr[v].aim += inc*M_PI;
    }
    for (i=1;i<=p->nodecount;i++) p->packR_ptr[i].rad=1.0;
    fillcurves(p);
    e_riffle(p,2000); /* repack */
    /* have to normalize, location/orientation of first face, size
       of some circle, or area? */
    fix_call(p,"");
    holonomy_trace(p,"flist",&mob);
    sprintf(bff," :: Mob ::  :: ");
    output_parse(p,outfp,bff,mob);
    fprintf(outfp,"\n");
  } /* end of while */
  fclose(outfp);

 BOMB:
  if (fp) fclose(fp);
  if (infp) fclose(infp);
  if (outfp){
    fprintf(outfp," ]\n"); /* close matlab output */
    fclose(outfp);
  }
  exit(0);
} /* main */

/* ====== library requires thes communication routines; 
user may want to modify ====== */

int emsg()
{
  printf(emsgbuf);return 1;
} /* emsg */

int msg()
{
  printf(msgbuf);return 1;
} /* msg */

int repack_activity_msg(char *datastr)
     /* report progress of repacking */
{
  strcpy(msgbuf,datastr);
  msg();
  return 1;
} /* repack_activity_msg */

int refresh_canvas(struct s_data *q)
     /* some drawing routines want to refresh */
{return 1;} /* refresh_canvas */

/* ================================================================= */



