#include "sa_head.h"
#include "Interface.h"

/**
 * NSpole is for performing various normalizations of packings on
 * the sphere. Calls are:
 *   NSpole N S [E] [t]
 * and for combinatorial 'antipodal' option
 *   NSpole -a N [t]
 */
int NSpole(struct p_data *p,char *datastr)
{
  int N,S,E=0,errflag=0,i,*genlist;
  double factor=1.0,Erad=0.0;
  complex Ectr;
  char *nextpoint;
  Mobius Mob;
  struct R_data *pR_ptr;

  if (p->hes<=0) return 0;
  pR_ptr=p->packR_ptr;
  Ectr.re=Ectr.im=0.0;
  nextpoint=datastr;
  stripsp(nextpoint);
  if (!strncmp(nextpoint,"-a",2)) { // use north pole and antipodal
  	grab_next(&nextpoint,next); // flush the '-a', must get N
  	if (!(N=grab_one_vert(p,&nextpoint))) return 0;
  	// find antipodal S
    for (i=1;i<=p->nodecount;i++) p->packK_ptr[i].util_flag=0;
    p->packK_ptr[N].util_flag=1;
    if (!(genlist=label_generations(p,-1,&S,&i)))
      return 0;
    free(genlist);
  }
  else { // standard N S
  	if (!(N=grab_one_vert(p,&nextpoint))
      || !(S=grab_one_vert(p,&nextpoint)) || N==S)
    	return 0;
  }

  // if next isn't flag, it must be E option
  if (grab_next(&nextpoint,next) && next[0]!='-'
   	&& (E=grab_one_vert(p,&nextpoint)) && E!=N && E!=S) {
	Ectr=pR_ptr[E].center;
	Erad=pR_ptr[E].rad;
  }
  // otherwise, look for -t option and factor
  else if (next[0]=='-' && next[1]=='t'
	   && sscanf(nextpoint,"%lf",&factor))
    {Ectr.re=Ectr.im=0.0;}
    
  // Now, find Mobius
  NS_mobius(pR_ptr[N].center,pR_ptr[S].center,Ectr,
		pR_ptr[N].rad,pR_ptr[S].rad,Erad,factor,&errflag,&Mob);
  if (!errflag) {
      apply_Mobius(p,"a",1,Mob);
      return 1;
  }
  else return 0;
} /* NSpole */

