#include "sa_types.h"
#include "sa_proto.h"

#define FRAC 50

/* Plot_cir_images.c 
This is designed to read a packing and to display a postscript
image of its circles after a specified function has been applied to them.
The original purpose was to show circles in an annulus using its 
intrinsic metric. */

char *emsgbuf,*msgbuf;
int repack_activity_msg(),refresh_canvas();
struct p_data *p;
Mobius Mob; /* global declaration needed by set_CPack_globals */
FILE *sysfp;

int main(int argc,char *argv[])
{
  int i,n;
  double c[181],s[181],rad,x,y,sc;
  complex u,v,w,z,cent;
  complex ONE={1,0.0};
  char filename[NAME_MAX],psname[NAME_MAX];

  set_CPack_globals();

  /* ----------- user code starts here */

  sprintf(msgbuf,"Plot_cir reads a circle packing and puts "
	  "out a Postscript image of them under a function hard coded "
	  "into the C code.");
  msg();

  if (argc<2) {
    sprintf(msgbuf,"Usage: Plot_cir {filename}.");
    emsg();
    exit(0);
  }
  filename[0]='\0';
  strncpy(filename,argv[1],255);

  /* read packing we are going to draw */

  p=(struct p_data *)malloc(sizeof(struct p_data));
  if ((fp=fopen(filename,"r"))) {
    p->sizelimit=r_get_pack_size(fp,p);
    fclose(fp);
  }
  if (!p->sizelimit || !(fp=fopen(filename, "r"))
      || (p->packK_ptr=
	  (struct K_data *)malloc((p->sizelimit+1)
				  *sizeof(struct K_data)))==NULL
      || (p->packR_ptr=
	  (struct R_data *)malloc((p->sizelimit+1)
				  *sizeof(struct R_data)))==NULL
      || !readpack(fp,p) ) {
    sprintf(msgbuf,"Plot_cir_images: read of packinge %s has failed.",
	    filename);
    emsg();
    if (fp) fclose(fp);
    exit(2);
  }
  if (fp) fclose(fp);

  /* open PostScript */

  strcpy(psname,filename);
  strcat(psname,".ps");
  fp=fopen(psname,"w");
  
  fprintf(fp,"%%!PS-Adobe-2.0 EPSF-2.0\n%%%%Title: %s\n",p->file_name);
  fprintf(fp,"%%%%Orientation: Portrait\n");
  post_size_settings(fp,p,1);
  fprintf(fp,"%%%%Pages: 1\n%%%%BeginSetup\n%%%%EndSetup\n%%%%"
	  "Magnification: 1.0000\n%%%%EndComments\n");
  fprintf(fp,"\n%% CirclePack dictionary ================\n");
  
  /* -------- define CPdict ------------------------------------------ */
  fprintf(fp,"/CPdict 256 dict def\nCPdict begin");
  
  fprintf(fp,"\n%% --- Standard abbreviations\n");
  fprintf(fp,"\t/cp {closepath} bind def\n\t/ef {eofill} bind "
	  "def\n\t/gr {grestore} bind def\n\t/gs {gsave} bind "
	  "def\n\t/sa {save} bind def\n\t/rs {restore} bind "
	  "def\n\t/l {lineto} bind def\n\t/m {moveto} bind "
	  "def\n\t/rm {rmoveto} bind def\n");
  fprintf(fp,"\t/n {newpath} bind def\n\t/s {stroke} bind "
	  "def\n\t/sh {show} bind def\n\t/slc {setlinecap} bind "
	  "def\n\t/slj {setlinejoin} bind def\n\t/slw "
	  "{setlinewidth} bind def\n\t/srgb {setrgbcolor} bind "
	  "def\n\t/rot {rotate} bind def\n\t/sc {scale} bind def\n");
  fprintf(fp,"\t/sd {setdash} bind def\n\t/ff {findfont} bind "
	  "def\n\t/sf {setfont} bind def\n\t/scf {scalefont} bind "
	  "def\n\t/sw {stringwidth} bind def\n\t/tr {translate} "
	  "bind def\n");
  /* -------- Unique to CirclePack abbreviations, etc. -------------- */
  fprintf(fp,"\n%% --- Special abbreviations\n");
  fprintf(fp,"   /sg {setgray} bind def\n   /a {arc} bind "
	  "def\n   /an {arcn} bind def\n");
  fprintf(fp,"   /c { 0 360 a s} bind def\t\t\t%% circle\n");
  fprintf(fp,"   /cc {0 360 a gs srgb s gr} bind def\t\t%% color circle\n");
  fprintf(fp,"   /d { 0 360 a gs sg fill gr s} bind def\t%% disc\n");
  fprintf(fp,"   /cd {0 360 a gs srgb fill gr s} bind def\t%% color disc\n");
  fprintf(fp,"   /mark {gs 0 360 a srgb fill gr} bind "
	  "def\t%%default mark symbol\n");
  fprintf(fp,"   /wht {1.0} bind def\t\t\t\t%% gray levels\n   "
	  "/gry {0.8} bind def\n  /drk {0.5} bind def\n   /blck "
	  "{0.0} bind def\n");
  fprintf(fp,"   /ourlinewidth {.004 mul setlinewidth} bind "
	  "def\n");
  fprintf(fp,"\nend\n%% end CirclePack dictionary "
	  "=================\n%%%%EndProlog\n"
	  "%%%%Page:1 1\nCPdict begin\n");
  fprintf(fp,"gsave\n  72 72 sc %% inches\n");
  fprintf(fp,"  4.250000 5.500000 tr\n  1 slc 1 slj\n  1 ourlinewidth");

  /* set up the x, y coords */

  double tinc=M_PI/FRAC;
  for (i=0;i<=2*FRAC;i++) {
    c[i]=cos(i*tinc);s[i]=sin(i*tinc);
  }

  /* cycle through circles */

  double h=.9995;
      

  for (n=1;n<=p->nodecount;n++) {
    rad=p->packR_ptr[n].rad;
    cent=p->packR_ptr[n].center;

    fprintf(fp,"\n %%circle %d\n gs n\n",n);
    for (i=1;i<=2*FRAC;i++) {
      z.re=cent.re+rad*c[i];
      z.im=cent.im+rad*s[i];

/* =====================  formula for w = f(z) ==================== */

/* 1. This function maps the unit disc conformally onto annulus; the
   parameter h has to do with the modulus of the annulus. */

/*      sc=2.0*M_PI/log((1+h)/(1-h)); 
      u=cLog(cdiv(cadd(ONE,z),csub(ONE,z)));
      v.re=(-1.0)*sc*u.im;
      v.im=sc*u.re;
      w=cExp(exp(v.re),v.im);
*/

/* 2. This function is merely the square function */

/*    w=cmult(z,z);*/

/* 3. Simple exponential. */

      w=cExp(exp(z.re),z.im);


      if (i==1) fprintf(fp,"%f %f m\n",w.re,w.im);
      else fprintf(fp,"%f %f l\n",w.re,w.im);
    }
    fprintf(fp,"closepath s\ngr\n");
  }

  /* close out and display the postscript file */

  fprintf(fp,"\nend\nshowpage\n");
  fclose(fp);
  sysfp=popen("/bin/sh","w");
  fprintf(sysfp," ghostview %s &\n",psname);
  pclose(sysfp);
  /* ----------- user code ends here */
  sprintf(msgbuf,"Plot_cir: PostScript output is in %s.",filename);
  msg();
  exit(0);
} /* main */


/* ====== communication routines; user may want to modify ====== */

int emsg()
{
  printf(emsgbuf);return 1;
} /* emsg */

int msg()
{
  printf(msgbuf);return 1;
} /* msg */

int repack_activity_msg(char *datastr)
     /* report progress of repacking */
{
  strcpy(msgbuf,datastr);
  msg();
  return 1;
} /* repack_activity_msg */

int refresh_canvas(struct s_data *q)
     /* some drawing routines want to refresh */
{return 1;} /* refresh_canvas */

/* ================================================================= */

