#include "cp_types.h"
#include "cp_graphics_proto.h"
#include "cp_proto.h"

/* Adjust a redchain to accommodate hints from a face list. 
ops=1: When suggested face list is used to create the redchain,
some of its blue faces will likely we missed, and this will
try to put them in. 

fixup: ops=2: wish to adjust a segment of redchain to change the 
fundamental domain. 

Return 0 on error. */

int adjust_redchain(struct p_data *p, struct Vertlist *facelist,
		    int ops)
{
  int wflag=0;
  struct Vertlist *ftrace,*hold,*endtrace;
  struct RedList *redlist,*rtrace,*ttrace;
  
  if (!facelist || !p->redfaces) return 0;
  if (ops==1) /* (only choice currently) assume facelist was prescribed
		 when redchain was created; fix up any blue faces that
		 were dropped. */
    {
      ftrace=hold=etrace=facelist;
      while (endtrace->next) endtrace=endtrace->next; /* ptr to end */
      /* must be a closed facelist for the option */
      if (facelist->v != endtrace->v) return 0; 

      /* create copy of redchain so original can be reinstated on
	 failure */
      redlist=ttrace=(struct Vertlist *)
	calloc((size_t)1,sizeof(struct Vertlist));
      ttrace->face=p->redfaces->face;
      rtrace=p->redfaces->next;
      while (rtrace!=p->redfaces)
	{
	  ttrace->next=(struct Vertlist *)
	    calloc((size_t)1,sizeof(struct Vertlist));
	  ttrace->next->prev=ttrace;
	  ttrace=ttrace->next;
	  ttrace->face=rtrace->face;
	  rtrace=rtrace->next;
	}
      ttrace->next=redlist;
      redlist->prev=ttrace;
      
      /* may need to adjust facelist; if first face not in redlist
	 or second face is blue (i.e. third = first), then shift 
	 facelist. */
      while ((ftrace!=hold || !(wflag++)) 
	     && ftrace->next && ftrace->next->next 
	     && ((ftrace->v == ftrace->next->next->v)
		 || !position_redlist(&redlist,ftrace->v)))
	{
	  endtrace=endtrace->next=(struct Vertlist *)
	    calloc((size_t)1,sizeof(struct Vertlist));
	  endtrace->v=ftrace->next->v;
	  hold=ftrace;
	  ftrace=ftrace->next;
	  free(hold);
	}
      if (ftrace==hold
	  || !position_redlist(&redlist,ftrace->v)) 
	/* didn't find a good face to start search or it wasn't in
	   redchain. */
	{
	  free_redfaces(&redlist);
	  return 0;
	}

      /* Match first/second faces of facelist to first/second face in
	 redchain; may need to reposition redchain, but check to prevent 
	 infinite loop. */
      
      rtrace=rhold=redlist;
      while (rtrace->next->face!=ftrace->next->v
	     || (!position_redlist(&rtrace,ftrace->v)
		 || (rtrace==rhold)));
      if (rtrace->next->face!=ftrace->next->v) 
	{
	  free_redfaces(&redlist);
	  return 0;
	}
      redlist=rhold=rtrace;
      
      /* travel through redchain finding any faces of facelist missing. */
      rtrace=redlist;
      hold=ftrace;
      while (hold)
	{
	  while (hold && (rtrace->face == hold->v))
	    {
	      rtrace=rtrace->next;
	      hold=hold->next;
	    }
	  if (hold && hold->next && !p->faces[hold->next->v].next_red
	      && (hold->next->v==rtrace->next->face)) 
	    /* not at end, so must have found mismatch; insert new
	       red face (if not already red). */
	    {
	      new=(struct RedList *)
		calloc((size_t)1,sizeof(struct RedList));
	      new->face=hold->v;
	      rtrace->next->prev=new;
	      new->next=rtrace->next;
	      new->prev=rtrace;
	      rtrace->next=new;
	    }
	  if (hold) hold=hold->next;
	}

      /* adjust redchain, adjust next_red's, redo edge-pairing */
      if (!catalog_faces(p))
	{
	  free_redfaces(&redlist);
	  return 0;
	}

      /* fixup: lost! don't know how to proceed because the
	 redface stuff is so involved and intermingles creation
	 of redchain and drawing order. */

      redface_comb_info(p,redlist,p->edge_pair);

    }
} /* adjust_redchain */
