#include "sa_types.h"
#include "sa_proto.h"

/* Convert data from an existing p_light structure (to the extent 
   possible) and put in existing packing p. 

   If flag is set, data is for the existing packing (only verify 
   that CHECK_COUNT and hes from p_light correspond to p); use 
   'orig_indices' of p_light pack to send data to the right spots in p. 

   Otherwise, clean out p and create a new packing from p_light data. 
   Create the vertex_map of p to store the original vertex numbers.
   (I don't know what will happen if the combinatorics of p_light are 
   too strange.) 

   Return 0 on error and flush out p or return pl vertcount. */

int try_contig_pair(struct p_data *p,int w,int vi,int vj);
int add_to_flower(struct p_data *p,int w, int u);
int *new_flower,*bit_flower,eptr;

int convert_from_p_light(struct p_data *p,struct p_light *pl,int flag)
{
  int i,j,k,n,u,v,w,m,indx,*iptr,bdry_count,var_count,num,new_indx;
  struct Edgelist *vtrace=NULL;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;

  if (flag) /* does this match the size and geometry of p? */
    {
      pR_ptr=p->packR_ptr;
      if (!p->status || pl->counts[4]==0 
	  || pl->counts[4]!=p->nodecount
	  || ((pl->counts[1]>0 && p->hes<=0)
	      || (pl->counts[1]==0 && p->hes!=0)
	      || (pl->counts[1]>0 && p->hes<=0)))
	return 0;
      /* read in new radii */
      for (i=1;i<=pl->counts[0];i++)
	pR_ptr[pl->orig_indices[i]].rad=pl->radii[i];
      if (pl->counts[5]>0) /* also have some aims specified in pl */
	{
	  for (i=1;i<=pl->counts[5];i++)
	    p->packR_ptr[pl->orig_indices[pl->aim_indices[i]]].aim
	      =pl->aims[i];
	}
      return pl->counts[0];
    }
  
  /* need to clear packing and enter new data */
  if (!p || !alloc_pack_space(p,pl->counts[0]+1,0))
    return 0;
  pK_ptr=p->packK_ptr;
  pR_ptr=p->packR_ptr;
  p->nodecount=pl->counts[0];
  var_count=pl->counts[2];
  bdry_count=pl->counts[0]-var_count;

  /* build flowers of 'var' verts first (these are given
     explicitly in p_light 'var_indices' list. */
  p->vertex_map=vtrace=(struct Edgelist *)
    calloc((size_t)1,sizeof(struct Edgelist));
  iptr=pl->var_indices+1;
  for (i=1;i<=var_count;i++){
    pK_ptr[i].num=*(iptr++);
    pK_ptr[i].flower=(int *)calloc((size_t)(pK_ptr[i].num+1),sizeof(int));
    for (j=0;j<=pK_ptr[i].num;j++)
      pK_ptr[i].flower[j]=*(iptr++); 
    if (pK_ptr[i].flower[0]=pK_ptr[i].flower[pK_ptr[i].num])
      pK_ptr[i].bdry_flag=0;
    else pK_ptr[i].bdry_flag=1;
    pR_ptr[i].rad=pl->radii[i];
    vtrace=vtrace->next=(struct Edgelist *)
      calloc((size_t)1,sizeof(struct Edgelist));
    vtrace->v=i;
    vtrace->w=pl->orig_indices[i];
  }
  if (!bdry_count) /* done: there are no 'fixed' verts */
    goto GOBACK;

  /* flowers for 'fixed' verts have to be reconstructed. First,
     we count how often each occurs in variable flowers */

  for (i=var_count+1;i<=p->nodecount;i++){
    pK_ptr[i].num=0;
    pK_ptr[i].bdry_flag=1; /* adjust later if turns out to be interior */
  }
  for (k=1;k<=var_count;k++){
    for (j=0;j<pK_ptr[k].num;j++)
      if ((m=pK_ptr[k].flower[j])>var_count) /* petal is 'fixed' */
	pK_ptr[m].num++;
    if ((m=pK_ptr[k].flower[pK_ptr[k].num])>var_count
	&& m!=pK_ptr[k].flower[0]) /* last petal already counted? */
      pK_ptr[m].num++;
  }
  /* might as well toss index of any 'fixed' having no 'var' neighbors. */
  num=0;
  for (i=var_count+1;i<=p->nodecount;i++){
    new_indx=i-num; /* i adjusted for previous removals */
    if (!pK_ptr[new_indx].num){  /* delete this index */
      for (v=1;v<=var_count;v++) /* adjust indices in all 'var' verts */
	for (j=0;j<=pK_ptr[v].num;j++)
	  if ((w=pK_ptr[v].flower[j])>new_indx)
	    pK_ptr[v].flower[j]--; 
      for (v=new_indx;v<(p->nodecount-num);v++){
	pK_ptr[v]=pK_ptr[v+1];
	pl->orig_indices[v]=pl->orig_indices[v+1];
      }
      num++;
    }
  }
  bdry_count -= num;
  p->nodecount -= num;
	
  /* now create the 'fixed' vert flower (without regard to petal order
     and with extra space for expansion). Only 'var' petals for now. */
  for (w=var_count+1;w<=p->nodecount;w++){
    pK_ptr[w].flower=iptr=(int *)
      calloc((size_t)(3*(pK_ptr[w].num+2)),sizeof(int));
    for (v=1;v<=var_count;v++){
      if (w==pK_ptr[v].flower[pK_ptr[v].num] /* last in open flower? */
	  && w!=pK_ptr[v].flower[0]){
	*(iptr)=v;
	iptr++;
      }
      else 
	for (j=0;j<pK_ptr[v].num;j++)
	  if (w==pK_ptr[v].flower[j]){
	    *(iptr)=v;
	    iptr++;
	    j=pK_ptr[v].num; /* kick out of for loop */
	  }
    }
  }

  /* Hard part: recreate the flowers of the remaining 'fixed' verts. */
      
  for (w=var_count+1;w<=p->nodecount;w++){ /* consider fixed w */
    num=pK_ptr[w].num; /* number of 'var' nghbs */
    new_flower=(int *)
      calloc((size_t)(3*(pK_ptr[w].num+2)),sizeof(int));
    bit_flower=(int *)
      calloc((size_t)(3*(pK_ptr[w].num+2)),sizeof(int));
    eptr=0;
    
    /* will first find all the other 'fixed' that should be
       nghbs of w (because they're nghbs of w in a flower of
       some neighboring 'var' vert v) */
    
    for (j=0;j<num;j++) /* loop through initial (ie. 'var') nghbs */
      {
	v=pK_ptr[w].flower[j];
	indx=nghb(p,v,w);
	if (indx==0){
	  if((u=pK_ptr[v].flower[1])>var_count
	    || (pK_ptr[v].flower[pK_ptr[v].num]==w
		&& (u=pK_ptr[v].flower[pK_ptr[v].num-1])>var_count))
	    pK_ptr[w].num=add_to_flower(p,w,u);
	}
	else if (indx==pK_ptr[v].num){
	  if((u=pK_ptr[v].flower[indx-1])>var_count)
	    pK_ptr[w].num=add_to_flower(p,w,u);
	}
	else{
	  if ((u=pK_ptr[v].flower[indx-1])>var_count)
	    pK_ptr[w].num=add_to_flower(p,w,u);
	  if ((u=pK_ptr[v].flower[indx+1])>var_count)
	    pK_ptr[w].num=add_to_flower(p,w,u);
	}
      }
    
    /* build new_flower from chunks of flower, copied over as 
       they fit in oriented fashion; mark in bit_flower those
       locations in old flower that have found their spots in the
       growing new flower.
       Have to look to variable nghbs of w to find consecutive
       petals for w; after the first pair is copied over, add 
       to either end, keep track of growing size with eptr. 
       (Have to avoid 'fixed' vertices having more than one fan 
       of petals; I think we are depending here on having 'cloned' 
       any 'fixed' vertices needed by 'var' flowers from >= two 
       sides.
       ?? can 'fixed' vert become interior, even though it was
       not allowed to change? I guess so; have to allow this,
       but should check into routines generating the light packings.*/
    
    for (j=0;j<num;j++){ /* search 'var' nghbs only */
      /* Need one contig pair to move to new_flower; after 
	 that, contig pairs for which one member can be adjoined 
	 to those already in the new_flower. */
      v=pK_ptr[w].flower[j];
      if ((indx=nghb(p,v,w))<0) /* v should be 'var' nghb */
	goto BOMB; 
      if (indx==0){ /* w is first (in flower of v), check last also */
	try_contig_pair(p,w,pK_ptr[v].flower[1],v);
	if (w==pK_ptr[v].flower[pK_ptr[v].num])
	  try_contig_pair(p,w,v,pK_ptr[v].flower[pK_ptr[v].num-1]);
      }
      else if (indx==pK_ptr[v].num)
	try_contig_pair(p,w,v,pK_ptr[v].flower[indx-1]);
      else{
	try_contig_pair(p,w,v,pK_ptr[v].flower[indx-1]);
	try_contig_pair(p,w,pK_ptr[v].flower[indx+1],v);
      }
    }
    
    for (k=0;k<pK_ptr[w].num;k++)
      if (!eptr || !bit_flower[k]){ /* missed one or more ==> some 
				       inconsistent around w; error */
	free(new_flower);
	free(bit_flower);
	goto BOMB;
      }
    pK_ptr[w].num=eptr-1;
    free(pK_ptr[w].flower);
    pK_ptr[w].flower=(int *)
      calloc((size_t)(pK_ptr[w].num+1),sizeof(int));
    for (j=0;j<=pK_ptr[w].num;j++)
      pK_ptr[w].flower[j]=new_flower[j];
    free(new_flower);
    pR_ptr[w].rad=pl->radii[w];
    /* add entry in vertex_map */
    vtrace=vtrace->next=(struct Edgelist *)
      calloc((size_t)1,sizeof(struct Edgelist));
    vtrace->v=w;
    vtrace->w=pl->orig_indices[w];
  } /* end of for loop on 'fixed' w */    

  /* should be done now --- though I can't garantee that there
     aren't combinatoric/oientation/connectivity inconsistencies. */

 GOBACK:
  p->status=1;
  choose_alpha(p);choose_gamma(p);
  if (!complex_count(p,0) || !facedraworder(p,0)
      || !fillcurves(p) || !set_aim_default(p))
    goto BOMB;
  if (pl->counts[5]>0) /* some aims given in pl */
    for (i=1;i<=pl->counts[5];i++)
      pR_ptr[pl->aim_indices[i]].aim=pl->aims[i];
  vtrace=p->vertex_map; /* first spot in vertex_map is empty */
  p->vertex_map=p->vertex_map->next;
  free(vtrace);
  return pl->counts[0];

 BOMB:
      alloc_pack_space(p,5000,0); /* clean out any corruption */
      p->status=0;
      return 0;
      
} /* convert_from_p_light */

/* ----------------- local utility routines ----------------- */  

int add_to_flower(struct p_data *p,int w, int u)
{
  int j,num;

  num=p->packK_ptr[w].num;
  for (j=0;j<num;j++) if (p->packK_ptr[w].flower[j]==u) return num;
  p->packK_ptr[w].flower[num]=u;
  p->packK_ptr[w].num++;
  return num+1;
}

int try_contig_pair(struct p_data *p,int w,int vi,int vj)
     /* Check if (vi,vj) (contig pair of petals from some 'var'
	vert, but in clockwise order) can be added to flower of w. 
	Return 1 on success, meaning this is first pair, or one 
	of vi vj is added at begin or end of flower. */    
{
  int i,j,k;

  if ((i=nghb(p,w,vi))<0 || (j=nghb(p,w,vj))<0) return 0;
  /* shouldn't happen; both vi and vj should be in (un-ordered) flower of w */

  if (!eptr){ /* first pair to new_flower */
    new_flower[0]=p->packK_ptr[w].flower[i];
    new_flower[1]=p->packK_ptr[w].flower[j];
    eptr += 2;
    bit_flower[i]=bit_flower[j]=1;
    return 1;
  }
  if (vj==new_flower[0]){ /* shift and add vi at front. */
    if (vi==new_flower[eptr-1]){ /* vi already at end: closed flower */
      p->packK_ptr[w].bdry_flag=0; 
      return 1;
    }
    for (k=eptr-1;k>=0;k--)
      new_flower[k+1]=new_flower[k];
    new_flower[0]=vi;
    eptr++;
    bit_flower[i]=1;
    return 1;
  }
  else if (vi==new_flower[eptr-1]){ /* add vj at end */
    if (vj==new_flower[0]){ /* vj already at beginning, closed flower */
      p->packK_ptr[w].bdry_flag=0; 
      return 1;
    }
    new_flower[eptr]=vj;
    eptr++;
    bit_flower[j]=1;
    return 1;
  }
  return 0;
} /* try_contig_pair */

  
