#include "sa_types.h"
#include "sa_proto.h"

/* Construct hex packings of annuli. Parameters in data string 
   are p, q, steps forming the fundamental loop for this doyle spiral, 
   and n for the number of additional loops to be added to each side 
   of the initial loop; so at end should have 2n+1 copies of 
   the fundamental loop, symmetric about the original loop. */

int doyle_annulus(struct p_data *p,char *datastr)
{
  int nvert,lvert,rvert,j,v,w,qq,pp,n,num;
  char *endptr;
  int hits=0,count=1,*newflower=NULL;

  if (sscanf(datastr,"%d %d %d",&pp,&qq,&n)<3) return 0;
  /* swap to make q larger, hence q>1. */
  if (qq<pp) {j=qq;qq=pp;pp=j;}
  if ((!pp && qq<3) || (pp+qq)<3 || n<1) {
    sprintf(msgbuf,"usage: doyle_annulus p q n. Check legal values (p,q).");
    emsg();
    return 0;
  }
  if (!seed_pack(p,0,6)) return 0;

  if (qq==2 && pp==2) { /* this is a special case */
    if (!enfoldcall(p,"3 7")) return 0;
    add_vert(p,9);
    add_vert(p,9);

    /* fix 5 */
    newflower=(int *)calloc(7,sizeof(int));
    newflower[0]=newflower[6]=1;
    newflower[1]=4;
    newflower[2]=10;
    newflower[3]=9;
    newflower[4]=12;
    newflower[5]=6;
    free(p->packK_ptr[5].flower);
    p->packK_ptr[5].flower=newflower;
    p->packK_ptr[5].num=6;
    p->packK_ptr[5].bdry_flag=0;

    /* fix up 9 */
    newflower=(int *)calloc(7,sizeof(int));
    newflower[0]=newflower[6]=5;
    newflower[1]=10;
    newflower[2]=7;
    newflower[3]=8;
    newflower[4]=11;
    newflower[5]=12;
    free(p->packK_ptr[9].flower);
    p->packK_ptr[9].flower=newflower;
    p->packK_ptr[9].num=6;
    p->packK_ptr[9].bdry_flag=0;

    /* fix up 10 */
    newflower=(int *)calloc(5,sizeof(int));
    newflower[0]=2;
    newflower[1]=7;
    newflower[2]=9;
    newflower[3]=5;
    newflower[4]=4;
    free(p->packK_ptr[10].flower);
    p->packK_ptr[10].flower=newflower;
    p->packK_ptr[10].num=4;
    p->packK_ptr[10].bdry_flag=1;

    /* fix up 4 */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=10;
    newflower[1]=5;
    newflower[2]=1;
    newflower[3]=3;
    free(p->packK_ptr[4].flower);
    p->packK_ptr[4].flower=newflower;
    p->packK_ptr[4].num=3;
    p->packK_ptr[4].bdry_flag=1;

    /* fix up 6 */
    newflower=(int *)calloc(5,sizeof(int));
    newflower[0]=8;
    newflower[1]=7;
    newflower[2]=1;
    newflower[3]=5;
    newflower[4]=12;
    free(p->packK_ptr[6].flower);
    p->packK_ptr[6].flower=newflower;
    p->packK_ptr[6].num=4;
    p->packK_ptr[6].bdry_flag=1;

    /* fix up 12 */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=6;
    newflower[1]=5;
    newflower[2]=9;
    newflower[3]=11;
    free(p->packK_ptr[12].flower);
    p->packK_ptr[12].flower=newflower;
    p->packK_ptr[12].num=3;
    p->packK_ptr[12].bdry_flag=1;
  }

    /* another special case, but there are problems in the
       combinatorics when trying to add generations, so I 
       have temporarily disabled this section. */  

  /*
  else if (pp==1 && qq==2) { 
    add_vert(p,7);
    add_vert(p,7);
  */
    /* fix 7 */
  /*
    newflower=(int *)calloc(7,sizeof(int));
    newflower[0]=newflower[6]=1;
    newflower[1]=6;
    newflower[2]=8;
    newflower[3]=9;
    newflower[4]=5;
    newflower[5]=2;
    free(p->packK_ptr[7].flower);
    p->packK_ptr[7].flower=newflower;
    p->packK_ptr[7].num=6;
    p->packK_ptr[7].bdry_flag=0;
  */
    /* fix 5 */
  /*
    newflower=(int *)calloc(7,sizeof(int));
    newflower[0]=newflower[6]=1;
    newflower[1]=4;
    newflower[2]=2;
    newflower[3]=7;
    newflower[4]=9;
    newflower[5]=6;
    free(p->packK_ptr[5].flower);
    p->packK_ptr[5].flower=newflower;
    p->packK_ptr[5].num=6;
    p->packK_ptr[5].bdry_flag=0;
  */
    /* fix 4 */
  /*
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=2;
    newflower[1]=5;
    newflower[2]=1;
    newflower[3]=3;
    free(p->packK_ptr[4].flower);
    p->packK_ptr[4].flower=newflower;
    p->packK_ptr[4].num=3;
    p->packK_ptr[4].bdry_flag=1;
  */
    /* fix 2 */
  /*
    newflower=(int *)calloc(5,sizeof(int));
    newflower[0]=3;
    newflower[1]=1;
    newflower[2]=7;
    newflower[3]=5;
    newflower[4]=4;
    free(p->packK_ptr[2].flower);
    p->packK_ptr[2].flower=newflower;
    p->packK_ptr[2].num=4;
    p->packK_ptr[2].bdry_flag=1;
  */
    /* fix 6 */
  /*
    newflower=(int *)calloc(5,sizeof(int));
    newflower[0]=8;
    newflower[1]=7;
    newflower[2]=1;
    newflower[3]=5;
    newflower[4]=9;
    free(p->packK_ptr[6].flower);
    p->packK_ptr[6].flower=newflower;
    p->packK_ptr[6].num=4;
    p->packK_ptr[6].bdry_flag=1;
  */
    /* fix 8 */
  /*
    newflower=(int *)calloc(3,sizeof(int));
    newflower[0]=9;
    newflower[1]=7;
    newflower[2]=6;
    free(p->packK_ptr[8].flower);
    p->packK_ptr[8].flower=newflower;
    p->packK_ptr[8].num=2;
    p->packK_ptr[8].bdry_flag=1;
  */
    /* fix 9 */
  /*
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=6;
    newflower[1]=5;
    newflower[2]=7;
    newflower[3]=8;
    free(p->packK_ptr[9].flower);
    p->packK_ptr[9].flower=newflower;
    p->packK_ptr[9].num=3;
    p->packK_ptr[9].bdry_flag=1;
    p->alpha=7;
  }
  end of disabled section */

  /* In general, start with p steps in orig direction. Note: I know what
     happens at each step: three circles are added counterclockwise
     to enclose nvert and make it a hex flower. Except for the first
     step, their vertex numbers are always nvert+2, nvert+3, nvert+4,
     this last equaling the new nodecount. So to continue on a straight
     line we let the new nvert be nodecount-1. */
  else {
    if (pp) {
      nvert=7;
      for (j=1;j<=pp;j++) {
	sprintf(buff,"3 %d",nvert);
	if (!enfoldcall(p,buff)) return 0;
	nvert=p->nodecount-1;
      }
      nvert=p->nodecount; /* set up for shallow left turn */
    }
    else nvert=2; 
    /* now, q-3 steps in this direction. */
    if (qq>3) for (j=1;j<=qq-3;j++) {
      sprintf(buff,"3 %d",nvert);
      if (!enfoldcall(p,buff)) return 0;
      nvert=p->nodecount-1;
    }
    else nvert=p->nodecount;
    
    /* now paste up the end/beginning; nvert comes in next to petal 5,
       opposite from vertex 2. Start with two new vertices, rvert, then 
       lvert (right/left of nvert as one looks from nvert toward 1) */
    add_vert(p,nvert);
    add_vert(p,p->packK_ptr[nvert].flower[0]);
    
    /* fix up vert 5 */
    newflower=(int *)calloc(7,sizeof(int));
    newflower[0]=newflower[6]=6;
    newflower[1]=1;
    newflower[2]=4;
    lvert=newflower[3]=p->packK_ptr[nvert].flower[0];
    newflower[4]=nvert;
    rvert=newflower[5]=p->packK_ptr[nvert].flower[4];
    free(p->packK_ptr[5].flower);
    p->packK_ptr[5].flower=newflower;
    p->packK_ptr[5].num=6;
    p->packK_ptr[5].bdry_flag=0;
    
    /* fix up vert nvert */
    newflower=(int *)calloc(7,sizeof(int));
    for (j=0;j<5;j++) newflower[j]=p->packK_ptr[nvert].flower[j];
    newflower[5]=5;
    newflower[6]=newflower[0];
    free(p->packK_ptr[nvert].flower);
    p->packK_ptr[nvert].flower=newflower;
    p->packK_ptr[nvert].num=6;
    p->packK_ptr[nvert].bdry_flag=0;
    
    /* fix up rvert */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=6;
    newflower[1]=5;
    newflower[2]=nvert;
    newflower[3]=nvert-1;
    free(p->packK_ptr[rvert].flower);
    p->packK_ptr[rvert].flower=newflower;
    p->packK_ptr[rvert].num=3;
    p->packK_ptr[rvert].bdry_flag=1;
    
    /* fix up lvert */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=p->packK_ptr[lvert].flower[0];
    newflower[1]=nvert;
    newflower[2]=5;
    newflower[3]=4;
    free(p->packK_ptr[lvert].flower);
    p->packK_ptr[lvert].flower=newflower;
    p->packK_ptr[lvert].num=3;
    p->packK_ptr[lvert].bdry_flag=1;
    
    /* fix up 4 */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=lvert;
    for (j=0;j<3;j++) newflower[j+1]=p->packK_ptr[4].flower[j];
    free(p->packK_ptr[4].flower);
    p->packK_ptr[4].flower=newflower;
    p->packK_ptr[4].num=3;
    p->packK_ptr[4].bdry_flag=1;
    
    /* fix up 6 */
    num=p->packK_ptr[6].num;
    newflower=(int *)calloc(num+2,sizeof(int));
    newflower[num+1]=rvert;
    for (j=0;j<=num;j++) newflower[j]=p->packK_ptr[6].flower[j];
    free(p->packK_ptr[6].flower);
    p->packK_ptr[6].flower=newflower;
    p->packK_ptr[6].num=num+1;
    p->packK_ptr[6].bdry_flag=1;
  }

  /* need to organize combinatorics */
  complex_count(p,0);
  facedraworder(p,0); 

  /* store the interior vertices as vlist */
  sprintf(next,"i");
  p->vlist=Node_link_parse(p,next,&endptr,&hits,
	   &Vlist,&Elist,&Flist,&region,pathlist,pathlength);

  /* Want 2n+1 loops symmetrically about the original, so have
     to add n-1 hex generations to each boundary. */
  for (j=1;j<=n-1;j++) {
    v=p->bdry_starts[1];
    w=p->bdry_starts[2];
    sprintf(next,"%d %d %d",v,v,6);
    count += add_layer(p,next,&hits);
    sprintf(next,"%d %d %d",w,w,6);
    count += add_layer(p,next,&hits);
  }
  double newrad=1.0/((double)(pp+qq));
  for (j=1;j<=p->nodecount;j++) p->packR_ptr[j].rad=newrad;
  fillcurves(p);
  return count;
} /* doyle_annulus */


    
    
    
