#include "sa_types.h"
#include "sa_proto.h"

/* see original 'new_e_riffle' in 'e_riffle_stuff.cpp' for comments.

   Various optional methods were tried with O3 optimization. So far,
   method_flag 4 performs best on intel with -O3. Should try additional
   optimization flags. */

typedef struct new_e_riffle_state  {
    int aimnum,*index,key;
    int sct, fct;
    int maxBadCuts;
    int minBadCuts;
    int sumBadCuts;
    int cntBadCuts;
    double m, c0;
    double ttoler, fact0;
    double *R1,*R2;
}  new_e_riffle_state;

/* (local) global variables: available to subroutines. */
/* fixup: can we define these JUST for this file? */

int e_aimnum,*e_indx;
double *e_rad,*e_pdel;
double e_mval[11],tval[11];
struct p_data *e_p;
struct R_data *e_pR_ptr;
struct K_data *e_pK_ptr;
double e_rif_routine(int method_flag);

int e_riffle_O3(struct p_data *p_in,int passes, int passNo, 
		    double *cutp, void **state)
{
  /* can compile to use alternate computation methods: */
  int method_flag=7; 

  int i,k,j, N, n,flag,*flower;
  int numBadCuts = 0;
  int count = 0;
  double mm,c0;
  double ftol, c1,act, pred, factor, lambda, mmax;
  double r, r2,m1,m2;
  double fbest,del,bet;
  double lmax;
  double rat,rr;
  new_e_riffle_state *localstate;

  /* These (and other) variables are restored from state variable. */
  int key,sct,fct;
  int maxBadCuts,minBadCuts,sumBadCuts,cntBadCuts;
  double m,ttoler,fact0 = 0.0,*R1,*R2;

  /* set up local-global variables */
  e_p=p_in; 
  e_rad=e_pdel=NULL;
  e_pR_ptr=e_p->packR_ptr;
  e_pK_ptr=e_p->packK_ptr;

  ftol = 0.0005;           /* tolerance for Type 2 SS */

  /* allocate new tmp space for radii/aims */
  e_rad=(double *)calloc((e_p->nodecount+1),sizeof(double));
  e_pdel=(double *)calloc((e_p->nodecount+1),sizeof(double));
  for (i=1;i<=e_p->nodecount;i++) {
    e_rad[i]=e_pR_ptr[i].rad;
    del=sin(e_pR_ptr[i].aim/(2*e_pK_ptr[i].num));
    e_pdel[i]=(1.0-del)/del;
  }

/* construct index and allocate space for storage of radius labels */

  if(passNo & PASS_ALLOC) {
    e_indx=(int *)calloc(5*(e_p->nodecount+1),sizeof(int));
    R1 = (double *)calloc((e_p->nodecount+1),sizeof(double));
    R2 = (double *)calloc((e_p->nodecount+1),sizeof(double));
    if(state != NULL) {
      localstate = (new_e_riffle_state *)
	calloc(1,sizeof(new_e_riffle_state));
      localstate->index = e_indx;
      localstate->R1 = R1;
      localstate->R2 = R2;
      *state = (void *) localstate;
    }
  } /* end of PASS_ALLOC case */

  if(passNo & PASS_INIT) {
    if(!(passNo & PASS_ALLOC)) {
      if(state == NULL) {
	fprintf(stderr, "error in new_e_riffle: passNo "
		"specifies PASS_INIT but not PASS_ALLOC, "
		"and state = NULL (aborting)\n");
	return -1;
      }
      localstate = (new_e_riffle_state *) *state;
      e_indx = localstate->index;
      R1 = localstate->R1;
      R2 = localstate->R2;
    }
    maxBadCuts = 0;
    minBadCuts = 0;
    sumBadCuts = 0;
    cntBadCuts = 0;
    e_aimnum = 0;

    /* list radii to be adjusted: aim>0 and 2 or more faces. */
    for (i=1;i<=e_p->nodecount;i++)
      if (e_pR_ptr[i].aim>0 && e_pK_ptr[i].num>1) {
	e_indx[e_aimnum++]=i;
      }
    if (e_aimnum==0){
      free(e_indx);e_indx=NULL;
      free(R1);free(R2);R1=R2=NULL;
      free(e_rad);free(e_pdel);e_rad=e_pdel=NULL;
      return count;
    }
        
    /* set up parameters */
    
    ttoler = 3*e_aimnum*toler; /* adjust tolerance */
    key = 1;                 /* which type of superstep to start with */
    m = 1;                   /* begining Type 1 multiplier */
    sct = 1;                 /* count of number of times for Type 1 SS */
    fct = 2;                 /* number of iters before a Type 2 SS */

    /* do one iteration to get started */
    c0 = e_rif_routine(method_flag);
    if (c0<0.0) 
      goto ABORT;
    if(state != NULL) {
      localstate->aimnum = e_aimnum;
      localstate->key = key;
      localstate->sct = sct, fct;
      localstate->maxBadCuts = maxBadCuts;
      localstate->minBadCuts = minBadCuts;
      localstate->sumBadCuts = sumBadCuts;
      localstate->cntBadCuts = cntBadCuts;
      localstate->m = m;
      localstate->c0 = c0;
      localstate->ttoler = ttoler;
      localstate->fact0 = fact0;
    }
  } /* end of PASS_INIT case */
  
  if(passNo & PASS_MIDDLE) {
    if(state != NULL) {
      localstate = (new_e_riffle_state *) *state;
      e_aimnum = localstate->aimnum;
      e_indx = localstate->index;
      key = localstate->key;
      sct = localstate->sct, fct;
      maxBadCuts = localstate->maxBadCuts;
      minBadCuts = localstate->minBadCuts;
      sumBadCuts = localstate->sumBadCuts;
      cntBadCuts = localstate->cntBadCuts;
      m = localstate->m;
      c0 = localstate->c0;
      ttoler = localstate->ttoler;
      fact0 = localstate->fact0;
      R1 = localstate->R1;
      R2 = localstate->R2;
    }
    /* ======== Begin Main Loop ============== */
    while ((c0 >ttoler && count<passes) || (count == 0 && state != NULL)) {
      /* fixup: block write? */
      for (i=1;i<=e_p->nodecount;i++) R1[i] = e_rad[i];
      
      /* do update using uniform label model */
      numBadCuts = 0;
      do  {   // Make sure factor < 1.0
	c1=e_rif_routine(method_flag);
	if (c1<0.0) 
	  goto ABORT;
	factor = c1/c0;       /* error reduction ratio */
	if(factor >= 1.0)  {
	  c0 = c1;
	  key = 1;
	  numBadCuts++;
	}
	if(numBadCuts > MAX_ALLOWABLE_BAD_CUTS)  
	  goto ABORT;
      }  while(factor >= 1.0);

      if(cntBadCuts == 0)  {
	maxBadCuts = numBadCuts;
	minBadCuts = numBadCuts;
	sumBadCuts = numBadCuts;
      }
      else  {
	maxBadCuts = (numBadCuts > maxBadCuts) ? numBadCuts : maxBadCuts;
	minBadCuts = (numBadCuts < minBadCuts) ? numBadCuts : minBadCuts;
	sumBadCuts += numBadCuts;
      }
      cntBadCuts++;
      
      /* superstep calculation */
      /* fixup: block write? */
      for (i=1;i<=e_p->nodecount;i++) R2[i] = e_rad[i];
    
      /* find maximum step can safely take */
      lmax = 10000;
      for (j=0;j<e_aimnum;j++) {      /* find max step */
	i = e_indx[j];
	r = e_rad[i];
	rat = r - R1[i];
	if (rat<0)
	  lmax = (lmax < (rr=(-r/rat))) ? lmax : rr;
      }
      lmax = lmax/2;
      
      /* do super step */
      if (key==1) {            /*  type 1  SS */
	lambda = m*factor;
	mmax = 0.75/(1-factor);               /* upper limit on m */
	m = (mmax < (mm=(1+0.8/(sct+1))*m)) ? mmax : mm;
      }
      else {                     /*  type 2 SS */
	if (sct>fct && fabs(factor-fact0)<ftol) {  /* try SS-2 */
	  lambda = factor/(1-factor);
	  sct = -1;
	}
	else
	  lambda = factor;                  /* do something */
      }
      lambda = (lambda>lmax) ? lmax : lambda;
      
      for (j=0;j<e_aimnum;j++) {   /* interpolate to new radius */
	i = e_indx[j];
	e_rad[i] += lambda*(e_rad[i]-R1[i]);
	if (e_rad[i]<0)
	  {e_rad[i]=.001;
	  goto ABORT;}
      }
      sct++;
      fact0 = factor;
    
      /* end of superstep */
    
      /* do update/check using uniform label model */
      c0=e_rif_routine(method_flag);
      if (c0<0.0) 
	goto ABORT;
         
      /* check results */
    
      pred = exp(lambda*log(factor));          /* predicted improvement */
      act = c0/c1;                   /* actual improvement */
      if (act<1) {                   /* did some good */
	if (act>pred) {           /* not as good as expected: reset */
	  m = 1;
	  sct = 0;
	  if (key==1) key = 2;
	}                       /* implied else: accept result */
      }
      else {                       /* reset to before superstep */
	m = 1;
	sct =0;
	for (i=1;i<=e_p->nodecount;i++) e_rad[i] = R2[i];
	c0 = c1;
	if (key==2) key = 1;
      }
      
      /* show activity on occasion */
      if (!(count % 10)) 
      repack_activity_msg("n");
      count++;
    } /* end of main while loop */
    if(state != NULL) {
      localstate->aimnum = e_aimnum;
      localstate->index = e_indx;
      localstate->key = key;
      localstate->sct = sct, fct;
      localstate->maxBadCuts = maxBadCuts;
      localstate->minBadCuts = minBadCuts;
      localstate->sumBadCuts = sumBadCuts;
      localstate->cntBadCuts = cntBadCuts;
      localstate->m = m;
      localstate->c0 = c0;
      localstate->ttoler = ttoler;
      localstate->fact0 = fact0;
      localstate->R1 = R1;
      localstate->R2 = R2;
    }
  } /* end of PASS_MIDDLE case */
  
  /* fixup: don't understand management of open files in 
     the various situations. Who free's the space? */
  if(passNo & PASS_LAST) {
    if(state != NULL) {
      localstate = (new_e_riffle_state *) *state;
      e_aimnum = localstate->aimnum;
      e_indx = localstate->index;
      key = localstate->key;
      sct = localstate->sct, fct;
      maxBadCuts = localstate->maxBadCuts;
      minBadCuts = localstate->minBadCuts;
      sumBadCuts = localstate->sumBadCuts;
      cntBadCuts = localstate->cntBadCuts;
      m = localstate->m;
      c0 = localstate->c0;
      ttoler = localstate->ttoler;
      fact0 = localstate->fact0;
      R1 = localstate->R1;
      R2 = localstate->R2;
      free(localstate);
    }
    if(sumBadCuts > 0)  {
      sprintf(msgbuf, "\nreturning:\n  maxBadCuts = %d;"
	      "  minBadCuts = %d;  avgBadCuts = %.8lf;"
	      "  final cut value = %le\n",
	      maxBadCuts,minBadCuts,
	      sumBadCuts / (singledouble) cntBadCuts,
	      (singledouble) c0);
      msg();
    }
  } /* end of PASS_LAST case */

  if(cutp != NULL)  *cutp = c0;
  free(e_indx);e_indx=NULL;
  free(R1);free(R2);R1=R2=NULL;
  free(e_pdel);e_pdel=NULL;
  for (i=1;i<=e_p->nodecount;i++) e_pR_ptr[i].rad=e_rad[i];
  free(e_rad);e_rad=NULL;
  /* note: angle sums have not been updated */
  return count;
 ABORT:
  if (e_indx) {free(e_indx);e_indx=NULL;}
  if (R1) {free(R1);free(R2);R1=R2=NULL;}
  if (e_pdel) {free(e_pdel);e_pdel=NULL;}
  if (e_rad) {
    for (i=1;i<=e_p->nodecount;i++) e_pR_ptr[i].rad=e_rad[i];
    free(e_rad);e_rad=NULL;
  }
  sprintf(msgbuf,"Aborting repack: radius went negative.");
  emsg();
  if(cutp != NULL)  *cutp = c0;
  return 0;
} /* e_riffle_O3 */

double e_rif_routine(int method_flag)
  /* Computationally intense part of repack. Return accum error
or -1.0 to abort. */
{
  int i,j,k,N,*flower;
  double bet,r,rr,r2,m1,m2,fbest,aae=0;

  for (j=0;j<e_aimnum;j++) {
    i = e_indx[j];     /* point to active node */
    r = e_rad[i];      /* get present label */
    N=e_pK_ptr[i].num;
    fbest=0.0;
    flower=e_pK_ptr[i].flower;

    /* experiments with different organizations */
    switch(method_flag) {

case 1: /* Method 1: original */
  {
    r2=e_rad[*flower];
    m2 = r2/(r+r2);
    for (k=1;k<=N;k++) {
      m1 = m2;
      r2 = e_rad[*(flower+k)];
      m2 = r2/(r+r2);
      fbest +=acos(1.0-2.0*m1*m2);
    }
    break;
  }
case 2: /* Method 2: original method if N>10; else, precompute. */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      for (k=0;k<=N;k++) {
	rr=e_rad[*(flower+k)];
	e_mval[k]=rr/(r+rr);
      }
      for (k=1;k<=N;k++)
	fbest +=acos(1.0-2.0*e_mval[k-1]*e_mval[k]);
    }
    break;
  }
case 3: /* Method 3: cut out loop */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      for (k=0;k<=N;k++) {
	rr=e_rad[*(flower+k)];
	e_mval[k]=rr/(r+rr);
      }
      switch(N) { /* fall through from different starting points */ 
case 10:
  fbest += acos(1.0-2.0*e_mval[9]*e_mval[10]);
case 9:
  fbest += acos(1.0-2.0*e_mval[8]*e_mval[9]);
case 8:
  fbest += acos(1.0-2.0*e_mval[7]*e_mval[8]);
case 7:
  fbest += acos(1.0-2.0*e_mval[6]*e_mval[7]);
case 6:
  fbest += acos(1.0-2.0*e_mval[5]*e_mval[6]);
case 5:
  fbest += acos(1.0-2.0*e_mval[4]*e_mval[5]);
case 4:
  fbest += acos(1.0-2.0*e_mval[3]*e_mval[4]);
case 3:
  fbest += acos(1.0-2.0*e_mval[2]*e_mval[3]);
case 2:
  {
    fbest += acos(1.0-2.0*e_mval[1]*e_mval[2]);
    fbest += acos(1.0-2.0*e_mval[0]*e_mval[1]);
    break;
  }
      } /* end of switch on N */
    }
    break;
  }
case 4: /* Method 4: separate computations */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      for (k=0;k<=N;k++) {
	rr=e_rad[*(flower+k)];
	e_mval[k]=rr/(r+rr);
      }
      switch(N) { /* fall through from different starting points */ 
case 10:
  {
    tval[10]=1.0-2.0*e_mval[9]*e_mval[10];
    fbest += acos(tval[10]);
  }
case 9:
  {
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
    fbest += acos(tval[9]);
  }
case 8:
  {
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    fbest += acos(tval[8]);
  }
case 7:
  {
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    fbest += acos(tval[7]);
  }
case 6:
  {
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest += acos(tval[6]);
  }
case 5:
  {
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest += acos(tval[5]);
  }
case 4:
  {
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest += acos(tval[4]);
  }
case 3:
  {
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest += acos(tval[3]);
  }
case 2:
  {
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest += acos(tval[2]);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest += acos(tval[1]);
    break;
  }
      } /* end of switch on N */
    }
    break;
  }
case 5: /* Method 5: (Seems best.) delay comp of acos */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      for (k=0;k<=N;k++) {
	rr=e_rad[*(flower+k)];
	e_mval[k]=rr/(r+rr);
      }
      switch(N) { 
case 10:
    tval[10]=1.0-2.0*e_mval[9]*e_mval[10];
case 9:
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
case 8:
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
case 7:
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
case 6:
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
case 5:
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
case 4:
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
case 3:
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
case 2:
  {
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
      } /* end of switch on N */
      for (k=1;k<=N;k++) fbest += acos(tval[k]);
    }
    break;
  }
case 6: /* Method 6: blend comp of mval. */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      switch(N) { 
case 10:
  {
    rr=e_rad[*(flower+10)];
    e_mval[10]=rr/(r+rr);
    rr=e_rad[*(flower+9)];
    e_mval[9]=rr/(r+rr);
    tval[10]=1.0-2.0*e_mval[9]*e_mval[10];
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 9:
  {
    rr=e_rad[*(flower+9)];
    e_mval[9]=rr/(r+rr);
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 8:
  {
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 7:
  {
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 6:
  {
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 5:
  {
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 4:
  {
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 3:
  {
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
case 2:
  {
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    break;
  }
      } /* end of switch on N */
      for (k=1;k<=N;k++) fbest += acos(tval[k]);
    }
    break;
  }
case 7: /* Method 7: further blend comp. */
  {
    if (N>10) { 
      r2=e_rad[*flower];
      m2 = r2/(r+r2);
      for (k=1;k<=N;k++) {
	m1 = m2;
	r2 = e_rad[*(flower+k)];
	m2 = r2/(r+r2);
	fbest +=acos(1.0-2.0*m1*m2);
      }
    }
    else {
      switch(N) { 
case 10:
  {
    rr=e_rad[*(flower+10)];
    e_mval[10]=rr/(r+rr);
    rr=e_rad[*(flower+9)];
    e_mval[9]=rr/(r+rr);
    tval[10]=1.0-2.0*e_mval[9]*e_mval[10];
    fbest+=acos(tval[10]);
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
    fbest+=acos(tval[9]);
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    fbest+=acos(tval[8]);
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    fbest+=acos(tval[7]);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest+=acos(tval[6]);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 9:
  {
    rr=e_rad[*(flower+9)];
    e_mval[9]=rr/(r+rr);
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    tval[9]=1.0-2.0*e_mval[8]*e_mval[9];
    fbest+=acos(tval[9]);
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    fbest+=acos(tval[8]);
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    fbest+=acos(tval[7]);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest+=acos(tval[6]);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 8:
  {
    rr=e_rad[*(flower+8)];
    e_mval[8]=rr/(r+rr);
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    tval[8]=1.0-2.0*e_mval[7]*e_mval[8];
    fbest+=acos(tval[8]);
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    fbest+=acos(tval[7]);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest+=acos(tval[6]);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 7:
  {
    rr=e_rad[*(flower+7)];
    e_mval[7]=rr/(r+rr);
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    tval[7]=1.0-2.0*e_mval[6]*e_mval[7];
    fbest+=acos(tval[7]);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest+=acos(tval[6]);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 6:
  {
    rr=e_rad[*(flower+6)];
    e_mval[6]=rr/(r+rr);
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    tval[6]=1.0-2.0*e_mval[5]*e_mval[6];
    fbest+=acos(tval[6]);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 5:
  {
    rr=e_rad[*(flower+5)];
    e_mval[5]=rr/(r+rr);
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    tval[5]=1.0-2.0*e_mval[4]*e_mval[5];
    fbest+=acos(tval[5]);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 4:
  {
    rr=e_rad[*(flower+4)];
    e_mval[4]=rr/(r+rr);
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    tval[4]=1.0-2.0*e_mval[3]*e_mval[4];
    fbest+=acos(tval[4]);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 3:
  {
    rr=e_rad[*(flower+3)];
    e_mval[3]=rr/(r+rr);
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    tval[3]=1.0-2.0*e_mval[2]*e_mval[3];
    fbest+=acos(tval[3]);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
case 2:
  {
    rr=e_rad[*(flower+2)];
    e_mval[2]=rr/(r+rr);
    rr=e_rad[*(flower+1)];
    e_mval[1]=rr/(r+rr);
    tval[2]=1.0-2.0*e_mval[1]*e_mval[2];
    fbest+=acos(tval[2]);
    rr=e_rad[*(flower)];
    e_mval[0]=rr/(r+rr);
    tval[1]=1.0-2.0*e_mval[0]*e_mval[1];
    fbest+=acos(tval[1]);
    break;
  }
      } /* end of switch on N */
    }
    break;
  }

	  } /* end of method_flag switch */

/* ==================== end of intense region ===================== */
 
    /* use the model to predict the next value */
    N = 2*e_pK_ptr[i].num;
    bet = sin(fbest/N);
    r2 = r*bet*e_pdel[i]/(1-bet);
    /* store as new radius label */
    if (r2<=0) 
      return (-1);
    e_rad[i] = r2;
    fbest -= e_pR_ptr[i].aim;
    aae += fbest*fbest;   /* accum abs error */
  } /* end of pass through vertices */
  
  return sqrt(aae);
} /* e_rif_routine */
