#include "sa_types.h"
#include "sa_proto.h"

/* =================== embedding check =====================

To see how much of p1 complex embeds in p2, starting from given
pair of neighbors, adding faces as possible. The result is not
unique, in general. For now, we are not very sophisticated; we
are actually checking for embedding of the faces. Might end up with
identifications for the 3 vertices of a face from p1 which are
not vertices of a face in p2.

a and A must be interiors of p1, p2, respectively. Return the 
count of matches. Send back Edgelist (v, w), w the index in p2 
corresponding to v in p1. 

Do our search on interior vertices of p1: check degree first, then
check that matched neigbhors are consistent. Cycle through two 
linked lists.

Keep vstat, Vstat arrays of vertex identifications for p1, p2. */

int embedding(struct p_data *p1, struct p_data *p2,int a,int b,
	      int A,int B,struct Edgelist **matches)
{
  int *vstat=NULL,*Vstat=NULL,count=0,u,U,num,v,V,i,indv,indV;
  struct Vertlist *big=NULL,*news=NULL,*btrace,*ntrace;
  struct Edgelist *etrace;
  struct K_data *pK1,*pK2;

  if (!p1->status || !p2->status 
      || (indv=nghb(p1,a,b))<0 || (indV=nghb(p2,A,B))<0 
      || p1->packK_ptr[a].bdry_flag || p2->packK_ptr[A].bdry_flag
      || p1->packK_ptr[a].num!=p2->packK_ptr[A].num)
     return 0;
  pK1=p1->packK_ptr;
  pK2=p2->packK_ptr;
  vstat=(int *)calloc((size_t)(p1->nodecount+1),sizeof(int));
  Vstat=(int *)calloc((size_t)(p2->nodecount+1),sizeof(int));
  big=btrace=(struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
  btrace->v=a;
  btrace=btrace->next=(struct Vertlist *)
    calloc((size_t)1,sizeof(struct Vertlist));
  btrace->v=b;
  vstat[a]=A;Vstat[A]=a;
  vstat[b]=B;Vstat[B]=b;

  /* entering main loop */
  while (big)
    {
      news=ntrace=(struct Vertlist *)
	calloc((size_t)1,sizeof(struct Vertlist));
      btrace=big;
      while (btrace)
	{
	  v=btrace->v;
	  V=vstat[v];
	  if (!pK1[v].bdry_flag && !pK2[V].bdry_flag
	      && (num=pK1[v].num)==pK2[V].num)
	    {
	      i=0;
	      while (i<num && !(U=vstat[(u=pK1[v].flower[i])])) i++;
	      if (i==num) goto OUTAHERE; /* error? */
	      indv=i;
	      indV=nghb(p2,V,U);
	      for (i=1;i<num;i++)
		{
		  u=pK1[v].flower[(indv+i)%num];
		  U=pK2[V].flower[(indV+i)%num];
		  if (!vstat[u] && !Vstat[U]) /* new identification */
		    {
		      vstat[u]=U;Vstat[U]=u;
		      ntrace=ntrace->next=(struct Vertlist *)
			calloc((size_t)1,sizeof(struct Vertlist));
		      ntrace->v=u;
		    }
		}
	    }
	  btrace=btrace->next;
	} /* end of inner while */
      ntrace=news->next;
      free(news);
      news=NULL;
      vert_free(&big);
      big=ntrace;
    }
  if (*matches) 
    {
      free(*matches);
      *matches=NULL;
    }
  *matches=etrace=(struct Edgelist *)
    calloc((size_t)1,sizeof(struct Edgelist));
  for (u=1;u<=p1->nodecount;u++)
    if ((U=vstat[u]))
      {
	etrace=etrace->next=(struct Edgelist *)
	  calloc((size_t)1,sizeof(struct Edgelist));
	etrace->v=u;
	etrace->w=U;
	count++;
      }
  etrace=(*matches)->next;
  free(*matches);
  *matches=etrace;
 OUTAHERE:
  free(vstat);
  free(Vstat);
  return count;
} /* embedding */
  
