#include "sa_types.h"
#include "sa_proto.h"

/* Set boundary radii of a packing based on the derivative e^{z^n};
n=1 is exponential, n=2 is "error function". There is a domain packing 
p1 and a range p2. Normally the domain is univalent; it is copied to 
p2 and each boundary radii of p2 are set to e^{z^n}*r, where r is its 
radius, z its center in p1. Return 0 on error. */

int erf_ftn(struct p_data *p1, struct p_data *p2,int n)
{
  int v,endv,wflag=0;
  complex z,w;

  if (n<1 || n>3 || !p1->status || p1->hes || !p1->num_bdry_comp 
      || (p1->euler!=1 && p1->euler!=2) || p1->genus!=0
      || !copy_pack(p1,p2)) {
    sprintf(msgbuf,"error: erf_ftn: p1 must be closed disc, n=1,2, or 3.");
    emsg();
    return 0;
  }
  v=endv=p1->bdry_starts[1];
  while ((v=p1->packK_ptr[v].flower[0])!=endv || !(wflag++)) {
    z=w=p1->packR_ptr[v].center;
    if (n>1) w=cmult(z,z); /* z^2 */
    if (n>2) w=cmult(w,z); /* z^3 */
    p2->packR_ptr[v].rad=exp(w.re)*p1->packR_ptr[v].rad;
  }
  return 0;
} /* erf_ftn */

	
    
