#include "sa_types.h"
#include "sa_proto.h"

/* routines to locate positions in open file */

long find_token(FILE *fp,long pos,char **string,char *token)
     /* starting at position pos, find token in file, return
	line following it. Return -1 on error or failure, else
	return position following returned string. */
{
  int n,m,i;
  char *buff,*hit;

  *string=NULL;
  if (!fp || (n=strlen(token))>34 || n==0
      || pos < 0 || fseek(fp,pos,SEEK_SET)) 
    return -1;
  buff=(char *)calloc((size_t)(BUFSIZE+2),sizeof(char));
  buff[BUFSIZE+1]='\0';

  fgets(buff,BUFSIZE,fp);
  hit=NULL;
  while (!(hit=strstr(buff,token)) && fgets(buff,BUFSIZE,fp))
    hit=NULL;
  if (hit) /* found token */
    {
      hit += strlen(token);
      i=0;
      do
	{
	  *(buff+i)=*(hit+i);
	  i++;
	} while (hit[i]!='\0' && hit[i]!='\n' && i<=BUFSIZE);
      *(buff+i)='\0';
      while (buff[(n=strlen(buff)-1)]=='\\' && (m=(BUFSIZE-n)) > 1
	     && fgets((buff+n+1),m,fp)) 
	/* line continuations? Replace \ by space and concatenate. */
	{
	  buff[n]=' ';
	  buff[strlen(buff)-1]='\0'; /* get rid of \n */
	}
      *string=buff;
      return (long)ftell(fp);
    }
  free(buff);
  return (long)(-1);
} /* find_token */
  
int find_name(FILE *fp,char *name,long *pos)
     /* search file from current location for 'name',
	look for key word at beginning of one of next
	5 lines: return 
	1 = NODECOUNT, 2 = CHECKCOUNT, 3 = PATH,
	4 = CMDS, -1 = error; 
	file position following name is given in 'pos',
        but file position is reset to original. */
{
  int len,i,flag=0;
  long save_pos,spot;
  char buff[BUFSIZE],next[BUFSIZE],*hit,*ptr;

  if (!fp || (len=strlen(name))<1 || len>128)
    return -1;
  save_pos=spot=ftell(fp);
  
  hit=NULL;
  while (!hit && fgets(buff,BUFSIZE,fp))
    {
      if ((hit=strstr(buff,name))) /* found name */
	{
	  spot=spot+(int)(hit-buff)+len;
	  fseek(fp,spot,SEEK_SET);
	  for (i=1;i<=5;i++)
	    {
	      ptr=&(buff[0]);
	      if (fgets(buff,BUFSIZE,fp) && grab_next(&ptr,next))
		{
		  if (!strncmp(next,"NODECOUNT",9))
		    flag=1;
		  else if (!strncmp(next,"CHECKCOUNT",10))
		    flag=2;
		  else if (!strncmp(next,"PATH",4))
		    flag=3;
		  else if (!strncmp(next,"CMDS",4))
		    flag=4;
		}
	      if (flag) i=6; /* bomb out */
	    }
	  if (!flag) /* didn't find a key word */
	      hit=NULL;
	  else fseek(fp,spot,SEEK_SET);
	}
      if (!hit) spot=ftell(fp);
    } /* end of while */
  fseek(fp,save_pos,SEEK_SET);
  if (flag)
    {
      *pos=spot;
      return flag;
    }
  return -1;
} /* find_name */
