#include "sa_types.h"
#include "sa_proto.h"

/* given 2 hyp centers/x_radii and x_radius of third, return hyp 
center and x_radius for third (latter changes only if negative 
to reflect eucl rad of horocycle). 
oj=inv dist of edge opposite circle j. No consistency check on 
first two circles or incompatiblities. iflag flags incompats (but not 
yet passed back). ol_flag set only if there may be nondefault overlaps.
Return 0 if there's an arithmetic error. */

int h_compcenter(complex z1,complex z2,complex *z3,
		 double x1,double x2,double *x3,
		 int ol_flag,double o1,double o2,double o3)
{
  int iflag=0;
  double sgn=1,cc,cc2,rad,ahc,sc,s,ac,r,side_p1,ss1,ss3,s1,s2,s3;
  double s13,acstuff,x13,x1p3;
  complex a,b,c,w1,w2,w3;

  s1=x_to_s_rad(x1);s2=x_to_s_rad(x2);s3=x_to_s_rad(*x3);
  a=z1;
  b=z2;
  if ((s1<=0) && (s2>0)) { /* second circle finite, first not */
    a=z2; b=z1; s=s1; s1=s2; s2=s; s=x1; x1=x2; x2 = s;
    s=o1; o1=o2; o2=s; sgn=(-1.0); 
    /* interchange order */
  }
  if (s1>0) { /* first is now finite */
    if (ol_flag) /* tougher computation */
      cc=h_cos_s_overlap(x1,x2,*x3,o1,o2,o3,&iflag);
    else
      cc=h_comp_x_cos(x1,x2,*x3);

    /* if third circle is finite */
    if (s3>0) {
      ss1=1-x1;
      ss3=1-*x3;
      x13 = x1*(*x3);
      x1p3 = x1+(*x3);
      s13=s1*s3;
      acstuff=(x1p3-x13)/(s13*(1+s13))
	-(2*x1p3-(1+o2)*x13)/(4*s13);
      if (acstuff<0.0)
	return 0;
      side_p1= acstuff + sqrt(acstuff*(acstuff+2));
      
      ahc=side_p1/(side_p1+2); /* abs value of hyp center */
      z3->re = cc*ahc;
      z3->im = sgn*sqrt(1-cc*cc)*ahc; /* center as if z1 at origin */
      
      if (a.re==b.re && a.im==b.im) z3->re *= ahc;

      *z3=mob_norm_inv(*z3,a,b);   /* move to right place */
      return (1);
    }

    /* else third circle is a horocycle */

    r=(1-s1)/(1+s1);
    sc=(r*r+1+2*r*o2)/(2*(1+r*o2)); /* abs value of eucl center */
    c.re=sc*cc;
    if ((cc2=cc*cc)>1.0) return 0;
    c.im=sc*sgn*sqrt(1-cc2);
    rad=1-sc; /* Now have c and its radius */
    w1.re=c.re-rad;
    w1.im=c.im;
    w2.re=c.re+rad;
    w2.im=c.im;
    w3.re=c.re;
    w3.im=c.im+rad; /* three points on the circle */
    w1=mob_norm_inv(w1,a,b);
    w2=mob_norm_inv(w2,a,b);
    w3=mob_norm_inv(w3,a,b);       /* 3 pts on new circle */
    circle_3(w1,w2,w3,&c,&rad);
    *x3=(-rad); /* store neg of eucl radius as x_rad */
    ac=cAbs(c);
    z3->re=c.re/ac;
    z3->im=c.im/ac; /* get hyp center on unit circle */
    return (1);
  } 
  /* remaining case: first two are horocycles. Expect -eucl radius as
     s-rad of target circle. */
  if (s3<=(-1.0)) /* not a valid eucl radius */
    return 0; 
  return h_horo_center(z1,z2,z3,-s2,x3,o1,o2,o3);
} /* h_compcenter */

