#include "sa_types.h"
#include "sa_proto.h"

/* Routine for experimenting with harmonic mappings. 

If h and g are analytic then f = h + conj(g) is harmonic. If |h'|
never vanishes and |h'|>|g'|, then f is orientation preserving. 
Certain other conditions will imply that f is one-to-one (which is 
generally intended when using the term "harmonic mapping" in
function theory). A good set of examples is h = identity and
g = (1/n)(conj(z))^n on the unit disc. (See Duren's book.)

For discrete version, just need two euclidean packings H and G
for same complex K so that the boundary radii of H are greater
than those of G and so H is locally univalent. 

This routine computes F = H + conj(G) (i.e., computes the resulting
complex number associated with each vertex), copies H into p2
and sets its centers based on H + conj(G).

The calling routine should do various other things:
* check that Hp and Gp are euclidean, have same nodecount.
* check that Hp is locally univalent.
* check that bdry radii of Hp are larger than those of Gp
* copy Hp into Fp.
* adjust the screen size of Fp to accommodate result
* send status to user.

May put other options in later, such as rotating and/or translating
H and/or G.

Return 0 on error.

*/

int h_g_bar(struct p_data *Hp, struct p_data *Gp, struct p_data *Fp)
{
  int j;
  struct R_data *Hptr,*Gptr,*Fptr;

  Hptr=Hp->packR_ptr;
  Gptr=Gp->packR_ptr;
  Fptr=Fp->packR_ptr;
  for (j=1;j<=Hp->nodecount;j++) {
    Fptr[j].center=cadd(Hptr[j].center, cconj(Gptr[j].center));
  }
} /* h_g_bar */


    
      
  


 

