#include "sa_types.h"
#include "sa_proto.h"

/* recompute centers using drawing order as in draw_in_order. 
Report all locations of vertex v. flag means to place first face. 
Note that this changes center data. */

int layout_report(struct p_data *p,int v,int flag)
{
  int start=0,move=0,j,nf,vert,vflag=1,ck=0,n0,indx;
  complex a,g,w,I,modw,newz,z,y;
  struct Centlist *centlist,*ctmp,*ctrace=NULL;
  struct R_data *pR_ptr=p->packR_ptr;

  centlist=NULL;
  if (v>0 && v<=p->nodecount) vflag=0;
  nf=p->first_face;	
  if (flag) place_face(p,nf,p->faces[nf].index_flag);
  for (j=0;j<3;j++) /* initial location of v */
    if (p->faces[nf].vert[j]==v) /* record first center */
      {
	start=j+1;
	ctmp=(struct Centlist *)calloc(1,sizeof(struct Centlist));
	ctmp->z=pR_ptr[v].center;
	if (!centlist) centlist=ctrace=ctmp;
	else ctrace=ctrace->next=ctmp;
      }
  while ( (nf=p->faces[nf].next_face)!=p->first_face 
	  && ck<2*p->facecount && nf>0 && nf <= p->facecount)
    {
      vert=p->faces[nf].vert[
			     (indx=(p->faces[nf].index_flag + 2) % 3)];
      n0=nghb(p,vert,p->faces[nf].vert[(indx+1)%3]);
      fancy_comp_center(p,vert,n0,n0,1,0,1,1,toler);
      if (!vflag && vert==v) 
	/* record center */
	{
	  move++;
	  ctmp=(struct Centlist *)calloc(1,sizeof(struct Centlist));
	  ctmp->z=pR_ptr[v].center;
	  if (!centlist) centlist=ctrace=ctmp;
	  else ctrace=ctrace->next=ctmp;
	}
      ck++;
    }
  if (nf!=p->first_face) /* some error */
    return 0; 
  if (!vflag && start && move) 
    /* if v in first face and has moved, give final location */
    {
      ctmp=(struct Centlist *)calloc(1,sizeof(struct Centlist));
      ctmp->z=pR_ptr[v].center;
      if (!centlist) centlist=ctrace=ctmp;
      else ctrace=ctrace->next=ctmp;
    }

  /* normalize */
  a=pR_ptr[p->alpha].center;
  g=pR_ptr[p->gamma].center;
  norm_any_pack(p,a,g);

  /* adjust locations reported for v because of normalization */
  if (centlist)
    {
      sprintf(msgbuf,"Locations of vertex %d: ",v);
      msg();
      ctrace=centlist;
      while (ctrace)
	{
	  z=newz=ctrace->z;
	  if (p->hes< 0) /* hyp */
	    newz=mob_norm(z,a,g);
	  else if (p->hes ==0) /* eucl */
	    {
	      w=csub(g,a);
	      I.re=0; I.im=cAbs(w);
	      modw=cdiv(I,w);
	      y=csub(z,a);
	      newz=cmult(modw,y);
	    }
	  sprintf(msgbuf,"% .10e + % .10e i ;",
		  (singledouble)newz.re,(singledouble)newz.im);
	  msg();
	  ctmp=ctrace;
	  ctrace=ctrace->next;
	  free(ctmp);
	}
    }
  return (vflag+start+move); /* error if this is zero */
} /* layout_report */

int layout_facelist(struct p_data *p,struct Vertlist **facelist,
		    int fix,int l_face)
/* Lay out faces from facelist. fix=1 means to locate faces 
successively (changing stored centers). facelist is cleared. 
l_face="live face" may be given. See also 'layout_draw_facelist'. */
{
	int n,v1,v2,v3,vv=0;
	double o1,o2,o3;
	struct Vertlist *trace;
	struct K_data *pK_ptr;
	struct R_data *pR_ptr;

	pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
	if ((*facelist)==NULL) return 0;
	trace=*facelist;
		/* find first */
	if (l_face>0 && l_face<=p->facecount) vv=l_face;
	do {
			/* if fix, draw successive ones as contig. */
	  if (fix)
	    {
	      if ( (n=nghb_tri(p,vv,trace->v))<0 )
		n=p->faces[trace->v].index_flag;
	      /* if not contig, use index_flag */
	      v1=p->faces[trace->v].vert[n];
	      v2=p->faces[trace->v].vert[(n+1)%3];
	      v3=p->faces[trace->v].vert[(n+2)%3];
	      
	      if (p->overlap_status) /* oj for edge opposite vj */
		{
		  o1=pK_ptr[v2].overlaps[nghb(p,v2,v3)];
		  o2=pK_ptr[v3].overlaps[nghb(p,v3,v1)];
		  o3=pK_ptr[v1].overlaps[nghb(p,v1,v2)];
		}
	      else o1=o2=o3=1.0;
	      any_compcenter(p,
		pR_ptr[v1].center,pR_ptr[v2].center,&pR_ptr[v3].center,
		pR_ptr[v1].rad,pR_ptr[v2].rad,&pR_ptr[v3].rad,o1,o2,o3);
			/* compute and store new center */
	    }
	  
	  while (trace!=NULL && trace->v==vv) trace=trace->next;
					/* skip repeats */
	  if (trace==NULL) break;
	  vv=trace->v;
	  trace=trace->next;
	} while (trace!=NULL);
	vert_free(facelist);
	return vv;
} /* layout_facelist */
