#include "sa_types.h"
#include "sa_proto.h"

/* ================= Some new redchain processing (8/06) ============ 
Want to make redchains for multiply connected surfaces more stable,
so, e.g., if one does an edge-flip, the whole ordering doesn't
(necessarily) get redone.

Idea: record the "red edge" list (Edgelist) of outside edges for
a given red chain; after whatever modification is made, an attempt
is made to first build the new redchain using the red edges. May
want options to force a check, eg, that all faces are hit.

*/

struct Edgelist *red_to_outlist(struct p_data *p,int mode) 
/* Given a packing, return outer (righthand side) closed edgelist
   around the redchain; return NULL on error. mode will be used
   in future for options.*/
{
  int v,w,nw,nlv,rv,lv,count=0,wflag=0;
  struct Edgelist *elist=NULL,*etrace;
  struct RedList *spot,*sptrace,*Spot;

  if (!p->status || !p->redfaces) return NULL;

  elist=etrace=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));

  // keep pointer
  Spot=spot=p->redfaces;

  // Getting started:
  // first face blue? put two edges in.
  if (spot->prev->face==spot->next->face) { // blue
    etrace->v=p->faces[spot->face].vert[(spot->v_flag+1)%3];
    etrace->w=p->faces[spot->face].vert[(spot->v_flag+2)%3];
    etrace=etrace->next=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
    etrace->v=w;
    etrace->w=lv=p->faces[spot->face].vert[spot->v_flag];
    spot=spot->next;
  }
  else {  // go until a redface has a red edge
    rv=red_share_vert(p,spot->prev);
    lv=red_share_vert(p,spot);
    while (spot && lv==rv) {
    	spot=spot->next;
    	lv=red_share_vert(p,spot);
    }

    etrace->v=rv;
    etrace->w=lv;
  }

  // reiterate
  while (spot && (!(wflag++) || spot!=Spot)) {
  // Go until you get a new edge (count is safty check)
    while (spot && count<10*p->facecount && (nlv=red_share_vert(p,spot))>-1 
    	&& nlv==lv) {
      spot=spot->next;
      if (spot==Spot) return elist;
      count++;
    }
    if (!spot || count>=10*p->facecount || nlv<0 || nlv==lv) { // some error
      edge_free(&elist);
      return NULL;
    }
    etrace=etrace->next=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
    etrace->v=lv;
    etrace->w=lv=nlv;
  }
  if (spot==Spot) { // done
    return elist;
  }
} /* red_to_outlist */

struct RedList *red_from_outlist(struct p_data *p,struct Edgelist *elist)
  /* Given an edgelist, try to create a redlist that will have this
     as its outer edge (right-hand side). Return NULL on error. */
{
  int u,f,dum,nv,nu,nf,f1,f2,ff,u1,u2,wflag=0,common,k,num,count=0,stop=0;
  struct Edgelist *etrace,*ltrace;
  struct RedList *rlist=NULL,*nextr,*backfan,*btrace;

  if (!p->status || !elist) return NULL;
  rlist=nextr=(struct RedList *)calloc(1,sizeof(struct RedList));

  // compare first/last faces to see if blue
  etrace=elist;
  f1=left_face(p,etrace->v,etrace->w,&u1);
  while (etrace->next) etrace=etrace->next;
  if (etrace->w!=elist->v) goto BOMB; // edgelist isn't closed.
  ff=left_face(p,etrace->v,etrace->w,&dum);
  if (ff==f1) { // yes? rejigger so last face is moved to first.
  	// find new last edge.
  	ltrace=elist;
  	while (ltrace && ltrace->next && ltrace->next->next) 
  		ltrace=ltrace->next;
  	ltrace->next=NULL; // make it last	
  	etrace->next=elist;
    elist=etrace; // make etrace first
  }

  // now continue (still could have blue f1, but first edge is hit first
  etrace=elist;
  f1=left_face(p,etrace->v,etrace->w,&u1);
  nextr->face=f1;
  etrace=etrace->next;
  f2=left_face(p,etrace->v,etrace->w,&u2);
  if (f1==f2) { // blue face
    etrace=etrace->next;
    f2=left_face(p,etrace->v,etrace->w,&u2);
  }
  common=etrace->v;

  // iterate through all the edges: come in with f1, u1, f2, u2, common.
  while (count<10*p->nodecount) {
    // have to build fan between f2 and f1:

    // f2 is first face in fan moving back to f1
    u=u2;
    backfan=btrace=(struct RedList *)calloc(1,sizeof(struct RedList));
    btrace->face=f2;
    num=p->packK_ptr[common].num;
    while (u!=u1) {
      k=(nghb(p,common,u)+1)%num;
      nu=p->packK_ptr[common].flower[k];
      nf=what_face(p,common,u,nu);
      RedList *tmp=btrace;
      btrace=btrace->prev=(struct RedList *)calloc(1,sizeof(struct RedList));
      btrace->next=tmp;
      btrace->face=nf;
      u=nu;
    }
    // hook this new list in, point to new end
    btrace->prev=nextr;
    nextr->next=btrace;
    nextr=backfan;
    count++;

	// this is the exit spot; backfan's face is a repeat
    if (stop) { // this was last; close up and return
	  nextr=nextr->prev;
	  nextr->next=rlist;
	  rlist->prev=nextr;
	  // fix redchain data, v_flags, pairings, etc.
	  redface_comb_info(p,rlist,p->edge_pair);
	  return rlist;
    }    	

    // set up for repeat
    f1=f2;
    u1=u2;
    common=etrace->w;
    // get next new face
    etrace=etrace->next;
    if (!etrace) { // need first edge to finish up
    	stop=1;
    	ltrace=elist; 
    }
    else ltrace=etrace;
    f2=left_face(p,ltrace->v,ltrace->w,&u2);
    if (f1==f2) { // blue?
      common=etrace->w;
      etrace=etrace->next;
      if (!etrace) { // need first edge to finish up
      	stop=1;
      	ltrace=elist;
      }
	  else ltrace=etrace;
      f2=left_face(p,ltrace->v,ltrace->w,&u2);
	  if (f2<0) goto BOMB;
    }
  } // end of while
 BOMB:
   free_redfaces(&rlist);
   return NULL;
} /* red_from_outlist */
	 
  
    
