#include "sa_types.h"
#include "sa_proto.h"

/* 'output' is a structure command to print various formated
   numerical results to files. Last string in the command should
   be filename, preceeded by '-a' if data is to be adjoined. */

static int digs=1;
struct p_data *q,*pp;

struct DataObj
 {
   int vfe;               /* 1=vertices, 2=faces, 3=edges */
   int code;              /* code for type of data */
   char *spec;            /* string object */
   struct DataObj *next;  /* linked list */
 };

int output_parse(struct p_data *p,FILE *fp,char *datastr,Mobius mob)
     /* Parse "output" calls. Assume fp is open, return count. */
{
  int i,count=0,vfe=0,Vfe=0,hits,code;
  complex tr2;
  char *nextpoint=datastr,next[BUFSIZE];
  char *loop_str=NULL,*spec,*suffix_str,*endptr,*tmpb,*part[3];
  struct DataObj *dataobj=NULL,*dtrace;
  struct Vertlist *vertlist=NULL,*facelist=NULL,*vtrace,*ftrace;
  struct Edgelist *edgelist=NULL,*etrace;  struct Pathlist *ptrace=NULL;


  digs=1;q=NULL;pp=NULL;
  while ((pow(10.,(double)digs))*toler < 0.1 
	 && digs<MAX_ACCUR) digs++;
  stripsp(nextpoint);
  tmpb=nextpoint;
  for (i=0;i<3;i++) { /* Must find three "::" breakers */
    if (!(tmpb=strstr(tmpb,"::"))) {
      sprintf(msgbuf,"output: error in locating some part; check format.");
      emsg();
      return 0;
    }
    part[i]=tmpb;
    tmpb += 2;
  }
  
/* ======= break out "prefix :: data :: loop :: suffix" */

  *(part[0])=*(part[1])=*(part[2])='\0';

/* prefix: up to first ::, could be empty, send for processing. */
  stripsp(nextpoint);
  if ((tmpb=find_literal(nextpoint,&endptr,0))) {
    fprintf(fp,"%s",tmpb);
    free(tmpb);
    tmpb=NULL;
  }

/* ================= parsing data/loop ===================== */
    
/* Must always have separator ::, but could be empty looping */

  nextpoint=part[1]+2;
  stripsp(nextpoint);
  if (!strlen(nextpoint)) { /* empty looping: parse/print. */
    nextpoint=part[0]+2;
    while (*nextpoint!='\0') { /* cycle through objects */
      stripsp(nextpoint);
      if (*nextpoint=='\"') { /* look for literal */
	if (!(spec=find_literal(nextpoint,&endptr,1))) { /* error */
	  sprintf(msgbuf,"output: missing '""'");
	  emsg();
	  goto LEAVING;
	}
	fprintf(fp,"%s",spec);
	nextpoint=endptr;
      }
      else if (grab_next(&nextpoint,next)) { /* look for what to print */
	switch(next[0]) {
	case 'M': {
	  if (!strncmp(next,"Mob",3)) { /* given mobius trans. */
	    CPputd(fp,mob.a.re,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.a.im,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.b.re,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.b.im,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.c.re,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.c.im,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.d.re,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,mob.d.im,digs,'e');
	    fprintf(fp," ");
	    count++;
	  }
	  break;
	}
	case 'P': {
	  if (!strncmp(next,"PATH",4)) { /* current path */
	    ptrace=pathlist;
	    while(ptrace) {
	      fprintf(fp," %lf %lf ",
		      (singledouble)ptrace->x,(singledouble)ptrace->y);
	      ptrace=ptrace->next;
	      count++;
	    }
	  }
	  break;
	}
	case 't': {
	  if (!strncmp(next,"tr",2)) { /* trace^2 of given mobius */
	    tr2=tsqr_mob(mob);
	    CPputd(fp,tr2.re,digs,'e');
	    fprintf(fp," ");
	    CPputd(fp,tr2.im,digs,'e');
	    fprintf(fp," ");
	    count++;
	  }
	  break;
	}
	case '#': {
	  if (!strncmp(next,"#N",2)) { /* packing nodecount */
	    fprintf(fp,"%d ",p->nodecount);
	    count++;
	  }
	  else if (!strncmp(next,"#F",2)) { /* packing facecount */
	    fprintf(fp,"%d ",p->facecount);
	    count++;
	  }
	  break;
	}
	} /* end of switch */
      }
      stripsp(nextpoint);
    } /* end of while */
  } 
  else { /* seems to be a loop */

    /* find loop_str */
    nextpoint=part[1]+2;
    stripsp(nextpoint);
    if(strlen(nextpoint)) {
      loop_str=(char *)calloc((size_t)(strlen(nextpoint)+1),sizeof(char));
      strcpy(loop_str,nextpoint);
    }
    
    dataobj=dtrace=(struct DataObj *)
      calloc((size_t)1,sizeof(struct DataObj));
    nextpoint=part[0]+2;
    stripsp(nextpoint);
    while (*nextpoint!='\0') { /* cycle through objects */
      vfe=0;
      stripsp(nextpoint);
      if (*nextpoint=='\"' && (spec=find_literal(nextpoint,&endptr,1))) { 
	/* this is a literal? */
	dtrace=dtrace->next=(struct DataObj *)
	  calloc((size_t)1,sizeof(struct DataObj));
	dtrace->code=0;
	dtrace->spec=spec;
	nextpoint=endptr;
	count++;
      }
      else if (*nextpoint=='\"') { /* error */
	sprintf(msgbuf,"output: error in some quoted string.");
	emsg();
	count=0;
	goto LEAVING;
      }
      else if (grab_next(&nextpoint,next) 
	       && (vfe=data_obj_parse(next,&code))) { /* this is data */
	dtrace=dtrace->next=(struct DataObj *)
	  calloc((size_t)1,sizeof(struct DataObj));
	dtrace->vfe=vfe;
	dtrace->code=code;
	count++;
      }
      if (vfe) Vfe=vfe;
    } /* end of while */
    dtrace=dataobj->next; /* didn't use first spot */
    free(dataobj);
    dataobj=dtrace;
    if (!dataobj) goto LEAVING;; /* didn't get any objects */
  }
  if(loop_str) { /* got loop_str */

    /* expect vertex list? */
    if (Vfe==1 && (vertlist=Node_link_parse(p,loop_str,&endptr,&hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength))) {
      vtrace=vertlist;
      while (vtrace) { /* loop through vertices */
	dtrace=dataobj;
	/* loop through objects until you find improper one. */
	while (dtrace && (dtrace->spec || dtrace->vfe==Vfe)) { 
	  if (dtrace->spec) { /* literal */
	    fprintf(fp,"%s",dtrace->spec);
	  }
	  else print_vert_obj(p,fp,dtrace->code,vtrace->v);
	  dtrace=dtrace->next;
	} /* done going through objects */
	vtrace=vtrace->next;
	count++;
      } /* done going through vertices */
      vert_free(&vertlist);
    }

    /* expect face list? */
    else if (Vfe==2 &&(facelist=Face_link_parse(p,loop_str,&endptr,&hits,
	 &Vlist,&Elist,&Flist,&region,pathlist,pathlength))) {
      ftrace=facelist;
      while (ftrace) { /* loop through faces */
	dtrace=dataobj;
	/* loop through objects until you find improper one. */
	while (dtrace && (dtrace->spec || dtrace->vfe==Vfe)) { 
	  if (dtrace->spec) { /* literal */
	    fprintf(fp,"%s",dtrace->spec);
	  }
	  else print_face_obj(p,fp,dtrace->code,ftrace->v);
	  dtrace=dtrace->next;
	} /* done going through objects */
	ftrace=ftrace->next;
	count++;
	} /* done going through faces */
      vert_free(&facelist);
    }

    /* expect edge list? */
    else if ((Vfe==3) && (edgelist=Node_pair_link(p,loop_str,&endptr,&hits,
	     &Vlist,&Elist,&Flist,&region,pathlist,pathlength))) {
      etrace=edgelist;
      while (etrace) { /* loop through edges */
	dtrace=dataobj;
	/* loop through objects until you find improper one. */
	while (dtrace && (dtrace->spec || dtrace->vfe==Vfe)) { 
	  if (dtrace->spec) { /* literal */
	    fprintf(fp,"%s",dtrace->spec);
	  }
	  else print_edge_obj(p,fp,dtrace->code,etrace->v,etrace->w);
	  dtrace=dtrace->next;
	} /* done going through objects */
	etrace=etrace->next;
	count++;
      } /* done going through edges */
      edge_free(&edgelist);
    }

    /* error */
    else { 
      sprintf(msgbuf,"output: didn't find proper list.");
      emsg();
    }
  } /* end of looping */
    /* finally, add on the suffix */
  nextpoint=part[2]+2;
  stripsp(nextpoint);
  if ((tmpb=find_literal(nextpoint,&endptr,0))) {
    fprintf(fp,"%s",tmpb);
    free(tmpb);
  }

 LEAVING:
  if (loop_str) free(loop_str);
  while (dataobj) { 
    dtrace=dataobj->next;
    if (dataobj->spec) free(dataobj->spec);
    free(dataobj);
    dataobj=dtrace;
  }
  return count;
} /* output_parse */

char *find_literal(char *datastr,char **endptr,int qflag)
     /* Process string as a literal. Take double quotes "" as "
	and convert control characters indicated by (single) '\'. 
	If qflag set, string must start/end with " and we pass endptr.
        Return pointer or NULL if not found or illegal. */
{
  int i=0,done=0;
  char *str=NULL,*tmp,*nextpoint,*ttmp,c,*begin;

  begin=datastr;
  while ((c=*begin)==' ' || c=='\n' || c=='\r' || c=='\t') begin++;
  if (qflag) {  /* must find start/end quotes (not doubled) */
    if (*begin!='\"') return NULL; /* must start with " */
    *endptr=begin=begin+1;
    while (!done) { /* search for end " */
      while (((c=**endptr))!='"' && (c!='\0')) (*endptr)++;
      if ((**endptr)=='\0') return NULL; /* no closing " */
      if (*((*endptr)+1)=='"') (*endptr) += 2; /* double "" */
      else {
	done=1; /* found end */
	**endptr='\0';
	(*endptr)++;
      }
    }
  }
  done=0;
  nextpoint=begin;
  while (!done) {  /* search for "", \, or end */
    while (((c=*nextpoint)!='\"') && (c!='\\') && (c!='\0')) nextpoint++;
    tmp=nextpoint;
    if (*(tmp)=='\"' && *(tmp+1)!='\"') return NULL; /* can't have single " */
    else if (*(tmp)=='\"' && *(tmp+1)!='\"') tmp += 2; /* double " */
    else if (*(tmp)=='\\') {  /* found \ */
      tmp++;
      if (*tmp=='\\') nextpoint++; /* double \\ */
      else if (*tmp=='n' || *tmp=='r' || *tmp=='t') { /* control code */
	if (*tmp=='n') *(tmp-1)='\n';
	else if (*tmp=='r') *(tmp-1)='\r';
	else if (*tmp=='t') *(tmp-1)='\t';
	ttmp=tmp;
	do {  /* shift string to fill hole */
	  *ttmp=*(ttmp+1);
	  ttmp++;
	} while (*(ttmp-1)!='\0'); 
	nextpoint=tmp;
      }
      else nextpoint++;
    }
    else if (*tmp=='\0') done=1;
  } /* end of while */
  if (nextpoint==begin) return str;
  str=(char *)calloc((size_t)(nextpoint-begin+1),sizeof(char));
  while ((begin+i)<nextpoint) {
    str[i]=begin[i];
    i++;
  }
  return str;
} /* find_literal */

int data_obj_parse(char *datastr,int *code)
   /* parse a string to see whether it applies to vertices/faces/edges
      (return 1, 2, 3, respectively), and find appropriate code. */
{
  switch(datastr[0])
    {
    case 'V':
      {
	if (!strncmp(datastr,"VI",2)) { /* circle index */
	  *code=1;
	  return 1;
	}
	else if (!strcmp(datastr,"VZ")) { /* circle center */
	  *code=2;
	  return 1;
	}
	else if (!strncmp(datastr,"VA",2)) { /* angle sum */
	  *code=3;
	  return 1;
	}
	else if (!strncmp(datastr,"VT",2)) { /* angle target */
	  *code=4;
	  return 1;
	}
	else if (!strncmp(datastr,"VD",2)) { /* circle degree */
	  *code=5;
	  return 1;
	}
	else if (!strncmp(datastr,"VC",2)) { /* circle color */
	  *code=6;
	  return 1;
	}
	else if (!strncmp(datastr,"VM",2)) { /* vertex map */
	  *code=7;
	  return 1;
	}
	else if (!strncmp(datastr,"VR",2)) { /* radii */ 
	  *code=8;
	  return 1;
	}
	else if (!strncmp(datastr,"VS",2)) { /* sharp ftn, map p to q.
						Must specify p and q
						i.e., ratio q/p of radii */
	  char c[2];
	  c[0]=datastr[2];c[1]='\0';
	  int pnum=atoi(c);
	  c[0]=datastr[3];c[1]='\0';
	  int qnum=atoi(c);
	  q=NULL;pp=NULL;
	  if (strlen(datastr)<4 || pnum<0 || qnum<0 || pnum>=3 || qnum>=3
	      || !packdata[pnum].status || !packdata[qnum].status) {
	    sprintf(msgbuf,"usage: output 'VS' option (sharp ftn) "
		    "requires pack numbers p and q.");
	    emsg();
	  }
	  pp=&packdata[pnum];
	  q=&packdata[qnum];
	  *code=9;
	  return 1;
	}
	else if (!strncmp(datastr,"Varg",4)) { /* center argument */
	  *code=10;
	  return 1;
	}

	break;
      }
    case 'F':
      {
	if (!strcmp(datastr,"FI")) { /* face index */
	  *code=1;
	  return 2;
	}
	else if (!strcmp(datastr,"FZ")) { /* corners */
	  *code=2;
	  return 2;
	}
	else if (!strcmp(datastr,"FC")) { /* face color */
	  *code=3;
	  return 2;
	}
	else if (!strcmp(datastr,"FA")) { /* face area */
	  *code=4;
	  return 2;
	}
	else if (!strcmp(datastr,"FV")) { /* face vertices */
	  *code=5;
	  return 2;
	}
	break;
      }
    case 'E':
      {
	if (!strcmp(datastr,"EV")) { /* edge vertices */
	  *code=1;
	  return 3;
	}
	else if (!strcmp(datastr,"EC")) { /* edge color */
	  *code=2;
	  return 3;
	}
	else if (!strcmp(datastr,"EL")) { /* edge length */
	  *code=3;
	  return 3;
	}
	break;
      }
    } /* end of switch */
  return 0; /* no legal code */
} /* data_obj_parse */
	    
int print_vert_obj(struct p_data *p,FILE *fp,int code,int v)
{
  int i=0;
  struct Edgelist *etrace;

  if (code==1) {fprintf(fp,"%d ",v);return 1;}
  if (code==2) {
    CPputd(fp,p->packR_ptr[v].center.re,digs,'e');
    fprintf(fp," ");
    CPputd(fp,p->packR_ptr[v].center.im,digs,'e');
    fprintf(fp," ");
    return 1;
  }
  if (code==3) {
    CPputd(fp,p->packR_ptr[v].curv,digs,'e');
    fprintf(fp," ");
    return 1;
  }
  if (code==4) {
    CPputd(fp,p->packR_ptr[v].aim,digs,'e');
    fprintf(fp," ");
    return 1;
  }
  if (code==5) {
    fprintf(fp,"%d ",p->packK_ptr[v].num+p->packK_ptr[v].bdry_flag);
    return 1;
  }
  if (code==6) {fprintf(fp,"%d ",p->packK_ptr[v].color);return 1;}
  if (code==7 && (etrace=p->vertex_map)) {  /* vertex map for v */
    while (etrace) {
      if (etrace->v==v) {
	fprintf(fp,"%d ",etrace->w);
	i++;
      }
      etrace=etrace->next;
    }
    if (i) return 1;
  }
  if (code==8) {
    CPputd(fp,radius(p,v),digs,'e');
    fprintf(fp," ");
    return 1;
  }
  if (code==9) {
    double shp;
    if (pp && v<q->nodecount) {
      shp=radius(q,v)/radius(pp,v);
      CPputd(fp,shp,digs,'e');
      fprintf(fp," ");
      return 1;
    }
  }
  if (p->hes<=0 && code==10) { /* center argument, eucl/hyp only */
    double ag=Arg(p->packR_ptr[v].center);
    CPputd(fp,ag,digs,'e');
    fprintf(fp," ");
    return 1;
  }
  return 0;
} /* print_vert_obj */

int print_face_obj(struct p_data *p,FILE *fp,int code,int f)
{
  f_data fptr=p->faces[f];

  if (code==1) { /* face index */
    fprintf(fp,"%d ",f);
    return 1;
  }
  if (code==2) { /* corner locations */
    fprintf(fp,"fixup");
    return 1;
  }
  if (code==3) { /* color */
    fprintf(fp,"%d",fptr.color);
    return 1;
  }
  if (code==4) { /* area */
    fprintf(fp,"fixup");
    return 1;
  }
  if (code==5) { /* vertices */
    fprintf(fp,"%d %d %d ",fptr.vert[0],fptr.vert[1],fptr.vert[2]);
    return 1;
  }
  return 0;
} /* print_face_obj */
							       
int print_edge_obj(struct p_data *p,FILE *fp,int code,int v,int w)
{
  if (code==1) { /* edge indices */
    fprintf(fp,"%d, %d ",v,w);
    return 1;
  }
  if (code==2) { /* color */
    fprintf(fp,"fixup");
    return 1;
  }
  if (code==3) { /* length */
    fprintf(fp,"fixup");
    return 1;
  }
  return 0;
} /* print_edge_obj */

int output_call(struct p_data *p,char *datastr)
{
  int i,append=0;
  char *ptr,filename[NAME_MAX],*name_ptr;
  FILE *fp=NULL;
  extern Mobius Mob;

  /* get filename from end of datastr */
  if (!(name_ptr=ptr_to_name_at_end(datastr,&append))
      || !grab_next(&name_ptr,filename) ) {
    sprintf(msgbuf,"output: did not find data and {filename} at end.");
    emsg();
    return 0;
  }
  if (append) fp=fopen(filename,"a");
  else fp=fopen(filename,"w");
  if (!fp) {
    sprintf(msgbuf,"output: failed to open file %s.",filename);
    emsg();
    return 0;
  }
  i=output_parse(p,fp,datastr,Mob);
  fclose(fp);
  return i;
} /* output_call */
