#include  "sa_types.h"
#include "sa_proto.h"

/* Create a new packing in which each component of identified 
"poison" vertices has been replaced by a single circle. The
intended use is to replace components of circles which are too 
deep in the packing to get packed correctly. 

   The parent (original) packing is lost, but the vertex_map 
maintains the links to the vertex indicesrent (original) packing. 
The parents boundaries will, in normal circumstances, remain 
unaffected if the vertices to be removed are separated from it. 

  The poison vertices are identified by util_flag<0. 'seed' is
a vert that should remain in the new packing (e.g., the alpha vert); 
if seed=0 (the usual situation), then choose a convenient point 
neighboring the parent's outer (i.e. first) boundary component.

*/

int plug_it_up(struct p_data *p,int seed)
{
  int hits=0,count=0,i,j,spot,*newold=NULL,bv,iv;
  char buf[64];
  struct Edgelist *trace;
  struct K_data *pK_ptr=p->packK_ptr,*poldK;
  struct p_data *pold=NULL;

/* seed? Must be interior vert for the call to cookie. */
  if (seed<0 || seed >p->nodecount || pK_ptr[seed].util_flag<0
      || pK_ptr[seed].bdry_flag)
    seed=0; /* given seed is poison or bdry; pick new one. */
  if (!seed) /* look for interior neighbor of vert in first bdry 
		comp, else first non-poison interior vert. */
    {
      if (p->num_bdry_comp) 
	{
	  if (pK_ptr[(bv=p->bdry_starts[1])].num>1 
	      && pK_ptr[(iv=pK_ptr[bv].flower[1])].util_flag>=0)
	    seed=iv; /* should be interior nghb */
	  if (!seed)
	    {
	      bv=pK_ptr[bv].flower[0];
	      while (!seed && bv!=p->bdry_starts[1])
		{
		  if (pK_ptr[bv].num>1 
		      && pK_ptr[(iv=pK_ptr[bv].flower[1])].util_flag>=0)
		    seed=iv;
		  bv=pK_ptr[bv].flower[0];
		}
	    }
	}
      if (!seed) /* no luck? choose smallest non-poison vertex */
	for (j=1;!seed && j<=p->nodecount;j++)
	  if (pK_ptr[j].util_flag>=0) 
	    seed=j;
    }

/* save original pack to identify its bdry */
  if (!(pold=init_packing()) || !copy_pack(p,pold)
      || !complex_count(pold,1)) return 0;
  poldK=pold->packK_ptr;
  
/* cookie */
  sprintf(buf,"-v %d",seed);
  if (!seed || !(cookie_cutter(p,buf,&newold)) 
      || !complex_count(p,0) || !p->num_bdry_comp) /* nothing to plug? */
    return 0;
  pK_ptr=p->packK_ptr;
  
/* plug holes not in the parent packing */
  for (i=1;i<=p->num_bdry_comp;i++)
    {
      /* check if this bdry has vert in bdry of original packing */
      if (!poldK[newold[(spot=p->bdry_starts[i])]].bdry_flag)
	{
	  hits=0;
	  j=pK_ptr[spot].flower[0];
	  while (!hits && j!=spot)
	    {
	      j=pK_ptr[j].flower[0];
	      if (poldK[newold[j]].bdry_flag)
		hits++;
	    }
	  if (!hits) /* nope, this is a new 'hole'; plug it */
	  if (!ideal_bdry_node(p,p->bdry_starts[i]))
	    return count;
	  count++;
	}
    } /* should have plugged all new holes */
  complex_count(p,0);
  facedraworder(p,0);
  set_aim_default(p);
  set_default_overlaps(p);

/* define vertex map */
  p->vertex_map=trace=(struct Edgelist *)
    calloc((size_t)1,sizeof(struct Edgelist));
  for (j=1;j<=p->nodecount;j++)
    if ((i=newold[j]))
      {
	trace=trace->next=(struct Edgelist *)
	  calloc((size_t)1,sizeof(struct Edgelist));
	trace->v=j;
	trace->w=i;
      }
  trace=p->vertex_map;
  p->vertex_map=p->vertex_map->next;
  free(trace);

/* clean up and leave */
  free(newold);
  free_p_data(&pold);
  return count;
} /* plut_it_up */

