#include "sa_types.h"
#include "sa_proto.h"

int post_any_polygon(FILE *fp,struct p_data *p,
		     struct Vertlist *vertlist,int eflag,int ecol,int col)
     /* calling routine must make sure first vertex is repeated at end
	if a closed polygon is desired. If the 'filled' option is
	specified, then a 'closepath' will be issued to close it up in
	any case. */
{
  int v,w;
  complex a,b;
  hyp_geodesic hg;
  struct Vertlist *vtrace;
  struct Centlist *centlist=NULL,*ctrace;
  struct Pathlist *plist=NULL;
  struct R_data *pR_ptr=p->packR_ptr;

  if (!(vtrace=vertlist) || !(vtrace->next)
      || (v=vtrace->v)<1 || v>p->nodecount) return 0;
  centlist=ctrace=(struct Centlist *)
    calloc((size_t)1,sizeof(struct Centlist));
  ctrace->z=pR_ptr[vtrace->v].center;
  vtrace=vtrace->next;
  while (vtrace)
    {
      if ((v=vtrace->v)>0 && v<=p->nodecount)
	{
	  ctrace=ctrace->next=(struct Centlist *)
	    calloc((size_t)1,sizeof(struct Centlist));
	  ctrace->z=pR_ptr[v].center;
	}
      vtrace=vtrace->next;
    }
  ctrace=centlist;
  a=centlist->z;
  fprintf(fp,"gs   %f %f m\n",(singledouble)a.re,(singledouble)a.im);
  if (p->hes<0)
    {
      while (ctrace && ctrace->next)
	{
	  a=ctrace->z;
	  b=ctrace->next->z;
	  if (a.re!=b.re || a.im!=b.im) /* eat repeats */
	    {
	      h_geodesic(a,b,&hg);
	      if (hg.line_flag) /* use straight line */
		{
		  fprintf(fp,"%f %f l\n",(singledouble)b.re,
			  (singledouble)b.im);
		}
	      else 
		{
		  if (hg.z1.re==a.re && hg.z1.im==a.im) 
				/* right direction*/
		    fprintf(fp,"%f %f %f %f %f a\n",
			    (singledouble)hg.c.re,
			    (singledouble)hg.c.im,(singledouble)hg.rad,
			    (singledouble)(hg.arg1*degPI),
			    (singledouble)(hg.arg2*degPI));
		  else fprintf(fp,"%f %f %f %f %f an\n",
			       (singledouble)hg.c.re,
			       (singledouble)hg.c.im,(singledouble)hg.rad,
			       (singledouble)(hg.arg2*degPI),
			       (singledouble)(hg.arg1*degPI));
		}
	    }
	  ctrace=ctrace->next;
	} /* end of while */
      if (ecol!=FG_COLOR) /* edge colors? */
	fprintf(fp,"%f %f %f srgb ",
		(singledouble)((double)red[ecol]/255),
		(singledouble)((double)green[ecol]/255),
		(singledouble)((double)blue[ecol]/255));
      if ((eflag & 2) && col!=FG_COLOR) 
	fprintf(fp,"cp\ngs %f %f %f srgb fill gr ",
		(singledouble)((double)red[col]/255),
		(singledouble)((double)green[col]/255),
		(singledouble)((double)blue[col]/255));
      else if (eflag & 2)
	fprintf(fp,"cp\ngs gry  sg fill gr\n"); 
      fprintf(fp,"s gr\n"); /* draw border */
      return 1;
    } /* end of hyp case */
  else if (p->hes==0)
    {
      while (ctrace && ctrace->next)
	{
	  a=ctrace->z;
	  b=ctrace->next->z;
	  if (a.re!=b.re || a.im!=b.im) /* eat repeats */
	    fprintf(fp,"%f %f l\n",(singledouble)b.re,(singledouble)b.im);
	  ctrace=ctrace->next;
	} /* end of while */
      if (ecol!=FG_COLOR) /* edge colors? */
	fprintf(fp,"%f %f %f srgb ",
		(singledouble)((double)red[ecol]/255),
		(singledouble)((double)green[ecol]/255),
		(singledouble)((double)blue[ecol]/255));
      if ((eflag & 2) && col!=FG_COLOR) 
	fprintf(fp,"cp\ngs %f %f %f srgb fill gr ",
		(singledouble)((double)red[col]/255),
		(singledouble)((double)green[col]/255),
		(singledouble)((double)blue[col]/255));
      else if (eflag & 2)
	fprintf(fp,"cp\ngs gry  sg fill gr\n"); 
      fprintf(fp,"s gr\n"); /* draw border */
      return 1;
    } /* end of eucl case */
  else if (p->hes>0) /* fixup: will sph 'fill' screw up if not convex? */
    {
      if (!(plist=sph_poly_list(p->screen,centlist))) return 0;
      post_s_convex(fp,plist,eflag,ecol,col);
      return 1;
    }
  return 0;
} /* post_any_polygon */

