#include "sa_types.h"
#include "sa_proto.h"

/* Printing transition info in matlab readable form. Put data in
" v w t(v,w)" form to save space; matlab can create matrix. Currently
for eucl; hyp formulae need to be updated; spherical being
computed by Matt Cathey. File already open. 'opt' for various
options: e.g., absorbing nodes, etc. 
Return count or 0 on failure. */

int print_transition(struct p_data *p,FILE *fp,int opt)
{
  int v,j,i,*absorb=NULL,digits=1,count=0;
  double t,nc;
  struct K_data *pK_ptr=p->packK_ptr;

  if (!p->status || p->hes!=0 || !fp) return 0;
  while ((pow(10.,(double)digits))*toler < 0.1 
	 && digits<MAX_ACCUR) digits++;
  fprintf(fp,"\n%% transition matrix, euclidean, nodecount %d\n\n"
	  "Tdata = [\n",
	  p->nodecount);
  /* identify absorbing nodes */
  absorb=(int *)calloc((size_t)(p->nodecount+2),sizeof(int));
  if (opt==1) { /* bdry are absorbing */
    for (i=1;i<=p->nodecount;i++)
      if (pK_ptr[i].bdry_flag) absorb[i]=1;
  }
  for (v=1;v<=p->nodecount;v++) {
    if (!absorb[v] && (nc=node_conductance(p,v))>0.0)
      for (j=0;j<(pK_ptr[v].num+pK_ptr[v].bdry_flag);j++) {
	t=edge_conduct(p,v,j)/nc;
	fprintf(fp," %d %d ",v,pK_ptr[v].flower[j]);
	CPputd(fp,t,digits,'e');
	fprintf(fp,"\n");
	count++;
      }
  }
  fprintf(fp,"  ]\n\ndisp('Data for transition matrix is in Tdata')\n");  
  /* finished */
  if (absorb) free(absorb);

  /* now add matlab code to create the matrix itself */
  /* fprintf(fp,"disp('press any key to continue')\npause\n"); */
  fprintf(fp,"clear T\n N=max(max(Tdata(:,1)),max(Tdata(:,2)));\n"
	  "T=zeros(N,N);\n"
	  "for j=1:N,\nT(Tdata(j,1),Tdata(j,2))=Tdata(j,3);\nend\n");
  return count;
} /* print_transition */

int transition_data_call(struct p_data *p,char *datastr) 
     /* put transition data in matlab form in file */
{
  int append=0,i;
  char *ptr,filename[NAME_MAX],*name_ptr;
  FILE *fp=NULL;
  
  if (!(name_ptr=ptr_to_name_at_end(datastr,&append))
      || !grab_next(&name_ptr,filename) ) {
    sprintf(msgbuf,"output: did not find data and {filename} at end.");
    emsg();
    return 0;
  }
  if (append) fp=fopen(filename,"a");
  else fp=fopen(filename,"w");
  if (!fp) {
    sprintf(msgbuf,"output: failed to open file %s.",filename);
    emsg();
    return 0;
  }
  i=print_transition(p,fp,1); /* opt=1 means bdry are absorbing;
				 may need other options in future */
  fclose(fp);
  return i;
} /* transition_data_call */
  
