#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>

#include "rand_mesh.h"

void mesh_storage_class::print_node(long node){
  printf("node %ld\t ndx:%ld ct::%ld fl:%ld hd:%ld\n", node, nodes[node].link_index, nodes[node].link_count, nodes[node].flags, nodes[node].link_head);
};

void mesh_storage_class::print_link(long link){
  printf("link %ld\t n:%ld  p:%ld n:%ld o:%ld \n", link, links[link].next, links[link].prev, links[link].node, links[link].owner);
};


void mesh_storage_class::print_summary_data(){
  printf("There are %ld nodes and %ld links.\n", node_count, link_count);
};

void mesh_storage_class::print_node_data(long node){
  long l, nlink;
  bool bordertest;
  
  bordertest = (nodes[node].flags==MBORDER_FLAG);
  l = nodes[node].link_count;

  long vertex, faces, head;
  
  vertex = node;
  if (bordertest) faces = l-1;
  else
    faces = l;
  
  printf("%ld\t%d\t", node, faces);
  // if boundary mark
  
  nlink = links[nodes[node].link_index].next;
  head = links[nlink].node;
  for(long j = 0; j<l; j++){
    // new -- only print list
    printf( " %ld", links[nlink].node);
    //    printf(" (link %ld, node %ld)", nlink, links[nlink].node);
    nlink = links[nlink].next;
  }
  
  if (!bordertest) printf(" %ld", head);
  printf("\n");
};


void mesh_storage_class::init(long n, long l){
  // make room for storage
  nodes = (node_struct *)new node_struct[n];
  links = (link_struct *)new link_struct[l];

  // now initialize data 
  free_node_count = n-1;
  node_count = 0;

  free_link_count = l-1;
  free_link_start = 1;
  link_count = 0;

  // initialize free chain of links
  link_struct templink;
  templink.owner = templink.node = 0; templink.prev = 0; templink.next = 2; 
  for (long i=1; i<free_link_count-1; i++) {
    links[i] = templink;
    templink.next++;
  }
  templink.next = 0;
  links[free_link_count-1] = templink;

  // initialize nodes data
  node_struct tempnode;
  tempnode.link_head = tempnode.link_count = tempnode.flags = tempnode.link_index = 0;
  for (long i=0; i<n; i++) {
    nodes[i]=tempnode;
  }

};
// void mesh_storage_class::delete_link(long node){};

long mesh_storage_class::new_link_index(){
  if (free_link_count <= 0){
    printf("freelink error\n"); exit(0); 
  }
  else {
    link_count++; free_link_count--;
    long tlink = free_link_start;
    free_link_start = links[free_link_start].next;
    links[tlink].owner = links[tlink].node = 0;
    links[tlink].next=tlink; links[tlink].prev = tlink;
    return(tlink);
  }  
};

void mesh_storage_class::front_link_insert(long owner, long node,  long link){
  long tlink = new_link_index();
  long nextlink = links[link].next;

  links[tlink].node = node; links[tlink].owner = owner;
  links[tlink].next = nextlink; links[tlink].prev = link;

  links[link].next = tlink; links[nextlink].prev = tlink;
  
  nodes[links[link].owner].link_count++;

  // printf("inserting front link from %ld to %ld at link %ld ::", owner, node, link);
  // print_link(tlink);	   
};

void mesh_storage_class::back_link_insert(long owner, long node, long link){
  long tlink = new_link_index();
  long prevlink = links[link].prev;
  link_struct newlink;
  newlink.node = node; newlink.owner = owner;
  newlink.next = link; newlink.prev = prevlink;
  links[tlink] = newlink;

  links[link].prev = tlink; links[prevlink].next = tlink;

  nodes[links[link].owner].link_count++;
  // printf("inserting back link from %ld to %ld at link %ld ::", owner, node, link);
  // print_link(tlink);
};

long mesh_storage_class::new_node_index(){
  long tnode;
 
  if(free_node_count > 0) {
    tnode = ++node_count; free_node_count--;
    // now add empty link for ring
    nodes[tnode].link_index = new_link_index();
    nodes[tnode].flags = nodes[tnode].link_count = 0;

    links[nodes[tnode].link_index].node = MROOT_LINK_FLAG;
    links[nodes[tnode].link_index].owner = tnode;
  }
  else {
    printf("Node allocation error\n"); exit (0);
  }
  // printf("Making new node :: node %ld :: ", tnode);
  // print_node(tnode);
  // printf("     root link_index :: "); 
  // print_link(nodes[tnode].link_index);
  return(tnode);

};
long mesh_storage_class::new_border_node(long fromnode, long tonode){
  long tnode;
  tnode = new_node_index();
  back_link_insert(tonode, tnode, nodes[tonode].link_index);
  front_link_insert(tnode, tonode, nodes[tnode].link_index);

  back_link_insert(tnode, fromnode, nodes[tnode].link_index);
  front_link_insert(fromnode, tnode, nodes[fromnode].link_index);

  nodes[tnode].flags = MBORDER_FLAG;

  // printf("making new border node between %ld and  %ld ::", fromnode, tonode);
  // print_link(tnode);	     
  return tnode;
};

void mesh_storage_class::set_interior(long node){
  // set node state to interior
  // ALSO set head link to interior?? NO!!!!  this is always a root link flag
  nodes[node].flags = MINTERIOR_FLAG;
};
long mesh_storage_class::get_first_link_node(long node){
  long tlink;
  tlink = links[nodes[node].link_index].next;
  return(links[tlink].node);
}; // returns primary link

int mesh_storage_class::border_test(long node){
  return( nodes[node].flags==MBORDER_FLAG);
}; //  

long mesh_storage_class::count_links(long node){
  return(nodes[node].link_count);
};

void tmesh_class::init(long nodes)
{
  // init storage class here!!
  mdata.init(2*nodes, 8*nodes);


  // now need to make the three nodes

  long node1, node2, node3;
  node1 = mdata.new_node_index();
  node2 = mdata.new_node_index();
  node3 = mdata.new_node_index();

  mdata.back_link_insert(node1, node3, mdata.nodes[node1].link_index); 
  mdata.front_link_insert(node1, node2, mdata.nodes[node1].link_index);
  mdata.back_link_insert(node2, node1, mdata.nodes[node2].link_index); 
  mdata.front_link_insert(node2, node3, mdata.nodes[node2].link_index);
  mdata.back_link_insert(node3, node2, mdata.nodes[node3].link_index); 
  mdata.front_link_insert(node3, node1, mdata.nodes[node3].link_index);

  node_count = 3;
  
}

long tmesh_class::get_link_count(){
  return (mdata.link_count - mdata.node_count); // theres one utility link per node
}

long tmesh_class::get_node_count(){
  return (mdata.node_count);
}

long tmesh_class::next_border_node(long node)
{ 
  // called by main -- here for consistency

  return(mdata.get_first_link_node(node));
}


long tmesh_class::insert_border_node(long node)
{
  long newnode;
  
  node_count++;
  // printf("tmesh::insert_border_node at node \n", node);
  // errors handled inside call below
  newnode = mdata.new_border_node(node, mdata.get_first_link_node(node));
  return(newnode);
}

int mesh_storage_class::seek_link(long node, long linknode){
  
  
  long linkiter, linkroot;;
  linkroot = nodes[node].link_index;
  linkiter = links[linkroot].next;
  while( (linkiter != linkroot) && (links[linkiter].node != linknode) ){
    linkiter = links[linkiter].next;
  }
  // printf("seeking link from %ld to %ld: error: %ld\n", node, linknode,links[linkiter].node == linknode );  
  return (links[linkiter].node != linknode);
};

long tmesh_class::join_border_node(long node)
{

  // must verify there are enough connections -- need three at next;
  // also be sure to not make degenerate! check to see that next is different!
  // for triangulations need to leave at least three boundary nodes

  long degen_test_ok = 1;
  // simply need to check the there is not alread a link to join node
  // if there is it will be the second link!
  // don't forget to check in both directions

  long nextnode = mdata.get_first_link_node(node);
  long nextnextnode = mdata.get_first_link_node(nextnode);


  // search for nextnode in link node list! Don't take a shortcut
  degen_test_ok = mdata.seek_link(node, nextnextnode); // returns 1 if found 0 otherwise

  if (degen_test_ok)
    {
      // printf ("joining: from %ld to %ld \n", node, nextnextnode);
      // printf("next next next %ld %ld\n", node, nextnextnode);
      // printf("tmesh::join_border_node. at node %ld\n", node);
      mdata.set_interior(nextnode);
      long tlink;
      mdata.front_link_insert(node, nextnextnode, mdata.nodes[node].link_index);
      mdata.back_link_insert(nextnextnode, node, mdata.nodes[nextnextnode].link_index);
      return (nextnextnode);
    }
  else
    {
      // printf("tmesh::join_border_node. unable to join \n");
      // printf("there are %ld nodes. and link between %ld and %ld\n", testcount, node, nextnextnode);
      return(node);
    }
  return(node);
}

void tmesh_class::output_node_data(long node)
{
  long l, nlink;
  l = mdata.nodes[node].link_count;
  printf("%ld:", node);
  nlink = mdata.links[mdata.nodes[node].link_index].next;
  for(long j = 0; j<l; j++){
    printf(" %ld,", nlink);
    nlink = mdata.links[nlink].next;
  }
  printf("\n");
}

void tmesh_class::output_text()
{
  printf("NODECOUNT:\t%ld\n\n", node_count);
  printf("GEOMETRY:\teuclidean\n\n");
  printf("FLOWERS:\n\n");

  // loop through the nodes
  // list the links
  for (long i=1; i<= node_count; i++) {
    mdata.print_node_data(i);
  }
  
  printf("\nEND\n");
    
};

int distribution()
{
  // This routine determines which action choices are SKIP ADD JOIN
  // need to add multiple operations and directional (left/right) to make more symetric

  long r_val = random();
  int step = (RAND_MAX) / 3;
  return (r_val/step + 1);
};

tmesh_class random_mesh(long size){

  tmesh_class mesh;
  mesh.init(size);

  long c_border_node = 1;
  int mesh_action;

  while (mesh.node_count < size) 
    {
      long save_node; // used for debug return info
      mesh_action = distribution();
      
      if (mesh.node_count <3) mesh_action = M_ADD;
      // prevent mesh from getting too small
      switch(mesh_action) {
      case M_SKIP: 
	save_node = c_border_node;
	c_border_node = mesh.next_border_node(c_border_node); 
	// printf("SKIP %d : %d\n",save_node, c_border_node); mesh.output_text();
	break;
      case M_ADD: 
	save_node = c_border_node;
	c_border_node = mesh.insert_border_node(c_border_node);
	// printf("ADD %d : %d\n",save_node, c_border_node); mesh.output_text();
	break;
      case M_JOIN:
	save_node = c_border_node;
	c_border_node = mesh.join_border_node(c_border_node); 
	// printf("JOIN %d : %d\n",save_node,c_border_node); mesh.output_text();
	break;
 }
      // 
      //printf("there are %ld nodes and %ld links.  ", mesh.node_count, mesh.link_count);
      //printf("there are %ld free nodes and %ld free links. \n ", mesh.free_node_count, mesh.free_link_count);

    }
  return mesh;
};
