#include "sa_types.h"
#include "sa_proto.h"

/* Read in color info from a named file. */

extern unsigned char red[],green[],blue[];
extern int color_ramp_size,bc_min,bc_max,bc_offset;

int read_rgb_file(char *rgb_filename)
{
  char keyword[81],count=0;
  int rd,gn,bl,indx,i;
  FILE *fp;

  if (!strlen(rgb_filename) || (fp=fopen(rgb_filename,"r"))==NULL)
    {
      fprintf(stderr,"\nerror: color file %s not found.",rgb_filename);
      return 0;
    }
  fprintf(stdout,"\nColor file %s is being read.",rgb_filename);
  while (fscanf(fp,"%63s",keyword)!=EOF)
    {
      stripsp(keyword);
      if( *keyword!='#' && *keyword!='\n' 
	  && *keyword!='\0' && *keyword!='\t')
	{

/* ramp size given? */
          if (!strncmp(keyword,"RAMP:",5) 
              && fscanf(fp,"%d",&color_ramp_size)
	      && (count=color_ramp_size)>1)
	    blue_to_red_ramp(1);

/* list of indices with red/green/blue levels? */
          else if (!strncmp(keyword,"LIST_IND_R_G_B:",15))
	    {
	      bc_min=NUM_COLORS;
	      bc_max=0;
	      while( (fgets(keyword,80,fp)!=NULL)
                   && (strcmp(keyword,"END")!=0) )
		{
		  sscanf(keyword,"%d %d %d %d",&indx,&rd,&gn,&bl);
		  if (indx>=0 && indx<=NUM_COLORS 
		      && rd>=0 && rd<=NUM_COLORS
		      && gn>=0 && gn<=NUM_COLORS 
		      && bl>=0 && bl<=NUM_COLORS)
		    {
		      red[indx]=rd;
		      green[indx]=gn;
		      blue[indx]=bl;
		      if (indx<bc_min) bc_min=indx;
		      if (indx>bc_max) bc_max=indx;
		      indx=-1; /* re-initialize index */
		      count++;
		    }
		}

	      if ((bc_offset=bc_min-bc_max-1)<0
		  && bc_offset > -NUM_COLORS/2
		  && bc_min > -bc_offset)
		{
		  for (i=bc_min;i<=bc_max;i++)
		    {
		      red[i+bc_offset]=
			(unsigned char)(.75*red[i]);
		      green[i+bc_offset]=
			(unsigned char)(.75*green[i]);
		      blue[i+bc_offset]=
			(unsigned char)(.75*blue[i]);
		    }
		}
	      else bc_offset=0;
	    }
	}
    } /* end of while */
  if (fp) fclose(fp);
  return count;
} /* read_rgb_file */
		       
