/* sa_call.c    CirclePack dependent, bureaucratic routines for
   calling more general routines in the libraries. */

#include "sa_head.h"
#include "Interface.h"

int pack_num(struct p_data *p)
     /* return pack number from pointer */
{
  int i;for (i=0;i<NUM_PACKS;i++) if (&packdata[i]==p) return i;return 0;
}/* pack_num */

struct Vertlist *node_link_parse(struct p_data *p,char *dpoint,
				 char **endptr,int *hits)
/* call to more general routine */
{
  return (Node_link_parse(p,dpoint,endptr,hits,&Vlist,&Elist,&Flist,
			  &region,pathlist,pathlength));
} /* node_link_parse */

struct Vertlist *face_link_parse(struct p_data *p,char *dpoint,
				 char **endptr,int *hits)
/* call to more general routine */
{
  return (Face_link_parse(p,dpoint,endptr,hits,&Vlist,&Elist,&Flist,
			  &region,pathlist,pathlength));
} /* face_link_parse */

struct Edgelist *node_pair_link(struct p_data *p,char *dpoint,
				char **endptr,int *hits)
/* call to more general routine */
{
  return (Node_pair_link(p,dpoint,endptr,hits,&Vlist,&Elist,&Flist,
			 &region,pathlist,pathlength));
} /* node_pair_link */

int cmd_search(char *datastr,int *shift)
     /* Process datastr to pick out full cmd; takes care of line 
continuations, etc. Shift gives end of cmd. */
{
  int flag=0,done=0,high_water=0;
  char *nextpoint;

  *shift=0;
  nextpoint=datastr;
  while (!done)
    {
      while (nextpoint[flag]!='\n' 
	     && nextpoint[flag]!='\0'
	     && nextpoint[flag]!='\\') flag++;
      if (nextpoint[flag]=='\0') 
	{
	  *shift=flag;
	  if (flag==0) return 0; /* vacuous command */
	  done=1;
	}
      else if (nextpoint[flag]=='\n')
	{
	  nextpoint[flag]='\0';
	  done=*shift=flag;
	  if (flag==0) return 0; /* vacuous command */
	}
      else /* continuation indicator */
	{
	  high_water=flag+1;
	  while (nextpoint[high_water]==' '
		 || nextpoint[high_water]=='\t') high_water++;
	  if (nextpoint[high_water]=='\n') flag=high_water+1;
	  else /* improper termination of cmd, but okay */
	    {
	      nextpoint[high_water]='\0';
	      *shift=high_water;
	      done=1;
	    }
	}
    }
  flag=0;
  while (flag<*shift) /* replace line breaks/continues by spaces */
    {
      if (nextpoint[flag]=='\\' || nextpoint[flag]=='\n') 
	nextpoint[flag]=' ';
      flag++;
    }
  return flag;
} /* cmd_search */

int match_call(struct p_data *p,char *datastr)
     /* call to put two packs in register */
{
  int qnum=0,v=0,w=0,V=0,W=0,err_flag=0;
  double err=0.0;
  char *ptr;
  Mobius M;
  struct p_data *q=NULL;
  
  ptr=datastr;
  stripsp(ptr);
  if (ptr[0]!='-' || ptr[1]!='q' || (qnum=atoi(ptr+2))<0 || qnum>=NUM_PACKS
      || sscanf(ptr+4,"%d %d %d %d",&v,&w,&V,&W)<2
      || v<1 || v>p->nodecount || w<1 || w>p->nodecount
      || V>(q=&packdata[qnum])->nodecount || W>q->nodecount
      || (V<1 && p->status && (V=vertex_translation(p->vertex_map,v))<1)
      || (W<1 && (W=vertex_translation(p->vertex_map,w))<1) )
    {
      sprintf(msgbuf,"match usage: -q{pack#} v w {V W}. "
	      "No pack number given or vertices improper.\n");
      emsg();
      return 0;
    }
  if (!match(p,q,v,w,V,W,&err,&err_flag,&M) || err_flag)
    {
      sprintf(msgbuf,"match: something went wrong in generating the "
	      "Mobius transformation.");
      emsg();
      return 0;
    }
  apply_Mobius(p,"a",1,M);
  return 1;
} /* match_call */

int blend_call(struct p_data *p,char *datastr)
{
  int qnum,n,v;
  char *nextpoint;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || strlen(next)<3
      || next[0]!='-' || next[1]!='q' || (qnum=atoi(next+2))<0
      || qnum>=NUM_PACKS || sscanf(nextpoint,"%d %d",&v,&n)!=2)
    {
      sprintf(msgbuf,"blend usage: -q{pack#} v n.");
      emsg();
      return 0;
    }
  return (blend(p,&packdata[qnum],v,n));
} /* blend_call */

int plug_call(struct p_data *p,char *datastr)
{
  int i,hits;
  char *nextpoint,*endptr;
  struct Vertlist *vlist=NULL,*trace;
  struct K_data *pK_ptr=p->packK_ptr;

  nextpoint=datastr;
  if (!(vlist=node_link_parse(p,nextpoint,&endptr,&hits)))
    return 0;
  for (i=1;i<=p->nodecount;i++) pK_ptr[i].util_flag=0;
  trace=vlist;
  while (trace)
    {
      pK_ptr[trace->v].util_flag=-1;
      trace=trace->next;
    }
  vert_free(&vlist);
  return (plug_it_up(p,0));
} /* plug_call */

int path_construct_call(struct p_data *p,char *datastr)
{
  int flag=0,v=0;
  char next[256],*nextpoint;

  nextpoint=datastr;
  stripsp(nextpoint);
  if (*nextpoint=='-')
    {
      grab_next(&nextpoint,next);
      if (next[1]=='i') flag=1;
    }
  if (!(v=grab_one_vert(p,&nextpoint)))
    return 0;
  if (Vlist) vert_free(&Vlist);
  Vlist=path_construct(p,flag,nextpoint);
  return 1;
} /* path_construct_call */

int get_put_data_call(struct p_data *p,char *datastr,int rev_flag)
/* to transfer various data from p to target pack. datastr
must begin with -q{pnum}, where pnum is the target pack, then have 
flags: -t means to use the vertex_map to translate indices (default
is not to translate), -{f} indicating desired data. */
{
  int qnum,trans_flag=0,hits=0;
  char *ptr,flag;
  char next[BUFSIZE];

  ptr=datastr;
  if (!grab_next(&ptr,next) || next[0]!='-' || next[1]!='q' 
      || (qnum=atoi(next+2)) <0 || qnum>=NUM_PACKS)
    {
      sprintf(msgbuf,"get/put_data usage: -q{pack#} -{flag} {v}. "
	      "No pack number given.\n");
      emsg();
      return 0;
    }
  stripsp(ptr);
  while (grab_next(&ptr,next) && next[0]=='-')
    {
      if ((flag=next[1])=='t') trans_flag=1; /* want to translate indices */
      else 
	if (get_put_data(p,&packdata[qnum],ptr,flag,
			 rev_flag,trans_flag))
	  hits++;
    } /* end of looking for data specifications */
  if (!hits) 
    {
      sprintf(msgbuf,"get/put_data usage: -q{pack#} -{flag} {v}. "
	      "The flag for 'translation' is -t, data flags are:\n"
	      "  r=radii,z=centers,a=aims,s=angle sums,c=circle "
	      "colors,m=mark,f=face colors.");
      emsg();
      return 0;
    }
  return (hits);
} /* get_put_data_call */
  

int cir_invert_call(struct p_data *p,char *datastr)
     /* map outside of one circle onto inside of second target circle */
{
  int err_flag=0,v1,v2,hits;
  double CPrad1,CPrad2;
  complex ctr1,ctr2;
  char *nextpoint,*endptr;
  struct Vertlist *vlist;
  Mobius mob;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next))
    return 0;
  if (next[0]=='-') /* option -u means target is the unit circle */
    {
      if (next[1]!='u') return 0; 
      else 
	{
	  ctr2.re=ctr2.im=0.0;
	  CPrad2=1.0;
	}
      if (!(vlist=node_link_parse(p,nextpoint,&endptr,&hits)))
	return 0;
      v1=vlist->v;
      vert_free(&vlist);
      if (p->hes<0) geom_to_e_pack_convert(p); 
               /* have to convert to euclidean */
      ctr1=p->packR_ptr[v1].center;
      CPrad1=p->packR_ptr[v1].rad;
      if (p->hes>0) /* spherical */
	s_to_e_data(ctr1,CPrad1,&ctr1,&CPrad1);
    }
  else /* target circle is vert v2 */
    {
      if (!(vlist=node_link_parse(p,nextpoint,&endptr,&hits))
	  || !(vlist->next) || (v1=vlist->v)==(v2=vlist->next->v)) 
	{
	  vert_free(&vlist); /* didn't get two distinct verts */
	  return 0; 
	}
      vert_free(&vlist);
      if (p->hes<0) geom_to_e_pack_convert(p); 
               /* have to convert to euclidean */
      ctr1=p->packR_ptr[v1].center;
      ctr2=p->packR_ptr[v2].center;
      CPrad1=p->packR_ptr[v1].rad;
      CPrad2=p->packR_ptr[v2].rad;
      if (p->hes>0) /* spherical */
	{
	  s_to_e_data(ctr1,CPrad1,&ctr1,&CPrad1);
	  s_to_e_data(ctr2,CPrad2,&ctr2,&CPrad2);
	}
    }
      
  mob=cir_invert(ctr1,CPrad1,ctr2,CPrad2,&err_flag);
  if (err_flag || !apply_Mobius(p,"a",1,mob)) 
    return 0;
  return 1;
} /* cir_invert_call */

int molify_call(struct p_data *p,char *datastr)
{
  int n=1,hits;
  char next[32],*nextpoint,*endptr;
  struct Vertlist *vertlist;

  nextpoint=datastr;
  stripsp(nextpoint);
  if (nextpoint[0]=='-' && nextpoint[1]=='i')
    {
      nextpoint += 2;
      grab_next(&nextpoint,next);
      if ((n=atoi(next))<1) n=1;
    }
  vertlist=node_link_parse(p,nextpoint,&endptr,&hits);
  n=molify_pack_centers(p,n,vertlist);
  vert_free(&vertlist);
  return n;
} /* molify_call */

int cookie_call(struct p_data *p,char *datastr)
{
  int i,j,w,orig_nodecount,v,ind,ocount=0;
  int *newold=NULL,*oldnew=NULL;
  double angle;
  char *nextpoint;
  struct xyzPoint *xyz_data=NULL;
  struct Overlaps *overlaps=NULL,*otrace,*btrace;
  struct Edgelist *trace;
  struct K_data *pK_ptr;

  pK_ptr=p->packK_ptr;
  nextpoint=datastr;
  orig_nodecount=p->nodecount;
  /* overlaps saved as linked list */
  if (p->overlap_status) 
    {
      overlaps=(struct Overlaps *)
	calloc(1,sizeof(struct Overlaps));
      otrace=overlaps;
      for (v=1;v<=p->nodecount;v++)
	for (j=0;j<(pK_ptr[v].num+pK_ptr[v].bdry_flag);j++)
	  if ( v<pK_ptr[v].flower[j]
	       && (angle=pK_ptr[v].overlaps[j])!=1.0 )
	    {
	      ocount++;
	      otrace=otrace->next=(struct Overlaps *)
		calloc(1,sizeof(struct Overlaps));
	      otrace->v=v;
	      otrace->w=pK_ptr[v].flower[j];
	      otrace->angle=angle;
	    }
      free_overlaps(p);
      if (!ocount) {free(overlaps);overlaps=NULL;} /* no overlaps */
    }
  /* put xyz data pointer aside to avoid loosing it in alloc_pack_space. */
  xyz_data=p->xyzpoint;
  p->xyzpoint=NULL;

  /* now do the cookie job */
  if (!(i=cookie_cutter(p,nextpoint,&newold))) return 0;
  pK_ptr=p->packK_ptr;

  if (newold) /* set up vertex_map */
    {
      p->vertex_map=trace=(struct Edgelist *)
	calloc((size_t)1,sizeof(struct Edgelist));
      for (j=1;j<=p->nodecount;j++)
	if ((w=newold[j]))
	  {
	    trace=trace->next=(struct Edgelist *)
	      calloc((size_t)1,sizeof(struct Edgelist));
	    trace->v=j;
	    trace->w=w;
	  }
      trace=p->vertex_map;
      p->vertex_map=p->vertex_map->next;
      free(trace);
    }
  p->xyzpoint=xyz_data; /* restore xyz data */
  if (p->xyzpoint && p->vertex_map) { /* reset xyz data */
    struct xyzPoint *new_xyz;
    new_xyz=(struct xyzPoint *)
      calloc((size_t)(p->nodecount+1),sizeof(struct xyzPoint));
    trace=p->vertex_map;
    while (trace) {
      new_xyz[trace->v]=p->xyzpoint[trace->w];
      trace=trace->next;
    }
    free(p->xyzpoint);
    p->xyzpoint=new_xyz;
  }
  if (overlaps && alloc_overlaps(p)) /* reset saved overlaps */
    {
      oldnew=(int *)calloc((size_t)(orig_nodecount+1),sizeof(int));
      for (i=1;i<=p->nodecount;i++)
	oldnew[newold[i]]=i;
      /* change indices to their new values, throw out data no longer
	 in the new complex. Remember, first overlap is empty, but
         we need to keep it as handle to linked list as we prune it. */
      btrace=overlaps;
      otrace=overlaps->next;
      while (otrace)
	{
	  if (!(otrace->v=oldnew[otrace->v])
	      || !(otrace->w=oldnew[otrace->w]))
	    {
	      /* throw out? reset back trace ptr */
	      if (!otrace->next) 
		{
		  btrace->next=NULL;
		  free(otrace);
		  otrace=NULL; /* done */
		}
	      else
		{
		  btrace->next=otrace->next;
		  free(otrace);
		  otrace=btrace->next;
		}
	    }
	  else
	    {
	      btrace=otrace;
	      otrace=otrace->next;
	    }
	}
      if (!overlaps->next)
	{
	  free(overlaps);
	  overlaps=NULL;
	}
      else /* remove empty first position in overlaps */
	{
	  btrace=overlaps->next;
	  free(overlaps);
	  overlaps=btrace;
	}
      otrace=overlaps;
      while(otrace)
	{
	  ind=nghb(p,otrace->v,otrace->w);
	  pK_ptr[otrace->v].overlaps[ind]=otrace->angle;
	  if (!pK_ptr[otrace->v].bdry_flag && ind==0)
	    pK_ptr[otrace->v].overlaps[pK_ptr[otrace->v].num]=otrace->angle;
	  ind=nghb(p,otrace->w,otrace->v);
	  pK_ptr[otrace->w].overlaps[ind]=otrace->angle;
	  if (!pK_ptr[otrace->w].bdry_flag && ind==0)
	    pK_ptr[otrace->w].overlaps[pK_ptr[otrace->w].num]=otrace->angle;
	  otrace=otrace->next;
	}
      if (oldnew) free(oldnew);
    }
  if ((otrace=overlaps))
    {
      while (otrace && otrace->next) 
	{
	  overlaps=otrace->next;
	  free(otrace);
	  otrace=overlaps;
	}
      free(overlaps);
    }
  if (newold) free(newold);
  return i;
} /* cookie_call */

/* ================= diagostic calls =========================== */
int diag_call(struct p_data *p,char *datastr)
/* selecting diagnostic routines:
   c = deep combinatoric check
   i = connected interior?
   h = hanging bdry verts?
   o <v> = placement data on v
   e <x> = first vertex placed using 2 circles with inv dist in error
           by more than x.
   E <x> = number of edges whose length is more than x times what
           it should be (based on radii and inv dist).
   r = check red chain
   R = red chain layout error?
   s = verts with radius < 2*(smallest radius)
   v n order by size of ratio length/ideal_length
   
 */
{
  int i,v,count=0,n,vertcount,err_count;
  int pnum=pack_num(p);
  double val;
  char *nextpoint;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) return 1; /* no request */
  do {
    for (i=strlen(next);i<10;i++) next[i]='\0'; /* clear old info */
    if (next[0]!='-') return count;
    if (next[1]=='c') { /* deep combinatoric check */
      if ((n=comb_deep_ck(p,&v))) {
	sprintf(msgbuf,"Diag: seems combinatorics are okay.");
	msg();
      }
      else {
	sprintf(msgbuf,"Diag: deep check suggests combinatoric "
		"error in p%d.\n Error code %d, seems to involve "
		"vert %d.",pnum,n,v);
	emsg();
      }
      count++;
    }
    if (next[1]=='i') { /* connected interior? */ 
      if (!diag_int_connected(p)) {
	sprintf(msgbuf,"Diag: p%d has a connected interior.",pnum);
	msg();
      }
      count++;
    }
    else if (next[1]=='h') { /* any hanging bdry verts? */
      if (!diag_hanging_verts(p)) {
	sprintf(msgbuf,
		"Diag: in p%d, all bdry verts have interior nghbs.",
		pnum);
	msg();
      }
      count++;
    }
    else if (next[1]=='s') /* smallest radius? */
      diag_small_rads(p);
    else if (next[1]=='R') { /* check red chain layout error */
      sprintf(msgbuf,
	      "Diag: in p%d, red chain layout error %f; see "
	      "/tmp/diagnostic_log for details.",
	      pnum,(singledouble)check_red_layout(p));
      emsg();
    }
    else if (next[1]=='r') { /* check red chain */
      if ((n=check_red_chain(p))) {
	sprintf(msgbuf,
		"Diag: in p%d, red chain seems ok; count is %d.",
		pnum,n);
	msg();
      }
      else {
	sprintf(msgbuf,
		"Diag: in p%d, corrupt red chain; see "
		"/tmp/diagnostic_log for details.",
		pnum);
	emsg();
      }
    }
    else if (next[1]=='o' && !(v=grab_one_vert(p,&nextpoint)))
      /* feedback placement info on v */
      count+=cir_placement_data(p,v);
    else if (next[1]=='e' && grab_next(&nextpoint,next) 
	     && sscanf(next,"%lf",&val)) /* placement errors */
      count+=placement_err(p,val);
    else if (next[1]=='E' && grab_next(&nextpoint,next)
	     && sscanf(next,"%lf",&val)) { /* layout edgelength errors */
      v=diag_layout_quality(p,&val,&vertcount,&err_count,4);
      sprintf(msgbuf,"diag E: %d length errors greater than %lf",
	      vertcount,val);
      msg();
      if (v) {
	sprintf(msgbuf,"  The worst vertex seemed to be %d.",v);
	msg();
      }
      count+=vertcount;
    }
    else if (next[1]=='v' && grab_next(&nextpoint,next)
	     && sscanf(next,"%d",&n) && n>0 && n<4) { /* sorting options */
      val=2.0;
      v=diag_layout_quality(p,&val,&vertcount,&err_count,n);
      sprintf(msgbuf,"diag v: worst error seems to be %lf at vertex %d.",
	      val,v);
      msg();
      count++;
    }
  }
  while (nextpoint!=NULL && grab_next(&nextpoint,next) );
  return count;		
} /* diag_call */

int reverse_orient_call(struct p_data *p)
/* call and process */
{
  if (!reverse_orient(p)) return 0;
  sprintf(buf,"fix -p%d -K -c",pack_num(p));
  handle_cmd(buf,&current_p);
  sprintf(buf,"fix -p%d",pack_num(p));
  handle_cmd(buf,&current_p);
  return 1;
} /* reverse_orient_call */

int add_circles_call(struct p_data *p,char *datastr)
/* call and process */
{
  int count;

  count=add_circles(p,datastr);
  if (count)
    {
      if (p->xyzpoint) {free(p->xyzpoint);p->xyzpoint=NULL;}
      if (!complex_count(p,0)) return 0;
      facedraworder(p,0);
      sprintf(msgbuf,"Added %d circles to p%d.",
	      count,pack_num(p));
      msg();
    }
  return count;
} /* add_circles_call */

int rect_ratio_call(struct p_data *p,char *datastr)
/* Report the ratio length/height of a eucl packing, presumably 
in form of a rectangle. Give the four corners in counterclockwise 
order, upper-left corner first. Return values:
  0: error
  1; okay
  -1: data of lengths/heights aren't consistent with a rectangle 
  (off by > 5%). 
The reported result is left-right length divided by top-bottom height.*/
{
  int val,pnum=pack_num(p);
  double aspect;

  val=rect_ratio(p,datastr,&aspect);
  if (!val) return 0;
  if (val<0)
    {
      sprintf(msgbuf,"rect_ratio: p%d doesn't seem to be a rectangle.",
	      pnum);
      emsg();
      return 0;
    }
  sprintf(msgbuf,"In p%d, length/height=%f.",pnum,(singledouble)aspect);
  msg();
  return 1;
} /* rect_ratio_call */

int riffle_call(struct p_data *p,int passes,int lpr,int flag)
/* call to compute radii of packing p. flag indicates which
   routine to use: 0=old reliable, 1=super-superstep, 2=superstep 
   (only  eucl+hyper for now). */
{
  int count,hr,min,sec;
  long tic,toc;

  tic=get_time();
  if (!flag) /* old reliable */
    {
      if (p->hes<0) count=old_h_riffle(p,passes);
      else if (p->hes==0) count=old_e_riffle(p,passes);
      else count=0; /* spherical algorithm not yet working */
    }
  else if (flag==1) /* super-superstep */
    {
      if (p->hes<0) {
	if (p->overlap_status) count=h_riffle(p,passes);
	else count=h_riffle_O3(p,passes); /* h_riffle_O3, no overlaps yet */
      }
      else if (p->hes==0) {
	if (p->overlap_status) count=e_riffle(p,passes);
	else count=e_riffle_O3(p,passes); /* e_riffle_O3, no overlaps yet */
      }
      else count=0; /* spherical algorithm not yet working */
    }
  else if (flag==2)
    {
      if(p->hes<0){ /* older superstep */
	count=h_riffle(p,passes);
      }
      else if(p->hes==0)  count=new_e_riffle(p,passes);
      else  count=0;  /* spherical algorithm not yet working */
    }
  if (count && lpr)
    {
      toc=get_time();
      time_to_hms(toc-tic,&hr,&min,&sec);
      sprintf(msgbuf," REPACK p%d; %d passes; (%d:%d:%d) (h:m:s). ",
      	pack_num(p),count,hr,min,sec);
      msg();
    }
  return count;
} /* riffle_call */

int gen_mark_call(struct p_data *p,char *datastr)
     /* call routine which records generations of vertices or faces
	as measured from 'seeds' (specified in datastr). Records
	generation in 'mark' and returns the last vertex marked. Command
	'-m n' tells it to stop at max generation n. */
{
  int last_vert=0,last_face=0,faceflag=0,mx,f;
  int Gmax=0,g0,g1,g2,gmax,gmin;
  char *nextpoint;
  struct K_data *pK_ptr=p->packK_ptr;

  nextpoint=datastr;
  stripsp(nextpoint);
  if (!strncmp(datastr,"-f",2)) /* mark faces: face is marked in
				   generation n if vertices are 
				   all generation n or all .le. n+1. */
    {
      faceflag=1;
      nextpoint += 2;
    }
  if (!(last_vert=gen_mark(p,nextpoint)))
    {
      sprintf(msgbuf,"Usage: gen_mark [-f] [-m n] {v..}");
      emsg();
      return 0;
    }
  mx=pK_ptr[last_vert].mark;
  if (faceflag)
    {
      for (f=1;f<=p->facecount;f++) p->faces[f].mark=0;
      for (f=1;f<=p->facecount;f++)
	{
	  gmax=gmin=0;
	  if ((g0=pK_ptr[p->faces[f].vert[0]].mark)
	      && (g1=pK_ptr[p->faces[f].vert[1]].mark)
	      && (g2=pK_ptr[p->faces[f].vert[2]].mark))
	    {
	      gmax=g0;
	      gmax=(g1>gmax) ? g1 : gmax;
	      gmax=(g2>gmax) ? g2 : gmax;
	      gmin=g0;
	      gmin=(g1<gmin) ? g1 : gmin;
	      gmin=(g2<gmin) ? g2 : gmin;
	    }
	  if (gmax>gmin) p->faces[f].mark=gmax-1; 
	  else if (gmax) p->faces[f].mark=gmin; /* gmax=gmin */
	  if (p->faces[f].mark>Gmax)
	    {
	      Gmax=p->faces[f].mark;
	      last_face=f;
	    }
	}
      if (last_face)
	{
	  sprintf(msgbuf,"gen_mark: last face marked = %d and its "
		  "generation is %d.",last_face,Gmax);
	  msg();
	}
      return last_face;
    }

  sprintf(msgbuf,"gen_mark: last vertex marked = %d and its "
	  "generation is %d.",last_vert,mx);
  msg();
  return last_vert;
} /* gen_mark_call */

int gen_cut_call(struct p_data *p,char *datastr)
/* cut out a simply connected subcomplex based on generations
   from a specified vertex (which is generation 1). 
   fixup: currently, only works for simply connected. */
{
  int v,gen;
  char *nextpoint=datastr,*mvpt;

  if (!p->status || p->locks) return 0;
  if (p->genus!=0 || (p->euler!=1 && p->euler!=2))
    {
      sprintf(msgbuf,"gen_cut: works only for simply connnected "
	      "complexes.");
      msg();
      return 1;
    }
  mvpt=next;
  if (!grab_next(&nextpoint,next) || !(v=grab_one_vert(p,&mvpt)) 
      || !sscanf(nextpoint,"%d",&gen) || gen<2) 
      return 0;
  else return (gen_cut(p,v,gen));
} /* gen_cut_call */

int spiral_call(struct p_data *p,char *datastr)
/* call Doyle spiral routine. Either set the radii of the alpha
flower's petals 0 and 1; or with -f flag, multiply the current ones 
by the two given factors. */
{
  int j,al,fflag=0;
  double a,b,rad;
  char *nextpoint;

  nextpoint=datastr;
  if (!p->status || !grab_next(&nextpoint,next)) return 0;
  if (next[0]=='-' && next[1]=='f') fflag=1;
  else nextpoint=datastr;
  if (sscanf(nextpoint,"%lf %lf",&a,&b)!=2
      || a<=0.0 ||  b<=0.0)
    {
      strcpy(msgbuf,"Usage: a b (or -f a b).");
      emsg();
      return 0;
    }
  if (fflag) /* set radii of first/second petals of alpha:
		use a, b as factors by which to change their
		ratio vis-a-vis alpha itself. */
    {
      al=p->alpha;
      rad=p->packR_ptr[al].rad;
      if (rad<=okerr) { /* this shouldn't happen */ 
	sprintf(msgbuf,"spiral: error in radius of alpha.");
	emsg();
	return 0;
      }
      j=p->packK_ptr[p->alpha].flower[0];
      a=a*p->packR_ptr[j].rad/rad;
      j=p->packK_ptr[p->alpha].flower[1];
      b=b*p->packR_ptr[j].rad/rad;
    }
  return (spiral(p,a,b));
} /* spiral_call */

int marking(struct p_data *p,char *datastr)
/* set mark on circles or faces */
{
  int count=0,i,vert,flg,hits,*gen_list,tick,nv,dum;
  char *nextpoint,*endptr;
  struct Vertlist *vertlist,*trace,*facelist;
  struct K_data *pK_ptr;
  
  pK_ptr=p->packK_ptr;
  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || next[0]!='-' 
      || !p->status) return 1;
  do
    {
if (next[0]!='-') return count;
else if (next[1]=='w') /* wipe out all marks */
 {
	for (i=1;i<=p->nodecount;i++) pK_ptr[i].mark=0;
	for (i=1;i<=p->facecount;i++) p->faces[i].mark=0;
	count++;
 }
else if (next[1]=='g') /* mark circles by generations from vert */
  /* vert is gen 1 */
  {
    if (!nextpoint || !(vert=grab_one_vert(p,&nextpoint)))
	vert=p->alpha; /* alpha = default */
    hits=0;
    if ((gen_list=label_seed_generations(p,vert,NULL,0,&hits,&dum)))
      {
	for (i=1;i<=p->nodecount;i++) 
	  pK_ptr[i].mark=gen_list[i];
	free(gen_list);gen_list=NULL;
      }
  }
else if (next[1]=='c' && next[2]=='o'
	 && vert_draw_order(p))          /* mark by drawing order */ 
  {
    tick=1;
    pK_ptr[p->alpha].mark=tick++;
    nv=pK_ptr[p->alpha].next_vert;
    while (nv>0 && nv!=p->alpha)
      {
	pK_ptr[nv].mark=tick++;
	nv=pK_ptr[nv].next_vert;
	count++;
      }
  }
else if (next[1]=='c') /* mark circles */
 {
	flg= (next[2]=='w'); /* 1==>wipe out */
	if ((vertlist=node_link_parse(p,nextpoint,
		&endptr,&hits))!=NULL)
	 {
		nextpoint=endptr;
		trace=vertlist;
		do
		 {
			vert=trace->v;
			if (flg) pK_ptr[vert].mark=0;
			else pK_ptr[vert].mark++;
			count++;
			trace=trace->next;
		 } while (trace!=NULL);
		vert_free(&vertlist);
	 }
	else if (flg)
	 {
		for (i=1;i<=p->nodecount;i++) pK_ptr[i].mark=0;
		count++;
	 }
 }
else if (next[1]=='f') /* mark faces */
 {
	flg= (next[2]=='w'); /* 1==>wipe out */
	if ((facelist=face_link_parse(p,nextpoint,&endptr,&hits))!=NULL)
	 {
		nextpoint=endptr;
		trace=facelist;
		do
		 {
			vert=trace->v;
			if (flg) p->faces[vert].mark=0;
			else p->faces[vert].mark++;
			count++;
			trace=trace->next;
		 } while (trace!=NULL);
		vert_free(&facelist);
	 }
	else if (flg)
	 {
		for (i=1;i<=p->facecount;i++) p->faces[i].mark=0;
		count++;
	 }
 }
	 } /* end of do */
	while (nextpoint!=NULL && grab_next(&nextpoint,next) );
	return count;		
} /* marking */

int norm_scale_call(struct p_data *p,char *datastr)
     /* normalize a eucl packing in various ways */
{
  int v,w,qnum;
  double factor,rad;
  char *nextpoint;

  nextpoint=datastr;
  if ( !p->status || p->hes>0 || p->hes<0) return 0;
  if (!grab_next(&nextpoint,next) || next[0]!='-') return 0;
  if (next[1]=='e')     /* scale p so rad of v in p equals
				rad of w in pack q */
    {
      if (sscanf(nextpoint,"%d %d %d",&qnum,&v,&w)!=3
	  || qnum<0 || qnum>=NUM_PACKS || !packdata[qnum].status
	  || v<1 || v>p->nodecount
	  || w<1 || w>packdata[qnum].nodecount
	  || (rad=p->packR_ptr[v].rad)<=okerr) return 0;
      factor=packdata[qnum].packR_ptr[w].rad/rad;
      return eucl_scale(p,factor);
    }
  return e_norm_scale(p,datastr); /* library routine handles rest
				   of options */
} /* norm_scale_call */

int set_aim_call(struct p_data *p,char *datastr)
/* Set the intended target angle sums. */ 
{ 
  int count=0,mode;
  char *nextpoint;

  nextpoint=datastr;
  grab_next(&nextpoint,next);
  if (next[0]=='-' && next[1]=='d') mode=0; /* default mode */
  else if (next[0]=='-' && next[1]=='c') mode=1;/* current mode */
  else if (next[0]=='-' && next[1]=='%') mode=2; /* add inc mode */
  else mode=3;
  if ((count=set_aims(p,datastr)))
    {
      if (mode==0) 
	sprintf(msgbuf,"Set %d `aims' of pack %d to default.",
		count,pack_num(p));
      else if (mode==1)
	sprintf(msgbuf,"Set %d `aims' of pack %d to "
		"their current angle sums.",
		count,pack_num(p));
      else if (mode==2)
	sprintf(msgbuf,"Incremented %d `aims' of pack %d.",
		count,pack_num(p));
      else if (mode==3) 
	sprintf(msgbuf,"Set %d `aims' of pack %d.",
		count,pack_num(p));
      msg();
      return count;
    }
  strcpy(msgbuf,"No aims adjusted.");
  emsg();
  return 0;
} /* set_aim_call */

int randize(struct p_data *p,char *datastr)
/* set random radii (default) or overlaps (or random changes) */
{ 
  int mode=0,count=0,vert,hits,flag=0,vv,ww,k;
  double rn,rand_value,lowr,highr,tmprad;
  char *nextptr,*lastptr,*endptr;
  struct Vertlist *vertlist=NULL,*trace;
  struct Edgelist *edgelist=NULL,*etrace;
  struct R_data *pR_ptr;
  
  pR_ptr=p->packR_ptr;
  lowr=10.0*okerr;highr=10; /* default radii range */
  nextptr=lastptr=datastr;
  grab_next(&nextptr,next);
  if (!strncmp(next,"-o",2)) { /* overlaps, not radii */
    flag=1;
    lowr=0.0; /* default overlap range */
    grab_next(&nextptr,next);
  }
  do {
    if (next[0]=='-') {
      if (next[1]=='r') { /* range */
	if (sscanf(nextptr,"%lf %lf",&lowr,&highr)!=2
	    || highr<lowr) {
	  if (count) fillcurves(p);
	  return count;
	}
	grab_next(&nextptr,next);
	grab_next(&nextptr,next); /* move past data */
	if (mode==1) {
	  if (lowr<okerr) lowr=okerr;
	  if (highr<lowr) highr=2*lowr;
	}
      }
      else if (next[1]=='f' && (!flag || p->overlap_status)) {
	/* increment (treat as factor) */
	mode=1;
	if (lowr<okerr) lowr=okerr;
	if (highr<lowr) highr=2*lowr;
      }
      else if (!flag && next[1]=='j') { /* jiggle radii */
	mode=1;
	lowr=.999;
	highr=1.0001;
      }
      lastptr=nextptr;
    }
    else if (!flag) {  /* adjust radii */
      if ((vertlist=node_link_parse(p,lastptr,&endptr,&hits))==NULL) {
	if (count) fillcurves(p);
	return count;
      } 
      /* default = no action */
      nextptr=lastptr=endptr;
      trace=vertlist;
      do {
	vert=trace->v;
	rn=gwrand(); /* pseudo-random in (0,1) */
	rand_value=lowr+(highr-lowr)*rn;
	if (mode==1) { /* multiply by factor */
	  if (p->hes>0) { /* sphere */
	    pR_ptr[vert].rad *= rand_value;
	    if (pR_ptr[vert].rad>=M_PI)
	      pR_ptr[vert].rad=M_PI-okerr;
	    count++;
	  }
	  if (p->hes<0) { /* hyperbolic */
	    if (pR_ptr[vert].rad < 0) { /* infinite radius */
	      if (rand_value < 1.0) /* set h = .5 */
		pR_ptr[vert].rad=1.0 - exp(-.25);
	    }
	    else  { /* x=1-exp((-1/2 ln(1-x)*rand_value)^2) */
	      tmprad=1.0-exp(-log(1.0-pR_ptr[vert].rad)*rand_value); 
	      if (tmprad > .999999) pR_ptr[vert].rad = -.1;
	      /* large? make infinite */
	      else pR_ptr[vert].rad = tmprad;  /* x-radii */
	    }
	    count++;
	  }
	  else { /* eucl */
	    pR_ptr[vert].rad *= rand_value;
	    count++;
	  }
	}
	else { /* random, w/o using current value */
	  if (p->hes>0) {
	    pR_ptr[vert].rad = rand_value;
	    if (pR_ptr[vert].rad>=M_PI)
	      pR_ptr[vert].rad=M_PI-okerr;
	    count++;
	  }
	  if (p->hes<0) { /* hyperbolic */
	    tmprad= 1.0 -exp((-2.0)*rand_value);	
	    if (tmprad > .999999) pR_ptr[vert].rad = -.1;
	    /* large? make infinite */
	    else pR_ptr[vert].rad = tmprad;			
	    count++;
	  }
	  else {
	    pR_ptr[vert].rad = rand_value;
	    count++;
	  }
	}
	trace=trace->next;
      } /* end of inner do */
      while (trace!=NULL);
      vert_free(&vertlist);
      if (count) fillcurves(p);
      return count;
    }
    else { /* adjust overlaps */
      if ((edgelist=node_pair_link(p,lastptr,&endptr,&hits))==NULL) {
	if (count) fillcurves(p);
	return count;
      }   /* default = no action */
      alloc_overlaps(p);
      nextptr=lastptr=endptr;
      etrace=edgelist;			
      do  {
	vv=etrace->v;
	ww=etrace->w;
	rn=gwrand(); /* pseudo-random in (0,1) */
	rand_value=lowr+(highr-lowr)*rn;
	if ((k=nghb(p,vv,ww))>=0) {
	  if (mode==1) /* factor */
	    set_single_overlap(p,vv,k,
			       rand_value*p->packK_ptr[vv].overlaps[k]);
	  else set_single_overlap(p,vv,k,rand_value);
	  count++;
	}
	etrace=etrace->next;
      } /* end of inner do */
      while (etrace!=NULL);
      edge_free(&edgelist);
      if (count) fillcurves(p);
      return count;
    }
  } /* end of do */
  while ((lastptr=nextptr)!=NULL && grab_next(&nextptr,next) );
  if (count) fillcurves(p);
  return count;		
} /* randize */

int max_pack(struct p_data *p,char *datastr)
/* shortcut for doing maximal packing */
{
  int n,pnum=pack_num(p);
  struct Vertlist *holdv,*holdf;
  struct Edgelist *holde,*holdvm;

  if (p->intrinsic_geom < 0) { /* hyperbolic case */
    if (p->hes >=0) {
      sprintf(buf,"geom_to_h -p%d",pnum);
      handle_cmd(buf,&current_p);
    }
    sprintf(buf,"set_aim -p%d -d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"set_rad -p%d 10.0 b",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"repack -p%d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"fix -p%d",pnum);
    handle_cmd(buf,&current_p);
    return 1;
  }
  if (p->intrinsic_geom == 0) { /* must be 1-torus */
    if (p->hes < 0 || p->hes > 0) {
      sprintf(buf,"geom_to_e -p%d",pnum);
      handle_cmd(buf,&current_p);
    }
    sprintf(buf,"set_aim -p%d -d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"repack -p%d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"fix -p%d",pnum);
    handle_cmd(buf,&current_p);
    return 1;
  }
  if (p->intrinsic_geom > 0) { /* sphere */
    /* can specify node at which to puncture, else max node */
    if (sscanf(datastr,"%d",&n) && n>0 && n<p->nodecount) {
      sprintf(buf,"swap -p%d %d %d",pnum,n,p->nodecount);
      handle_cmd(buf,&current_p);
    }
    else n=p->nodecount;
    /* save lists to restore later */
    if (holdv=p->vlist) p->vlist=NULL;
    if (holde=p->elist) p->elist=NULL;
    if (holdf=p->flist) p->flist=NULL;
    if (holdvm=p->vertex_map) p->vertex_map=NULL;
    
    sprintf(buf,"puncture -p%d %d",pnum,p->nodecount);
    handle_cmd(buf,&current_p);
    if (p->hes >=0) {
      sprintf(buf,"geom_to_h -p%d",pnum);
      handle_cmd(buf,&current_p);
    }
    sprintf(buf,"set_aim -p%d -d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"set_rad -p%d 5.0 b",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"repack -p%d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"fix -p%d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"geom_to_s -p%d",pnum);
    handle_cmd(buf,&current_p);
    sprintf(buf,"add_ideal -p%d",pnum);
    handle_cmd(buf,&current_p);
    if (n != p->nodecount) {
      sprintf(buf,"swap -p%d %d %d",pnum,n,p->nodecount);
      handle_cmd(buf,&current_p);
    }
    if (n==p->alpha || nghb(p,n,p->alpha) >= 0)
      n = p->gamma; /* try to avoid normalization problem */
    sprintf(buf,"NSpole -p%d %d %d",pnum,p->alpha,n);
    handle_cmd(buf,&current_p);
    /* restore lists */
    p->vlist=holdv;p->elist=holde;p->flist=holdf;p->vertex_map=holdvm;
    return 1;
  }
  return 0;
} /* max_pack */

int repack_call(struct p_data *p,char *data)
{
  int pass,n,hits,count=0;
  char *nextpoint,*endptr=NULL;
  struct Vertlist *vertlist,*trace;

  stripsp(data);
  nextpoint=data;
  pass=totalpasses;
  if (p->overlap_status) n=0; /* old repack routine */
  else if (p->hes==0) n=2; /* eucl; default to 'new_e_riffle' */
  /* fixup: new_h_riffle not dependable yet. */
  else n=1;
  while (grab_next(&nextpoint,next)){
    if (next[0]=='-'){
      if (next[1]=='v'){ /* just certain verts */
	if ( (vertlist=node_link_parse(p,
	      nextpoint,&endptr,&hits))!=NULL){
	  trace=vertlist;
	  do {
	    if (p->packR_ptr[trace->v].aim>=0){
	      if (p->hes<0)
		h_riffle_vert(p,trace->v);
	      else if (p->hes==0)
		e_riffle_vert(p,trace->v);
	      else if (p->hes>0)
		s_riffle_vert(p,trace->v);
	      count++;
	    }
	    trace=trace->next;
	  } while (trace!=NULL);
	  vert_free(&vertlist);
	}
	else{
	  if (p->hes<0) 
	    h_riffle_vert(p,p->active_node);
	  else if (p->hes==0)
	    e_riffle_vert(p,p->active_node);
	  else if (p->hes>0)
	    s_riffle_vert(p,p->active_node);
	  count++;
	}
	return (count);
      }
      else if (next[1]=='n') n=1; /* superstep (hyp/eucl) */
      else if (p->hes<=0 && next[1]=='t') n=2; /* superstep2 (hyp/eucl) */
      else if (next[1]=='o') 
	n=0; /* use old reliable routines */
    }
    else if (sscanf(next,"%d",&pass)!=1 || pass < 1 || pass > 100010)
      pass = totalpasses;
  } /* end of while */
/*  if (p->nodecount>1000){
    int tracking_code=(int)rand();
    bprintf(msg_out_fp,"<msg type=\"SA_PROGRESS\">\n"
	    "  <pack_id> %d </pack_id>\n"
	    "  <progress status=\"on\" tracking_code=\"%d\"> "
	    "repack </progress>\n"
	    "</msg>\n",
	    pack_num(p),tracking_code);
    riffle_call(p,pass,1,n);
    bprintf(msg_out_fp,"<msg type=\"SA_PROGRESS\">\n"
	    "  <pack_id> %d </pack_id>\n"
	    "  <progress status=\"off\" tracking_code=\"%d\"></progress>\n"
	    "</msg>\n",pack_num(p),tracking_code);
  } */
  riffle_call(p,pass,1,n);
  return 1;
} /* repack_call */

int sa_reset_screen(struct p_data *p)
     /* get canvas info from GUI, send screen info. */
{
	p->screen->box.lx=p->screen->box.ly=-1.1;
	p->screen->box.rx=p->screen->box.ry=1.1;
	return (send_setViewBox(p));
} /* sa_reset_screen */

int enfoldcall(struct p_data *p,char *datastr)
{
  int n,m,vert,i,count=0,enfold_flag=0,hits;
  struct K_data *pK_ptr=p->packK_ptr;
  char *node_ptr,*endptr;
  struct Vertlist *vertlist,*trace;

  node_ptr=datastr;
  if (!grab_next(&node_ptr,next) || sscanf(next,"%d",&n)!=1
      || n<0 || n>(MAX_PETALS-3)) return 0;
  if ( (vertlist=node_link_parse(p,node_ptr,&endptr,&hits)) != NULL)
    {
      trace=vertlist;
      do {
	vert=trace->v;
	if (pK_ptr[vert].bdry_flag)
	  {
	    m=MAX_PETALS-pK_ptr[vert].num-1;
	    m=(n<m)? n:m;
	    for (i=1;i<=m;i++) add_vert(p,vert);
	    enfold_flag=enfold(p,vert);
	    complex z=p->packR_ptr[pK_ptr[vert].flower[0]].center;
	    complex w=p->packR_ptr[pK_ptr[vert].flower[pK_ptr[vert].num-1]].center;
	    drawEdge(pack_num(p),z.re,z.im,w.re,w.im);
	    count++;
	  }
	trace=trace->next;
      } while (trace!=NULL && enfold_flag);
      vert_free(&vertlist);
    }
  if (count)
    {
      /* if (!pK_ptr[p->beta].bdry_flag) choose_beta(p);*/
      complex_count(p,0);
      facedraworder(p,0);
      fillcurves(p);
    }
  return count;
} /* enfoldcall */

int adjoincall(struct p_data *p,char *datastr)
     /* return 0 on error */
{
  int pnum1,pnum2,v1,v2,n,n1,n2,*oldnew=NULL;
  int v,j,vb,vert,b1,b2,vv,ww,offset=0,over_flag=0;
  double angle;
  struct p_data *p1,*p2;
  struct K_data *pK_ptr1;
  struct Overlaps *overlaps=NULL,*trace;
  struct Vertlist *vtrace;
  struct Edgelist *etrace;
  char *nextpoint,*newstr;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum1)
      || pnum1<0 || pnum1>=NUM_PACKS
      ||!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum2)
      || pnum2<0 || pnum2>=NUM_PACKS
      || !(v1=grab_one_vert(p,&nextpoint)) ||  !(v2=grab_one_vert(p,&nextpoint)) 
      || !grab_next(&nextpoint,next))
    return 0;
  n=0;
  newstr=(char *)calloc((size_t)strlen(next),sizeof(char));
  if ((next[0]!='(' && !sscanf(next,"%d",&n))
      || (next[0]=='(' && (!strncpy(newstr,next,strlen(next))
			   || !paren_parse(&newstr,&b1,&b2)))) return 0;
  p1=&packdata[pnum1];p2=&packdata[pnum2];
  pK_ptr1=p1->packK_ptr;
  if (!n) /* get n from 'paren' style input */
    {
      if (b1<=0 || b2>p1->nodecount 
	  || !pK_ptr1[b1].bdry_flag) return 0;
      n=1;
      vert=b1;
      vb=pK_ptr1[vert].flower[pK_ptr1[vert].num];
      while (vb!=b2 && vb!=vert && n< p1->nodecount)
	{
	  vb=pK_ptr1[vb].flower[pK_ptr1[vb].num];
	  n++;
	}
      if (n>=p1->nodecount) return 0;
    }
  if (pnum1!=pnum2) offset=p1->nodecount;
  if ((n1=bdry_comp_count(p1,v1))<n || n<0) 
    /* signal to identify as many edges as possible */
    /* fixup: needs work for all possible cases ?? */
    {
      if (n1<1 || (n2=bdry_comp_count(p2,v2))<1)
	return 0;
      if (n1==n2) n=n1-1;
      else n= (n1<n2) ? n1:n2;
    }

  if ( p1->overlap_status || p2->overlap_status )
    over_flag=1;

  /* overlaps of pnum1 saved as linked list */
  if (p1->overlap_status) 
    {
      overlaps=(struct Overlaps *)
	calloc(1,sizeof(struct Overlaps));
      trace=overlaps;
      for (v=1;v<=p1->nodecount;v++)
	for (j=0;j<(pK_ptr1[v].num+pK_ptr1[v].bdry_flag);j++)
	  {
	    if ( v<pK_ptr1[v].flower[j]
		 && (angle=pK_ptr1[v].overlaps[j])!=1.0 )
	      {
		trace->v=v;
		trace->w=pK_ptr1[v].flower[j];
		trace->angle=angle;
		trace=trace->next=(struct Overlaps *)
		  calloc(1,sizeof(struct Overlaps));
	      }
	  }
    }

  /* now try to adjoin */
  if (adjoin(p1,p2,v1,v2,n,&oldnew))
    {
      if (p1->xyzpoint) {free(p1->xyzpoint);p1->xyzpoint=NULL;}
      complex_count(p1,1);

      if (p1->flist) vert_free(&p1->flist);
      if (pnum1==pnum2 && p1->vlist)
	{
	  vtrace=p1->vlist;
	  while (vtrace)
	    {
	      vtrace->v=oldnew[vtrace->v];
	      vtrace=vtrace->next;
	    }
	}
      if (pnum1==pnum2 && p1->elist)
	{
	  etrace=p1->elist;
	  while (etrace)
	    {
	      etrace->v=oldnew[etrace->v];
	      etrace->w=oldnew[etrace->w];
	      etrace=etrace->next;
	    }
	}

      /* restablish saved overlaps */
      if(over_flag && alloc_overlaps(p1))
	{
	  if (!offset && overlaps) /* self-adjoin, use new indices */
	    {
	      trace=overlaps;
	      while (trace && trace->next!=NULL)
		{
		  vv=oldnew[trace->v];
		  ww=oldnew[trace->w];
		  set_single_overlap(p1,vv,nghb(p1,vv,ww),trace->angle);
		  trace=trace->next;
		  free(overlaps);
		  overlaps=trace;
		}
	      free(overlaps);
	    }
	  else if (overlaps) /* reset pnum1 overlaps */
	    {
	      trace=overlaps;
	      while (trace && trace->next!=NULL)
		{
		  set_single_overlap(p1,trace->v,
			      nghb(p1,trace->v,trace->w),trace->angle);
		  trace=trace->next;
		  free(overlaps);
		  overlaps=trace;
		}
	      free(overlaps);
	      overlaps=NULL;
	    }
	  if ( offset && p2->overlap_status ) 
	    /* new overlaps from p2? */		
	    {
	      for(v=1;v<=p2->nodecount;v++)
		{
		  vv=oldnew[v];
		  for(j=0;j<p2->packK_ptr[v].num+
			p2->packK_ptr[v].bdry_flag;j++)
		    if (v<p2->packK_ptr[v].flower[j])
		      {
			ww=oldnew[p2->packK_ptr[v].flower[j]];
			if ((angle=p2->packK_ptr[v].overlaps[j])!=1.0)
			  set_single_overlap(p1,vv,nghb(p1,vv,ww),angle);
		      }
		}
	    }
	}


      fillcurves(p1);
      facedraworder(p1,0);
      set_aim_default(p1);
      p1->active_node=p1->beta; 
      sprintf(msgbuf,"adjoin %d %d %d %d %d;",
	      pnum1,pnum2,v1,v2,n);
      if (oldnew) free(oldnew);
      if ((trace=overlaps))
	{
	  while (trace && trace->next) 
	    {
	      overlaps=trace->next;
	      free(trace);
	      trace=overlaps;
	    }
	  free(overlaps);
	}
      msg();return 1;
    }
  strcpy(msgbuf,"Something went wrong with 'adjoin' routine.");
  emsg();
  if (oldnew) free(oldnew);
  if ((trace=overlaps))
    {
      while (trace && trace->next) 
	{
	  overlaps=trace->next;
	  free(trace);
	  trace=overlaps;
	}
      free(overlaps);
    }
  return 0;
} /* adjoincall */

int zip_call(struct p_data *p,char *datastr) 
/* A special case of 'adjoin'; p1=p2 and v1=v2; one is 'zipping' up one bdry
 * component (like reverse of 'slit'). Other options the same, so n<0 means 
 * do as many as possible; if the bdry component has an odd number of edges, then 
 * have to scale back to leave 3 unpasted edges. For other cautions, see 'adjoin', 
 * e.g., losing overlap data. */
{
	int b,m,v,n,pnum;
	char cmd[128];
	char *nextpoint=datastr;
	
	nextpoint=datastr;
	if (!p->status || !(v=grab_one_vert(p,&nextpoint))
		|| !p->packK_ptr[v].bdry_flag || (sscanf(nextpoint,"%d",&n)!=1))
	    return 0;
	b=bdry_comp_count(p,v);
	m=b/2;
	if (b!=2*m) m=(b-1)/2;
	if (n<0 || n>=m) n=m; // full bdry; but check for compatibility
	pnum=pack_num(p);
	sprintf(cmd,"%d %d %d %d %d",pnum,pnum,v,v,n);
//printf(cmd);
//sprintf(msgbuf," zip with vertex %d",v);
//debugmsg();
	return (adjoincall(p,cmd));
} /* zip_call */
	
int color_circles(struct p_data *p,int f,char *datastr)
{
  int cp,i,ccode,hits;
  double base;
  struct K_data *pK_ptr;
  struct Vertlist *vertlist,*trace;
  char *nextpoint,*endptr;

  pK_ptr=p->packK_ptr;
  nextpoint=datastr;
  if (!p->status) return 0;
  switch(f)
    {
    case 5: /* fill with foreground */
      {
	for (i=1;i<=p->nodecount;i++) pK_ptr[i].color=FG_COLOR;
 	return 1;
      }
    case 4: /* fill with background */
      {
	for (i=1;i<=p->nodecount;i++) pK_ptr[i].color=BG_COLOR;
 	return 1;
      }
    case 1: /* color by radius */
      {
	return (radius_col_comp(p));
      }
    case 3: /* color by comparison */
      {
	if (!sscanf(nextpoint,"%d",&cp) || cp<0 || cp>=NUM_PACKS
	    || !packdata[cp].status
	    || packdata[cp].hes!=p->hes) return 0;
	if (p->hes<0) {h_compare_cl(p,&packdata[cp]);return 1;}
	else if (p->hes==0) {e_compare_cl(p,&packdata[cp]);return 1;}
	return 0;
      }
    case 6: /* specified color, circles */
      {
	if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&ccode) 
	    || ccode<0 || ccode>255) return 0;
	if ((vertlist=node_link_parse(p,nextpoint,&endptr,&hits))==NULL) 
	  return 1;
	trace=vertlist;
	while (trace!=NULL)
	  {
	    pK_ptr[trace->v].color=ccode;
	    trace=trace->next;
	  }
	vert_free(&vertlist);
	return 1;	
      }
    case 7: /* take colors from pack cp */
      {
	if (!sscanf(nextpoint,"%d",&cp) || cp<0 || cp>=NUM_PACKS
	    || !packdata[cp].status) cp=current_p;
	for (i=1;i<=p->nodecount;i++)
	  if (i<=packdata[cp].nodecount) 
	    pK_ptr[i].color=packdata[cp].packK_ptr[i].color;
	return 1;
      }
    case 8: /* compare model curvature to value: red=pos, blue=neg */
      {
	if (!sscanf(nextpoint,"%lf",&base)) base=0.0;
	else
	  {
	    if (base<=0) base=(-1.0)*sqrt((-1.0)*base);
	    else base=sqrt(base);
	  }
	color_kappa(p,base);
      }
    } /* end of switch */
  return 0;
} /* color_circles */
	
int color_faces(struct p_data *p,int f,char *datastr)
{
  int cp,i,ccode,hits,vc[3];
  double kmax;
  struct Vertlist *vertlist,*trace;
  struct xyzPoint *xyz_list=NULL;
  char *nextpoint,*endptr,filename[NAME_MAX],*bufr=NULL;
  FILE *fpxyz=NULL;

  if (!p->status) return 0;
  nextpoint=datastr;
  switch(f)
    {
    case 5: /* foreground */
      {
	for (i=1;i<=p->facecount;i++) p->faces[i].color=FG_COLOR;
 	return 1;
      }
    case 4: /* background */
      {
	for (i=1;i<=p->facecount;i++) p->faces[i].color=BG_COLOR;
 	return 1;
      }
    case 1: /* color by area */
      {
	return (area_col_comp(p));
      }
    case 3: /* color by area comparison */
      {
	if (!sscanf(nextpoint,"%d",&cp) || cp<0 || cp>=NUM_PACKS
	    || !packdata[cp].status
	    || packdata[cp].hes!=p->hes) return 0;
	if (p->hes<0) {h_compare_area(p,&packdata[cp]);return 1;}
	else if (p->hes==0) {e_compare_area(p,&packdata[cp]);return 1;}
	return 0;
      }
    case 6: /* specified color, faces */
      {
	if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&ccode) 
	    || ccode<0 || ccode>199) return 0;
	if ((vertlist=face_link_parse(p,nextpoint,&endptr,&hits))==NULL) 
	  return 1;
	trace=vertlist;
	while (trace!=NULL)
	  {
	    p->faces[trace->v].color=ccode;
	    trace=trace->next;
	  }
	vert_free(&vertlist);
	return 1;	
      }
    case 7: /* take colors from pack cp */
      {
	if (!sscanf(nextpoint,"%d",&cp) || cp<0 || cp>=NUM_PACKS
	    || !packdata[cp].status) cp=current_p;
	for (i=1;i<=p->facecount;i++)
	  if (i<=packdata[cp].facecount) 
	    p->faces[i].color=packdata[cp].faces[i].color;
	return 1;
      }
    case 9: /* take colors from vertices: 2/3 equal colors wins,
	       else take that of vert[0] */
      {
	if ((vertlist=face_link_parse(p,nextpoint,&endptr,&hits))==NULL) 
	  return 1;
	trace=vertlist;
	while (trace)
	  {
	    for (i=0;i<3;i++)
	      vc[i]=p->packK_ptr[p->faces[trace->v].vert[i]].color;
	    p->faces[trace->v].color=vc[0];
	    if (vc[1]==vc[2]) p->faces[trace->v].color=vc[1];
	    trace=trace->next;
	  }
	vert_free(&vertlist);
	return 1;	
      }
    case 8: /* quasiconformal distortion vis-a-vis another pack
	       or a file of xyz data. */
      {
	if (!grab_next(&nextpoint,next)) return 1;
	if (!strncmp(next,"-n",2)) /* compare to named file of (x,y,z)
				      coordinates */
	  {
	    if (!grab_next(&nextpoint,filename) 
		|| !(fpxyz=fopen(filename,"r"))
		|| !(xyz_list=read_xyz(fpxyz,nextpoint,&hits)) || !hits)
	      {
		sprintf(msgbuf,"color -f: error in reading (x,y,z) coords "
			"from given file %s.",filename);
		emsg();
		if (fpxyz) fclose(fpxyz);
		return 1;
	      }
	    if (fpxyz) fclose(fpxyz);
	    color_qc_xyz(p,xyz_list,hits,nextpoint,&kmax);
	    free(xyz_list);
	    sprintf(msgbuf,"Dilatation max, from %s --> p%d (for specified "
		    "faces) is %f.",filename,pack_num(p),(singledouble)kmax);
	    msg();
	    return 1;
	  }
	/* compare to named (x,y,z) data in the script file */	
	/* fixup:
	else if (!strncmp(next,"-s",2)) 
	  {
	    if (!grab_next(&nextpoint,filename) 
		|| !(bufr=find_script_data(filename,4))
		|| !(xyz_list=read_xyz_buf(bufr,nextpoint,&hits)) || !hits)
	      {
		sprintf(msgbuf,"color -f: error in reading (x,y,z) coords "
			"named %s from script file.",filename);
		emsg();
		if (bufr) free(bufr);
		return 1;
	      }
	    if (bufr) free(bufr);
	    color_qc_xyz(p,xyz_list,hits,nextpoint,&kmax);
	    free(xyz_list);
	    sprintf(msgbuf,"Dilatation max, from %s --> p%d (for specified "
		    "faces) is %f.",filename,pack_num(p),kmax);
	    msg();
	    return 1;
	  }
	*/
	else if (!strncmp(next,"-q",2)  /* comparison to pack cp */
		 && grab_next(&nextpoint,next)
		 && sscanf(next,"%d",&cp)
		 && cp>=0 && cp<NUM_PACKS
		 && packdata[cp].status
		 && color_qc(p,&packdata[cp],nextpoint,&kmax)
		 ) {
	  sprintf(msgbuf,"Dilatation max for p%d --> p%d (specified "
		  "faces) is %f.",pack_num(p),cp,(singledouble)kmax);
	  msg();
	  return 1;
	}
	else if (!strncmp(next,"-x",2)) { /* compare to own xyz data */
	  if (!p->xyzpoint) {
	    sprintf(msgbuf,"color: no xyz data for this packing.");
	    emsg();
	    return 0;
	  }
	  color_qc_xyz(p,p->xyzpoint,p->nodecount,nextpoint,&kmax);
	  sprintf(msgbuf,"Dilatation max, from p%d xyz_data --> p%d (for "
		  "specified faces) is %f.",
		  pack_num(p),pack_num(p),(singledouble)kmax);
	  msg();
	  return 1;
	}
      }
    } /* end of switch */
  return 0;
} /* color_faces */

int map_list_call(struct p_data *p,char *datastr)
/* to form Vlist or Flist using the 'vertex_map' of pack p to 
translate indices. For faces, can only do this if target pack 
is specified with -q flag. Flag -1 means to translate using the 
inverse of vertex_map. Get settings and call map_list. */
{
  int mode=1,qnum=-1,rev_flag=0;
  char *nextpoint;
  struct p_data *q;

  if (!p->status || !p->vertex_map) return 0;
  nextpoint=datastr;
  q=p;
  if (grab_next(&nextpoint,next) && next[0]=='-')
    {
      if (next[1]=='f') mode=2; /* faces */
      else if (next[1]=='c') mode=1; /* circles */
      else if (next[1]=='q') /* map to pack qnum */
	{
	  if (sscanf(next+2,"%d",&qnum) && qnum>=0 && qnum<NUM_PACKS
	      && packdata[qnum].status) 
	    q=&packdata[qnum];
	}
      else if (next[1]=='1') rev_flag=1; /* use inverse map */
    }
  else nextpoint=datastr;
  return (map_list(p,q,nextpoint,mode,rev_flag));
  /* rest of datastr for specifying indices */
} /* map_list_call */

int adjacency_call(struct p_data *p,char *datastr,int pflag)
{
  /* if pflag, prepend path */
  int n,flag=0;
  char *nextpoint,filename[NAME_MAX];
  FILE *fp;

  filename[0]='\0';
  if (pflag) strcpy(filename,packing_dir_name);
  nextpoint=datastr;
  stripsp(nextpoint);
  if (strcmp(nextpoint,"-t")) { /* want transition matrix */
    flag=1;
    nextpoint +=2;
  }
  if (!grab_next(&nextpoint,next)) return 0; /* no filename? */
  strcpy(filename,next);
  if (!(fp=fopen(filename,"w"))) return 0;
  n=adjacency(fp,p,flag);
  fclose(fp);
  return n;
} /* adjacency_call */

int write_multi_call(struct p_data *p,char *datastr,int call,int pflag)
     /* call=0, 'write_vert_list', or call=1 or 2, 'write_select_r_c'; 
	if pflag, prepend path. */
{
  char *nextpoint,filename[NAME_MAX];

  filename[0]='\0';
  if (pflag) strcpy(filename,packing_dir_name);
  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) return 0; /* no filename? */
  strcat(filename,next);
  if (!call) return (write_vert_list(p,filename,nextpoint));
  else if (call==1 || call==2) 
    return (write_select_r_c(p,filename,nextpoint,call));
  else return 0;
} /* write_multi_call */

int write_light_call(struct p_data *p,char *datastr,int flag,int pflag)
     /* write pack data out to file in p_light form; flag determines
      raw (flag) or keyword (!flag) form; if pflag, prepend path. */
{
  int i;
  char *nextpoint,filename[NAME_MAX];
  struct K_data *pK_ptr;
  struct p_light *pl=NULL;
  FILE *fp;

  pK_ptr=p->packK_ptr;
  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) /* no filename? */
    goto filename_error;
  filename[0]='\0';
  if (pflag) strcpy(buff,packing_dir_name);
  strcat(buff,next);
  strcpy(filename,buff); /* now have full pathname */
  for (i=1;i<=p->nodecount;i++) /* mark interior */
    {
      if (pK_ptr[i].bdry_flag) pK_ptr[i].util_flag=0;
      else pK_ptr[i].util_flag=1;
    }
  if (!(fp=fopen(filename,"w")) 
      || !(pl=convert_to_p_light(p)) 
      || !write_light(fp,pl,flag))
    {
      if (fp) fclose(fp);
      if (pl) free_p_light(&pl);
      return 0;
    }
  free_p_light(&pl);
  fclose(fp);
  return 1;

filename_error: 
  sprintf(buf,"Need filename with the 'write_light' command.");
  emsg();
  return 0;
} /* write_light_call */

int write_parse(struct p_data *p,char *datastr,int pflag)
/* parse options for writing packings. Caution, no confirmation 
   requested. Default is cgirzv. Note: all options in one
   continuous string directly after '-'. If pflag, prepend path.

A append (requires additional options)

c combinatorics
g geometry
i non-default inv_dist and aims
r radii (g needed)
z centers
a angle sums
v vertex_map
l non-default lists: verts/faces/edges
o non-default colors
f plot_flags which are .le. 0 (e.g. means poorly placed)
t edge-pairing mobius

m minimum = cgri
M max = cgriazsoflv
*/
{
  int act=00137,i,n,c,append_flag=0;
  int script_flag=0;
  char *nextpoint,filename[NAME_MAX];
  long fspot;
  FILE *fpp,*tmpfp;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) /* no filename? */
    goto filename_error;
  filename[0]='\0';
  if (next[0]=='-') /* here come the options */
    {
      act=0;
      i=1;
      while ((c=next[i++])!='\0')
	{
	  switch(c)
	    {
	    case 'A': {append_flag=1;break;}
	    case 'c': {act |= 00001;break;}
	    case 'g': {act |= 00002;break;}
	    case 'i': {act |= 00004;break;}
	    case 'r': {act |= 00010;break;}
	    case 'z': {act |= 00020;break;}
	    case 'a': {act |= 00040;break;}
	    case 'v': {act |= 00100;break;}
	    case 'l': {act |= 00200;break;}
	    case 'o': {act |= 00400;break;}
	    case 'f': {act |= 01000;break;}
	    case 's': { // write to the script file
	    	script_flag=1;
	    }
	    case 't': {
	      if (p->edge_pair[1].edge) act |= 04000;
	      break;
	     }

	    case 'm': {act |= 00017;break;}
	    case 'M': {act |= 01777;break;}
	    case 'x': {
	      if (p->xyzpoint) act |= 02000;}
	    }
	} /* done with - option */
      if (!grab_next(&nextpoint,next)) /* get filename */
	goto filename_error;
	
	/* writing to script: have to ask Java to
	   adjust to the temporary filename and 
	   write to a regular file: it then records the name
	   for later incorporation. There's a special callback
	   here sending info to Java; new options will be added
	   as needed. */
	  if (script_flag) {
	  	scriptWriteFile(act,append_flag,1,next); // 1=packing
	  	return 1;
	  }
	  
      if (append_flag)
	{
	  if (!act) return 0; /* 'append' option needs additional options */
	  strcpy(filename, next);
	  if (pflag) strcpy(buff,packing_dir_name);
	  strcat(buff,filename);
	  strcpy(filename,buff); /* now have full pathname */
	  if ((fpp=fopen(filename,"r+"))!=NULL 
	      && position_before(fpp,"END")
	      && (tmpfp=fopen("/tmp/hold_tail","w+")))
	    /* open and position named file for read/write */
	    {
	      fspot=ftell(fpp);
	      while ((c=getc(fpp))!=EOF) putc(c,tmpfp);
	      fseek(fpp,fspot,SEEK_SET);
      	      n=writepack(fpp,p,act,1);
	      fprintf(fpp,"\n");
	      rewind(tmpfp);
	      while ((c=getc(tmpfp))!=EOF) putc(c,fpp);
	      if (tmpfp) fclose(tmpfp);
	      if (fpp) fclose(fpp);
	      return n;
	    }
	  else
	    {
	      if (fpp) fclose(fpp);
	      sprintf(msgbuf,"Couldn't open %s for 'append'.",
		      filename);
	      emsg();
	      return 0;
	    }
	}
    }

  /* fixup: when do I fall through to here? May lack path name */
  strcpy(p->file_name,next);
  *buff='\0';
  if (pflag) {
    strcpy(buff,packing_dir_name);
  }
  strcat(buff,next);
  strcpy(filename,buff); /* now have full pathname */

   if (n=writecall(p,filename,act)) {
    sprintf(msgbuf,"wrote p%d to file %s. (flags=%d).",
	    pack_num(p),filename,act);
    msg();
  }
  return n;

filename_error: 
  sprintf(buf,"Need filename with the 'write' command.");
  emsg();
  return 0;
} /* write_parse */

int create_parent_call(struct p_data *p,char *datastr)
{
  int N=1;
  char *nextpoint,filename[NAME_MAX];
  FILE *pfp=NULL;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) goto BACK;
  filename[0]='\0';
  *buff='\0';
  strcpy(filename,packing_dir_name);
  strcat(filename,next); /* now have full pathname */
  if (grab_next(&nextpoint,next))
    sscanf(next,"%d",&N); /* get the number of children (if given) */
  if (!(pfp=fopen(filename,"w")) || !create_parent(p,pfp,N)) goto BACK;
  fclose(pfp);
  return 1;

 BACK:
  if (pfp) fclose(pfp);
  sprintf(msgbuf,"create_parent failed");
  emsg();
  return 0;
} /* create_parent_call */

int set_overlapscall(struct p_data *p,char *datastr)
/* set overlaps or inversive distances. Return 0 on error. */
{
  int j,v,count=0,hits=1,nb,i;
  double ang,inv_d,upper;
  char *nextpoint,*endptr,*bufr=NULL,name[NAME_MAX];
  struct Edgelist *edgelist,*track;
  struct Vertlist *vertlist,*trace;
  struct K_data *pK_ptr;
  struct xyzPoint *xyz_list=NULL;
  FILE *fpxyz=NULL;

  pK_ptr=p->packK_ptr;
  nextpoint=datastr;
  stripsp(nextpoint);
  if (nextpoint[0]=='-' && nextpoint[1]=='x') { /* set using 3-space
						 xyz data. */
    nextpoint += 2;
    if (!grab_next(&nextpoint,next)
	|| next[0]!='-') {
      if (!p->xyzpoint) return 0;
      else {
	(xyz_list=p->xyzpoint);
	hits=p->nodecount;
      }
    }
    else if (next[1]=='n') { /* xyz data from filename */
      if (!grab_next(&nextpoint,next)
	  || !sscanf(next,"%127s",name) || !(fpxyz=fopen(name,"r"))
	  || !(xyz_list=read_xyz(fpxyz,nextpoint,&hits)) || !hits) {
	if (xyz_list) free(xyz_list);
	if (fpxyz) fclose(fpxyz);
	return 0;
      }
      if (fpxyz) fclose(fpxyz);
    }
    /* xyz data from script */
    /* fixup:
    else if (next[1]=='s') {
      if (!grab_next(&nextpoint,name) 
	  || !(bufr=find_script_data(name,4))
	  || !(xyz_list=read_xyz_buf(bufr,nextpoint,&hits)) || !hits) {
	if (bufr) free(bufr);
	if (xyz_list) free(xyz_list);
	return 0;
      } 
      free (bufr);
    }
    */
    set_xyz_overlaps(p,xyz_list,hits,nextpoint,1);
    if (xyz_list!=p->xyzpoint) free(xyz_list);
    return 1;
  }
  if (nextpoint[0]=='-' && nextpoint[1]=='h') { /* if we have xyz data, 
						   set as though each
						   radius was 1/2 of 
						   shortest edge from it */
    set_xyz_overlaps(p,p->xyzpoint,p->nodecount,nextpoint,2);
    return 1;
  }
  if (nextpoint[0]=='-' && nextpoint[1]=='d') { /* set to default */
    free_overlaps(p);
    fillcurves(p);
    sprintf(msgbuf,"Overlap angles for pack %d set to 0.",
	    pack_num(p));
    msg();
    return 1;
  }
  if (nextpoint[0]=='-' && nextpoint[1]=='c') { /* set current */
    nextpoint[0]=nextpoint[1]=' ';stripsp(nextpoint);
    nextpoint += notspace(nextpoint);
    if (nextpoint[0]=='\0') {nextpoint--;nextpoint[0]='a';}
    if ((edgelist=node_pair_link(p,nextpoint,
				 &endptr,&hits))!=NULL)	{
      if (!alloc_overlaps(p)) return 0;
      track=edgelist;
      do {
	nb=nghb(p,track->v,track->w);
	set_single_overlap(p,track->v,nb,
			   comp_inv_dist(p,track->v,track->w));
	/* note: inv_dist routine not very robust */
	track=track->next;
	count++;
      } while (track!=NULL);
      edge_free(&edgelist);
    }
    fillcurves(p); /* not yet robust for inversive dist */
    sprintf(msgbuf,"Set %d overlap angles for p=%d "
	    "to current values.",count,pack_num(p));
    msg();
    return count;
  }
  if (nextpoint[0]=='-' && nextpoint[1]=='t' && p->overlap_status) {
    /* truncate inv_dist above (??? fixup later for overlaps also) */
    nextpoint[0]=nextpoint[1]=' ';stripsp(nextpoint);
    if (!sscanf(nextpoint,"%lf",&upper) || upper<1.0) {
      sprintf(msgbuf,"set_over: -t <x>, truncated at x, x>1.");
      emsg();
      return count;
    }
    for (v=1;v<=p->nodecount;v++) {
      for (i=0;i<=p->packK_ptr[v].num;i++) 
	if (p->packK_ptr[v].overlaps[i]>upper) {
	  p->packK_ptr[v].overlaps[i]=upper;
	  count++;
	}
    }
    sprintf(msgbuf,"Cut %d/2 inversive distances down to max of %lf.",
	    count,upper);
    msg();
    return count;
  }
  inv_d=1.0; /* just in case */
  if ( !grab_next(&nextpoint,next)
       || (next[0]=='*' && (!sscanf(next+1,"%lf",&ang)
			    || ang<0.0))
       || (next[0]!='*' && (!sscanf(next,"%lf",&ang)
/* ??			|| ang>0.5 || ang<0.0)) ) */
/* part of attempt to allow overlaps greater than pi/2 */
			    || ang>1.0 || ang<0.0)) )
    {
      strcpy(msgbuf,"Specify first: angle in [0,1] or '*' "
	     "and inv dist [0,infty].");
      emsg();
      return 0;
    }
  if (next[0]!='*') 
    {
      if (ang>0.5)
	{
	  strcpy(msgbuf,"WARNING: overlap > 0.5 (Pi) specified; "
		 "behavior not guaranteed.\n repack with -o option.");
	  emsg();
	}
      inv_d=cos(ang*M_PI);	/* overlap */
    }
  else inv_d=ang;				/* separated */
  nextpoint += notspace(nextpoint);
  if (nextpoint[0]=='-' && nextpoint[1]=='f') /* petals of given flowers */
    {
      nextpoint[0]=nextpoint[1]=' ';stripsp(nextpoint);
      if (nextpoint[0]=='\0' || (vertlist=node_link_parse(p,nextpoint,
				&endptr,&hits))==NULL) 
	return count;
      {
	if (!alloc_overlaps(p)) return 0;
	trace=vertlist;
	do
	  {
	    v=trace->v;
	    for (j=0;j<pK_ptr[v].num+pK_ptr[v].bdry_flag;j++)
	      set_single_overlap(p,v,j,inv_d);		
	    trace=trace->next;
	    count++;
	  } while (trace!=NULL);
	vert_free(&vertlist);
      }
      fillcurves(p);
      return count;
    }			
  if (nextpoint[0]=='\0') {nextpoint--;nextpoint[0]='a';}
  if ((edgelist=node_pair_link(p,nextpoint,&endptr,&hits))!=NULL)
    {
      if (!alloc_overlaps(p)) return 0;
      track=edgelist;
      do {
	nb=nghb(p,track->v,track->w);
	set_single_overlap(p,track->v,nb,inv_d);
	track=track->next;
	count++;
      } while (track!=NULL);
      edge_free(&edgelist);
    }
  fillcurves(p);
  if (inv_d<=1.0 && inv_d>=0.0)
    { 
      sprintf(msgbuf,"Set %d overlap angle(s) for p=%d to %f pi.", 	
	      count,pack_num(p),(singledouble)acos(inv_d)/(singledouble)M_PI);
      msg();
    }
  else if (inv_d>1.0)
    { 
      sprintf(msgbuf,"Set %d inv_distance(s) for p=%d to %f.", 	
	      count,pack_num(p),(singledouble)inv_d);
      msg();
    }
  return count;
} /* set_overlapscall */

int set_postfile_call(int flag,char *datastr)
     /* flag=0 means print file, else custom postscript file */
{
  char filename[NAME_MAX];

  if (sscanf(datastr,"%s",filename))
    {
      if (!flag)
	{
	  strcpy(print_file_name,filename);
	}
      else
	{
	  strcpy(ps_file_name,filename);
	}
      return 1;
    }
  else return 0;
} /* set_postfile_call */

int find_help_packrc(char *argv0)
/* Locate help and set CPHelp, locate and read `packrc.txt' file. 
Check in order: current directory, home directory, CPHOME/doc (for
help), or basename of the CirclePack call. */
{
  int i,n=0;
  char CPenv[NAME_MAX+1],RC[NAME_MAX+1],*ptr=NULL;

  CPHelp[0]='\0';
  CPenv[NAME_MAX]=RC[NAME_MAX]='\0';

      /* clear the function keys */
  buff[BUFSIZE-1]='\0'; 
  for (i=0;i<=10;i++) ftn_keys[i]=NULL;
 
      /* find environment */
  if (!(ptr=getenv("CPHOME"))) /* check first for env */
    {
      n=strlen(argv0);
      while (n>0 && argv0[n]!='/') n--;
      if (n==0) strncpy(CPenv,home_dir,NAME_MAX);
      else
	{
	  if (n>NAME_MAX-1) n=NAME_MAX;
	  strncpy(CPenv,argv0,n);
	}
    }
  else strncpy(CPenv,ptr,NAME_MAX);

      /* find and read system default CPpackrc */
  strncpy(RC,CPenv,NAME_MAX);
  if (ptr && strlen(RC)<(NAME_MAX-5)) strcat(RC,"/etc");
  if (strlen(RC)<(NAME_MAX-10)) strcat(RC,"/CPpackrc");
  if (stat_file(RC)) read_packrc(RC); 
      
       /* find and read individual packrc.txt*/
  strcpy(RC,"./packrc.txt"); /* current ?*/
  if (!stat_file(RC))
    {
      strncpy(RC,home_dir,NAME_MAX);
      if (strlen(RC)<(NAME_MAX-8)) strcat(RC,"/packrc.txt"); /* home dir? */
    }
  if (stat_file(RC)) read_packrc(RC);
  /* mh note: if no packrc.txt, then read_packrc is never called and no warning
     given about missing packrc.txt file; following 5 lines added by mh */
  else
    {
      fprintf(stderr,
	      "CirclePack did not find your 'packrc.txt' file. Continue.\n");
    }

       /* find cp_help.info if CPHelp not set in packrc calls */
  if (strlen(CPHelp)==0)
    {
      strcpy(CPHelp,"./cp_help.info"); /* current ?*/
      if (!stat_file(CPHelp))
	{
	  strncpy(CPHelp,home_dir,(NAME_MAX-15));
	  strcat(CPHelp,"/cp_help.info"); /* home dir? */
	  if (!stat_file(CPHelp))
	    {
	      strncpy(CPHelp,CPenv,(NAME_MAX-15));
	      if (ptr) strcat(CPHelp,"/doc");
	      strcat(CPHelp,"/cp_help.info");
	    }
	  if (!stat_file(CPHelp))
	    { 
	      fprintf(stderr,"CirclePack did not "
		      "find 'cp_help.info'. Continue.\n");
	    }
	}
    }

  return 1;
} /* find_help_packrc */

int read_packrc(char *filename)
{
  FILE *fp;
  char *keyword=NULL,*next=NULL,*ptr;
  int cs,i;

  if (!(fp=fopen(filename,"r")))
    {
      fprintf(stderr,
	      "CirclePack did not find your 'packrc.txt' file. Continue.");
      return 0;
    }
  keyword=(char *)malloc(256);
  next=(char *)malloc(256);
  while (fgets(keyword,BUFSIZE+255,fp)!=NULL)
    {
      stripsp(keyword);
      if( *keyword!='#' && *keyword!='\n' && *keyword!='\0')
	{
	  ptr=keyword;
	  grab_next(&ptr,next);	
	  /* directory name for reading and writing files: not longer
	   * used: java sends full pathnames. */
	  if (!strncmp(next,"CP_PATH",7)) {
	    stripsp(ptr);
	    strcpy(next,".");
	    if (*ptr!='#')
	      grab_next(&ptr,next);
	    strcpy(packing_dir_name,next);
	  }
	  /* postscript file */
	  else if (!strncmp(next,"PS_FILE",7) 
		   && grab_next(&ptr,next) && (strcspn(next,"#")>0) )
	    {
	      strcpy(print_file_name,next);
	    }
	  /* custom postscript file */
	  else if (!strncmp(next,"CUSTOM_FILE",11) 
		   && grab_next(&ptr,next) && (strcspn(next,"#")>0) )
	    {
	      strcpy(ps_file_name,next);
	    }
	  /* print command */
	  else if (!strncmp(next,"PRINT_CMD",9))
	    {
	      i=strcspn(ptr,"#\t\r\0\n");
	      strncpy(next,ptr,i); next[i]='\0'; stripsp(next);
	      strcpy(print_cmd,next);
	    }
	  /* function key pre-defined cmd strings */
	  else if (!strncmp(next,"F1_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[1]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[1],buff);
	    }
	  else if (!strncmp(next,"F2_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[2]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[2],buff);
	    }
	  else if (!strncmp(next,"F3_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[3]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[3],buff);
	    }
	  else if (!strncmp(next,"F4_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[4]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[4],buff);
	    }
	  else if (!strncmp(next,"F5_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[5]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[5],buff);
	    }
	  else if (!strncmp(next,"F6_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[6]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[6],buff);
	    }
	  else if (!strncmp(next,"F7_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[7]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[7],buff);
	    }
	  else if (!strncmp(next,"F8_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[8]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[8],buff);
	    }
	  else if (!strncmp(next,"F9_KEY:",7)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[9]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[9],buff);
	    }
	  else if (!strncmp(next,"F10_KEY:",8)
		   && grab_line(buff,&ptr,BUFSIZE-1))
	    {
	      ftn_keys[10]=
		(char *)calloc((size_t)(strlen(buff)+1),sizeof(char));
	      strcpy(ftn_keys[10],buff);
	    }
	  
	  /* help file name */
	  else if (!strncmp(next,"CP_HELP_FILE",12))
	    {
	      if (grab_next(&ptr,next) 
		  && sscanf(next,"%s",CPHelp))
		{
		  if (CPHelp[0]=='~')
		    {
		      strcpy(file_name,home_dir);
		      strcat(file_name,CPHelp+1);
		      strcpy(CPHelp,file_name);
		    }
		}
	      if (CPHelp[0]!='\0' || !stat_file(CPHelp)) CPHelp[0]='\0'; 
	      /* look elsewhere back in calling routine */
	    }
	}
    } /* end of while */
  free(keyword);
  free(next);
  fclose(fp);
  return 1;
} /* read_packrc */


/* ============================ hex walking ======================== */

int hex_parallel_call(struct p_data *p,int v,int w,int n,complex *cmod,
		      int show)
/* Try to draw parallelogram (treat complex as though it were hex). 
show=1 means to draw faces and to compute complex modulus of 
parallelogram. */
{
	int hold,hold1,hold2,count,v1,v2,v3,v4;
	complex z0,z1,z2,zz;
	struct Vertlist *facelist,*ftrace;
	struct Vertlist *dlist,*curve1,*ctrace1,*curve2,*ctrace2;

	cmod->re=cmod->im=0.0;
	if (show) {
	  dlist=try_hex_pg(p,v,w,n,&v1,&v2,&v3,&v4);
	  sa_layout_facelist(p,7,&dlist,1,0,1); /* fill with bg */
			/* note: clears list*/
	}
	ftrace=facelist=try_hex_pg(p,v,w,n,&v1,&v2,&v3,&v4);
	while (ftrace && ftrace->next) ftrace=ftrace->next;
	if (ftrace && ftrace->v!=facelist->v) /* didn't close up */
	 {vert_free(&facelist); return 0;}
/* create lists to get to corners */
	ftrace=facelist;
	if (ftrace!=NULL) {
	  curve1=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	  ctrace1=curve1;
	  ctrace1->v=ftrace->v;
	}
	else {vert_free(&facelist); return 0;}
	ftrace=ftrace->next;
	count=1;
	while (ftrace!=NULL && count < (2*n)-1) {
 	  ctrace1=ctrace1->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	  ctrace1->v=ftrace->v;
	  count++;
	  ftrace=ftrace->next;
	}
	if (!ftrace || !(ftrace->next)) 
	 {vert_free(&facelist);vert_free(&curve1);return 0;}
	hold1=ctrace1->v;  /* have to save this second-to-last face */
	ctrace1=ctrace1->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	ctrace1->v=ftrace->v;
	ctrace1=ctrace1->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	ctrace1->v=hold=curve1->v; /* close up */
	if (show) {
	  sa_layout_facelist(p,7,&curve1,1,0,1); /* clears curve1 */
	  z0=p->packR_ptr[v1].center; /* first corner */
	  z1=p->packR_ptr[v2].center; /* second corner */
	}
/* second curve list */
	curve2=ctrace2=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
 	ctrace2->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	ctrace2->v=hold; /* want start/end at same face as curve1 */
	ctrace2=ctrace2->next;
	ctrace2->v=hold2=ftrace->v;
	ftrace=ftrace->next;
	count=1;
	while (ftrace!=NULL && count < (2*n)-1) {
	  ctrace2=ctrace2->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	  ctrace2->v=ftrace->v;
	  count++;
	  ftrace=ftrace->next;
	}
	if (!ftrace) 
	 {vert_free(&facelist);vert_free(&curve1);vert_free(&curve2);return 0;}
	ctrace2=ctrace2->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	ctrace2->v=hold1;
	ctrace2=ctrace2->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	ctrace2->v=hold2;
	ctrace2=ctrace2->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	ctrace2->v=curve2->v; /* takes some doing to close up */
	if (show) {
	  sa_layout_facelist(p,7,&curve2,1,0,1); /* clears curve2*/
	  z2=p->packR_ptr[v3].center; /* third corner */
	}
/* later, may want to save curve1, curve2 as generators */
	vert_free(&curve1);vert_free(&curve2);
/* now to compute modulus */
	if (show) {
	  zz=csub(z1,z0);
	  if (cAbs(zz)> .0000001)
		*cmod=cdiv(csub(z2,z1),zz);
	  else {cmod->re=0.0;cmod->im=0.0;}
	}
	return 1;
} /* hex_parallel_call */
	
int j_ftn_call(struct p_data *p,char *datastr)
     /* call to build complex for classical j_function */
{
  int n0,n1,maxsize;

  if (sscanf(datastr,"%d %d %d",&n0,&n1,&maxsize)<3
      || n0<1 || n1<1 || maxsize< n0+n1+3) {
    sprintf(msgbuf,"Usage: j_ftn n0 n1 maxsize. n0, n1 are desired orders,"
      " maxsize limits size");
    emsg();
    return 0;
  }
  return (build_j_function(p,n0,n1,maxsize));
} /* j_ftn_call */

 
int h_dist_call(char *datastr)
     /* hyp distance in disc or uhp */
{
  int uhp_flag=0;
  double dist;
  complex z,w;

  stripsp(datastr);
  if (datastr[0]=='-')
    {
      if (datastr[1]!='u') return 0;
      datastr[0]=' ';
      datastr[1]=' ';
      uhp_flag=1;
    }
  sscanf(datastr,"%lf %lf %lf %lf",&z.re,&z.im,&w.re,&w.im);
  dist=h_dist(z,w,uhp_flag);
  if (fabs(dist)<m_toler)
    {
      sprintf(msgbuf,"h_dist: points essentially identical.");
      msg();
      return 0;
    }
  else if (dist<(-1)*m_toler)
    {
      sprintf(msgbuf,"h_dist: one or more points on ideal bdry");
      msg();
      return 0;
    }
  else
    {
      sprintf(msgbuf,"h_dist = %f ",(singledouble)dist);
      msg();
      return 1;
    }
} /* h_dist_call */

int focus_call(struct p_data *p,char *datastr)
     /* change screen view so specified point, -v <x y>, or specified
	vert, -v <v>, or active node (default) is at the center. 
	-vs <x> <v> option also sets screen width to 2*x*rad(v) in hyp/eucl.
	0 on error. */
{
  int v=0,front;
  double rad=0.0,x=0.0,hx,hy;
  complex ctr;
  char *ptr;
  struct s_data *q=p->screen;

  /* use given x,y location. For sphere, this is actually the
     projected point, (y,z), on the view. */
  if ((ptr=after_char(datastr,"z"))) { 
    if(sscanf(ptr,"%lf %lf",&ctr.re,&ctr.im)!=2) return 0;
  }
  /* use center of specified vertex */
  else if ((ptr=after_char(datastr,"v"))) {
    if (*ptr=='s' && (!grab_next(&(++ptr),next) 
		     || !sscanf(next,"%lf",&x)
		     || x<=0.0))
      return 0;
    if ((v=grab_one_vert(p,&ptr)) || (v=p->active_node)) {
      if (p->hes<0) 
		h_to_e_data(p->packR_ptr[v].center,p->packR_ptr[v].rad,&ctr,&rad);
      else if (p->hes==0) {
		ctr=p->packR_ptr[v].center;
		rad=p->packR_ptr[v].rad;
      }
      else {
		ctr=ss_view(p->screen,
		    p->packR_ptr[v].center,1,&front);
		if (!front) return 1;
		ctr=s_pt_to_visual_plane(ctr);
      }
    }
    else return 0;
  }

  if (x>0.0 && rad>0.0) { /* set_screen size */
    sprintf(buff,"-h %lf",2.0*x*rad);
    set_screen(p,buff);
  }
  hx=(q->box.rx-q->box.lx)*0.5;
  hy=(q->box.ry-q->box.ly)*0.5;
  q->box.lx=ctr.re-hx;
  q->box.rx=ctr.re+hx;
  q->box.ly=ctr.im-hy;
  q->box.ry=ctr.im+hy;
  return 1;
} /* focus_call */

int call_torus_t(struct p_data *p,char *datastr)
{
  int flag=0;
  complex tau;

  tau=torus_t(p,&flag);
  if (!flag)
    {
      sprintf(msgbuf,"The modulus of p%d is %.12f %.12f.",
	      pack_num(p),(singledouble)tau.re,(singledouble)tau.im);
      msg();
      return 1;
    }
  switch (flag)
  {
  case '1': 
    {
      sprintf(msgbuf,"torus_t: p%d is not a torus,",pack_num(p));
      break;
    }
  case '2':
    {
      sprintf(msgbuf,"torus_t: no edge-pairings available.");
      break;
    }
  default:
    {
      sprintf(msgbuf,"torus_t: some error.");
      break;
    }
  } /* end of switch */
  emsg();
  
  return 0;
} /* call_torus_t */

int gen_islands_call(struct p_data *p,char *datastr)
{
  int seed;

  seed=p->packK_ptr[p->bdry_starts[1]].flower[1];
  return (gen_islands(p,seed,datastr));
} /* gen_islands_call */

int slit_complex_call(struct p_data *p,char *datastr)
{
  int val,hits;
  struct Vertlist *vlist=NULL;
  char *endptr;

  if (!(vlist=Node_link_parse(p,datastr,&endptr,&hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength))
      || !vlist->next)
    return 0;
  val=slit_complex(p,vlist);
  vert_free(&vlist);
  return val;
} /* slit_complex_call */

int set_screen_call(struct p_data *p,char *datastr)
     /* need to intervene only in case of -q option (copying
	size of another pack) */
{
  int pnum;
  char *spot;

  if (!(spot=strstr(datastr,"-q"))) return set_screen(p,datastr);
  /* fixup: if -q is found, rest of options are ignored */
  spot+=2;
  if (!sscanf(spot,"%d",&pnum) || pnum<0 || pnum>NUM_PACKS
      || !packdata[pnum].status || !packdata[pnum].screen)
    return 0;
  p->screen->box=packdata[pnum].screen->box;
  p->screen->pix_per_unit
    =(int)((p->screen->pix_box.ry-p->screen->pix_box.ly)
	   /(p->screen->box.ry-p->screen->box.ly));
  return 1;
} /* set_screen_call */

int dilatation_call(struct p_data *p,struct p_data *q,double *kmax,int *face)
     /* Compute dilatation of map between two packings. The packings
must be euclidean for now; The call returns the maximal dilatation 
after running through all the faces. */
{
  int f;
  double dil;
  struct R_data *pR_ptr2,*pR_ptr;

  /* fixup: need datastr specifying faces, default to all. */

  *kmax=1.0;*face=1;
  if (!p->status || !q->status || p->hes || q->hes
      || p->facecount!=q->facecount) return 0;
  pR_ptr=p->packR_ptr;
  pR_ptr2=q->packR_ptr;
  for (f=1;f<=p->facecount;f++)
    {
      face_dilatation(p,q,f,&dil);
      if (dil>(*kmax))
	{
	  *face=f;
	  *kmax = dil;
	}
    }
  sprintf(msgbuf,"Maximal dilatation between p%d and p%d "
	  "is %f at face %d.",pack_num(p),pack_num(q),(singledouble)*kmax,(int)*face);
  msg();
  return *face;
} /* dilatation_call */
    
int face_error_call(struct p_data *p,char *datastr)
     /* read crit>1.0 and facelist; set plot_flag >0  for faces in
facelist (default to all faces) having max ratio 
     (dist between circles)/(intended dist between circles)
which is > crit or .lt. 1/crit. Return 0 on error.*/
{
  int hits,count;
  double crit;
  char *endptr,*nextpoint;
  struct Vertlist *facelist=NULL;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || !sscanf(next,"%lf",&crit)
      || crit<=1.0)
    {
      sprintf(msgbuf,"usage: face_error {x} {f..} with x .gt. 1.");
      emsg();
      return 0;
    }
  stripsp(nextpoint);
  if (*nextpoint=='\0') strcpy(nextpoint,"a\0"); /* default to all */
  facelist=face_link_parse(p,nextpoint,&endptr,&hits);
  if (!facelist) hits=p->facecount;
  if ((count=face_error(p,crit,facelist))<0) /* error */
    {
      if (facelist) vert_free(&facelist);
      sprintf(msgbuf,"face_error: some error in checking faces.");
      emsg();
      return 0;
    }
  if (facelist) vert_free(&facelist);
  sprintf(msgbuf,"face_error: found %d poor faces out of %d.",
	  count,hits);
  msg();
  return 1;
} /* face_error_call */

int pin_face_call(struct p_data *p,char *datastr)
     /* Identify most reliably placed vertex of a face; useful with
	'face_error' to fix ill-placed faces in layout. Return 0 on
	error. */
{
  int v,hits;
  char *endptr;
  struct Vertlist *facelist=NULL;
  
  if (!(facelist=face_link_parse(p,datastr,&endptr,&hits))
      || !(v=pin_face(p,facelist->v)))
    {
      if (facelist) vert_free(&facelist);
      return -1;
    }
  sprintf(msgbuf,"pin_face: most reliably place vert for "
	  "face %d is v=%d.",facelist->v,v);
  msg();
  vert_free(&facelist);
  return v;
} /* pin_face_call */
  

int brooks_call(struct p_data *p,char *datastr)
     /* Creating Brooks packings of quadrilaterals; return 0 on error. */
     /* fixup: have to get capability to read data from script, indicated
	with a new "-s" flag (see cp_call.c). */
{
  int vert,top_vert,count=0,flag=1,c,colr=0,fileflag=1;
  char *nextpoint,command[128],next[BUFSIZE];
  char datasource[BUFSIZE];
  struct Vertlist *VH_list,*vtrace;
  FILE *fp;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) return 0;
  if (next[0]=='-') /* options */
    {
      if (strchr(next,'a') || strchr(next,'A')) flag=2; /* alternating */
      if (strchr(next,'c')) colr=1; /* colors */
      if (strchr(next,'s')) fileflag=0; // data is here, not in file
    }
  else nextpoint=datastr;

  /* now get vert, top_vert, and filename/data */
  if (sscanf(nextpoint,"%d %d %1024s",&vert,&top_vert,datasource)!=3
      || vert<1 || vert>p->nodecount || p->packK_ptr[vert].bdry_flag
      || p->packK_ptr[vert].num!=4 || nghb(p,vert,top_vert)<0) {
    sprintf(msgbuf,"usage: brooks v w {filename} (or data HVHH...).");
    emsg();
    return 0;
  }
  VH_list=vtrace=(struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));

  if (fileflag) { // open/read file for data
    if (!(fp=fopen(datasource,"r")) ) {
      sprintf(msgbuf,"brooks: failed to find file %s.",datasource);
      emsg();
      return 0;
    }
    while (fp && (fscanf(fp,"%127s",command)!=EOF)) {
      if (!strcmp(command,"BROOKS_SEQUENCE:")) { /* found key word */
	while (fp && (c=fgetc(fp)) && c!=EOF
	       && (c=='V' || c=='v' || c=='H' || c=='h'
		   || c==' ' || c=='\t' || c=='\n')) {
	  if (c=='V' || c=='v') { /* vertical */
	    vtrace=vtrace->next=(struct Vertlist *)
	      calloc((size_t)1,sizeof(struct Vertlist));
	    vtrace->v=1;
	    count++;
	  }
	  else if (c=='H' || c=='h') { /* horizontal */ 
	    vtrace=vtrace->next=(struct Vertlist *)
	      calloc((size_t)1,sizeof(struct Vertlist));
	    vtrace->v=2;
	    count++;
	  }
	  else if (c=='E') { /* end */
	    fclose(fp);
	    fp=NULL;
	  }
	} /* end of while */
	if (fp) fclose(fp);
      }
    } /* end of outer while */
    if (fp) fclose(fp);
  }
  else { // read data directly
  	char C;
    int n=strlen(datasource);
    count=0;
    vtrace=VH_list;
    while (count<n) {
      C=datasource[count];
      if (C=='V' || C=='H') {
	vtrace=vtrace->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	vtrace->v=1;
	count++;
      }
      else if (C=='H' || C=='h') { /* horizontal */ 
	vtrace=vtrace->next=(struct Vertlist *)
	  calloc((size_t)1,sizeof(struct Vertlist));
	vtrace->v=2;
	count++;
      }
      else break;
    } /* end of while */
  }
  
  if (VH_list==vtrace) { /* didn't find any */
    sprintf(msgbuf,"brooks: didn't find any V's and H's.");
    emsg();
    free(VH_list);
    return 0;
  }
  else { /* found sequence; pack it. */
    vtrace=VH_list->next; /* didn't fill first spot */
    VH_list=vtrace;
    count=brooks_construct(p,vert,top_vert,VH_list,flag,colr);
    vert_free(&VH_list);
    if (count) {
      sprintf(msgbuf,"Add %d vertices for Brooks parameter "
	      "at vertex %d.",count,vert);
      msg();
    }
    return count;
  }

  if (VH_list) free(VH_list);
  return 0;
} /* brooks_call */	  

int locate_call(struct p_data *p,char *datastr)
/* Find circles/faces containing given point. Currently, only options
are to send list of labels for circles/faces for display. */
{
  int flag=1,count=0,dum;
  int msg_flag=3;// bit 1=on canvas, 2=in scratch window, 3=both (default)
  double a;
  complex z;
  char *nextpoint,next[BUFSIZE];
  struct Vertlist *list=NULL;

  nextpoint=datastr;
  /* manditory option flag; may augment options later */
  if (!grab_next(&nextpoint,next) || next[0]!='-') return 0;
  switch(next[1]) {
    case 'c': {flag=1;break;} /* circles */
    case 'f': {flag=2;break;} /* faces */
  }
  if (!sscanf(nextpoint,"%lf %lf",&(z.re),&(z.im))) return 0;
  if (p->hes>0) { // convert to real sph point 
  	if ((a=(z.re*z.re+z.im*z.im))>=1.0) return 0; // not in sphere 
    // so, find true point 
    z=ss_view(p->screen,proj_vec_to_sph(sqrt(1.0-a),z.re,z.im),0,&dum); 
  }
  if (flag==2 && (list=tri_search(p,z))) { /* faces */
    count=labellist(p,list,msg_flag,1);
  }
  else if (flag==1 && (list=cir_search(p,z))) { /* circles */
    count=labellist(p,list,msg_flag,0);
    p->active_node=list->v;
  }
  vert_free(&list);
  return count;
} /* locate_call */

int read_colortable(struct p_data *p,char *datastr)
     /*   "opts" bit encoding: 
     bit 1--> tell Java GUI to read colors (for this canvas) from file
     bit 2--> send back/foreground
     if bit 1 is off, send local colors. */
{
  int count=0,opts=3,sflag=0;
  char name[64],filename[NAME_MAX],next[256],*nextpoint;
  FILE *fpcol=NULL;

  filename[0]=name[0]='\0';
  nextpoint=datastr;
  sprintf(name,"default_clr.tbl");
  while (grab_next(&nextpoint,next) && next[0]=='-') {
    /* color table name, must come first */
    if (next[1]=='n') {
      if (!grab_next(&nextpoint,next) || !sscanf(next,"%63s",name))
	return 0;
    }

    /* local mode: read color file locally, then send to GUI */
    else if (next[1]=='l') opts &= 2; 
    /* set colors here in SA (so they are right in postscript output) */
    else if (next[1]=='s') sflag=1;
  }
  if (sflag && strlen(name)) {
	read_rgb_file(name);
  }
  if (strlen(next))   /* last string is filename */
    sscanf(next,"%256s",filename); 
} /* read_colortable */

int embed_call(struct p_data *p,char *datastr)
/* for embedding combinatorics of one packing in another */
{
  int pnum,a,b,A,B,count=0;
  struct p_data *p2;
  struct Edgelist *matches=NULL;

  if (sscanf(datastr,"%d %d %d %d %d",&pnum,&a,&b,&A,&B)!=5
      || pnum<0 || pnum>NUM_PACKS || a<1 || a>p->nodecount
      || b<1 || b>p->nodecount || A<1 || B<1 
      || !packdata[pnum].status || A>packdata[pnum].nodecount
      || B>packdata[pnum].nodecount)
    return 0;
  p2=&packdata[pnum];
  count=embedding(p,p2,a,b,A,B,&matches);
  if (count) {
	if (p->vertex_map) edge_free(&p->vertex_map);
  	p->vertex_map=matches;
    return 1;
  }
  return 0;
} /* embed_call */

int flip_priority_call(struct p_data *p,char *datastr)
  /* 1/2004: this is special routine for Jeff Berliss's work on edge
     flipping. Idea is to return list of potential edge flips
     to make to move towards hex pattern. */
{
  /* dump former edge list */
  if (p->elist) edge_free(&(p->elist)); 
  p->elist=flip_priority(p,datastr);
  if (p->elist) return 1;
  return 0;
} /* flip_priority_call */
    
int set_radii_call(struct p_data *p,char *datastr)
  /* set radii. */
{
  int n;
  double radius;
  char *nextpoint;

  nextpoint=datastr;
  stripsp(nextpoint);
  n=strcspn(nextpoint," ");
  strncpy(buf,nextpoint,n);
  *(buf+n)='\0';
  if (n==0 ||strpbrk(buf,".")==NULL 
      || sscanf(buf,"%lf",&radius)!=1 ) {
	strcpy(msgbuf,"Indicate real value for radius "
		"first (with decimal pt).");
	emsg();
    return 0;
  }
  if (p->hes>=0 && radius<=0) {
	strcpy(msgbuf,"Radius can't be negative for eucl "
		"and spherical packs.");
	emsg();
    return 0;
  }
  return (set_radii(p,radius,nextpoint+n+1));
} /* set_radii_call */

int beurling_call(struct p_data *p,char *datastr)
/* see beurling_flow.cpp for the meaning of the flags/options */
{
  int qnum,opt=0;
  double x=1.0,xpnt=1.0;
  complex w0={0.0,0.0};
  char *nextpoint;

  /* call: -q<n> [-t <x> -e <a> -s <w0> -k <opt>] */

  nextpoint=datastr;stripsp(nextpoint);
  if (!grab_next(&nextpoint,next) || strncmp(next,"-q",2)
      || (qnum=atoi(next+2))<0 || qnum>=NUM_PACKS ) {
    sprintf(msgbuf,"usage: beurling -q<n> [-t <x> -k <opt>].");
    emsg();
    return 0;
  }
  stripsp(nextpoint);
  while(*nextpoint=='-') { /* more options */
    if (nextpoint[1]=='t') { /* interpolation parameter */
      next[0]='\0';
      grab_next(&nextpoint,next);
      stripsp(nextpoint);
      if (!sscanf(nextpoint,"%lf",&x) || x<0.0) {
	sprintf(msgbuf,"usage: beurling, -t <x>, illegal x; has been reset to 1.0");
	emsg();
	x=1.0;
      }
    }
    else if (nextpoint[1]=='e') { /* exponent */
      next[0]='\0';
      grab_next(&nextpoint,next);
      stripsp(nextpoint);
      if (!sscanf(nextpoint,"%lf",&xpnt) || xpnt>0.0) {
	sprintf(msgbuf,"usage: beurling, -e <a>, must have a>0.0.");
	emsg();
      }
    }
    else if (nextpoint[1]=='k') { /* option */
      next[0]='\0';
      grab_next(&nextpoint,next);
      stripsp(nextpoint);
      if (!sscanf(nextpoint,"%d",&opt) || opt<0) {
	sprintf(msgbuf,"usage: beurling, -k <option>, option>=0.");
	emsg();
	opt=0;
      }
    }
    else if (nextpoint[1]=='s') { /* translate by w0. */
      next[0]='\0';
      grab_next(&nextpoint,next);
      stripsp(nextpoint);
      if (sscanf(nextpoint,"%lf %lf",&(w0.re),&(w0.im))!=2) {
	sprintf(msgbuf,"usage: beurling, -s <w0>, must give w0.re and w0.im.");
	emsg();
      }
    }
    else grab_next(&nextpoint,next);
  }
  
  return beurling_flow(p,&packdata[qnum],x,xpnt,w0,opt);
} /* beurling_call */

int weld_call(char *datastr)
/* Parse datastr and call weld_using_map() (this does everything),
   weld() with weldlist already in file, readweldmap() to read
   real weldmap into some pathlist for later welding use (note: it
   is not a normal closed path in the plane). Return 0 if error. */

/* fixup: for now, only call weld_using_map */
{
  int pnum1,pnum2,v1,v2,opt_flag=0;
  char weldmapfile[NAME_MAX],*nextpoint,*spot;
  struct p_data *p,*q;
	
  nextpoint=datastr;
  weldmapfile[0]='\0';
  if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum1)
      || pnum1<0 || pnum1>=NUM_PACKS 
      ||!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum2)
      || pnum2<0 || pnum2>=NUM_PACKS)
    goto BOMB;
  p=&packdata[pnum1];
  q=&packdata[pnum2];
  if (!(v1=grab_one_vert(p,&nextpoint))
      || !(v2=grab_one_vert(q,&nextpoint))) 
    goto BOMB;

  /* fixup: need to parse the various options here, send option flag.
     for now, just do full boundary, map packing. */
  
  if ((spot=strstr(nextpoint,"-w"))) { /* do the adjoin operation
					  also -- results in p */
    *spot=' ';
    *(spot+1)=' ';
    *(spot+2)=' ';
    opt_flag |=0010; /* set bit */
  }
  /* is there a filename? */
  strcpy(weldmapfile,"\0"); /* empty path = identity map. */
  if ((spot=strstr(nextpoint,"-f"))) {
    *spot=' ';
    *(spot+1)=' ';
    if (grab_next(&spot,next) && strlen(next)>2)
	sscanf(next,"%s",weldmapfile);
  }

  if (weld_using_map(p,q,v1,v2,-1,-1,weldmapfile,opt_flag)>0) { 
    /* fixup: should read in count of edges to be welded from p and q;
       '-1' indicates to use all. */ 
    complex_count(p,0);complex_count(q,0);
    fillcurves(p);fillcurves(q);
    facedraworder(p,0);facedraworder(q,0);
    set_aim_default(p);set_aim_default(q);
    p->active_node=p->beta; q->active_node=q->beta;
    sprintf(msgbuf,"weld: seemed to succeed.");
    msg();
    return 1;
  }

 BOMB:
  strcpy(msgbuf,"weld: usage: p q v w (for now).");
  emsg();
  return 0;
} /* weld_call */

int h_g_bar_call(char *datastr)
  /* creating harmonic functions f=h+conj(g). (See h_g_bar.cpp.) */
{
  int j,hit=0;
  struct p_data *Hp=&(packdata[0]),*Gp=&(packdata[1]),*Fp=&(packdata[2]);

  if (!Hp->status || !Gp->status || Hp->hes !=0 || Gp->hes != 0
      || Hp->nodecount != Gp->nodecount || !copy_pack(Hp,Fp)) {
    sprintf(msgbuf,"h_g_bar error: packs euclidean? same size?");
    emsg();
    return 0;
  }
  struct K_data *HK_ptr=Hp->packK_ptr;
  struct K_data *GK_ptr=Gp->packK_ptr;
  struct R_data *HR_ptr=Hp->packR_ptr;
  struct R_data *GR_ptr=Gp->packR_ptr;

  for (j=1;(j<=Hp->nodecount && !hit);j++) {
    /* bdry radii must be larger in Hp */
    if (HK_ptr[j].bdry_flag) {
      if (GR_ptr[j].rad>HR_ptr[j].rad) hit++;
    }
    /* Hp should be locally univalent */
    else if (fabs(HR_ptr[j].curv-2.0*M_PI)>.0001) {
      hit++;
    }
  }
  if (hit || !h_g_bar(Hp,Gp,Fp)) {
    sprintf(msgbuf,"h_g_bar error: Some problem with data.");
    emsg();
    return 0;
  }
  sprintf(buf,"set_screen -p2 a");
  handle_cmd(buf,&current_p);
  sprintf(msgbuf,"h_g_bar: p2 contains the data for the harmonic map.");
  msg();
  return 1;
} /* h_g_bar_call */

int Map_call(char *datastr)
/* Concerns Java "Map" call to map_activity */
{
	int pnum,qnum;
	char *nextpoint=datastr;
	
	if (strstr(datastr,"-x")) return (closeMap_call());
	if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum)
	|| pnum<0 || pnum>NUM_PACKS 
	|| !grab_next(&nextpoint,next) || !sscanf(next,"%d",&qnum)
	|| qnum<0 || qnum>NUM_PACKS)
	 return 0;
   	  int ans =map_activity(&packdata[pnum],&packdata[qnum],nextpoint);
   	  return (ans);
} /* Map_call */

/**
 * set line thickness for postscript; if jflag is set, then report
 * to java also. Set up to exchange sizes in 'pixels'. For a screen of
 * with w, a 'pixel' is w/(2.2*272). (So variable 'line_thick' represents
 * one pixel.
 */
int set_line_thick(struct p_data *p,int jpixels,int jflag) {
	line_thick=jpixels;
	ps_linewidth=(.2+line_thick)*PS_UNIT_LINEWIDTH;
	if (custom_fp) /* if custom print file is open, set new 'ourlinewidth' */
	    fprintf(custom_fp,"%f ourlinewidth\n",(singledouble)ps_linewidth);
	if (jflag) line_thick_mode(p,line_thick);
	return 1;
}
