/* sa_cmd.c  to parse and call routines sent by GUI. */

#include "sa_head.h"
#include "Interface.h"

int cmdno;

int sort_cmd(const char *cmd_stream,int *targ_p)
/* pick off individual commands in a stream. Apply to pack targ_p. 
Return positive number if at least one parsing succeeded. 
(4/03) Strings from the Java front end will generally have just
one command at a time, and may be proceeded by "cmdno n" command,
which attaches a number tag to the command.
Note: can have nested calls to this when executing named script cmds. */
{
  int count=0,flag=0,n;
  char *command,local_stream[BUFSIZE],*token_ptr;

  cmdno=1;
  strncpy(local_stream,cmd_stream,BUFSIZE-1);
  local_stream[BUFSIZE-1]='\0';
  token_ptr=local_stream;
  while (*(token_ptr)==';' || *(token_ptr)==' ') token_ptr++;
  stripsp(token_ptr);
  if ( (n=strcspn(token_ptr,";"))>0) {
	/* get first command. fixup: command might be just white space */
   	command=token_ptr;
    /* position pointer for next command */
    token_ptr += n;
    if (*(token_ptr)==';') {
	  *(token_ptr)='\0';
	  token_ptr += 1;
	  stripsp(token_ptr);
	}
	while (*(token_ptr)==';' || *(token_ptr)==' ') token_ptr++;
      /* process first command */
      stripsp(command);
      if ( (flag=count=(handle_cmd(command,targ_p)!=0) ) 
	   && *command!='!' && !cmd_search_depth) 
	strcpy(cmd_history,command);
      while ( flag && (n=strcspn(token_ptr,";"))>0 ) {
	    flag=0;
	    /* get current command */
	    command=token_ptr;
	    /* position pointer for next command */
	    token_ptr += n;
	    if (*(token_ptr)==';') {
	      *(token_ptr)='\0';
	      flag=1;
	      token_ptr += 1;	      
	      stripsp(token_ptr);
	    }
	    while (*(token_ptr)==';' || *(token_ptr)==' ') token_ptr++;
	    /* process current command */
	    stripsp(command);
	    if (!handle_cmd(command,targ_p)) {
	      flag=0;
	      continue;
	    }
	    if (!cmd_search_depth && *command!='!') {
	      if (count) {
		    strcat(cmd_history,";");
		    strcat(cmd_history,command);
		  }
	      else strcpy(cmd_history,command);
	    }
	    count++;
      } /* end of while */
  }
// sprintf(msgbuf,"%s",cmd_stream);
//  msg();
  return count;
} /* sort_cmd */

int handle_cmd(char *command,int *targ_p)
/* takes care of a single command only */
{
  int pnum,count,hflag=0,i;
  char *dpoint;
	
  if (*targ_p<0 || *targ_p>=NUM_PACKS) *targ_p=current_p;
  stripsp(command);
  dpoint=command;
  if (!grab_next(&dpoint,cmd)) return 0; /* command was empty */
  if (*dpoint==' ') dpoint++; /* should be a blank next */
  stripsp(dpoint);
  if (strncmp(dpoint,"-p",2)) pnum=*targ_p;
  else /* pack number is specified */
    {
      dpoint +=2;
      if (!grab_next(&dpoint,next) 
	  || (next[0]!='-' && strlen(next)>1)) return 0;
      if (next[0]=='a' || next[0]=='-') pnum=current_p;
      else if ((i=atoi(next))>=0 && i<NUM_PACKS) pnum=i;
      else return 0; /* no pack specified */
    }
  strcpy(data,dpoint); /* rest is data */
  stripsp(data);
  return (count=parse_cmd(&packdata[pnum],&hflag,targ_p));
} /* handle_cmd */

int parse_cmd(struct p_data *p,int *hflag,int *targ_p)
 /* set hflag to 1 to put this command (assuming successful) 
into the history file (2/2002: I removed the history file
functionality). targ_p is the target packing, passed through 
from sort_cmd in case sort_cmd is called recursively. */
{
  int v,w,p1,spec_num,i=0,count=0,vert,n;
  int pass,d,j,jj,hits,qnum,pnum=pack_num(p);
  char *dpoint=NULL,*nextpoint=NULL,*endptr=NULL;
  char filename[NAME_MAX],named_cmd[64];
  double radius,factor,ang,dum,x,y;
  struct Vertlist *vertlist=NULL,*trace;
  struct Edgelist *holde=NULL,*elist=NULL;
  struct Vertlist *holdv=NULL;
  struct s_data *q;
  complex ctr,a;
  Mobius *newMob=NULL;
  Pix_box new_pixbox;

  dpoint=data;

/* ================================================================== */
  /* UNLOCKED cases (locked packings skip these commands) */
/* ================================================================== */

  if (!p->locks)
    {

      switch (cmd[0]) /* unlocked, either GUI or cmd mode */
	{
case 'a':
 {
				/* ============ add barycenter ======== */
   if (!strncmp(cmd,"add_b",5))
     {
       return (*hflag=add_barys(p,data));
     }
   
				/* ============ add circle ============ */
   if (!strncmp(cmd,"add_cir",7))
     {
       return (*hflag=add_circles_call(p,data));
     }
				/* ========= add vert to edge =========== */
   if (!strncmp(cmd,"add_e",5))
     {
       return (*hflag=split_flips(p,data,1));
     }
				/* ========== add generations ======= */
   if (!strncmp(cmd,"add_gen",7)) {
   	return (add_generation(p,data));
     	
 /*      if (sscanf(data,"%d %d",&n,&d)!=2 
	   || n<1 || d<3 || d>MAX_PETALS) return 0;
       jj=1;
       for (i=1;i<=n;i++)
	 {
	   if (!p->num_bdry_comp)
	     {
	       sprintf(msgbuf,"No boundary component "
		       "for 'add_gen'.");
	       emsg();
	       goto BREAK_OUT;
	     }
	   else for (j=1;j<=p->num_bdry_comp;j++)
	     {
	       sprintf(next,"%d %d %d",p->bdry_starts[j],p->bdry_starts[j],d);
	       count += add_layer(p,next,&hits);
	       jj *= hits;
	     }
	 }
     BREAK_OUT: 
       if (!jj)
	 {
	   sprintf(msgbuf,"add_gen: Some circles "
		   "exceeded %d degree.",d);
	   emsg();
	 }
       return (*hflag=count);	*/	
     } 
				/* ========== add ideal vertex ======= */
   if (!strncmp(cmd,"add_ideal",9))
     {
       return (add_ideal_call(p,data));
     }
				/* ============== add layers ========== */
   if (!strncmp(cmd,"add_lay",7))
     {
       j=add_layer(p,data,&hits);
       if (!hits) {
       	sprintf(msgbuf,"add_layer: some circles exceeded the specified degree");
       	emsg();
       }
       return (*hflag=j);
     }
						     /* ========== adjust_rad ========= */
   if (!strncmp(cmd,"adjust_r",8)) {
   	return adjust_radii(p,data);
   }
                               /* =========== alt_rect ============= */
   if (!strncmp(cmd,"alt_rect",8)) {
     return (alt_rect_pack(p,data));
   }
				/* ============= set alpha ========== */
   if (!strncmp(cmd,"alpha",5) )
     {
       if ((holdv=node_link_parse(p,data,&endptr,&hits))!=NULL)
	 {
	   v=holdv->v;
	   vert_free(&holdv);
	   return (set_alpha(p,v));
	 }
       else return 0;
     }
   break;
 }
case 'b':
 {
                                /* ======= bary_refine ============ */
   if (!strncmp(cmd,"bary_r",6)) {
     return(bary_refine(p));
   }
                                /* ======= beurling_flow ============ */
   if (!strncmp(cmd,"beurl",5)) {
     return(beurling_call(p,data));
   }
				/* ============== set beta =========== */
   if (!strncmp(cmd,"beta",4) )
     {
       if (sscanf(data,"%d",&vert)==1) 
	 return (set_beta(p,vert));
       else return 0;
     }
   				/* ============== brooks =========== */
   if (!strncmp(cmd,"brooks",6) )
     {
       return (brooks_call(p,data));
     }
   break;
 }
case 'c':
 {
				/* ============== cir_invert ======= */
   if (!strncmp(cmd,"cir_invert",10))
     {
       return (*hflag=cir_invert_call(p,data));
     }
				/* ============== compute: centers ======= */
   if (!strncmp(cmd,"compute_cent",12))
     {
       return (comp_pack_centers(p,0,0,2,okerr));
     }
				/* ============== cookie cutting ======= */
   if (!strncmp(cmd,"cookie",6))
     {
       return (*hflag=cookie_call(p,data));
     }
   break;
 }
case 'd':
 {
				/* ============== double ========== */
   if (!strncmp(cmd,"double",12))
     {
       if (!double_K(p,data))
	 {
	   sprintf(msgbuf,"doubling failed.");
	   emsg();
	   return 0;
	 }
       else return (*hflag=1);
     }
                                /* ======= doyle_annulus ========== */
   if (!strncmp(cmd,"doyle_a",7)) {
     return doyle_annulus(p,data);
   }
   break;
 }
case 'e':
 {
				/* ============ enclose ============ */
   if (!strncmp(cmd,"enclose",7))
     {
       if ((*hflag=enfoldcall(p,data))) return 1;
       return 0;
     }
				/* ========== e_to_h (outdated) ======== */
   if (!strncmp(cmd,"e_to_h",6))
     {
       if (p->hes<0)
	 {strcpy(msgbuf,"Pack is already hyperbolic.");
	 emsg();return 1;}
       geom_to_h_call(p,data);
       return (*hflag=1);
     }
				/* ========== e_to_s (outdated) ========== */
   if (!strncmp(cmd,"e_to_s",6))
     {
       if (p->hes>0)
	 {strcpy(msgbuf,"Pack is already spherical.");emsg();return 1;}
       geom_to_s_call(p,data);
       return (*hflag=1);
     }
                               /* ========== erf ================== */
   if (!strncmp(cmd,"erf",3)) {
     int i,j,n;
     if (sscanf(data,"%d %d %d",&i,&j,&n)==3 
	 && i>=0 && j>=0 && i<NUM_PACKS && j<NUM_PACKS)
       return (erf_ftn(&packdata[i],&packdata[j],n));
     else return 0;
   }
   break;
 }
case 'F':
 {
				/* ==== redo all housekeeping ======== */
   if (!strncmp(cmd,"Fix",3))
     {
       return (*hflag=fix_call(p," -F"));
     }
   break;
 }
case 'f':
 {
				/* == fix: comp centers and-or ang sums === */
   if (!strncmp(cmd,"fix",3))
     {
       return (*hflag=fix_call(p,data));
     }
				/* ============= flat_hex ========= */
   if (!strncmp(cmd,"flat_h",6))
     {
       return (*hflag=flat_hex(p,data));
     }
                                /* ============= flip_priority ========= */
   if (!strncmp(cmd,"flip_pr",7)) {
     return (*hflag=flip_priority_call(p,data));
   }
				/* ============= flip edge ========= */
   if (!strncmp(cmd,"flip",4))
     {
       return (*hflag=split_flips(p,data,2));
     }
   break;
 }
case 'g':
 {
                                /* ========== gen_cut ================ */
   if (!strncmp(cmd,"gen_cut",7))
     {
       return (*hflag=gen_cut_call(p,data));
     }
                                /* ========== gen_cut ================ */
   if (!strncmp(cmd,"gen_islands",11))
     {
       return (gen_islands_call(p,data));
     }

				/* ========== geom_to_[hes] ============= */
   if (!strncmp(cmd,"geom_to_",8)) {
     int pnum=pack_num(p);
     if (!strncmp(cmd+8,"h",1)) {
       if (p->hes<0)
	 {strcpy(msgbuf,"Pack is already hyperbolic.");emsg();return 1;}
       if (geom_to_h_call(p,data)) { // changed geometry
       sprintf(msgbuf,"Pack %d converted to hyperbolic data.",pnum);
       switch_geom(p);
       }
       else sprintf(msgbuf,"Pack %d conversion to hyperbolic failed",pnum);
     }
     else if (!strncmp(cmd+8,"e",1)) {
       if (p->hes==0)
	 {strcpy(msgbuf,"Pack is already euclidean.");emsg();return 1;}
       if (geom_to_e_call(p,data)) { // changed geometry
       sprintf(msgbuf,"Pack %d converted to euclidean data.",pnum);
       switch_geom(p);
       }
       else sprintf(msgbuf,"Pack %d conversion to eucl failed",pnum);
     }
     else if (!strncmp(cmd+8,"s",1)) {
       if (p->hes>0)
	 {strcpy(msgbuf,"Pack is already spherical.");emsg();return 1;}
       if (geom_to_s_call(p,data)) { // changed geometry
       sprintf(msgbuf,"Pack %d converted to spherical data.",pnum);
       switch_geom(p);
       send_sph_view(p);
       }
       else sprintf(msgbuf,"Pack %d conversion to spherical failed",pnum);
     }
     msg();
     set_pack_label(p);
     sa_send_geom(p);
     return (*hflag=1);
   }

   if (!strncmp(cmd,"get_xy",6)) {
     return (get_xyz_data(p,data));
   }
   break;
 }
case 'h':
 {
 			    /* ========== h_g_bar ======== */
   if (!strncmp(cmd,"h_g_b",5)) {
   		return (*hflag=h_g_bar_call(""));
   }
				/* ========== h_to_e (outdated) ======== */
   if (!strncmp(cmd,"h_to_e",6))
     {
       if (p->hes==0)
	 {strcpy(msgbuf,"Pack is already euclidean.");emsg();return 1;}
       geom_to_e_call(p,data);
       return (*hflag=1);
     }
				/* ============ hex_refine ======== */
   if (!strncmp(cmd,"hex_refine",10))
     {
       if ((*hflag=hex_refine(p))) return 1;
       return 0;
     }
				/* ============ hex_walk ======== */
   if (!strncmp(cmd,"hex_walk",8))
     {
       if (sscanf(data,"%d %d %d",&v,&w,&n)!=3) 
	 {strcpy(msgbuf,"Usage: hex_walk v w n");emsg();return 0;}
       if ((*hflag=hex_parallel_call(p,v,w,n,&a,1))) 
	 {
	   sprintf(msgbuf,"hex_walk: pack %d modulus = %f+i%f; "
		   "abs val = %f",
		   pnum,(singledouble)a.re,(singledouble)a.im,
		   (singledouble)cAbs(a));
	   msg();
	   return 1;
	 }
       return 0;
     }
   break;
 }
case 'i':
 {
   				/* ========== infile_* ================ */
/* these are calls for data stored in the current script file, if there is one. */
   if (!strncmp(cmd,"infile_",7)) {
   	if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
   	if (!strncmp(cmd,"infile_read",10)) return callInfileRead(filename,1);
   	if (!strncmp(cmd,"infile_path",10)) return callInfileRead(filename,2);
   	if (!strncmp(cmd,"infile_cmds",10)) return callInfileRead(filename,3);
   	if (!strncmp(cmd,"infile_tree",10)) return callInfileRead(filename,4);
   	if (!strncmp(cmd,"infile_xyz",10)) return callInfileRead(filename,5);
   	if (!strncmp(cmd,"infile_raw",10)) return callInfileRead(filename,6);
   	else return 0;
   }
   break;
 }
case 'j':
 {
				/* ========== j_ftn ============= */
   if (!strncmp(cmd,"legal",5)) {
     struct Edgelist *elist=NULL;
     if ((elist=legal_edges(p,data))) { /* yes, found some illegals */
       edge_free(&Elist);
       Elist=elist;
     }
     return 1;
   }
   if (!strncmp(cmd,"j_ftn",5))
     {
       return (*hflag=(j_ftn_call(p,data)>0));
     }
   break;
 }
case 'l':
 {
 					/* == layout (new, equal to 'fix') === */
   if (!strncmp(cmd,"layout",6)) {
       return (*hflag=fix_call(p,data));
   }
 	                               /* ============= legal ============= */
   if (!strncmp(cmd,"legal",5)) {
     struct Edgelist *elist=NULL;
     if ((elist=legal_edges(p,data))) { /* yes, found some illegals */
       edge_free(&Elist);
       Elist=elist;
     }
     return 1;
   }
   break;
 }
case 'm':
  {
				/* ========== max_pack ============= */
    if (!strncmp(cmd,"max_p",5))
      {
	return (max_pack(p,data));
      }
				/* ============ mend ============== */
    if (!strncmp(cmd,"mend",4) )
      {
	return (*hflag=mend_call(p,data,totalpasses));
      }
				/* ===== molify_centers ============= */
    if (!strncmp(cmd,"molify_cent",11))
      {
	return (molify_call(p,data));
      }
    break;
  }
case 'M':
 {
				/* ========== Mobius_inv ============= */
   if (!strncmp(cmd,"Mobius_inv",10))
     {
       return (apply_Mobius(p,data,-1,Mob));
     }
				/* ========== Mobius ============= */
   if (!strncmp(cmd,"Mobius",6)) // use default Mobius
     {
       return (apply_Mobius(p,data,1,Mob));
     }
   break;
 }
case 'N':
 {
				/* ========== NSpole ============= */
   if (!strncmp(cmd,"NSpole",6))
     {
       return (NSpole(p,data));
     }
   break;
 }
case 'n':
 {
 				/* ========== norm_scale =========== */
   if (!strncmp(cmd,"norm_scale",10))
     {
       norm_scale_call(p,data);
       return 1;
     }
   break;
 }
case 'p':
 {
 				/* =========== pfacered ============ */
   if (!strncmp(cmd,"pfacered",8))
     {
       return (pfacered(p));
     }	
				/* =========== plug ============== */
   if (!strncmp(cmd,"plug",4))
     {
       return (plug_call(p,data));
     }
				/* =========== poison ============ */
   if (!strncmp(cmd,"poison",6))
     {
       set_poison(p,data);
       return 1;
     }
   				/* =========== pre_flip ============ */
   if (!strncmp(cmd,"pre_flip",8))
     {
       if (r_flip) {free(r_flip);r_flip=NULL;}
       if (b_flip) {free(b_flip);b_flip=NULL;}
       b_flip=(struct Edgelist *)
	 calloc(1,sizeof(struct Edgelist));
       b_flip->v=b_flip->w=0;
       sprintf(msgbuf,"Choose vertices to 'flip' with left-mouse");
       msg();
       return 1;
     }
				/* =========== proj_to_s ============ */
   if (!strncmp(cmd,"proj",4))
     {
       if (proj_max_to_s(p,data)){
	 fillcurves(p);
	 set_aim_default(p);
	 p->screen->unitcircle=0;
	 sprintf(msgbuf,"Pack %d converted to spherical data.   ",pnum);msg();
	 set_pack_label(p);
	 sa_send_geom(p);
	 send_sph_view(p);
	 unitcircle_mode(p,0);
	 equator_mode(p,0);
	 return 1;
       }
       return 0;
     }
				/* =========== puncture ============ */
   if (!strncmp(cmd,"puncture",8))
     {
       return (puncture_call(p,data));
     }
   break;	
 }
case 'r':
 {
 				/* ====== random_pack ================== */
   if (!strncmp(cmd,"random_p",8)) {
   		return (random_pack_call(p,data));
   } 				
				/* ====== read = Caution, no confirm === */
   if (!strncmp(cmd,"read",4) 
       && strncmp(cmd,"read_p",6) && strncmp(cmd,"read_c",6))
     {
       if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
       strcpy(buff,packing_dir_name);
       strcat(buff,filename);
       strcpy(filename,buff); /* now have full pathname */
       return (*hflag=readcall(p,filename));
     }
				/* ========= reorient ======= */
   if (!strncmp(cmd,"reorient",8))
     {
       return (*hflag=reverse_orient_call(p));
     }		
				/* ========== repack ======== */
   if (!strncmp(cmd,"repack",6))
     {
       return(repack_call(p,data));
     }
				/* ========= remove bdry circle ======== */
   if (!strncmp(cmd,"rm_cir",6))
     {
       return (*hflag=remove_circle(p,data));
     }		
				/* ============ remove edge ============ */
   if (!strncmp(cmd,"rm_edge",7))
     {
       if (sscanf(data,"%d %d",&v,&w)!=2) return 0;
       return (*hflag=remove_edge(p,v,w));
     }
				/* ============ remove quad vert ======== */
   if (!strncmp(cmd,"rm_quad",7))
     {
       return (*hflag=remove_quad_vert(p,data));
     }
   break;
 }
case 'R':
 {
				/* === Read (directory specified) ====== */

   if (!strncmp(cmd,"Read",4) && strncmp(cmd,"Read_p",6)) {
     if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
     return (*hflag=readcall(p,filename));
   }
   break;
 }
case 's':
 {
                                /* ============= scale_aims ====== */
   if (!strncmp(cmd,"scale_aim",9)) {
     return (scale_aims_call(p,data));
   }
                                /* ============== scale_rad ====== */
   if (!strncmp(cmd,"scale_rad",9)) {
     nextpoint=data;
     if (!grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum)
	 || pnum<0 || pnum>NUM_PACKS || &packdata[pnum]==p
	 || packdata[pnum].nodecount!=p->nodecount
	 || packdata[pnum].hes!=p->hes) {
       sprintf(msgbuf,"usage: scale_rad: must give compatible pack q");
       emsg();
       return 0;
     }
     return (scale_rad_call(p,&packdata[pnum],nextpoint));
   }

				/* ========== eucl scaling ====*/
   if (!strncmp(cmd,"scale",5))
     {
       if (!p->status) return 0;
       if (!strlen(data) || !sscanf(data,"%lf",&factor) 
	   || factor==0)
	 factor=eucl_factor;
       eucl_scale(p,factor);
       return (*hflag=1);
     }
				/* ========== seed ======== */
   if (!strncmp(cmd,"seed",4))
     {
       if (seed_start_call(p,data)) return 1;
       else return 0;
     }
				/* ========== set_aim ======= */
   if (!strncmp(cmd,"set_aim",7)) {
       return (*hflag=set_aim_call(p,data));
   }
   				/* ============= set_function ========= */
   if (!strncmp(cmd,"set_func",8)) {
   	   return (set_ftn_text(data));
   }  				
				/* ============= set_overlaps ==========*/
   if (!strncmp(cmd,"set_over",8))
     {
       return (*hflag=set_overlapscall(p,data));
     }
				/* ========== set_radii ===== */
   if (!strncmp(cmd,"set_rad",7))
     {
       return (*hflag=set_radii_call(p,data));
     }
				/* ===== set_random radii/overlaps = */
   if (!strncmp(cmd,"set_rand",8))
     {
       return (*hflag=randize(p,data));
     }
     			/* ====== set_ratio ============== */
   if (!strncmp(cmd,"set_ratio",10)) {
   	int p0=-1,p1=-1;
   	sscanf(data,"%d %d",&p0,&p1);
   	if (p0<0 || p0>=NUM_PACKS || p1<0 || p1>=NUM_PACKS || p0==p1) return 0;
   	int ans=set_ratio(&packdata[p0],&packdata[p1]);
   	return ans;
   }
   
				/* ========== set_center ===== */
   if (!strncmp(cmd,"set_center",7)) {
     nextpoint=data;
     stripsp(data);
     if (!grab_next(&nextpoint,next) || !sscanf(next,"%lf",&(ctr.re))
	 || !grab_next(&nextpoint,next) || !sscanf(next,"%lf",&(ctr.im))) {
       strcpy(msgbuf,"Usage: x, y coords must be given first.");
       emsg();
       return 0;
     }
     if (p->hes< 0) {  /* hyp case */
       if (cAbs(ctr)>=1.0) {
	 strcpy(msgbuf,"Center must be in disc for "
		"hyperbolic packs.");
	 emsg();
	 return 0;
       }
     }
     else if (p->hes>0) { /* sph case */
       if (ctr.im<0.0 || ctr.im>M_PI) {
	 strcpy(msgbuf,"imaginary part must be in [0,pi] for "
		"spherical packs.");
	 emsg();
	 return 0;
       }
     }
     return (*hflag=set_centers(p,ctr,nextpoint));
   }
                               /* ======== skeleton ============= */
   if (!strncmp(cmd,"skel",4)) {
     return (skeleton_call(p,data));
   }
				/* ========== slit ================= */
   if (!strncmp(cmd,"slit",4))
     {
       return (*hflag=slit_complex_call(p,data));
     }
				/* ========== special command ======== */
   if (!strncmp(cmd,"special",7))
     {
       if (!sscanf(cmd+7,"%d",&spec_num)) 
	 {
	   strcpy(msgbuf,"Can't read special number.");
	   emsg();
	   return 0;
	 }
       stripsp(data);
       if (*data=='\0') return 0; 
       if (spec_num==1) 
	 {
	   sprintf(buff,"%d ",pnum);
	   strcat(buff,data);
	   return (*hflag=special1(buff));
	 }
       if (spec_num==2) return (*hflag=special2(data));
       if (spec_num==3) return (*hflag=special3(data));
       if (spec_num==4) return (*hflag=special4(data));
       if (spec_num==5) return (*hflag=special5(data));
       if (spec_num==6) return (*hflag=special6(data));
       if (spec_num==7) return (*hflag=special7(data));
       if (spec_num==8) return (*hflag=special8(data));
       if (spec_num==9) return (*hflag=special9(data));
       else return 0;
     }
				/* ========== spiral ======== */
   if (!strncmp(cmd,"spiral",6))
     {
       return (*hflag=spiral_call(p,data) );
     }
				/* ========== sq_fit ======== */
   if (!strncmp(cmd,"sq_fit",6)) {
     return square_fit(p);
   }
                                /* ===== sq_grid_layers ======= */
   if (!strncmp(cmd,"sq_grid_lay",11)) {
     return (sq_grid_lay_call(p,data));
   }
                                /* ===== sq_grid_overlaps ======= */
   if (!strncmp(cmd,"sq_grid_ov",10)) {
     return (sq_grid_overlaps(p));
   }
				/* ==========  swap nodes ========== */
   if (!strncmp(cmd,"swap",4))
     {
       return (*hflag=swap_nodes_call(p,data));
     }
   break;
 }
case 't':
 {
				/* ====== tent ======= */
   if (!strncmp(cmd,"tent",4)) {
     return (tent_call(p,data));
   }
   if (!strncmp(cmd,"tree_p",6)) {
   	 return (tree_pack_call(p,data));
   }
   break;
 }
case 'u':
 {
				/* ====== unflip ======= */
   if (!strncmp(cmd,"unflip",6))
     {
       return (*hflag=split_flips(p,data,3));
     }
                                /* ========== unweld ========= */
   if (!strncmp(cmd,"unweld",6)) {
      nextpoint=data;
      if (strlen(nextpoint)<3 
	  || !grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum)
	  || pnum<0 || pnum>NUM_PACKS
	  || !grab_next(&nextpoint,next) || !sscanf(next,"%d",&qnum)
	  || qnum<0 || qnum>=NUM_PACKS || qnum==pnum) {
	sprintf(msgbuf,"unweld: check help file.");
	emsg();
	return 0;
      }
      return (*hflag=
	      unweld_call(&packdata[pnum],&packdata[qnum],nextpoint));
   }
   break;
 }
case 'v':
 {
				/* ====== vert_map_off ======= */
   if (!strncmp(cmd,"vert_map_off",12))
     {
       edge_free(&(p->vertex_map));
       return (1);
     }
   break;
 }
case 'w':
 {
                                /* ========== weld ========= */
   if (!strncmp(cmd,"weld",4) && strncmp(cmd,"weld_",5)) { 
     /* (distinguish from 'weld_map') */
     return (weld_call(data));
   }
   break;
 }
case 'z':
 {
 	/* ============== zip ============ */
 	if (!strncmp(cmd,"zip",3)) {
 		return (zip_call(p,data));
 	}
 	break;
 }

	} /* end of unlocked switch */
    } /* end of if: unlocked only */



/* =======================================================
     LOCKED and UNLOCKED packings both go through here 
==========================================================*/

  switch (cmd[0]) /* ALL cmds go through here */
    {
case '[':
 {
 	return (runSpecialCmd(cmd));
 }
case 'a':
 {
				/* ========== active pack =========== */
   if (!strncmp(cmd,"act",3))
     {
       if (strlen(data) && sscanf(data,"%d",&n) 
	   && n>=0 && n<NUM_PACKS) {
		chg_active_pack(n); 
		/* ??? not sure I want this command. If targ_p is not 
			the address of current_p, then we might not want to
			change it just because we changed the active packing. */
		 // *targ_p=n; 
		 return 1;
	   }
       return 0;
     }
				/* ============ adjacency ============ */
   if (!strncmp(cmd,"adjac",5))
     {
       if ((*hflag=adjacency_call(p,data,1))) return 1;
       return 0;
     }
				/* ============ adjoin ============ */
   if (!strncmp(cmd,"adjoin",6))
     {
       if ((*hflag=adjoincall(p,data))) return 1;
       return 0;
     }
                                /* =========== aspect ============= */
   if (!strncmp(cmd,"aspect",6)) {
     return (aspect(p,data));
   }
   break;
 }
case 'A':
 {
				/* ============ Adjacency ============ */
   if (!strncmp(cmd,"Adjac",5)) {
     if ((*hflag=adjacency_call(p,data,0))) return 1;
     return 0;
   }
   break;
 }
case 'b':
 {
				/* ============ blend ============ */
   if (!strncmp(cmd,"blend",5))
     {
       if ((*hflag=blend_call(p,data))) return 1;
       return 0;
     }
   break;
 }
case 'c':
 {
				/* ============ center pt ============ */
   if (!strncmp(cmd,"center_point",12))
     {
       if (sscanf(data,"%lf %lf",&(ctr.re),&(ctr.im))!=2) return 0;
       return (center_point(p,ctr));
     }
				/* ============ center vert ============ */
   if (!strncmp(cmd,"center_vert",11))
     {
       if ((holdv=node_link_parse(p,data,&endptr,&hits))!=NULL)
	 {
	   v=holdv->v;
	   vert_free(&holdv);
	 }
       else v=p->active_node;
       return (center_point(p,p->packR_ptr[v].center));
     }
				/* ============ change directory ====== */
   if (!strcmp(cmd,"cd")) 
     {
       buf[0]='\0';
       stripsp(data);
       if (!strlen(data)) strcpy(next,"~\0");
       else sscanf(data,"%s",next);
       if (next[0]=='~') 
	 {
	   sprintf(buf,"%s/",home_dir);
	   next[0]=' ';
	   stripsp(next);
	 }
       strcat(buf,next);
       if (chdir(buf))
	 {
	   sprintf(msgbuf,"No directory %s.",buf);
	   emsg();
	 }
       getcwd(buf,BUFSIZE);
       working_dir=(char *)malloc((strlen(buf)+2)*sizeof(char));
       strcpy(working_dir,buf);
       sprintf(buf,"CirclePack (dir; %s)",working_dir);
       return 1;
     }
                                /* ==== cmdno ============ */
   if (!strncmp(cmd,"cmdno",5)) {
     if (sscanf(data,"%d",&cmdno)<=0) cmdno=-1;
     return 1;
   }
				/* === set colors: circles/tris  ==== */
   if (!strncmp(cmd,"color",5))
     {
       nextpoint=data;
       stripsp(data);
       if (!grab_next(&nextpoint,next) || next[0]!='-') return 0;
       if (next[1]=='c') /* color circles */
	 {
	   data[0]=' ';data[1]=' ';
	   stripsp(data);
	   if (!strncmp(data,"bg",2)) 
	     return (color_circles(p,4,"\0"));
	   if (!strncmp(data,"fg",2)) 
	     return (color_circles(p,5,"\0"));
	   if (!strncmp(data,"rad",3))
	     return (color_circles(p,1,"\0"));
	   if (!strncmp(data,"-q",2)) /* compare to q */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_circles(p,3,data));
	     }
	   if (!strncmp(data,"-p",2)) /* copy from pack p */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_circles(p,7,data));
	     }
	   if (!strncmp(data,"-k",2)) /* by curvature */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_circles(p,8,data));
	     }
	   else return (color_circles(p,6,data));
	 }
       if (next[1]=='f') /* color faces */
	 {
	   data[0]=' ';data[1]=' ';
	   stripsp(data);
	   if (!strncmp(data,"bg",2)) 
	     return (color_faces(p,4,"\0"));
	   if (!strncmp(data,"fg",2)) 
	     return (color_faces(p,5,"\0"));
	   if (!strncmp(data,"area",4))
	     return (color_faces(p,1,"\0"));
	   if (!strncmp(data,"-q",2)) /* compare to q */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_faces(p,3,data));
	     }
	   if (!strncmp(data,"-p",2)) /* copy from pack p */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_faces(p,7,data));
	     }
	   if (!strncmp(data,"qc",2)) /* qc dilatation */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_faces(p,8,data));
	     }
	   if (!strncmp(data,"-v",2)) /* copy from vert colors */
	     {
	       *data=' ';*(data+1)=' ';
	       return (color_faces(p,9,data));
	     }
	   else return (color_faces(p,6,data));
	 }
       else return 0;
     }
				/* ========== copy ======== */
   if (!strncmp(cmd,"copy",4)) 
     {
       if (!sscanf(data,"%d",&p1)) return 0;
       sprintf(buff,"%d %d",pnum,p1);
       copy_packcall(buff);
       return (*hflag=1);
     }
				/* ========== count ======== */
   if (!strncmp(cmd,"count",5))
     {
       return(count_call(p,data));
     }
                                /* ==== create_parent ====== */
   if (!strncmp(cmd,"create_p",8)) {
     return(create_parent_call(p,data));
   }
   break;
 }
case 'd':
 {
				/* ========== diagnostic routines ========*/
   if (!strncmp(cmd,"diag",4))
     {
       return (diag_call(p,data));
     }
 				/* ========== drawing operations ========*/
   if (!strncmp(cmd,"disp",4))
     {
       return (sa_display_call(p,data));
     }
  
				/* ========== delay ============== */
   if (!strncmp(cmd,"delay",5))
     {
       if (sscanf(data,"%d",&n) && n>0) timedelay((double)n);
       return 1;
     }
                                /* ======= doyle_point ========== */
   if (!strncmp(cmd,"doyle_p",7)) {
     return doyle_point_call(p,data);
   }
   break;
 }
case 'e':
 {
                                /* ============ embed =============== */
   if (!strncmp(cmd,"embed",5))
     {
       return (embed_call(p,data));
     }
                                /* =========== encircle ============= */
   if (!strncmp(cmd,"encirc",6)) {
     struct Edgelist *elist=NULL;
     if(elist=encircle_call(p,data)) {
       if (p->elist) edge_free(&(p->elist));
       p->elist=elist;
       return 1;
     }
     return 0;
   }
 				/* ============ exec  =============== */
   if (!strncmp(cmd,"exec",4))
     {
       int dum;
       return (*hflag=init_remote(data,&dum));
     }
 				/* ============ export_cmd ============ */
   if (!strncmp(cmd,"export_cmd",10)) {
     return (cmd_export(p,data));
   }
   				/* ============ eucl_write ============ */
   if (!strncmp(cmd,"eucl_write",10))
     {
       if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
       strcpy(buff,packing_dir_name);
       strcat(buff,filename);
       strcpy(filename,buff); /* now have full pathname */
       return (*hflag=writeeucldata(p,filename));
     }	
   break;
 }
case 'E':
 {
				/* === Eucl_write (directory specified) ====== */
   if (!strncmp(cmd,"Eucl_write",10))
     {
       if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
       return (*hflag=writeeucldata(p,filename));
     }	
   break;
 }
case 'f':
 {
                                /* ======= face_error ====== */
   if (!strncmp(cmd,"face_err",8))
     {
       return (face_error_call(p,data));
     }
                                /* ===== fexec ======= */
   if (!strncmp(cmd,"fexec",5))
     {
       return (*hflag=fexec_call(data));
     }
	                        /* ==== set 'next_face' indices ==== */
   if (!strncmp(cmd,"f_next",6))
     {
       return (set_face_next(p,data));
     }
				/* ==== set 'index_flag' of faces ==== */
   if (!strncmp(cmd,"f_index",7))
     {
       return (set_face_index(p,data));
     }
   if (!strncmp(cmd,"f_red",5))
     {
       return (set_face_red(p,data));
     }
   if (!strncmp(cmd,"f_rnext",7))
     {
       return (set_face_rnext(p,data));
     }
   if (!strncmp(cmd,"focus",5))
     {
       return (focus_call(p,data));
     }
   break;
 }
case 'g':
 {
				/* ============ set gamma ============== */
   if (!strncmp(cmd,"gamma",5) )
     {
       if (sscanf(data,"%d",&vert)==1) 
	 return (set_gamma(p,vert));
       else return 0;
     }
                                /* ========== gen_mark ================ */
   if (!strncmp(cmd,"gen_mark",8))
     {
       return (*hflag=gen_mark_call(p,data));
     }
                                /* ========== get_data ================ */
   if (!strncmp(cmd,"get_data",8))
     {
       return (*hflag=get_put_data_call(p,data,0));
     }
   break;
 }
case 'h':
 {
                                /* ============ h_dist ================ */
   if (!strncmp(cmd,"h_dist",6))
     {
       return (h_dist_call(data));
     }
                                /* ========== holonomy_trace ======== */
   if (!strncmp(cmd,"holonomy",8)) {
     return (holonomy_trace(p,data,&Mob));
   }
   break;
 }
case 'i':
 {
 	/* ================== inline_ =============== */
 	// inline_read, inline_path, etc. are handled by CirclePack 
 	if (!strncmp(cmd,"inline_",7)) {
 		return (runSpecialCmd(cmd));
 	}
 				/* ========== import_cmd ======== */
   if (!strncmp(cmd,"import_cmd",10)) {
     return (cmd_import(data));
   }
   break;
 }
case 'l':
 {
                                /* =========== locate =========== */
   if (!strncmp(cmd,"locate",6))
     {
       return locate_call(p,data);
     }
   break;
 }
case 'M':
 {
 	                            /* ======= Map (between packs) ===== */
   if (!strncmp(cmd,"Map",3)) {
   	return (Map_call(data));
   }	
 }
case 'm':
 {
                                /* ===== map circle/face indices ==== */
   if (!strncmp(cmd,"map",3))
     {
       return (map_list_call(p,data));
     }
				/* ====== mark circles/faces ======== */ 
   if (!strncmp(cmd,"mark",4))
     {
       marking(p,data);
       return 1;
     }
				/* ====== match ================= */ 
   if (!strncmp(cmd,"match",5))
     {
       return (match_call(p,data));
     }
				/* ====== msg_remote: to process === */ 
/* fixup: Need WIN32 code
 */
 
/*#ifndef WIN32
   if (!strncmp(cmd,"msg_r",5) || !strncmp(cmd,"m_r",3) 
       || !strncmp(cmd,"mr",2))
     {
       nextpoint=data;
       stripsp(data);
       if (sscanf(data," %s %[^\n]1000",next,msgbuf)<2) 
	 return 0; 
       if ((v=atoi(next))) 
	 {
	   if (v>=NUM_PROC)
	     {
	       w=-1;
	       for (i=0;i<NUM_PROC;i++)
		 if (my_remote[i].pid==v) {w=i;break;}
	       if (w<0) return 0;
	     }
	   else if (v>=0 && v<NUM_PROC && my_remote[v].pid) w=v;
	 }
       else 
	 {
	   w=-1;
	   for (i=0;i<NUM_PROC;i++)
	     if (strcmp(my_remote[i].moniker,next)==0) 
	       {w=i;break;}
	   if (w<0) return 0;
	 }
       msg_remote(w);
       return 1;		
     }
#endif */
				/* ========== display a msg ====== */ 
   if (!strncmp(cmd,"msg",3))
     {
       strncpy(msgbuf,data,300);
       msg();
       return 1;
     }
   break;
 }
case 'n':
 {
				/* ========== nan ======== */
   if (!strncmp(cmd,"nan",3))
     {
       return (nan_check(p,data));
     }
				/* ========== set active node ======== */
   if (!strncmp(cmd,"node_active",11) )
     {
       if (sscanf(data,"%d",&vert)==1 && vert>0 
	   && vert<=p->nodecount)
	 {
	   p->active_node=vert;
	   return 1;
	 }
       else return 0;
     }
   break;
 }
case 'o':
 {
                                /* =============== outlist =========== */
   if (!strncmp(cmd,"outlist",7)) {
   		Edgelist *red_outer_edges=red_to_outlist(p,1);
   		if (red_outer_edges==NULL) return 0;
   		edge_free(&p->elist);
   		p->elist=red_outer_edges;
		return 1;
   }                                
                                /* =============== output =========== */
   if (!strncmp(cmd,"output",6))
     {
       return output_call(p,data);
     }
   if (!strncmp(cmd,"open",4)) { // deprecated: 'open' no longer needed
   	  return 1;
   }
   break;
 }
case 'P':
 {
				/* =========== Paginate ============== */
   if (!strncmp(cmd,"Pag",3))
     {
       nextpoint=data;
       /* default to custom file name */
       if (!grab_next(&nextpoint,next)) strcpy(next,ps_file_name); 
       return (*hflag=renumber_ps_Pages(next));
     }
   break;
 }
case 'p':
 {
                                /* ========== path_to_elist ========== */
   if (!strncmp(cmd,"path_to_elist",13)) {
   	 Edgelist *elist;
     if ((elist=path_to_elist(p,pathlist))) {
       edge_free(&(p->elist));
       p->elist=elist;
       return 1;
     }
     return 0;
   }
                                /* ============== pdata =============== */
   if (!strncmp(cmd,"pdata",5))
     {
       return (print_inq_data(p,1,data));
     }
				/* =========== pair_mob ============== */
   if (!strncmp(cmd,"pair_m",6))
     {
       return (*hflag=apply_pair_mobius(p,data));
     }
				/* =========== path_construct  ======= */
   if (!strncmp(cmd,"path_c",6)) 
     {
       return (path_construct_call(p,data));
     }
				/* =========== path_Mobius  ======= */
   if (!strncmp(cmd,"path_M",6))
     {
       return (path_Mobius(Mob,pathlist,1));
     }
				/* =========== pin_face  ======= */
   if (!strncmp(cmd,"pin_f",5))
     {
       return (pin_face_call(p,data));
     }

				/* =========== print (postscript) ======= */
   if (!strncmp(cmd,"print",11)) 
     {
       if (!p->status)
	 {
	   strcpy(msgbuf,"Note, pack is empty.");
	   msg();
	   return 0;
	 }
       return (*hflag=post_canvas(p));
     }
				/* ============ custom printing ========*/
   if (!strncmp(cmd,"post",4))
     {
       n=print_call(&custom_fp,p,ps_file_name,
		    data,user_name,CPVersion,print_cmd);
       return (*hflag=n);
     } 
				/* ============ print working directory =*/
   if (!strncmp(cmd,"pwd",3))
     {
       getcwd(buf,BUFSIZE);
       sprintf(msgbuf,"\nCurrent directory: %s.",buf);
       msg();
       return 1;	 	
     }
                                /* ========== put_data ================ */
   if (!strncmp(cmd,"put_data",8))
     {
       return (*hflag=get_put_data_call(p,data,1));
     }
   break;		
 }
case 'q':
 {
                                /* =========== qc_dil =============== */
   if (!strncmp(cmd,"qc_dil",6))
     {
       nextpoint=data;
       if (!grab_next(&nextpoint,next) 
	   || !sscanf(nextpoint,"%d",&pnum) || pnum<0 || pnum>=NUM_PACKS)
	 return 0;
       return (dilatation_call(p,&packdata[pnum],&dum,&n));
     }
				/* ======== quit (no confirmation) ==== */
   if (!strncmp(cmd,"quit",4) || !strncmp(cmd,"exit",4) )
     { quit(0);return 0;}
   break;
 }
case 'r':
 {
				/* ============ read_colors ============ */
   if (!strncmp(cmd,"read_col",8))
     {
       return (read_colortable(p,data));
     }
				/* ============ read_path ============ */
   if (!strncmp(cmd,"read_path",9)) 
     {
       if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
       strcpy(buff,packing_dir_name);
       strcat(buff,filename);
       strcpy(filename,buff); /* now have full pathname */
       return (read_path_file(filename));
     }
                                /* ============ red_from_elist =============== */
   if (!strncmp(cmd,"red_from_el",11)) {
   		RedList *rlist=red_from_outlist(p,p->elist);
   		if (rlist==NULL) return 0;
   		free_redfaces(&p->redfaces);
   		p->redfaces=rlist;
		return 1;
   }
				/* ============ rotate ============ */
   if (!strncmp(cmd,"rotate",6))
     {
       if (sscanf(data,"%lf",&ang)!=1) return 0;
       return (rotate(p,ang*M_PI));
     }
				/* ============ rect_orient ========== */
   if (!strncmp(cmd,"rect_orient",11))
     {
       sprintf(buf,"-h %s",data);
       return (e_norm_scale(p,buf));
     }
				/* ============ rect_ratio ========== */
   if (!strncmp(cmd,"rect_ratio",10))
     {
       return (rect_ratio_call(p,data));
     } 	
     				/* ============ refresh ========== */
   if (!strncmp(cmd,"refresh",7))
     {
       return (refresh()); // refreshes java canvasses
     } break;
 }
case 'R':
 {
					/* ======= Read_path (directory given) ======= */
   if (!strncmp(cmd,"Read_path",9)) // assume java sends full pathname
     {
       if (sscanf(data,"%s",filename)!=1) return 0; /* blank */
       return (read_path_file(filename));
     } break;
 }
case 's':
 {
 	/* java_fixup: call for loading script */
				/* ==== open/load script file  ==== */
   if (!strncmp(cmd,"script",6)) {
     if (!sscanf(data,"%s",next)) return 0;
   	 return sa_read_script(next);
   }
			/* ============ set ============== */
   
   if (!strncmp(cmd,"set_",3))
     {
       cmd[0]=cmd[1]=cmd[2]=cmd[3]=' ';
       stripsp(cmd);
       if (!strncmp(cmd,"accur",5) &&
	   sscanf(data,"%lf",&dum)==1 && dum>0 && dum<.10)
	 {
	   if (dum<0.0000000000000001) dum=0.0000000000000001;
	   /* represents max of about 15 decimal accur */
	   toler=dum;
	   if (okerr>(.01)*toler) okerr = (.01)*toler;
	   return 1;
	 }
       if (!strncmp(cmd,"active",6) 
	   && (vertlist=node_link_parse(p,data,&endptr,&hits)))
	 {
	   p->active_node=vertlist->v;
	   vert_free(&vertlist);
	   return 1;
	 }
       /* set display option flags */
       if (!strncmp(cmd,"disp_flags",9)
	   && sscanf(data,"%d %d",&i,&j)==2)
	 {
	   p->screen->cflag_opt=i;
	   p->screen->fflag_opt=j;
	   return 1;
	 }
       if (!strncmp(cmd,"pix_box",7) /* new canvas pixel size */
	     && sscanf(data,"%d %d %d %d",&new_pixbox.lx,
		       &new_pixbox.ly,&new_pixbox.rx,&new_pixbox.ry)==5
	     && new_pixbox.rx>new_pixbox.lx && new_pixbox.ry>new_pixbox.ly)
	 {
	   q=p->screen;
	   q->pix_box=new_pixbox;
	   q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
				 /(q->box.ry-q->box.ly));
	   return 1;
	 }
       if (!strncmp(cmd,"cycle",5) &&
	   sscanf(data,"%d",&i)==1 && i>0 && i<100010)
	 {
	   totalpasses=i;
	   return 1;
	 }
       if (!strncmp(cmd,"dir",3)) 
	 {
	   i=(int)strcspn(data,"\n");
	   if (i==0) packing_dir_name[0]='\0';
	   else
	     {
	       strncpy(packing_dir_name,data,i);
	       packing_dir_name[i]='/';
	       packing_dir_name[i+1]='\0';
	     }
	   return 1;
	 }
       if (!strncmp(cmd+1,"list",4)) /* various lists */
	 {
	   if (*cmd=='e' || *cmd=='E') /* elist */
	     {
	       holde=node_pair_link(p,data,&endptr,&hits);
	       if (*cmd=='e')
		 {
		   edge_free(&(p->elist));
		   p->elist=holde;
		 }
	       else
		 {
		   edge_free(&Elist);
		   Elist=holde;
		 }
	       return 1;
	     }
	   else if (*cmd=='v' || *cmd=='V') /* vlist */
	     {
	       holdv=node_link_parse(p,data,&endptr,&hits);
	       if (*cmd=='v')
		 {
		   vert_free(&(p->vlist));
		   p->vlist=holdv;
		 }
	       else
		 {
		   vert_free(&Vlist);
		   Vlist=holdv;
		 }
	       return 1;
	     }
	   else if (*cmd=='f' || *cmd=='F') /* flist */
	     {
	       holdv=face_link_parse(p,data,&endptr,&hits);
	       if (*cmd=='f')
		 {
		   vert_free(&(p->flist));
		   p->flist=holdv;
		 }
	       else
		 {
		   vert_free(&Flist);
		   Flist=holdv;
		 }
	       return 1;
	     }
	 }
       if (!strncmp(cmd,"toler",5) &&
	   sscanf(data,"%lf",&dum)==1 && dum>0 && dum<.10)
	 {
	   toler=dum;
	   return 1;
	 }
       if (!strncmp(cmd,"err",3) &&
	   sscanf(data,"%lf",&dum)==1 && dum>0 && dum<.10)
	 {
	   if (dum<=(.01)*toler) {okerr=dum;return 1;}
	   return 0;
	 }
       if (!strncmp(cmd,"iter",4) &&
	   sscanf(data,"%d",&i)==1 && i>0 && i<15)
	 {
	   iterates=i;
	   return 1;
	 }
       if (!strncmp(cmd,"post",4) &&
	   sscanf(data,"%d",&i)==1 && i>0 && i<8)
	 {
	   sc_print_size=i;
	   return 1;
	 }
       if (!strncmp(cmd,"brush",10) 
	   && sscanf(data,"%d",&i)==1 && i>=0 && i<=12) 
	 {
	 	return (set_line_thick(p,i,1)); // set and send to Java 
	 }
	 if (!strncmp(cmd,"lpr",3)) { // set command for printing
	 	char *ct;
	 	if ((ct=strchr(cmd+4,' '))) {
	 		strcpy(print_cmd,ct);
	 		return 1;
	 	}
	 	return 0;
	 }
	 if (!strncmp(cmd,"ps_v",4)) { // set ps_viewer
	 	char *ct;
	 	if ((ct=strchr(cmd+4,' '))) {
	 		strcpy(ps_viewer,ct);
	 		return 1;
	 	}
	 	return 0;
	 }
	 if (!strncmp(cmd,"ps",2)) { // set ps_file_name, custom postscript
	 	char *ct;
	 	if ((ct=strchr(cmd+4,' '))) {
	 		strcpy(ps_file_name,ct);
	 		return 1;
	 	}
	 	return 0;
	 }
       if (!strncmp(cmd,"eucl_scale",10))
	 {
	   if (sscanf(data,"%lf",&dum)==1 && dum>=.5 && dum<=1.5)
	     {
	       eucl_factor=dum;
	       /* fixup:
	       if (!cmd_mode) xv_set(Eucl_factor,PANEL_VALUE,
				     (int)(dum*100.0),0);
	       */
	       return 1;
	     }
	 }
	 
       if (!strncmp(cmd,"Mobius",6)) {
		   newMob=(Mobius *)malloc(sizeof(Mobius));
		   if ((n=sscanf(data,"%lf %lf %lf %lf %lf %lf "
			"%lf %lf %d",
			 &(newMob->a.re),&(newMob->a.im),
			 &(newMob->b.re),&(newMob->b.im),
			 &(newMob->c.re),&(newMob->c.im),
			 &(newMob->d.re),&(newMob->d.im), 
			 &(newMob->flip)))>=8 ) {
				complex deter=det(*newMob);
		     	if (cAbs(deter)>okerr) {
					if (n==8) newMob->flip=0; /* default */	
					Mob=*newMob;
					free(newMob);
					return 1;
		     	}
	     		return 0;
			 }
		   else {
		       sprintf(msgbuf,"Mobius usage: "
			       "8 reals (and a 1 to reverse "
		    	   "orientation).");
		       emsg();
		       free(newMob);
	    	   return 0;
		   }
       }
	 
       if (!strncmp(cmd,"sphere_view",11) || !strncmp(cmd,"sv",2)) {
	    return (sphere_view(p,data));
       }
       if (!strncmp(cmd,"sph_pts",7) && sscanf(data,"%d",&i)==1 
	   && i>10 && i<500) {
	    num_plot=i;
	    return 1;
	   }
       if (!strncmp(cmd,"screen",6)) {
       	return (set_screen_call(p,data));
	   }
       if (!strncmp(cmd,"custom",6)) {
	   	return (set_postfile_call(1,data));
       }
       if (!strncmp(cmd,"print",5)) {
	    return (set_postfile_call(0,data));
       }
       if (!strncmp(cmd,"plot_flags",10)) {
	    return (set_plot_flags(p,data));
       }
	   return 1;
     } /* end of "set_" commands */
	 			/* ========== show red chain ======== */
   if (!strncmp(cmd,"show_",5)) {
     cmd[0]=cmd[1]=cmd[2]=cmd[3]=cmd[4]=' ';
     if (!strncmp(cmd+5,"red",3)) 
       return (*hflag=pfacered(p));
     else if (!strncmp(cmd+5,"vlist",5))
       return (showlists(p,0));
     else if (!strncmp(cmd+5,"elist",5))
       return (showlists(p,1));
     else if (!strncmp(cmd+5,"flist",5))
       return (showlists(p,2));
   }
   break;
 }
case 't':
 {
				/* ====== test_repack ======= */
   if (!strncmp(cmd,"test_re",7))
     {
       return (*hflag=call_test_repack(p,data));
     }
				/* ====== torus_t =========== */
   if (!strncmp(cmd,"torus_t",7))
     {
       return (call_torus_t(p,data));
     }
				/* ====== transition =========== */
   if (!strncmp(cmd,"trans",5)) 
     {
       return (transition_data_call(p,data));
     }
   break;
 }
case 'v':
 {
   				/* ====== vert_order ======= */
   if (!strncmp(cmd,"vert_ord",8))
     {
       return (vert_draw_order(p));
     }
   break;
 }
case 'w':
  {
				/* ====== weld_map =========== */
    if (!strncmp(cmd,"weld_map",8)) {
      nextpoint=data;
      if (strlen(nextpoint)<10 
	  || !grab_next(&nextpoint,next) || !sscanf(next,"%d",&pnum)
	  || pnum<0 || pnum>NUM_PACKS
	  || !grab_next(&nextpoint,next) || !sscanf(next,"%d",&qnum)
	  || qnum<0 || qnum>=NUM_PACKS || qnum==pnum) {
	sprintf(msgbuf,"weld_map: check help file.");
	emsg();
	return 0;
      }
      return (*hflag=
	      weld_map_call(&packdata[pnum],&packdata[qnum],nextpoint));
    }
				/* ====== write_carrier ====== */
   if (!strncmp(cmd,"write_car",9))
     {
       return (write_carrier_file(p,data));
     }
				/* ====== write_path ======= */
    if (!strncmp(cmd,"write_path",10))
      {
		nextpoint=data;
		if (!grab_next(&nextpoint,next) 
	    	|| sscanf(next,"%s",filename)!=1) 
		  return 0;
    	strcpy(buff,packing_dir_name);
 		strcat(buff,filename);
		strcpy(filename,buff); /* now have full pathname */
		return (*hflag=write_path_file(filename));
      }
				/* ====== write_select_radii ======= */
    if (!strncmp(cmd,"write_select_radii",18)) {
      return (write_multi_call(p,data,1,1));
    }
				/* ====== write_select_centers ======= */
    if (!strncmp(cmd,"write_select_centers",20)) {
      return (write_multi_call(p,data,2,1));
    }
				/* ====== write_vert_list ======= */
    if (!strncmp(cmd,"write_vert",10))
      {
	return (*hflag=write_multi_call(p,data,0,1));
      }
                                /* ==== write_light_raw
				   (ie., p_light format w/o 
				   key words ==== */
    else if (!strncmp(cmd,"write_light_raw",15))
      {
	return (*hflag=write_light_call(p,data,1,1));
      }
                                /* ==== write_light =====  */
    else if (!strncmp(cmd,"write_light",11))
      {
	return (*hflag=write_light_call(p,data,0,1));
      }
                                /* ==== write =====  */
    else if (!strncmp(cmd,"write",5))
      {
	return (*hflag=write_parse(p,data,1));
      }
    
    break;
  }
case 'W':
 {
				/* ====== Write_select_radii ======= */
    if (!strncmp(cmd,"Write_select_radii",18)) {
      return (write_multi_call(p,data,1,0));
    }
				/* ====== Write_select_centers ======= */
    if (!strncmp(cmd,"Write_select_centers",20)) {
      return (write_multi_call(p,data,2,0));
    }
				/* ====== Write_vert_list ======= */
    if (!strncmp(cmd,"Write_vert",10))
      {
	return (*hflag=write_multi_call(p,data,0,0));
      }
                                /* ==== write_light_raw
				   (ie., p_light format w/o 
				   key words ==== */
    else if (!strncmp(cmd,"Write_light_raw",15))
      {
	return (*hflag=write_light_call(p,data,1,0));
      }
                                /* ==== write =====  */
    else if (!strncmp(cmd,"Write",5))
      {
	return (*hflag=write_parse(p,data,0));
      }
    break;
  }
case '!':
  {
				/* ======= repeat last ====== */
    if (!strncmp(cmd,"!!",2) && !cmd_search_depth) 
      /* note: not permitted when executing named script cmd */ 
      {
	sort_cmd(cmd_history,targ_p);
	return 1;
      }
    break;
  }
case '?':
 {
				/* ============ inquiries =========== */
   if (!strncmp(cmd,"?vert",5)) 
     return (print_inq_data(p,2,data));
   if (!strncmp(cmd,"?flower",7))
     return (print_inq_data(p,7,data));
   if (!strncmp(cmd,"?rad",4))
     return (print_inq_data(p,8,data));
   if (!strncmp(cmd,"?ang_sum",8))
     return (print_inq_data(p,9,data));
   if (!strncmp(cmd,"?aim",4))
     return (print_inq_data(p,10,data));
   if (!strncmp(cmd,"?face",5))
     return (print_inq_data(p,11,data));
   if (!strncmp(cmd,"?over",5))
     return (print_inq_data(p,12,data));
   if (!strncmp(cmd,"?cent",5))
     return (print_inq_data(p,13,data));
   if (!strncmp(cmd,"?alt_rad",8))
     return (print_inq_data(p,14,data));
   if (!strncmp(cmd,"?kap",4))
     return (print_inq_data(p,15,data));
   if (!strncmp(cmd,"?antip",6))
     return (print_inq_data(p,16,data));
   if (!strncmp(cmd,"?screen",7))
     return (print_inq_data(p,20,data));
   if (!strncmp(cmd,"?param",6))
     return (print_inq_data(p,21,data));
   if (!strncmp(cmd,"?pk_stat",7))
     return (print_inq_data(p,25,data));
   if (!strncmp(cmd,"?bdry_dist",10))
     return (print_inq_data(p,26,data));
   if (!strncmp(cmd,"?edge_p",7)) /* edge_pairing */
     return (print_inq_data(p,27,data));
   if (!strncmp(cmd,"?ratio_ftn",10))
     return (print_inq_data(p,28,data));
   if (!strncmp(cmd,"?conduct",8))
     return (print_inq_data(p,29,data));
   if (!strncmp(cmd,"?bdry_length",12))
     return (print_inq_data(p,31,data));
   if (!strncmp(cmd,"?energy",7))
     return (print_energy(p));
   if (!strncmp(cmd,"?script",8))
     return (print_inq_data(p,33,data));
   if (!strncmp(cmd,"?e_area",7))
     return (print_inq_data(p,34,data));
   if (!strncmp(cmd,"?map_rev",8))
     return (print_inq_data(p,36,data));
   if (!strncmp(cmd,"?map",4))
     return (print_inq_data(p,35,data));
   if (!strncmp(cmd,"?edge_x",7))
     return (print_inq_data(p,37,data));
   if (!strncmp(cmd,"?Mob",4)) /* current Mobius */
     return (print_inq_data(p,38,data));

   break;
 }
    } /* end of switch */

  return 0;
} /* parse_cmd */
  
