/* sa_screen.cpp = screen procedures and graphic ftns for CPack */

#include "Interface.h"
#include "sa_head.h"

/*
Idea is to gather display commands into package(s), each indicating what
packing canvas is target and perhaps options (like "display one-at-a-time").
*/

/* Drawing flag scheme: flags indicate drawing instructions
   for objects -- circles/lines/faces. Bits set as follows:
      1  ->  draw object?
      2  ->  fill? (4 and 16 imply 2, also)
      4  ->  off = foreground, on = background 
             (applies to interior, not border, overriden by bit 16)
      8  ->  border color? (default=foreground, else recorded color)
      16 ->  interior color? (default set by bit 4, on --> recorded color)
      32 ->  display label?

Eg.  flag=3: filled object, in foreground
     flag=9: open object, border in (recorded) color
             (for 'edge', this gives colored edge)
     flag=19: filled in color, border in foreground
     flag=27: filled, border and interior in color
     flag=15: filled with background, border in color
     flag=32: label only (face or circle)

Normally, flag for each type of object; often passed on to
subroutine, so may need color code with it:
Eg. (cflag, ccol, ecol) for circle flag, int color, border color.
*/

int sa_display_call(struct p_data *p,char *datastr)
{
	// first set the right world coordinates
	int count=0,vert,tracking_code=0;
	int eflag,fflag,dflag,nflag,cflag,ccol;
	int i,j,col,ecol,fcol,hes,length,hits,pnum=pack_num(p);
	int dspeed,v,w,indx,front;
	double dotsize,r;
	char *nextpoint,*endptr,next[64];
	struct Vertlist *vertlist=NULL,*trace,*facelist=NULL;
	struct Edgelist *edgelist=NULL,*track,*eelist=NULL,*eetrack;
	struct RedList *rtrace;
	complex ctr;
	CPXPoint *Xptr=NULL;
	struct s_data *q;

	if (!(p->status)) return 1;
	q=p->screen;
	dspeed=draw_speed; 
	nextpoint=datastr;
	if (!grab_next(&nextpoint,next) || next[0]!='-') {
		// no flags so use stored circle and face options
	   sa_disp_screen(p,q->cflag_opt,q->fflag_opt);
	   return 1;
	}
	hes=p->hes;
	do {
for (i=strlen(next);i<10;i++) next[i]='\0'; /* clear old info */
if (next[0]!='-') {
    return count;
}
if (next[1]=='w') { /* clear canvas (-w) or clear/refresh (-wr) */
    sa_clear_canvas(p,0);
    if (next[2]=='r') sa_disp_screen(p,q->cflag_opt,q->fflag_opt);
    count++;
}
if (next[1]=='u') { /* unit disc */
	if (p->hes==0) drawUnitCircle(p); // eucl
	else if (p->hes>0) {  // sph
		complex ctr;
		ctr.re=ctr.im=0.0;
		ctr=ss_view(p->screen,ctr,1,&dum_int);
		drawEquator(p,ctr.re,ctr.im);
	}
	count++;
}
else if (next[1]=='c') { /* draw circles */
    ecol=col=FG_COLOR;
    cflag=1;
    if (strstr(next+2,"c")) cflag |= 8; /* color object borders */
    if (strstr(next,"n")) cflag |= 32;  /* labels */
    if (strstr(next,"f")) cflag |= 2;   /* want filled circles */
    if (strstr(next,"bg")) cflag |= 6;  /* fill with background */
    else if (strstr(next,"fg"));
    else if (cflag & 2) cflag |= 16;    /* default, use recorded color */
    if (cflag & 4) col=BG_COLOR;

    if ((vertlist=node_link_parse(p,nextpoint,
       &endptr,&hits))!=NULL) {
		int loc_count=0;	

		nextpoint=endptr;
		trace=vertlist;
		do {
		  do {
		    vert=trace->v;
		    if (cflag & 16) col=p->packK_ptr[vert].color;
		    if (cflag & 8) {
				ecol=p->packK_ptr[vert].color;
				if (ecol>=bc_min && ecol<=bc_max) ecol+=bc_offset;
		    }
		    sa_draw_circle(p,vert,cflag,ecol,col,0);
		    loc_count++;count++;
		    trace=trace->next;
		  } while (loc_count<DISP_GRP && trace!=NULL); 
		  loc_count=0;
		} while (trace!=NULL); /* outer do loop */
		vert_free(&vertlist);
       }
    else { /* display whole pack using cflag */
		sa_disp_screen(p,cflag,0);
		count++;
    }
}
/* post small dot instead of circle. */
/* fixup:
else if (next[1]=='d' && p->hes<=0) {
   dotsize=(0.005)*(p->screen->box.ry-p->screen->box.ly); 
   if (next[2]=='c') dflag=1; 
   else dflag=0;
   if ((vertlist=node_link_parse(p,nextpoint,&endptr,&hits))!=NULL)
     {
       nextpoint=endptr;
       trace=vertlist;
       do 
	 {
	   vert=trace->v;
	   count += draw_any_dot(p,vert,dotsize,dflag,1,dspeed);
	   trace=trace->next;
	 } 
       while (trace!=NULL);
       vert_free(&vertlist);
     }
   else for (i=1;i<=p->nodecount;i++) 
     count += draw_any_dot(p,i,dotsize,dflag,1,dspeed);
 }
*/
/* show only small dot instead of circle, sphere only.*/
/* fixup:
else if (next[1]=='d' && p->hes>0) 
  {
    for (i=1;i<=p->nodecount;i++)
      if (p->packK_ptr[i].plot_flag>0)
	{
	  ccol=p->packK_ptr[i].color;
	  ctr=ss_view(p->screen,p->packR_ptr[i].center,1,&dum_int);
	  if (cos(ctr.re)>0) 
	    {
	      ctr.re=sin(ctr.im)*sin(ctr.re);
	      ctr.im=cos(ctr.im);
	      circle(p->screen,ctr,.005,7,ccol,ccol,dspeed);
	    }
	}
  }
*/
else if (next[1]=='C' || next[1]=='B') {
/* draw circles or circles/faces, recomputed in drawing order */
	if (strstr(next,"c")) cflag=fflag=8; /* color object borders */
    else cflag=fflag=0;
    if (strstr(next,"f")) {cflag += 2;fflag += 2;} /* filled */
    if (strstr(next,"bg")) {cflag += 6;fflag += 6;} /* fill with background */
    else if (strstr(next,"fg"));
    else if (cflag & 2) {cflag += 16;fflag += 16;}
    cflag += 1; /* always want circles */
    if (next[1]=='B') fflag += 1; /* also faces */
    if (strstr(next,"n")) {cflag += 32;fflag += 32;} /* labels */
    stripsp(nextpoint);
    if (*nextpoint=='r') { /* want just red circles/faces */
	/* if B, then red faces first */
		if ((fflag & 1)
		    && (facelist=face_link_parse(p,nextpoint,&endptr,&hits))!=NULL) {
		    nextpoint=endptr;
		    count += sa_layout_facelist(p,fflag,&facelist,1,0,1);
		    }
	else nextpoint++;
printf(" got here with -C r\n");
	/* now the circles: here use data in redfaces by segment */
	vert=0;
    int safty=0; // emergency exit
	for (i=1;p->edge_pair[i].edge;i++) {
printf(" now in loop, i=%d",i);
	    /* first (corner) circle black */
	    rtrace=p->edge_pair[i].edge;
	    fcol=ecol=FG_COLOR; /* corner circles black */
	    if (rtrace->next->face==rtrace->prev->face
		    && p->edge_pair[i].edge_indx!=rtrace->v_flag) {
	      /* blue face, want end of second edge, so get circle 
		 center from prev face */
			ctr=rtrace->prev->center;
			if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
			r=rtrace->prev->rad;
			if (cflag & 32) vert=p->faces[rtrace->prev->face].
			  vert[rtrace->prev->v_flag];
			sa_draw_circle_data(pnum,r,ctr,cflag,ecol,fcol,vert,1);
		    }
	    else {
			ctr=rtrace->center;
			if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
			r=rtrace->rad;
			if (cflag & 32) 
			  vert=p->faces[rtrace->face].vert[rtrace->v_flag];
			sa_draw_circle_data(pnum,r,ctr,cflag,ecol,fcol,vert,1);
	    }
	    /* now keep going until end */
	    fcol=ecol=p->edge_pair[i].color;
	    if (ecol>=bc_min && ecol<=bc_max) ecol+=bc_offset;
	    rtrace=next_red_edge(p,1,rtrace,indx,&indx);
	    indx=(indx+1) % 3;
printf("Entering while, p->redfaces->face is %d\n",p->redfaces->face);	    
	    while (!(rtrace->corner_flag[indx] & 2) && safty<3*p->facecount) {
	    	safty++;
		 if (rtrace->next->face==rtrace->prev->face
		     && indx!=rtrace->v_flag) {
		   /* blue face, second edge; get data from prev edge */
		     ctr=rtrace->prev->center;
		     if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
		     r=rtrace->rad;
		     if (cflag & 32) vert=p->faces[rtrace->prev->face].
		       vert[rtrace->prev->v_flag];
		     }
		 else {
		     ctr=rtrace->center;
		     if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
		     r=rtrace->rad;
		     if (cflag & 32) 
		       vert=p->faces[rtrace->face].vert[rtrace->v_flag];
		 }
		 sa_draw_circle_data(pnum,r,ctr,cflag,ecol,fcol,vert,1);
		 rtrace=next_red_edge(p,1,rtrace,indx,&indx);
		 indx=(indx+1) % 3;
	    } /* end of while */
	    /* last (corner) circle black */
	    fcol=ecol=FG_COLOR; /* corner circles black */
	    if (rtrace->next->face==rtrace->prev->face
		&& indx!=rtrace->v_flag) {
	      /* blue face, second edge; get data from prev face */
			ctr=rtrace->prev->center;
			if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
			r=rtrace->prev->rad;
			if (cflag & 32) vert=p->faces[rtrace->prev->face].
				vert[rtrace->prev->v_flag];
		}
	    else {
			ctr=rtrace->center;
			if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
			r=rtrace->rad;
			if (cflag & 32) 
			  vert=p->faces[rtrace->face].vert[rtrace->v_flag];
	    }
	    sa_draw_circle_data(pnum,r,ctr,cflag,ecol,fcol,vert,1);
	} /* end of i for-loop through segments */
    } /* end of drawing just red */
    else count += sa_draw_in_order(p,cflag,fflag,0);
} 
else if (next[1]=='e') {/* draw edges */
    if (next[2]=='e' && (vertlist=node_link_parse(p,nextpoint,
       &endptr,&hits))!=NULL) { /* hex extended edges */
	 /* set 'progress' mode? */
		nextpoint=endptr;
		trace=vertlist;
		while (trace && (v=trace->v)
	       && trace->next && (w=trace->next->v)) { /* get pairs */
		    trace=trace->next;
		    if ((eelist=get_extended_edge(p,v,w,256))!=NULL) {
			eetrack=eelist;
			do {
			    count+=sa_draw_edge(p,eetrack->v,
					eetrack->w,FG_COLOR);
			    eetrack=eetrack->next;
			} while (eetrack!=NULL);
			edge_free(&eelist);
		    }
	       }
		vert_free(&vertlist);
       }
    else if ((edgelist=node_pair_link(p,nextpoint,&endptr,&hits))
	     !=NULL) {
		nextpoint=endptr;
		track=edgelist;
		do {
		  count+=sa_draw_edge(p,track->v,track->w,FG_COLOR);
		  track=track->next;
		} while (track!=NULL);
		edge_free(&edgelist);
	     }
    else { /* draw all edges (do by using faces, set fflag) */
		sa_disp_screen(p,0,q->fflag_opt);
		count++;
    }
} /* finished with edges */
else if (next[1]=='f' || next[1]=='F') { /* draw faces */
	ecol=col=FG_COLOR;
    fflag=1;
    if (strstr(next+2,"c")) fflag |= 8; /* color object borders */
    if (strstr(next,"n")) fflag |= 32;  /* labels */
    if (strstr(next+2,"f")) fflag |= 2;   /* another f? want filled circles */
    if (strstr(next,"bg")) fflag |= 6;  /* fill with background */
    else if (strstr(next,"fg"));
    else if (fflag & 2) fflag |= 16;    /* default, use recorded color */
    if (fflag & 4) col=BG_COLOR;

    if ((facelist=face_link_parse(p,nextpoint,&endptr,&hits))!=NULL) {
       	nextpoint=endptr;
        if (next[1]=='F')  /* locate faces as they're drawn */ 
	 	count += sa_layout_facelist(p,fflag,&facelist,1,0,1);
       	else {
		   count += sa_layout_facelist(p,fflag,&facelist,0,0,1);
       	}
    }
	else if (next[1]=='F') { /* use draworder, but recompute centers */
     count += sa_draw_in_order(p,0,fflag,0);
	}
    /* fixup: why do this? it converts batch display into individual
       display */
   else if ((fflag & 2) && !(fflag & 16)) { /* filled, but not color */
       ecol=col=FG_COLOR;
       for (i=1;i<=p->facecount;i++) {
		   if (fflag & 8) {
		       ecol=p->faces[i].color;
	    	   if (ecol>=bc_min && ecol<=bc_max) ecol+=bc_offset;
		   }
		   if (fflag & 4) col=BG_COLOR;
		   count += sa_draw_face(p,i,fflag,ecol,col,1);
       }
   }
   else {
       sa_disp_screen(p,0,fflag);
       count++;
   }
} /* finished with faces */
else if (next[1]=='R' 
  && p->edge_pair[2].edge) { /* specified paired edges in non-simply 
			       connected cases; draw thick & in color */
	if (strstr(next,"p")) dflag=1; /* draw mated pairs */
    else dflag=0;
    nflag=8; /* flag of options for draw_bdry_edge; 8=color */
    if (strstr(next,"n")) nflag |= 4; /* label segments */
    if (strstr(next,"c")) nflag |= 16; /* draw numbered circles also */
    length=0;
    while (p->edge_pair[length+1].edge) length++;
    /* fixup: should we know number of segments from euler/etc? */
    line_thick_mode(p,4); /* set thicker line */
    if ((vertlist=node_link_parse(p,nextpoint,
       &endptr,&hits))!=NULL) {
		nextpoint=endptr;
		trace=vertlist;
		while (trace && (j=trace->v)>0 && j <= length) {
		    count += sa_draw_bdry_seg(p,j,nflag,p->edge_pair[j].color);
	    	if (dflag && (i=index_of_paired_edge(p,j)))
		      sa_draw_bdry_seg(p,i,nflag,p->edge_pair[i].color);
		    trace=trace->next;
		}
       }
    else for (i=1;p->edge_pair[i].edge;i++)
      sa_draw_bdry_seg(p,i,nflag,p->edge_pair[i].color);
    vert_free(&vertlist);
    line_thick_mode(p,1); /* reset line thickness */
    /* fixup: should restore ORIGINAL thickness */
  }
else if (next[1]=='s') { /* polygon through vertlist */
    col=FG_COLOR;
    eflag=1;
    if (strstr(next,"bg")) {col=BG_COLOR;eflag += 6;}
    else if (strstr(next,"fg")) eflag +=2;
    else if (strstr(next,"f")) eflag += 18;

    if ((vertlist=node_link_parse(p,nextpoint,
       &endptr,&hits))!=NULL) {
		trace=vertlist;
		while(trace && trace->next) {trace=trace->next;} /* find end? */
		if (trace && trace->v!=vertlist->v) { /* repeat first vert at end */
		    trace->next=(struct Vertlist *)
		      calloc((size_t)1,sizeof(struct Vertlist));
	    	trace->next->v=vertlist->v;
		}
		nextpoint=endptr;
		trace=vertlist;
		sa_polygon(p,vertlist,eflag,FG_COLOR,col);
		vert_free(&vertlist);
       }
}
else if (next[1]=='n') { /* label */
	if (next[2]=='f') { /* faces */
		if ((facelist=face_link_parse(p,nextpoint,&endptr,&hits))!=NULL
		    || (facelist=face_link_parse(p,"a",&endptr,&hits))!=NULL) {
		    labellist(p,facelist,1,1);
		    vert_free(&facelist);
	    	count++;
		    }
	}
    else if (next[2]=='c') { /* circles */
		if ((vertlist=node_link_parse(p,nextpoint,&endptr,&hits))!=NULL
		  || (vertlist=node_link_parse(p,"a",&endptr,&hits))!=NULL) {
		    labellist(p,vertlist,1,0);
		    vert_free(&vertlist);
	    	count++;
		  }
    }
    else if ( 			/* 'l' or 'z' */
      (
        (
	  next[2]=='z'  		/* given point */
		&& grab_next(&nextpoint,next)
		   && sscanf(next,"%lf",&(ctr.re))
		&& grab_next(&nextpoint,next)
		   && sscanf(next,"%lf",&(ctr.im))
		&& !(i==0)
	)
	||
	(
	  next[2]=='l'  		/* or center of circle */
	  && (i=grab_one_vert(p,&nextpoint))
	)
      )
      && grab_next(&nextpoint,next) 
      && strlen(strncpy(buf,next,32))!=0 
      ) {
		/* put string either at cent of i or at given ctr. */
		strcat(buf,"\0");
		if (i>0) {
		    ctr=p->packR_ptr[i].center;
		    if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
		}
		drawIndex(pnum,ctr.re,ctr.im,i,1); 
		count++;
      }
} /* finished with labeling */

/* display the global path */
else if (next[1]=='g') {
	drawPath(pack_num(p),pathlength);
	count++;
}
else if (next[1]=='r' && next[2]=='c' && (rtrace=p->redfaces)) {
  /* test: draw circles from redlist data */
    ecol=col=FG_COLOR;
    cflag=1;
    if (strstr(next+3,"c")) cflag += 8; /* color object borders */
    if (strstr(next,"n")) cflag += 32; /* labels */
    if (strstr(next,"f")) cflag += 2; /* want filled circles */
    if (strstr(next,"bg")) cflag += 6;  /* fill with background */
    else if (strstr(next,"fg"));
    else if (cflag & 2) cflag += 16; /* default, use recorded color */
    if (cflag & 4) col=BG_COLOR;

    i=0;
    while (rtrace!=p->redfaces || !(i++)) {
		vert=p->faces[rtrace->face].vert[rtrace->v_flag];
		if (cflag & 16) col=p->packK_ptr[vert].color;
		if (cflag & 8) {
		    ecol=p->packK_ptr[vert].color;
		    if (ecol>=bc_min && ecol<=bc_max) ecol+=bc_offset;
		}
		ctr=rtrace->center;
		if (p->hes>0) ctr=ss_view(p->screen,ctr,1,&front);
		count += sa_draw_circle_data(pnum,rtrace->rad,ctr,
					     cflag,ecol,col,vert,1);
		rtrace=rtrace->next;
    }
}
    
else if (next[1]=='r') { /* reset */
    sa_reset_screen(p);
    count++;
}

else if (next[1]=='x') { /* display coord axes */
	setAxisMode(pnum,1); // set true
}

} /* end of do */
	while (nextpoint!=NULL && grab_next(&nextpoint,next) );
	return count;		
} /* sa_display_call */

int sa_disp_screen(struct p_data *p,int cflag,int fflag)
/* display pack p, standard flag options. */
{
  int i,ccol=FG_COLOR,col=FG_COLOR;
  struct K_data *pK_ptr=p->packK_ptr;

  if (!p->status)
    {sprintf(msgbuf,"pack %d is empty",pack_num(p));msg();return 1;}
  if (cflag) {
      for (i=1;i<=p->nodecount;i++) {
	 	if (cflag & 16) col=p->packK_ptr[i].color;
		if (cflag & 8) {
			ccol=p->packK_ptr[i].color;
			if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
		}
		sa_draw_circle(p,i,cflag,ccol,col,0);
      }
  }
  if (fflag) {
	for (i=1;i<=p->facecount;i++) {
		if (fflag & 16) col=p->faces[i].color;
		if (fflag & 8) {
		  ccol=p->faces[i].color;
		  if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
		}
		sa_draw_face(p,i,fflag,ccol,col,0);
	}
  }
  return 1;
} /* sa_disp_screen */


int sa_draw_in_order(struct p_data *p,int cflag,int fflag,int pl)
/* draw all faces and/or all circles in drawingorder, but draw and 
recompute as you go along. E.g., will give 'ghosts'. Standard flag 
arrangement. If pl is set, then place first face; else, leave in 
current location. */
{
	int nf,ind,a,b,c,ck=0,col,ccol,v,n0,indx;

   /* handle first face */
	nf=p->first_face;
	ccol=col=FG_COLOR;
	if (pl) place_face(p,nf,p->faces[nf].index_flag);
	if (cflag & 1) { /* circles? */
	   ind=p->faces[nf].index_flag;
	   a=p->faces[nf].vert[ind];
	   b=p->faces[nf].vert[(ind+1) % 3];
	   c=p->faces[nf].vert[(ind+2) % 3];
	   if (cflag & 16) col=p->packK_ptr[a].color;
	   if (cflag & 8) {
	       ccol=p->packK_ptr[a].color;
	       if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	   }
	   sa_draw_circle(p,a,cflag,ccol,col,1);
	   if (cflag & 16) col=p->packK_ptr[b].color;
	   if (cflag & 8) {
	       ccol=p->packK_ptr[b].color;
	       if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	   }
	   sa_draw_circle(p,b,cflag,ccol,col,1);
	   if (cflag & 16) col=p->packK_ptr[c].color;
	   if (cflag & 8) {
	       ccol=p->packK_ptr[c].color;
	       if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	   }
	   sa_draw_circle(p,c,cflag,ccol,col,1);
	}
	ccol=col=FG_COLOR;
	if (fflag & 1) { /* faces? */
	    if (fflag & 16) col=p->faces[nf].color;
	    if (fflag & 8) {
			ccol=p->faces[nf].color;
			if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	    }
	    sa_draw_face(p,nf,fflag,ccol,col,1);
	}
   /* now the rest */
	while ( nf>0 && nf <= p->facecount 
		&& (nf=p->faces[nf].next_face)!=p->first_face
		&& ck<2*p->facecount ) {
	   v=p->faces[nf].vert[(indx=(p->faces[nf].index_flag+2)%3)];
	   n0=nghb(p,v,p->faces[nf].vert[(indx+1)%3]);
	   fancy_comp_center(p,v,n0,n0,1,0,1,1,toler);
	   col=ccol=FG_COLOR;
	   if (cflag & 1) { /* circle? */
	       c=p->faces[nf].vert[(p->faces[nf].index_flag + 2) % 3];
	       if (cflag & 16) col=p->packK_ptr[c].color;
	       if (cflag & 8) {
			   ccol=p->packK_ptr[c].color;
			   if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	       }
	       sa_draw_circle(p,c,cflag,ccol,col,1); 
				/* new circle */
	   }
	   ccol=col=FG_COLOR;
	   if (fflag & 1) { /* faces? */
	       if (fflag & 16) col=p->faces[nf].color;
	       if (fflag & 8) {
			   ccol=p->faces[nf].color;
			   if (ccol>=bc_min && ccol<=bc_max) ccol+=bc_offset;
	       }
	       sa_draw_face(p,nf,fflag,ccol,col,1);
	   }
	   ck++;
		}
	if (nf!=p->first_face) /* some error */
		return 0;
	return 1;
} /* sa_draw_in_order */

int refresh_canvas(struct s_data *q)
{
  /* fixup: stand-in to get by compiler */
  return 1;
} /* refresh_canvas */

