#include "sa_types.h"
#include "sa_proto.h"

int scale_aims(struct p_data *p,double x,struct Vertlist *vlist)
     /* Reset aims of vlist based on their current angle sum, aim, 
and specified factor x. If aim is positive,
    aim(v)=angle sum(v) + x*[aim(v)-angle sum(v)].
Return count of adjustments.
     */
{
  int count=0,flag,v;
  struct Vertlist *vtrace;
  struct R_data *pR_ptr=p->packR_ptr;

  if (!p->status || !vlist || x < 0.0) return 0;
  vtrace=vlist;
  while (vtrace) {
    v=vtrace->v;
    if (pR_ptr[v].aim>0) {
      if (p->hes<0)
	h_anglesum_overlap(p,v,pR_ptr[v].rad,&pR_ptr[v].curv,&flag);
      else if (p->hes>0)
	s_anglesum_overlap(p,v,pR_ptr[v].rad,&pR_ptr[v].curv,&flag);
      else 
	e_anglesum_overlap(p,v,pR_ptr[v].rad,&pR_ptr[v].curv,&flag);
      pR_ptr[v].aim = pR_ptr[v].curv-x*(pR_ptr[v].curv-pR_ptr[v].aim);
      count++;
    }
    vtrace=vtrace->next;
  }
  return 1;
} /* scale_aims */    

int scale_rad(struct p_data *p,struct p_data *q,double x,
	      struct Vertlist *vlist)
  /* Reset radii of p to interpolate between current values and
values in q by factor: 
     rad_p(v)=rad_p(v) + x*[rad_p(v)-rad_q(v)].
Infinite radii in hyperbolic are treated as value 5.
Return count of adjustments.*/
{
  int v,count=0;
  double rp,rq;
  struct Vertlist *vtrace;
  struct R_data *pR_ptr=p->packR_ptr,*qR_ptr=q->packR_ptr;

  if (!p->status || !q->status || !vlist || x < 0.0) return 0;
  vtrace=vlist;
  while (vtrace) {
    v=vtrace->v;
    rp=pR_ptr[v].rad;
    rq=qR_ptr[v].rad;
    if (p->hes<0) {
      if (rp<0) rp=1-exp(-2.0*5.0);
      if (rq<0) rq=1-exp(-2.0*5.0);
    }
    pR_ptr[v].rad = rp+x*(rq-rp);
    count++;
    vtrace=vtrace->next;
  }
  return 1;
} /* scale_rad */    

int scale_aims_call(struct p_data *p,char *datastr)
{
  int hits;
  double factor;
  char *nextpoint,*endptr;
  struct Vertlist *vlist=NULL;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || !sscanf(next,"%lf",&factor)
      || factor < 0.0
      || !(vlist=Node_link_parse(p,nextpoint,&endptr,&hits,
				 &Vlist,&Elist,&Flist,&region,pathlist,
				 pathlength))) {
    sprintf(msgbuf,"usage: scale_aims x {v..}");
    emsg();
    return 0;
  }
  hits=scale_aims(p,factor,vlist);
  vert_free(&vlist);
  return hits;
} /* scale_aims_call */

int scale_rad_call(struct p_data *p,struct p_data *q,char *datastr)
{
  int hits;
  double factor;
  char *nextpoint,*endptr;
  struct Vertlist *vlist=NULL;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || !sscanf(next,"%lf",&factor)
      || factor <= 0.0
      || !(vlist=Node_link_parse(p,nextpoint,&endptr,&hits,
				 &Vlist,&Elist,&Flist,&region,pathlist,
				 pathlength))) {
    sprintf(msgbuf,"usage: scale_rad q x {v ..}");
    emsg();
    return 0;
  }
  hits=scale_rad(p,q,factor,vlist);
  vert_free(&vlist);
  return hits;
} /* scale_rad_call */


