#include "sa_types.h"
#include "sa_proto.h"

/* Routines for setting overlaps (this includes inversive distances). */

int set_xyz_overlaps2(struct p_data *p,struct xyzPoint *xyz_list,
		     int count,char *datastr,int flag)
     /* Use xyz data to set overlaps of pack p. Allocate space if
necessary. In future, datastr and flag will indicate options; for
now, find max/min edge lengths in xyz data, set overlaps as though
circle radii were all min/2. Return count of overlaps set, 0 on error. */
{
  int i,j,k;
  double miN=1000000000.0,maX=0.0,ovlp,dist,rad,ovlpmax=okerr,ovlpmin=1000;
  struct K_data *pK_ptr=p->packK_ptr;

  if (!p->status || count<=0 || !xyz_list) return 0;

  /* compute maX/miN distances */
  for (i=1;i<=p->nodecount;i++)
    for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++)
      if ((k=pK_ptr[i].flower[j])>i
	  && i<= count && k<=count)
	  {
	    dist=xyz_dist(xyz_list[i], xyz_list[k]);
	    miN = (dist<miN) ? dist : miN;
	    maX = (dist>maX) ? dist : maX;
	  }
  if (miN==1000000000.0 || maX==0.0 || miN < okerr) return 0; /* ?? */
  rad=miN/2.0; /* assume all circles have this radius when
		  computing overlaps */
  /* compute and store overlaps */
  if (!alloc_overlaps(p)) return 0;
  for (i=1;i<=p->nodecount;i++)
    for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++)
      if ((k=pK_ptr[i].flower[j])>i
	  && i<= count && k<=count )
	{
	  ovlp=xyz_inv_dist(xyz_list[i],xyz_list[k],rad,rad);
	  set_single_overlap(p,i,j,ovlp);
	  ovlpmax = (ovlp>ovlpmax) ? ovlp : ovlpmax;
	  ovlpmin = (ovlp<ovlpmin) ? ovlp : ovlpmin;
	}
  sprintf(msgbuf,"Set overlaps using euclidean 3-space"
	  "distances.\n"
	  "  side lengths: max = %f, max/min = %f\n"
	  "  radius %f\n"
	  "  overlaps: max = %f, min =%f.",
	  (singledouble)maX,(singledouble)(maX/miN),(singledouble)rad,(singledouble)ovlpmax,(singledouble)ovlpmin);
  msg();
  return 1;
} /* set_xyz_overlaps */
