#include "sa_types.h"
#include "sa_proto.h"

/* Process various changes in screen settings. */

//#include "Interface.h"

int set_screen(struct p_data *p,char *datastr)
{
  int i,v,count=1,dum;
  double x,y,h,hh,factor,maxh,maxw,dist;
  singledouble l_x,l_y,r_x,r_y;
  char *nextpoint,next[BUFSIZE],buff[256];
  struct s_data *q;
  
  q=p->screen;
  nextpoint=datastr;
  if (!grab_next(&nextpoint,next) || next[0]!='-') goto OUTAHERE;
  do {
    if (next[0]!='-') goto OUTAHERE;
    
    // explicit box
    if (next[1]=='b' 
	&& (dum=sscanf(nextpoint,"%lf %lf %lf %lf",&l_x,&l_y,&r_x,&r_y))==4
	&& l_x<r_x && l_y<r_y) { /* new real canvas size */
      q->box.lx=l_x;q->box.ly=l_y;
      q->box.rx=r_x;q->box.ry=r_y;
      count++;
    }

	// incremental transformation (eg., dragging mouse)
	else if (next[1]=='i' && sscanf(nextpoint,"%lf %lf",&x,&y)==2) {
		q->box.lx += x;
		q->box.rx += x;
		q->box.ly += y;
		q->box.ry += y;
		count++;
	}

	// change to default canvas size, but keep current center
    else if (next[1]=='d' || (next[1]=='a' && p->hes<0)) { 
      x=(q->box.rx+q->box.lx)/2.0;
      y=(q->box.ry+q->box.ly)/2.0; 
      factor=((double)q->pix_box.rx)/((double)q->pix_box.ry);
      /* same apperture, smallest size = 1.1 */
      if (factor<=1) hh=1.1;
      else hh=1.1*factor;
      q->box.rx=hh;q->box.lx=-hh;
      if (factor<=1) hh=1.1/factor;
      else hh=1.1;
      q->box.ry=hh;q->box.ly=-hh;
      q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
			    /(q->box.ry-q->box.ly));
      count++;
    }

    /* center screen on x y coord or vert v */
    else if (next[1]=='c' || next[1]=='z' || next[1]=='v') {
      i=0;
      if (next[1]=='v' && (v=grab_one_vert(p,&nextpoint))) {
	x=p->packR_ptr[v].center.re;
	y=p->packR_ptr[v].center.im;
	i=1;
      }
      else if ((next[1]=='c' || next[1]=='z') && grab_next(&nextpoint,next)
	       && sscanf(next,"%lf",&x)==1 && grab_next(&nextpoint,next)
	       && sscanf(next,"%lf",&y)==1 ) {
	i=1;
      }
      if (i) {
	if (p->hes>0) { /* spherical, put x,y at center: reset, 
			   then rotate for theta then for phi. */
	  sprintf(buff,"0.0 0.0 %f",(singledouble)(-x/M_PI));
	  sphere_view(p,buff);
	  sprintf(buff,"-i 0.0 %f 0.0",(singledouble)(y/M_PI-.5));
	  sphere_view(p,buff);
	}
	else { 
	  h=(q->box.ry-q->box.ly)/2.0;
	  q->box.ry=y+h; q->box.ly=y-h;
	  hh=(q->box.rx-q->box.lx)/2.0;
	  q->box.rx=x+hh; q->box.lx=x-hh;
	  q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
				/(q->box.ry-q->box.ly));
	}
	count++;
      }
    }

    /* set height (and width) */
    else if (next[1]=='h' || next[1]=='w') {
      if (grab_next(&nextpoint,next) && sscanf(next,"%lf",&h)==1 && h>0.0) {
	hh = h/2.0;
	y=(q->box.ry+q->box.ly)/2.0;
	q->box.ry=y+hh; q->box.ly=y-hh;
	x=(q->box.rx+q->box.lx)/2.0;
	hh=h*(q->pix_box.rx)/q->pix_box.ry/2.0;
	q->box.rx=x+hh; q->box.lx=x-hh;
	q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
			      /(q->box.ry-q->box.ly));
	count++;
      }
    }

    /* mult height by factor*/
    else if (next[1]=='f') { 
      if (grab_next(&nextpoint,next)
	    && sscanf(next,"%lf",&factor)==1
	    && factor > 0.01) {
	h=q->box.ry-q->box.ly;
	hh = h*factor/2.0;
	y=(q->box.ry+q->box.ly)/2.0;
	q->box.ry=y+hh;q->box.ly=y-hh;
	x=(q->box.rx+q->box.lx)/2.0;
	hh=hh*(q->pix_box.rx)/q->pix_box.ry;
	q->box.lx=x-hh;q->box.rx=x+hh;
	q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
			      /(q->box.ry-q->box.ly));
	count++;
      }
    }

    /* for eucl packs, adjust size to show all circles */
    else if (next[1]=='a' && p->hes==0) {
      x=(q->box.rx+q->box.lx)/2.0;
      y=(q->box.ry+q->box.ly)/2.0;
      maxw = fabs(fabs(p->packR_ptr[1].center.re)+
		  p->packR_ptr[1].rad-x);
      maxh = fabs(fabs(p->packR_ptr[1].center.im)+
		  p->packR_ptr[1].rad-y);
      for (i=2;i<=p->nodecount;i++) {
	dist=fabs(fabs(p->packR_ptr[i].center.re)+
		  p->packR_ptr[i].rad-x);
	if (dist>maxw) maxw=dist;
	dist=fabs(fabs(p->packR_ptr[i].center.im)+
		  p->packR_ptr[i].rad-y);
	if (dist>maxh) maxh=dist;
      }
      if (maxh>okerr && maxw>okerr) {
	factor=((double)q->pix_box.rx)/((double)q->pix_box.ry);
	if (maxw/maxh>factor) {
	  hh=(1.1)*maxw;
	  q->box.rx=x+hh;q->box.lx=x-hh;
	  hh = hh/factor;
	  q->box.ry=y+hh;q->box.ly=y-hh;
	  q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
				/(q->box.ry-q->box.ly));
	  count++;
	}    
	else {
	  hh=(1.1)*maxh;
	  q->box.ry=y+hh;q->box.ly=y-hh;
	  hh = hh*factor;
	  q->box.rx=x+hh;q->box.lx=x-hh;
	  q->pix_per_unit=(int)((q->pix_box.ry-q->pix_box.ly)
				/(q->box.ry-q->box.ly));
	  count++; 
	}
      }
      else if (next[1]=='a') /* wrong geometry but not an error */
	count++;
    }
  } /* end of do */
  while (nextpoint!=NULL && grab_next(&nextpoint,next) );
 OUTAHERE:
  if (count) send_setViewBox(p); // have new box sent to java 
  return count;
} /* set_screen */
