#include "sa_types.h"
#include "sa_proto.h"

/* Set the 3x3 matrices for transforms of the sphere for purposes
   of graphing. These are in 'screen' data; disp_trans goes from
   real world x,y,z, to apparent screen x,y,z; its inverse is just
   its transpose and does the reverse identification. */

int sphere_view(struct p_data *p,char *datastr)
{
  int i,j;
  char c;
  double ax=0.0,ay=0.0,az=0.0;
  double M[3][3],iM[3][3];
  struct s_data *q=p->screen;
  
  stripsp(datastr);
  /* must check is '-' indicates command or negative sign */
  if (*datastr=='-' && (c=*(datastr+1))!='.' && (c<'0' || c>'9') ) {
    /* default settings */
    if (*(datastr+1)=='d') {
      for (i=0;i<3;i++) {
		for (j=0;j<3;j++) q->disp_trans[i][j]=0.0;
		q->disp_trans[i][i]=1.0;
      }
      sph_rotation(0.0,-0.1*M_PI,-0.03*M_PI,q->disp_trans);
    }

    /* use inputs as x, y, z increments to current matrix. */
    else if (*(datastr+1)=='i') { 
      if (sscanf(datastr+2,"%lf %lf %lf",&ax,&ay,&az)>0) 
		sph_rotation(ax*M_PI,ay*M_PI,az*M_PI,q->disp_trans);
    }

    /* read 3x3 transformation matrix, store and store inverse. */
    else if (*(datastr+1)=='t') {
      if (sscanf(datastr+3,"%lf %lf %lf %lf %lf %lf %lf %lf %lf",
		 &M[0][0],&M[0][1],&M[0][2],
		 &M[1][0],&M[1][1],&M[1][2],
		 &M[2][0],&M[2][1],&M[2][2])!= 9
	  || fabs(1.0-Determinant(M))>.0001 ) {
		sprintf(msgbuf,"set_sv error: -t option requires entries of "
			"3x3 orthogonal matrix.");
		emsg();
		return 0;
	  }
      for (i=0;i<3;i++) for (j=0;j<3;j++) q->disp_trans[i][j]=M[i][j];
	  if (*(datastr+2)!='q') // not in quiet mode, report new view
		send_sph_view(p);
	  return 1;
    }
  }

  /* use inputs as x, y, z of new matrix */
  if (sscanf(datastr,"%lf %lf %lf",&ax,&ay,&az)>0) {
   	for (i=0;i<3;i++) { /* reset to default */
		for (j=0;j<3;j++) q->disp_trans[i][j]=0.0;
		q->disp_trans[i][i]=1.0;
   	}
   	sph_rotation(ax*M_PI,ay*M_PI,az*M_PI,q->disp_trans);
  }
	return 1;
} /* sphere_view */
