#include "sa_types.h"
#include "sa_proto.h"

/* Given a boundary edge-path (from startv to endv), we add a layer
of "square grid" type to the complex. That is, you add matching circles
to each of these boundary circles and you add ball bearings to the
resulting interstices. Return new startv/endv indexes. Return 0 on error.*/

int sq_grid_lay_call(struct p_data *p,char *datastr)
{
  int i,v,w,hits,j,n=1,flag=0,count,corn[4],J,nv;
  int *bdrylist=NULL;
  char *nextpoint,*endptr,next[BUFSIZE];
  struct Vertlist *vlist=NULL,*vtrace;
  struct Edgelist *segs=NULL,*etrace;

  nextpoint=datastr;

  /* any flags? */

  if (grab_next(&nextpoint,next) && next[0]=='-') {
    /* -n flag, number of full layers */
    if (next[1]=='n' && p->num_bdry_comp 
	&& sscanf(next+2,"%d",&n) && n>0) {
      v=w=p->bdry_starts[1];
      flag=1;
    }
    /* -r flag; add one layer with four designated corners */
    else if (next[1]=='r') {
      if (!p->num_bdry_comp 
	  || !((p->euler==1 || p->euler==2) && p->genus==0)
	  || !(vlist=Node_link_parse(p,nextpoint,&endptr,&hits,
		     &Vlist,&Elist,&Flist,&region,pathlist,pathlength))
	  || hits<4 ) {
	vert_free(&vlist);
	return 0;
      }
      j=0;
      vtrace=vlist;
      while (vtrace && j<4) {
	corn[j]=vtrace->v;
	if (!p->packK_ptr[vtrace->v].bdry_flag) return 0;
	j++;
	vtrace=vtrace->next;
      }
      vert_free(&vlist);
      flag=2;
    }
  }

  /* else look for start/end vertices */

  else if (!(vlist=Node_link_parse(p,datastr,&endptr,&hits,
	      &Vlist,&Elist,&Flist,&region,pathlist,pathlength)) 
      && p->num_bdry_comp 
      && (p->euler==1 || p->euler==2) && p->genus==0) {  
             /* no vertices given; if simply connected, add full 
		layer to boundary */
    v=w=p->bdry_starts[1];
    n=1;
    flag=1;
  }
  else if (!(vlist->next)) { /* only got one vertex */
    vert_free(&vlist);
    return 0;
  }

  /* else, get the two points */
  else {
    v=vlist->v;
    w=vlist->next->v;
    if (!p->packK_ptr[v].bdry_flag || !p->packK_ptr[w].bdry_flag)
      return 0;
    vert_free(&vlist);
  }

  /* now branch for action: flag=1,  */

  if (!flag) return (sq_grid_layer(p,&v,&w));

  if (flag==1) { /* full layers */
    count=0;
    for (j=1;j<=n;j++) {
      w=v;
      if (sq_grid_layer(p,&v,&w)) count++;
      else return count;
    }
  }

  if (flag==2) { /* have four corners */
    /* need to find right order, check conditions, etc. */
    sprintf(next,"b(%d,%d)",corn[0],corn[0]);
    if (!(vlist=Node_link_parse(p,next,&endptr,&hits,
		&Vlist,&Elist,&Flist,&region,pathlist,pathlength)))
      return 0;
    j=0;
    vtrace=vlist;
    while (vtrace) {
      j++;
      vtrace=vtrace->next;
    }
    bdrylist=(int *)calloc((size_t)(j+1),sizeof(int));
    vtrace=vlist;
    J=0;

    /* mark the corners */
    while (vtrace) {
      bdrylist[J]=v=vtrace->v;
      if (v==corn[0] || v==corn[1] || v==corn[2] || v==corn[3]) 
	bdrylist[J] *= -1;
      vtrace=vtrace->next;
      J++;
    }
    vert_free(&vlist);

    /* make edgelist of four bdry segments between corners */
    if (bdrylist[0]>0) { /* first one should be marked */
      free(bdrylist);
      return 0;
    }
    corn[0]=0;
    segs=etrace=(struct Edgelist *)calloc((size_t)1,sizeof(struct Edgelist));
    etrace->v=-bdrylist[0];
    for (i=1;i<4;i++) {
      j=corn[i-1]+1;
      while (bdrylist[j]>0 && j<J) j++;
      if (j<corn[i-1]+3 || j==J) {
	edge_free(&segs);
	free(bdrylist);
	return 0; /* must have >=2 edges, four corners */
      }
      corn[i]=j;
      etrace->w=-bdrylist[j];
      etrace=etrace->next=(struct Edgelist *)
	calloc((size_t)1,sizeof(struct Edgelist));
      etrace->v=-bdrylist[j];
    }
    etrace->w=-bdrylist[0]; 

    /* now, add layers to every one of the four segments */
    etrace=segs;
    while (etrace) {
      v=etrace->v;
      w=etrace->w;
      sq_grid_layer(p,&v,&w);
      etrace=etrace->next;
    }
    edge_free(&segs);

    /* now, add extra circles to fill out each corner */
    for (i=0;i<4;i++) {
      add_vert(p,-bdrylist[corn[i]]);
      nv=p->nodecount;
      enfold(p,-bdrylist[corn[i]]);
      add_vert(p,nv);
      enfold(p,nv);
      corn[i]=p->nodecount;
    }
    sprintf(msgbuf,"New corner indices are %d %d %d %d.",
	    corn[0],corn[1],corn[2],corn[3]);
    msg();
    free(bdrylist);
    complex_count(p,0);
    facedraworder(p,0);
    return 1;
  }
  return 1;  // POSSIBLE PROBLEM HERE
} /* sq_grid_lay_call */
  
int sq_grid_layer(struct p_data *p,int *startv,int *endv)
{
  int nextv,bb,n,cflag=0,count=0,v,edv;

  if (*startv==*endv) cflag=1;
  edv=*endv;
  /* get started; duplicate first circle */
  nextv=p->packK_ptr[*startv].flower[0];
  add_vert(p,nextv);
  bb=p->nodecount;
  add_vert(p,bb);
  count++;
  *startv=p->nodecount;
  add_vert(p,bb);
  enfold(p,bb);

  /* move along the edge */
  v=nextv;
  nextv=p->packK_ptr[v].flower[0];
  while (nextv!=edv) {
    add_vert(p,nextv);
    bb=p->nodecount;
    enfold(p,v);
    add_vert(p,bb);
    count++;
    enfold(p,bb);
    v=nextv;
    nextv=p->packK_ptr[v].flower[0];
  }

  /* wrap up, we're done */
  if (!count) /* too short to close, so leave open. */
    cflag=0;
  add_vert(p,nextv);
  bb=p->nodecount;
  enfold(p,v);
  if (cflag) {
    enfold(p,nextv);
    enfold(p,bb);
    *endv=*startv;
    return count;
  }
  add_vert(p,bb);
  *endv=p->nodecount;
  enfold(p,bb);
  count++;
  return count;
} /* sq_grid_layer */




