/* tenting.cpp   5/2005 */
#include "sa_types.h"
#include "sa_proto.h"

/* "Tenting" is a construction step for building complexes motivated
by a question of type that Phil Bowers asked. Tenting on a sequence
of boundary edges involves adding faces to each edge and adding
edges between the new vertices forming neighboring peaks. One
can use this, e.g., in place of "add_layer" or "add_gen" if you
don't want degrees to effect the results. 

We process edge list in maximally connected ordered fragments.
(Note: incoming list of edges might be changed.) */ 

int tenting(struct p_data *p,struct Edgelist **inlist)
{
  int v,w,first_vert,count=0,tick=0;
  struct Edgelist *elist,*etrace,*etrc,*elast;
  struct K_data *pK_ptr=p->packK_ptr;

  elist=*inlist;
  etrace=elast=elist;
  if (!etrace) return 0; /* no list */

  /*  Ensure edges are bdry with verts in order. */

  while (etrace) {
    if (pK_ptr[(v=etrace->v)].bdry_flag 
	&& pK_ptr[(w=etrace->w)].bdry_flag
	&& (pK_ptr[v].flower[0]==w) || (pK_ptr[w].flower[0]==v)) {
      count++;
      /* switch order? */
      if (pK_ptr[w].flower[0]==v) {
	etrace->v=w;
	etrace->w=v;
      }
      elast=etrace;
      etrace=etrace->next;
    }
    /* else, need to remove from list */
    else if (elast==etrace) { /* still at beginning of list? */
      elist=elist->next;
      free(etrace);
      etrace=elast=elist;
    }
    else { 
      etrc=etrace;
      elast->next=etrace=etrace->next;
      free(etrc);
    }
  } /* end of while */

  if (!count) /* nothing left? */
    return 0; 

  /* Save first vertex */

  /* ================================ */
  /* Now, loop through max connected segments. */

  count=0;
  first_vert=elist->v;
  etrace=elist;
  while (etrace) {
    tick=0;
    w=etrace->w;
    /* edge already done? */
    if (etrace->w!=pK_ptr[etrace->v].flower[0]) {
      etrace=etrace->next;
    }
    else {
      /* add tent */
      if (!add_vert(p,etrace->w)) return count; /* error? */
      pK_ptr=p->packK_ptr;
      pK_ptr[p->nodecount].plot_flag=1;
      /* keep going along contiguous edges */
      count++;
      tick++;
      while ((etrace=etrace->next) && etrace->v==w 
	     && pK_ptr[etrace->v].flower[0]==(w=etrace->w)) {
	if (!add_vert(p,etrace->w) || !enfold(p,etrace->v)) /* error? */
	  return count;
	pK_ptr=p->packK_ptr;
	pK_ptr[p->nodecount].plot_flag=1;
	count++;
      }
    }

    if (!etrace && tick && w==first_vert) { /* done. connect to first? */
      if (!enfold(p,w)) return 0;
      count++;
    }
  } /* end of while */
  return count;
} /* tenting */

int tent_call(struct p_data *p,char *datastr)
{
  int hits,count;
  char *nextpoint,*endptr;
  struct Edgelist *elist=NULL;

  nextpoint=datastr;
  stripsp(datastr);
  if (!strlen(nextpoint)) {
    nextpoint=(char *)calloc(15,sizeof(char));
    nextpoint[0]='b';
    nextpoint[1]='\0';
  }
  if (!(elist=Node_pair_link(p,nextpoint,&endptr,&hits,
			     &Vlist,&Elist,&Flist,&region,
			     pathlist,pathlength))) {
    sprintf(msgbuf,"usage: tenting {e..}");
    emsg();
    return 0;
  }
  count=tenting(p,&elist);
  complex_count(p,0);facedraworder(p,0);
  strcpy(nextpoint,"fix -c\0");
  fix_call(p,nextpoint);
  return count;
} /* test_call */
  
