#include "sa_types.h"
#include "sa_proto.h"

/* Computes the moduli of a torus. Return 0 on error. Set flag=1 if
complex is not a torus, flag=2 if edge-pairings not recorded. */

complex torus_t(struct p_data *p,int *flag)
{
  int count=0,j;
  complex zreturn,m_one;
  complex r1,r12,r2,r22,r3,r32,w;  
  complex W[7];
  Mobius mob;

  *flag=0;
  zreturn.re=zreturn.im=0.0;
  m_one.re=-1.0;m_one.im=0.0;

  // Determine if this is a torus. 
 
  if (p->num_bdry_comp>0 || p->genus!=1)
    {
      *flag=1;
      return zreturn;
    }

/* Compute array of values cdiv(mob.a, mob.b) for each 
   edge-pair[i].mob. */
         
  j=1;
  while(p->edge_pair[j].edge)
    {
      mob=p->edge_pair[j].mob;
      W[j]=cdiv(mob.b,mob.a);
      j++;
    }
  if (j==1)
    {
      *flag=2;
      return zreturn;
    }
  

  /*----Calculate ratios and select candidate in upper half plane ------*/

  r1=cdiv(W[1],W[2]);
  r12=cdiv(W[2],W[1]);
  r2=cdiv(W[3],W[2]);
  r22=cdiv(W[2],W[3]);
  r3=cdiv(W[3],W[4]);
  r32=cdiv(W[4],W[3]);

  /*get initial tau in upper half plane*/
  
  if (r1.im>0)         
    w=r1;
  
  else if (r12.im>0)
    w=r12;
  
  else if (r2.im>0)
    w=r2;
  
  else if (r22.im>0)
    w=r22;
  
  else if (r3.im>0)
    w=r3;
  
  else if (r32.im>0)
    w=r32;
  
  else 
    return zreturn;

  w.re=w.re-floor(w.re)-floor(2*(w.re-floor(w.re)));
  while (count<1000 && (w.im<sqrt(1-w.re*w.re)))
    {
      count++;
      w=cdiv(m_one,w);
      w.re=w.re-floor(w.re)-floor(2*(w.re-floor(w.re)));
    }
  if (count==1000) return zreturn;
  /* fixup: change output to take advantage of doubledouble, if
     available. */
  sprintf(msgbuf,"The computed value of tau is "
	  "(%.12f) + (%.12f)i.",(singledouble)w.re,(singledouble)w.im);
  msg();
  return(w);
} /* torus_t */
      


    





