// FILE tree_pack_command.c
// 080106 Modified to randomize expansion

#include "sa_head.h"
#include "sa_types.h"
#include "Interface.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <sys/time.h>

// meaning of StrTok keys 
#define SEPARATOR 1
#define PACKNUM 2
#define COMMAND 3
#define STRING 4
#define INTDATA 5
#define FLOATDATA 6

const long MROOT_LINK_FLAG = -1;
const long MBORDER_FLAG = 1;
const long MINTERIOR_FLAG = 0;
const long DEFAULT_LINK_ALLOCATION = 8;

struct Edgelist *tree_edges;

typedef struct {
  // links are stored in ccw fashion
  long flags;
  long id;
  long left_index; // note indexes are mid-cursor points
  long right_index;
  long link_count;
  long link_allocation;
  long *link_list;
} node_struct;

typedef struct {
  long alpha;
  long gamma;
  long node_count;
  long free_node_count;
  long next_free_node;
  node_struct *nodes;
} mesh_struct;

typedef struct border_node_struct{
  long parent_index;
  long index;
  long state; // is it active or not?
  struct border_node_struct *prev;
  struct border_node_struct *next;
}border_node;

typedef struct {
  struct border_node_struct *mark;
  long border_count;
  long active_count;
} border_ring_struct;

struct {
  mesh_struct * msp;
  mesh_struct * tsp;
} memory_record_struct;

struct {
  char infile_name[256];
  char outfile_name[256];
  int center;
} cmd_arg;

mesh_struct *new_mesh(long n);
mesh_struct *new_empty_mesh (long n);
long next_border_node(mesh_struct *msp, long n);
int g_link_list(node_struct *nsp);
void left_link_insert (node_struct *nsp, long n);
void right_link_insert (node_struct *nsp, long n);
long insert_border_node(mesh_struct *msp, long node);
int check_for_link(node_struct *nsp, long n);
long join_border_node(mesh_struct *msp, long node);
void output_mesh_text ( mesh_struct *vmp);
void CP_output_mesh_text (mesh_struct *vmp);
int store_result(struct p_data *p,mesh_struct *msp);
int new_tree_pack(struct p_data *p,char *filename,char *datastr);
int init_border_struct (border_ring_struct *brsp, mesh_struct *msp, long c);
void border_ring_insert_prev ( border_ring_struct *brsp, long parent, long id);
int border_ring_delete( border_ring_struct * brsp );
int expand_border_node (border_ring_struct *brsp, mesh_struct *tsp, mesh_struct *msp, border_node *bnsp);
void iterate_border_ring(border_ring_struct * brsp);
void randomize_border_ring_mark(border_ring_struct * brsp);
// 080106 added randomize_border_ring_mark()

void cleanup_memory();
mesh_struct *process_tree_file(const char * s);
void init_memory_record();
void post_memory(mesh_struct *msp, mesh_struct *tsp);

int tree_pack_call(struct p_data *p,char *datastr)
{
  int N=0,len,allocation;
  char *options=NULL;
  tree_edges=NULL;

  StrTok *strTok=string_tokenizer(datastr,&allocation,&N);
  if (!strTok || N<1 || strTok[N-1].key!=STRING || !strTok[N-1].str || !(len=strlen(strTok[N-1].str))) {
  	if (strTok) tok_free(&strTok,allocation);
  	return 0;
  }
  strncpy(cmd_arg.infile_name,strTok[N-1].str,256);
  // should have name now, but is it 'infile'?
  if (N>1 && strTok[N-2].key==COMMAND 
  && strTok[N-2].str && !strncmp(strTok[N-2].str,"infile",6)) {
	// force call from Java using temp filename
  	int result=callInfileRead(cmd_arg.infile_name,4);
  	tok_free(&strTok,allocation);
  	return result;
  }
  // 'options' is null; in future may need construction options;
  //   current center and verbose options default to 0.
  int result=new_tree_pack(p,cmd_arg.infile_name,options);
  if (options) free(options);
  tok_free(&strTok,allocation);
  if (!result) cleanup_memory();
  return result;
} /* tree_pack_call */

int new_tree_pack(struct p_data *p,char *filename,char *datastr)
/*Flags will be provided in the future. Return 0 on error. */
{
  int treepack_status = 0;

  init_memory_record();
  // process tree
  // create initial tree info -- fill into mesh_struct; not quite a tri yet
  mesh_struct *tsp, *msp;
  if ((tsp = process_tree_file(filename))==NULL 
  	|| (msp = process_tree_file(filename))==NULL) // need a copy of original tree
  	return 0;
  post_memory(msp, tsp); // so we can cleanup later
  // need to fix up msp for use -- there's a better way to organize this
  // this is initialization stuff that should be done elsewhere
  {
    long j, k;
    for (k=0; k<(*msp).node_count; k++) {
      long lndx, rndx;
      lndx = (*msp).nodes[k].left_index;
      rndx = (*msp).nodes[k].right_index;
      for (j=lndx; j<rndx; j++) ((*msp).nodes[k].link_list)[j]=0;
      (*msp).nodes[k].right_index = lndx;
      (*msp).nodes[k].link_count =0;
      (*msp).nodes[k].flags =0;
    }
  }
  
  if (treepack_status != 0) { //file error?
    return 0;
  }

  // build border ring
  // choose head node (need to be interior)

  border_ring_struct *brsp;
  brsp = (border_ring_struct *)calloc(1,sizeof(border_ring_struct));

  long center=0;
  if (msp->alpha) {
    center = msp->alpha-1; // adjust for offset in indices!!!
  }

  // now adjust center
  { // is the center an interior i.e. more than one link?
    if ( (*tsp).nodes[center].link_count < 2) {
      node_struct *nsp;
      nsp = &(*tsp).nodes[center];
      long ndx = (*nsp).left_index;
      long save_ndx = -1;
      long center_status =0;
      while (ndx < (*nsp).right_index) {
	if ( (*tsp).nodes[(*nsp).link_list[ndx]].link_count > 1) {
	  center_status =1;
	  save_ndx = ndx;
	  break;
	}
	else ndx++;
      }
      if (!center_status) { // error all nodes are exterior
	cleanup_memory(); // ?????? tsp, nsp);
	return 0; // exit(0);
      }
      center = (*nsp).link_list[save_ndx];     
      msp->alpha=center+1; // adjust for offset??
    }
  }
  
  init_border_struct (brsp, tsp, center);

  // loop until all are inactive i.e. until no more expansion possible.
  while ((*brsp).active_count > 0) {
    //CP_output_mesh_text(msp);	
    if ( ((*((*brsp).mark))).state ) {
      // printf("expanding node %ld\n", ((*((*brsp).mark))).index+1);
      expand_border_node (brsp, tsp, msp, (*brsp).mark);
      // prev routine adjusts active node counts !!!
    } else {
    	iterate_border_ring(brsp);
     }
    // 080106 modified
    randomize_border_ring_mark(brsp);
  }
  store_result(p,msp); 
//  CP_output_mesh_text(msp);
  cleanup_memory();
  return 1; 
} /* new_tree_pack */

mesh_struct *process_tree_file(const char * s){
  mesh_struct *msp;

  FILE * fp;
  char command[256];
  struct Edgelist *etrace;
  long node_count;

  fp = fopen (s, "r");
  if (fp == NULL) {
	return NULL;
    // unable to open file 
  }
  
  // search for NODECOUNT
  while ((fscanf(fp, "%127s", command) != EOF) 
	 && strcmp(command,"NODECOUNT:")!= 0 );
  if (strcmp(command,"NODECOUNT:")!= 0) return NULL;

  // prepare msp for values
  fscanf(fp, "%ld", &node_count);
  msp = new_empty_mesh(node_count);
  int dum;
  
  // now sort through the data
  while ( (fscanf(fp,"%127s",command)!=EOF) && (strcmp(command,"END")!=0) ) {
	if (strcmp(command,"ALPHA/BETA/GAMMA:")==0)
		fscanf(fp,"%d %d %d",&(msp->alpha),&dum,&(msp->gamma));
	else if (strcmp(command,"FLOWERS:")==0) {
    // begin reading node data (remember off-by-one rule)
	    long k, c;
    	tree_edges=etrace=(struct Edgelist*)calloc(1,sizeof(struct Edgelist));
    	for (k=0; k<node_count; k++){ // do this for each node
 			long nodeid;
	      	long faces;
      		long read_count;
      		long link_count;

      		fscanf(fp, "%ld %ld", &nodeid, &faces);
      		read_count = faces +1;
      
      		node_struct *nsp;
      		nsp = &(*msp).nodes[nodeid-1];
      		(*nsp).id = nodeid;
      		long tempid, saveid;
      		for (c=0; c<read_count-1; c++) {
				fscanf(fp, "%ld", &tempid);
				right_link_insert(nsp, tempid-1);
				// save edge
				etrace=etrace->next=(struct Edgelist*)calloc(1,sizeof(struct Edgelist));
				etrace->v=nodeid;
				etrace->w=tempid;
      		}
      		// have one more
      		fscanf(fp, "%ld", &tempid);
      		if ((read_count > 1)&&(tempid == ( (*nsp).link_list[(*nsp).left_index]+1) )) {
				(*nsp).flags=0;
				// ignore this value
      		}
      		else {
				(*nsp).flags=1;
				right_link_insert(nsp, tempid-1);
				// save edge
				etrace=etrace->next=(struct Edgelist*)calloc(1,sizeof(struct Edgelist));
				etrace->v=nodeid;
				etrace->w=tempid;
      		}
		    (*msp).node_count++;
		    // printf ("Node %ld, faces %ld: ", nodeid, faces);
    	}
	}
  }
  
  // first edge spot is empty
  etrace=tree_edges->next;
  free(tree_edges);
  tree_edges=etrace;
  
  return (msp);
} /* process_tree_file */

void post_memory(mesh_struct *msp, mesh_struct *tsp) {
  memory_record_struct.msp = msp;
  memory_record_struct.tsp = tsp;
} /* post_memory */

void init_memory_record() {
  memory_record_struct.msp = NULL;
  memory_record_struct.tsp = NULL;
} /* init_memory_record */

void free_mesh_struct(mesh_struct *msp) {
  if (msp==NULL) return;
  long k;
  void *tempp;
  for (k=0; k<(*msp).node_count; k++) {
    tempp = (*msp).nodes[k].link_list;
    if (tempp != NULL) free(tempp);
  }
  tempp = (*msp).nodes;
  if (tempp != NULL) free(tempp);
} /* free_mesh_struct */

void cleanup_memory() {
  // deallocate all memory for java CP since cant use exit
  free_mesh_struct(memory_record_struct.msp);
  free_mesh_struct(memory_record_struct.tsp);
  edge_free(&tree_edges);
} /* cleanup_memory */

long next_border_node(mesh_struct *msp, long n){
  long *nodelinks;
  long leading_index;
  long newid;
  nodelinks = (*msp).nodes[n].link_list;
  leading_index = (*msp).nodes[n].left_index;
  newid = nodelinks[leading_index];
  return (newid);
} /* next_border_node */

int g_link_list(node_struct *nsp)
  /* return 0 on error */
{
  long newsize, leading_count;
  newsize = (*nsp).link_allocation + DEFAULT_LINK_ALLOCATION;
  leading_count = DEFAULT_LINK_ALLOCATION / 2;

  long * newlink_list;
  newlink_list = (long *)calloc(newsize,sizeof(long));
  if (newlink_list == NULL) { //allocation error
    cleanup_memory();
    return 0; // exit(0);
  }
  
  long k;
  for (k=(*nsp).left_index; k<(*nsp).right_index; k++) newlink_list[k + leading_count] = (*nsp).link_list[k];
  (*nsp).left_index += leading_count;
  (*nsp).right_index += leading_count;
  (*nsp).link_allocation = newsize;
  free( (*nsp).link_list);
  (*nsp).link_list = newlink_list;
  return 1;
} /* g_link_list */

void left_link_insert (node_struct *nsp, long n){
  // printf ("LLI: at node %ld to node %ld\n", (*nsp).id, n+1);
  if ((*nsp).left_index > 0) { // OK to insert
    (*nsp).left_index--;
    (*nsp).link_list[(*nsp).left_index] = n;
    (*nsp).link_count++;
  }
  else {
    g_link_list(nsp);
    left_link_insert(nsp, n);
  }
} /* left_link_insert */

void right_link_insert (node_struct *nsp, long n){
  if ((*nsp).right_index < (*nsp).link_allocation) { // OK to insert
    (*nsp).link_list[(*nsp).right_index] = n;
    (*nsp).right_index++;
    (*nsp).link_count++;
  }
  else {
    g_link_list(nsp);
    right_link_insert(nsp, n);
  }
} /* right_link_insert */

// ?????? seems this is never called ?????? is this a rand_tree routine?
long insert_border_node(mesh_struct *msp, long node){
  // newnode // dual links // be sure to mark border
  // return -1 on error
  long newid;
  newid = (*msp).next_free_node++;(*msp).free_node_count--;(*msp).node_count++;
  if ((*msp).free_node_count < 0) { // allocation error
    cleanup_memory();
    return -1; // exit(0);
  }
  node_struct *nsp;
  nsp = & (*msp).nodes[newid];
  (*nsp).id = newid + 1; // correct for the offset
  (*nsp).left_index  = (*nsp).right_index = DEFAULT_LINK_ALLOCATION / 2;
  (*nsp).flags = MBORDER_FLAG;

  // now connect the dots
  long nextnode;
  nextnode = next_border_node(msp, node);
  left_link_insert(nsp, nextnode);
  right_link_insert(nsp, node);
  right_link_insert( &(*msp).nodes[nextnode], newid);
  left_link_insert( &(*msp).nodes[node], newid);
  return (newid);
};

int check_for_link(node_struct *nsp, long n) {
  long status = 0;
  long k;
  for (k=(*nsp).left_index; k<(*nsp).right_index; k++) status |= ((*nsp).link_list[k]==n);

  return (status);
};

long join_border_node(mesh_struct *msp, long node){
  // sink node // make the links // 
    long nextnode; 
  long nextnextnode;
  nextnode = next_border_node(msp, node);
  nextnextnode = next_border_node(msp, nextnode);
  if (!check_for_link( &(*msp).nodes[node], nextnextnode)) { //we are OK
    right_link_insert(&(*msp).nodes[nextnextnode], node);
    left_link_insert(&(*msp).nodes[node], nextnextnode);
    (*msp).nodes[nextnode].flags = MINTERIOR_FLAG;
  }
  else { // already a link there
  };
  return (nextnextnode);
};

mesh_struct *new_mesh( long n ){
  long k; // utility

  // make storage and setup data
  mesh_struct * msp;
  msp = (mesh_struct *)calloc(1,sizeof(mesh_struct));
  if (msp == NULL) { //allocation error
    cleanup_memory();
    return NULL; // exit(0);
  }

  (*msp).nodes = (node_struct *)calloc(n,sizeof(node_struct));
  if ((*msp).nodes == NULL) { //allocation error
    cleanup_memory();
    return NULL; // exit(0);
  }
  (*msp).node_count =0;
  (*msp).free_node_count = 0;
  (*msp).next_free_node = 0;
  (*msp).free_node_count = n;
  
  // now set up each link list
  
  for (k=0; k<n; k++) {
    node_struct * nsp;
    nsp = &(*msp).nodes[k];
    (*nsp).link_list = (long *)calloc(DEFAULT_LINK_ALLOCATION,sizeof(long));
    if ((*nsp).link_list == NULL) { // allocation error
      cleanup_memory();
      return NULL; // exit(0);
    }
    (*nsp).id = k+1;
    (*nsp).flags =0;
    (*nsp).link_count = 0; (*nsp).link_allocation =0;
    (*nsp).link_allocation =DEFAULT_LINK_ALLOCATION;
    (*nsp).left_index = (*nsp).right_index = DEFAULT_LINK_ALLOCATION/2;
  }

  // always start with three vertices

  for (k=0; k<3; k++) {
    (*msp).nodes[k].flags = MBORDER_FLAG;
  };

  left_link_insert(&(*msp).nodes[0], 1);
  right_link_insert(&(*msp).nodes[0], 2);
  left_link_insert(&(*msp).nodes[1], 2);
  right_link_insert(&(*msp).nodes[1], 0);
  left_link_insert(&(*msp).nodes[2], 0);
  right_link_insert(&(*msp).nodes[2], 1);
 
  msp->node_count = 3;
  (*msp).next_free_node = 3;
  (*msp).free_node_count -= 3;

  return(msp);
};

mesh_struct *new_empty_mesh( long n ){
  long k; // utility

  // make storage and setup data
  mesh_struct * msp;
  msp = (mesh_struct *)calloc(1,sizeof(mesh_struct));
  if (msp == NULL) { //allocation error
    cleanup_memory();
    return NULL; // exit(0);
  }

  (*msp).nodes = (node_struct *)calloc(n,sizeof(node_struct));
  if ((*msp).nodes == NULL) { //allocation error
    cleanup_memory();
    return NULL; // exit(0);
  }
  (*msp).node_count =0;
  (*msp).next_free_node = 0;
  (*msp).free_node_count = n;
  
  // now set up each link list
  for (k=0; k<n; k++) {
    node_struct * nsp;
    nsp = &((*msp).nodes[k]);
    (*nsp).link_list = (long *)calloc(DEFAULT_LINK_ALLOCATION,sizeof(long));
    if ((*nsp).link_list == NULL) { // allocation error
      cleanup_memory();
      return NULL; // exit(0);
    }
    (*nsp).id = k+1;
    (*nsp).flags =0;
    (*nsp).link_count = 0;
    (*nsp).link_allocation =DEFAULT_LINK_ALLOCATION;
    (*nsp).left_index = (*nsp).right_index = DEFAULT_LINK_ALLOCATION/2;
  }

  (*msp).node_count = 0;
  (*msp).next_free_node = 0;
  (*msp).free_node_count -= 0;
  return(msp);
};

void output_mesh_text ( mesh_struct *msp){
  printf("NODECOUNT:\t%ld\n\n", (*msp).node_count);
  printf("GEOMETRY:\teuclidean\n\n");
  printf("FLOWERS:\n\n");

  // loop through the nodes
  // list the links
  long i,k;
  for (i=0; i< (*msp).node_count; i++) {
    printf ("outputing node %ld:%ld :", i,(*msp).nodes[i].id);

    for (k=(*msp).nodes[i].left_index; k < (*msp).nodes[i].right_index; k++) {
      printf("%ld ", (*msp).nodes[i].link_list[k]);
    }
      printf("\n");
  }
  printf("\nEND\n");
};

int store_result(struct p_data *p,mesh_struct *msp)
  /* generate the packing with the resulting combinatorics
     and max pack it */
{
  int *newflower=NULL;
  empty_pack(p);
  alloc_pack_space(p,msp->node_count+1,1);
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;

  p->nodecount=msp->node_count;
  p->status=1;
  p->hes=-1; // make it hyperbolic
  // loop through the nodes, reading off the links
  int i,j,k;
  for (i=0;i<msp->node_count; i++) {
    int faces;
    faces = msp->nodes[i].link_count;
    if (msp->nodes[i].flags == MBORDER_FLAG) faces--;
    newflower=(int *)calloc((size_t)(faces+1),sizeof(int));
    int save_left_index = msp->nodes[i].left_index;
    j=0;
    for (k=msp->nodes[i].left_index; k < msp->nodes[i].right_index; k++) {
      newflower[j++]=msp->nodes[i].link_list[k]+1;
    }
    if (msp->nodes[i].flags != MBORDER_FLAG) { // close up
      newflower[j]=newflower[0];
      pK_ptr[i+1].bdry_flag=0;
    }
    else pK_ptr[i+1].bdry_flag=1;
    pK_ptr[i+1].num=faces;
    pK_ptr[i+1].flower=newflower;
  }

  // fix up the packing
  for (i=1;i<=p->nodecount;i++) pR_ptr[i].rad=.5;
  complex_count(p,1);
  choose_alpha(p);choose_beta(p);choose_gamma(p);
  facedraworder(p,0);
  p->active_node=p->alpha;
  sprintf(p->file_name,"random_%d.p",p->nodecount);
  new_pack_setup(p);

// max_pack
  int dum;
  int pnum=pack_num(p);
  sprintf(buf,"max_pack -p%d",pnum);
  handle_cmd(buf,&dum);
  p->elist=tree_edges;
  tree_edges=NULL;
} /* store_result */

void CP_output_mesh_text ( mesh_struct *msp){
  printf("NODECOUNT:\t%ld\n\n", (*msp).node_count);
  printf("GEOMETRY:\teuclidean\n\n");
  printf("FLOWERS:\n\n");

  // loop through the nodes
  // list the links
  long i,k;
  for (i=0; i< (*msp).node_count; i++) {
    long faces;
    faces = (*msp).nodes[i].link_count;
    if ((*msp).nodes[i].flags == MBORDER_FLAG) faces--;
    printf ("%ld\t%ld\t",(*msp).nodes[i].id, faces);

    long save_left_index = (*msp).nodes[i].left_index;
    for (k=(*msp).nodes[i].left_index; k < (*msp).nodes[i].right_index; k++) {
      printf("%ld ", (*msp).nodes[i].link_list[k]+1);
    }
    if ( (faces > 0   ) && (!(*msp).nodes[i].flags == MBORDER_FLAG))
      printf("%ld ", (*msp).nodes[i].link_list[save_left_index]+1);
    printf("\n");
  }
  printf("\nEND\n");
};

int init_border_struct (border_ring_struct *brsp, mesh_struct *msp, long c)
  /* return 0 on error */
{
  (*brsp).mark = NULL;
  (*brsp).border_count = (*brsp).active_count = 0;

  // center already adjusted and in proper index format (one less for c strings)
  border_node *newbs;
  newbs = (border_node *)calloc(1,sizeof(border_node));
  if (newbs == NULL) { 
    cleanup_memory();
    return 0; // exit(0);
  }
  (*newbs).parent_index = -1;
  (*newbs).index = c;
  (*newbs).state = 1;

  (*newbs).prev = newbs; (*newbs).next = newbs;
  (*brsp).mark = newbs; // (*brsp).head = newbs;
  (*brsp).border_count++; (*brsp).active_count++;
  return 1;
};

void border_ring_insert_prev (  border_ring_struct *brsp, long parent, long id){
  border_node *bnsp, *mark_prev_bnsp, *mark_bnsp;
  bnsp = (border_node *)calloc(1,sizeof(border_node));
  mark_bnsp = (*brsp).mark;
  mark_prev_bnsp = (*((*brsp).mark)).prev;

  (*bnsp).parent_index = parent;
  (*bnsp).index = id;
  (*bnsp).state = 1; (*brsp).active_count++;(*brsp).border_count++;// setting to active
  (*bnsp).next = (*brsp).mark;
  (*bnsp).prev = mark_prev_bnsp;
  (*mark_bnsp).prev = bnsp;
  (*mark_prev_bnsp).next = bnsp;
};

int expand_border_node(border_ring_struct *brsp,mesh_struct *tsp,
mesh_struct *msp,border_node *bnsp)
  /* return 0 on error */
{
  (*bnsp).state = 0; (*brsp).active_count--; // we are processing this node
  // now add each child in order excepting parent
  long first_round_marker = 0;

  long save_parent;
  save_parent = (*bnsp).parent_index;

  long tempndx, left_ref_index;
  tempndx = (*bnsp).index;
  left_ref_index = (*tsp).nodes[tempndx].left_index;
  long right_ref_index;
  right_ref_index = (*tsp).nodes[tempndx].right_index;

  long add_count;
  long *newlist;
  add_count = (*tsp).nodes[(*bnsp).index].link_count;
  newlist = (long *)calloc (add_count,sizeof(long));

  { // preparing newlist for addition
    long * orig_link_list;
    long startndx, stopndx;
    orig_link_list = (*tsp).nodes[(*bnsp).index].link_list;
    startndx =(*tsp).nodes[(*bnsp).index].left_index;
    stopndx = (*tsp).nodes[(*bnsp).index].right_index;
    
    if (save_parent >= 0) { // ie not the first node !!
      add_count--; // there will be a parent link to ignore
      // set up to save links in order
      
      long k;
      long currndx =0;
      long savendx =0;
	// cycle through indexes to find the parent
      for (k=startndx; k<stopndx; k++) {
	if (orig_link_list[k] == save_parent) savendx = k;
      }
      if (savendx<stopndx)
	for (k=savendx+1; k<stopndx; k++) {
	  newlist[currndx++] = orig_link_list[k];	  
	}
      for (k=startndx; k<savendx; k++) {
	newlist[currndx++] = orig_link_list[k];	  	
      }
    } else { // we have the original list -- nothing to be excluded
      long k;
      for (k=startndx; k<stopndx; k++) 
	newlist[k-startndx] = orig_link_list[k];
    
      first_round_marker = 1; // !!!! needed below
    }
  }
  
  long curr_left_ref = (*((*bnsp).prev)).index;  
  if (curr_left_ref == (*bnsp).index) curr_left_ref = newlist[add_count-1];
  long curr_right_ref = (*((*bnsp).next)).index;  
  if (curr_right_ref == (*bnsp).index) curr_right_ref = newlist[0];

  long k;
  for(k=0; k <add_count ; k++){
  border_ring_insert_prev(brsp, (*bnsp).index, newlist[k]); 
  (*msp).nodes[newlist[k]].flags =1;

  // insert should set to active for us
  // this ordering should be OK.  is this backwards? dont think so
  // fixed ordering 06262006

  // there are border nodes now -- need to add edges to the mesh
  // spokes down and up
  }

  if (add_count >1) {
    for (k=0; k<add_count; k++) { 
      left_link_insert(&(*msp).nodes[newlist[k]], (*bnsp).index);
      right_link_insert(&(*msp).nodes[(*bnsp).index], newlist[k]);
    }
    for (k=0; k<add_count; k++) { 
      left_link_insert(&(*msp).nodes[curr_left_ref], newlist[k]);
      right_link_insert(&(*msp).nodes[newlist[k]], curr_left_ref);
            
      curr_left_ref = newlist[k];
    }
    if (!first_round_marker) {
    // have a final link pair to add
      left_link_insert(&(*msp).nodes[curr_left_ref], curr_right_ref);
      right_link_insert(&(*msp).nodes[curr_right_ref], curr_left_ref);
    }
    // all the links are in place now

    // now delete mark!! -- be sure to set interior
    long del_id;
    del_id = (*((*brsp).mark)).index;
    (*msp).nodes[del_id].flags =0;
    border_ring_delete(brsp);
  } else {
    // no need to delete -- just increment mark
    iterate_border_ring(brsp);
  }
  free(newlist);
} /* expand_border_node */

int border_ring_delete (border_ring_struct * brsp)
  /* return 0 on error */
{
  if ((*brsp).mark != NULL) {
    border_node *bnp, *prev_bnp, *next_bnp;
    bnp = (*brsp).mark;
    next_bnp = (*bnp).next;
    prev_bnp = (*bnp).prev;
    if (prev_bnp == bnp) { // deleting rest of list
      (*brsp).mark = NULL;
      free(bnp);
    }
    else {
      (*prev_bnp).next = (*bnp).next;
      (*next_bnp).prev = prev_bnp;
      free(bnp);
    }
    (*brsp).mark = next_bnp;
    (*brsp).border_count--;
  }
  else { // error if here!!
    cleanup_memory();
    return 0; //exit(0);
  }
  return 1;
};

void iterate_border_ring(border_ring_struct * brsp){
  (*brsp).mark = (*((*brsp).mark)).next;
}

// 080106 added the following to get rid of bias
void randomize_border_ring_mark(border_ring_struct * brsp) {
  // here we will randomly select the mark from possible active border nodes
  if ( (*brsp).active_count > 1 ) { // only do something if there is a choice
    // pick a random active node
    long temp;
    temp = random();
    temp = temp % (*brsp).active_count;
    // now need to count through these
    border_node *mark = (*brsp).mark;
     while ( temp > 1 ) { // find next border node
      --temp;
      mark = (*mark).next;
      while  ( (*mark).state != 1) mark = (*mark).next;
    }
    (*brsp).mark = mark;
  } // else do nothing 
}
