#include "sa_types.h"
#include "sa_proto.h"

/* Goal is to split a complex into two complexes based on a
given closed edgepath, as if this path were a welding curve to
be analyzed. Can't use normal routines like cookie because they may
prune bdry verts to get a legal complex --- we need to keep the
boundary in tact. We just build the complexes directly so the 
edge path forms the boundary (and we build the vertex_map). Have to 
call this twice, once for inside, once for outside.

oi_flag: 0 = outside, 1 = inside.

For outside we add ideal vertex.
Swap nodes so that 2 is on bdry and is the gamma vert.

Note this distroys whatever was in pack q.
Return 0 on error, nodecount of new packing otherwise.

*/


int unweld(struct p_data *p,struct p_data *q,
	   struct Edgelist *elist,int oi_flag)
{
  int n,j,v,vv,vert,in_count,f,click,flips,num,hits;
  int ind1,ind2,a,b,k,new_alp;
  int *newflower=NULL,*targ=NULL;
  char *endptr,*nextpoint;
  struct Vertlist *vlist;
  struct Vertlist *curlist,*oldlist,*otrace,*ctrace;
  struct Edgelist *etrace;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;

  nextpoint=(char *)calloc(10,sizeof(char));
  strcpy(nextpoint,"g");
  if (!p->status) return 0;

  /* We mark in 'targ' all verts to the left of the given edgelist. */
  /* ------------------------------------------------------- */

  targ=(int *)calloc((p->nodecount+2),sizeof(int));
  etrace=elist;
  while(etrace->next) {
    targ[etrace->v]=targ[etrace->w]=1;
    etrace=etrace->next;
  }
  b=etrace->v;
  etrace=elist;
  curlist=ctrace=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
  while(etrace) {
    v=etrace->v;
    a=etrace->w;
    ind1=nghb(p,v,a);
    ind2=nghb(p,v,b);
    if (ind1<0 || ind2<0 || ind1==ind2) {
      free(targ);
      vert_free(&curlist);
      return 0;
    }
    if (ind2<ind1) ind2=ind2+pK_ptr[v].num;
    if (ind2>ind1+1) 
      for (j=ind1+1;j<=ind2;j++) 
	if (!targ[(k=pK_ptr[v].flower[j%pK_ptr[v].num])]) {
	  targ[k]=1;
	  ctrace=ctrace->next=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  ctrace->v=k;
	}
    etrace=etrace->next;
    b=v;
  }
  ctrace=curlist; /* first spot was empty */
  curlist=curlist->next;
  free(ctrace);
  
  oldlist=curlist;
  while(oldlist && oldlist->v) { /* keep alternating lists until 
				    old one is empty */
    curlist=ctrace=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
    otrace=oldlist;
    while(otrace) {
      for (j=0;j<=(pK_ptr[(v=otrace->v)].num+pK_ptr[otrace->v].bdry_flag);j++)
	if (!targ[(k=pK_ptr[v].flower[j%pK_ptr[v].num])]) {
	  targ[k]=1;
	  ctrace=ctrace->next=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  ctrace->v=k;
	}
      otrace=otrace->next;
    }
    vert_free(&oldlist);
    ctrace=curlist; /* first spot was empty */
    curlist=curlist->next;
    free(ctrace);
    oldlist=curlist;
  }
  vert_free(&oldlist);

  /* ------------------------------------------------------- */

  /* In case we want the outside instead of the inside: */
  if (!oi_flag) { 
    for (v=1;v<=p->nodecount;v++) {
      if(targ[v]) targ[v]=0;
      else targ[v]=1;
    }
    /* include vertices along elist also */
    etrace=elist;
    while(etrace) {
      targ[etrace->v]=targ[etrace->w]=1;
      etrace=etrace->next;
    }

    /* for disc, have to add an ideal vertex */
    if (p->num_bdry_comp>1) {
      sprintf(msgbuf,"unweld: in 'outside' case, p must be "
	      "combinatorial.sphere or disc.");
      free(targ);
      emsg();
      return 0;
    }
    if (p->num_bdry_comp==1) {
      vlist=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
      vlist->v=p->bdry_starts[1];
      if (!add_ideal(p,vlist)) {
	sprintf(msgbuf,"unweld: outside case, failed to add ideal "
		"vertex.");
	free(targ);
	emsg();
	vert_free(&vlist);
	return 0;
      }
      else {
	sprintf(msgbuf,"unweld: Note that an ideal vertex has been "
		"added to the parent complex.");
	msg();
      }
      vert_free(&vlist);
      pK_ptr=p->packK_ptr;
      pR_ptr=p->packR_ptr;
      v=p->nodecount;
      targ[v]=1;
      pR_ptr[v].rad=1.0;
      pR_ptr[v].center.re=pR_ptr[v].center.im;
      pK_ptr[v].plot_flag=1;
    }
  }

  /* At this point, inside or outside are set in 'targ' */
  /* -------------------------------------------------------- */

  /* give new index numbers to the marked vertices */
  in_count=0;
  for (n=1;n<=p->nodecount;n++) 
      if (targ[n]) targ[n]=++in_count;
  if (!oi_flag) new_alp=targ[p->nodecount];
  else new_alp=targ[p->alpha];
  
  if (!alloc_pack_space(q,p->nodecount,0)) {
    free(targ);
    sprintf(msgbuf,"unweld: error in allocating packing space.");
    emsg();
    return 0;
  }
  
  q->nodecount=in_count;
  for (n=1;n<=q->nodecount;n++) q->packK_ptr[n].util_flag=0;

  /* create the flowers for q */
  
  /* first, do those along edge -- sharp turns could yield 
     illegal combinatorics which will show up later */

  etrace=elist;
  while(etrace->next) etrace=etrace->next;
  if (oi_flag) b=etrace->v;
  else a=etrace->v;
  etrace=elist;
  while(etrace) {
    v=etrace->v;
    if (oi_flag) a=etrace->w;
    else b=etrace->w;
    ind1=nghb(p,v,a);
    ind2=nghb(p,v,b);
    if (ind1<0 || ind2<0 || ind1==ind2) {
      sprintf(msgbuf,"unweld: combinatoric problem.");
      emsg();
      goto BOMB;
    }
    if (ind2<ind1) ind2=ind2+pK_ptr[v].num;
    vv=targ[v];
    q->packK_ptr[vv].num=num=ind2-ind1;
    q->packK_ptr[vv].bdry_flag=1;
    newflower=(int *)calloc((size_t)(num+1),sizeof(int));
    for (j=ind1;j<=ind2;j++) {
      k=pK_ptr[v].flower[j%pK_ptr[v].num];
      newflower[j-ind1]=targ[k];
    }
    if (q->packK_ptr[vv].flower) free(q->packK_ptr[vv].flower);
    q->packK_ptr[vv].flower=newflower;
    q->packK_ptr[vv].util_flag=1;
    etrace=etrace->next;
    if (oi_flag) b=v;
    else a=v;
  }

  /* now do the rest, all should be interior */

  for (n=1;n<=p->nodecount;n++) 
    if ((vert=targ[n]) && !q->packK_ptr[vert].util_flag) { 
      /* this is one of the marked vertices but not on edge-list */
      if (pK_ptr[n].bdry_flag) { /* should be interior */
	sprintf(msgbuf,"unweld: combinatoric problem.");
	emsg();
	goto BOMB;
      }
      newflower=(int *)calloc(((num=pK_ptr[n].num)+1),sizeof(int));
      for (j=0;j<=num;j++) {
	k=targ[pK_ptr[n].flower[j]];
	newflower[j]=k;
      }
      if (q->packK_ptr[vert].flower) free(q->packK_ptr[vert].flower);
      q->packK_ptr[vert].flower=newflower;
      q->packK_ptr[vert].bdry_flag=0;
      q->packK_ptr[vert].num=pK_ptr[n].num;
    } /* end of if and for */

  /* -------------------------------------------------- */
  /* Now do the final details to get q in shape */
  /* create the vertex_map for q */
  
  if (q->vertex_map) q->vertex_map;
  q->vertex_map=etrace=(struct Edgelist *)
    calloc(1,sizeof(struct Edgelist));
  for (n=1;n<=p->nodecount;n++) 
    if ((v=targ[n])) {
      etrace=etrace->next=(struct Edgelist *)
	calloc(1,sizeof(struct Edgelist));
      etrace->v=v;
      etrace->w=n;
    }
  
  /* set default data for q --- in preparation for max packing */
  q->status=1;
  if (oi_flag) strcpy(q->file_name,"outside");
  else strcpy(q->file_name,"inside");
  q->hes=p->hes;
  for (n=1;n<=p->nodecount;n++) 
    if ((v=targ[n])) {
      q->packR_ptr[v].rad=pR_ptr[n].rad;
      q->packR_ptr[v].center=pR_ptr[n].center;
      q->packK_ptr[v].plot_flag=1;
    }
  v=targ[elist->v];
  if (q->packK_ptr[v].flower[0]==q->packK_ptr[v].flower[q->packK_ptr[v].num]
      || !swap_nodes(q,v,2)) {
    sprintf(msgbuf,"unweld: some error with bdry vert %d or 2.",v);
    emsg();
  }
  free(targ);
  q->gamma=2;
  if (new_alp==2) choose_alpha(q);
  else {
    q->alpha=new_alp;
    choose_alpha(q);
  }
  complex_count(q,1);
  facedraworder(q,0);
  set_aim_default(q);
  return in_count;
  
 BOMB:
  if (targ) free(targ);
  alloc_pack_space(q,5000,0);
  return 0;
} /* unweld */


int unweld_call(struct p_data *p,struct p_data *q,char *datastr)
{
  int flag=1,hits,count=0;
  char *endptr,*nextpoint=datastr;
  struct Edgelist *elist=NULL;

  stripsp(nextpoint);
  if (!strncmp(nextpoint,"-o ",3)) {
    flag=0;
    grab_next(&nextpoint,next);
  }
  if (!(elist=Node_pair_link(p,nextpoint,&endptr,&hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength))
      || hits<3) {
    sprintf(msgbuf,"unweld: usage: p q [-o] {e..}.");
    emsg();
    return 0;
  }
  if ((count=unweld(p,q,elist,flag))) {
    sprintf(msgbuf,"unweld returned a packing with nodecount %d.",count);
    msg();
    return count;
  }
  return 0;
} /* unweld_call */
